Package: mnirs
Title: Muscle Near-Infrared Spectroscopy Processing and Analysis
Version: 0.7.0
Authors@R: 
    person("Jem", "Arnold", , "jem.arnold@gmail.com", role = c("aut", "cre", "cph"),
           comment = c(ORCID = "0000-0003-3908-9447"))
Description: Read, process, and analyse data from muscle near-infrared
    spectroscopy (mNIRS) devices. Import raw data from file and return
    time-series data and metadata. Standardised methods for cleaning,
    filtering, transforming, and analysing mNIRS data. Custom plot theme
    and colour palette. Intended for mNIRS researchers and practitioners
    in exercise physiology, sports science, and clinical practice.
License: MIT + file LICENSE
URL: https://jemarnold.github.io/mnirs/,
        https://github.com/jemarnold/mnirs
BugReports: https://github.com/jemarnold/mnirs/issues
Depends: R (>= 4.1)
Imports: cli, data.table, lifecycle, readxl, rlang, stats, tibble,
        tidyselect
Suggests: dplyr, ggplot2, knitr, quarto, scales, signal, testthat (>=
        3.0.0), zoo
VignetteBuilder: quarto
Config/Needs/website: quarto, tidyverse
Config/roxygen2/version: 8.0.0
Config/testthat/edition: 3
Encoding: UTF-8
NeedsCompilation: no
Packaged: 2026-08-01 04:13:09 UTC; Jem
Author: Jem Arnold [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0003-3908-9447>)
Maintainer: Jem Arnold <jem.arnold@gmail.com>
Repository: CRAN
Date/Publication: 2026-08-01 05:10:07 UTC
Built: R 4.6.0; ; 2026-08-01 07:17:20 UTC; unix

Package: mnirs
Title: Muscle Near-Infrared Spectroscopy Processing and Analysis
Version: 0.6.5
Authors@R: 
    person("Jem", "Arnold", , "jem.arnold@gmail.com", role = c("aut", "cre", "cph"),
           comment = c(ORCID = "0000-0003-3908-9447"))
Description: Read, process, and analyse data from muscle near-infrared
    spectroscopy (mNIRS) devices. Import raw data from .csv or .xls(x)
    files and return time-series data and metadata. Includes standardised
    methods for cleaning, filtering, and pre-processing mNIRS data for
    subsequent analysis. Also includes a custom plot theme and colour
    palette. Intended for mNIRS researchers and practitioners in exercise
    physiology, sports science, and clinical rehabilitation with minimal
    coding experience required.
License: MIT + file LICENSE
URL: https://jemarnold.github.io/mnirs/,
        https://github.com/jemarnold/mnirs
BugReports: https://github.com/jemarnold/mnirs/issues
Depends: R (>= 4.1)
Imports: cli, data.table, lifecycle, readxl, rlang, stats, tibble,
        tidyselect
Suggests: dplyr, ggplot2, knitr, quarto, scales, signal, testthat (>=
        3.0.0), zoo
VignetteBuilder: quarto
Config/Needs/website: quarto, tidyverse
Config/testthat/edition: 3
Encoding: UTF-8
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-06-02 02:50:43 UTC; Jem
Author: Jem Arnold [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0003-3908-9447>)
Maintainer: Jem Arnold <jem.arnold@gmail.com>
Repository: CRAN
Date/Publication: 2026-06-02 03:10:02 UTC
Built: R 4.6.0; ; 2026-06-02 06:45:23 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(plot,mnirs)
S3method(print,mnirs)
export(breaks_timespan)
export(by_label)
export(by_lap)
export(by_sample)
export(by_time)
export(create_mnirs_data)
export(example_mnirs)
export(extract_intervals)
export(filter_butter)
export(filter_butterworth)
export(filter_ma)
export(filter_mnirs)
export(filter_moving_average)
export(format_hmmss)
export(palette_mnirs)
export(read_mnirs)
export(replace_invalid)
export(replace_missing)
export(replace_mnirs)
export(replace_outliers)
export(resample_mnirs)
export(rescale_mnirs)
export(scale_color_mnirs)
export(scale_colour_mnirs)
export(scale_fill_mnirs)
export(shift_mnirs)
export(theme_mnirs)
importFrom(cli,cli_abort)
importFrom(cli,cli_inform)
importFrom(cli,cli_warn)
importFrom(cli,col_blue)
importFrom(cli,qty)
importFrom(lifecycle,deprecated)
importFrom(rlang,"!!!")
importFrom(rlang,"!!")
importFrom(rlang,"%||%")
importFrom(rlang,":=")
importFrom(rlang,.data)
importFrom(rlang,check_installed)
importFrom(rlang,enquo)
importFrom(stats,coef)
importFrom(stats,median)
importFrom(stats,rnorm)
importFrom(stats,setNames)
importFrom(tibble,tibble)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# mnirs <a href="https://jemarnold.github.io/mnirs/"><img src="man/figures/mnirs-hex.svg" align="right" height="139" alt="mnirs website" /></a>

<!-- badges: start -->

[![CRAN
downloads](https://cranlogs.r-pkg.org/badges/grand-total/mnirs)](https://cran.r-project.org/package=mnirs)
[![Lifecycle:
experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)
[![R-CMD-check](https://github.com/jemarnold/mnirs/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/jemarnold/mnirs/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/jemarnold/mnirs/graph/badge.svg)](https://app.codecov.io/gh/jemarnold/mnirs)
<!-- badges: end -->

*{mnirs}* contains standardised, reproducible methods for reading,
processing, transforming, and analysing data from muscle near-infrared
spectroscopy (mNIRS) devices. Intended for mNIRS researchers and
practitioners in exercise physiology, sports science, and clinical
practice.

## Installation

``` r
install.packages("mnirs")
```

You can install the development version of *`mnirs`* from
[GitHub](https://github.com/jemarnold/mnirs) with:

``` r
# install.packages("pak")
pak::pak("jemarnold/mnirs")
```

## Citation

<https://cran.r-universe.dev/mnirs/citation>

`<package manuscript coming soon>`

## Online App

A very basic implementation of this package is hosted at
<https://jemarnold-mnirs-app.share.connect.posit.cloud/> and can
currently be used for reading and processing mNIRS data.

[![mnirs processing shiny
app](https://raw.githubusercontent.com/jemarnold/mnirs/main/man/figures/README-mnirs-app.gif)](https://jemarnold-mnirs-app.share.connect.posit.cloud/)

## Usage

A more detailed vignette for common usage can be found on the package
website: [Reading and Cleaning Data with
{mnirs}](https://jemarnold.github.io/mnirs/articles/reading-mnirs-data.html)

*`mnirs`* is designed for mNIRS data, but it can be used to read, clean,
and process other time series datasets which require many of the same
processing steps. Enjoy!

### `read_mnirs()` Read data from file

``` r
library(ggplot2) ## for plotting
library(mnirs)

## {mnirs} includes sample files from a few mNIRS devices
example_mnirs()
#> [1] "artinis_intervals.xlsx"  "moxy_intervals.csv"     
#> [3] "moxy_ramp.xlsx"          "portamon-oxcap.xlsx"    
#> [5] "train.red_intervals.csv"

## rename channels in the format `renamed = "original_name"`
## where "original_name1" should match the file column name exactly
data_raw <- read_mnirs(
    file_path = example_mnirs("moxy_ramp"), ## call an example data file
    nirs_channels = c(
        smo2_left = "SmO2 Live",    ## identify and rename channels
        smo2_right = "SmO2 Live(2)"
    ),
    time_channel = c(time = "hh:mm:ss"), ## date-time format will be converted to numeric
    event_channel = NULL,           ## leave blank if unused
    sample_rate = NULL,             ## if blank, will be estimated from time_channel
    add_timestamp = FALSE,          ## omit a date-time timestamp column
    zero_time = TRUE,               ## recalculate time values from zero
    keep_all = FALSE,               ## return only the specified data channels
    verbose = TRUE                  ## show warnings & messages
)
#> ! Estimated `sample_rate` = 2 Hz.
#> ℹ Define `sample_rate` explicitly to override.
#> Warning in read_mnirs(): ! Duplicate or irregular `time_channel` samples detected.
#> ℹ time = 211.59 and 1183.6.
#> ℹ Re-sample with `mnirs::resample_mnirs()`.

## Note the above info message that sample_rate was estimated correctly at 2 Hz 👆
## ignore the warnings about irregular sampling for now, we will resample later

data_raw
#> # A tibble: 2,202 × 3
#>     time smo2_left smo2_right
#>    <dbl>     <dbl>      <dbl>
#>  1 0            54         68
#>  2 0.560        54         68
#>  3 1.11         54         66
#>  4 1.66         54         66
#>  5 2.21         54         66
#>  6 2.76         54         66
#>  7 3.31         57         67
#>  8 3.86         57         67
#>  9 4.41         57         67
#> 10 4.96         57         67
#> # ℹ 2,192 more rows

## note the `time_labels` plot argument to display time values as `h:mm:ss`
plot(
    data_raw,
    time_labels = TRUE,
    na.omit = FALSE
)
```

<img src="man/figures/README-read_mnirs-1.png" alt="" width="100%" />

### Metadata stored in `mnirs` data frames

``` r
## view metadata
attr(data_raw, "nirs_channels")
#> [1] "smo2_left"  "smo2_right"

attr(data_raw, "time_channel")
#> [1] "time"

attr(data_raw, "sample_rate")
#> [1] 2
```

### `resample_mnirs()`: Resample data

``` r
data_resampled <- resample_mnirs(
    data_raw,            ## blank channels will be retrieved from metadata
    time_channel = time, ## channels can be left blank or specified explicitly
    sample_rate = NULL,  ## blank by default will be retrieved from metadata
    resample_rate = 2,   ## blank by default will resample to `sample_rate`
    method = "linear"    ## linear interpolation across resampled indices
)
#> ℹ Output is resampled at 2 Hz.

## note the altered "time" values from the original data frame 👇
data_resampled
#> # A tibble: 2,419 × 3
#>     time smo2_left smo2_right
#>    <dbl>     <dbl>      <dbl>
#>  1   0        54         68  
#>  2   0.5      54         68  
#>  3   1        54         66.4
#>  4   1.5      54         66  
#>  5   2        54         66  
#>  6   2.5      54         66  
#>  7   3        55.3       66.4
#>  8   3.5      57         67  
#>  9   4        57         67  
#> 10   4.5      57         67  
#> # ℹ 2,409 more rows
```

### `replace_mnirs()`: Replace local outliers, invalid values, and missing values

``` r
data_cleaned <- replace_mnirs(
    data_resampled,     ## blank channels will be retrieved from metadata
    invalid_values = 0, ## known invalid values in the data
    invalid_above = 90, ## remove data spikes above 90
    outlier_cutoff = 3, ## recommended default value
    width = 7,          ## window to detect and replace outliers/missing values
    method = "linear"   ## linear interpolation over `NA`s
)

plot(data_cleaned, time_labels = TRUE)
```

<img src="man/figures/README-replace_mnirs-1.png" alt="" width="100%" />

### `filter_mnirs()`: Digital filtering

``` r
data_filtered <- filter_mnirs(
    data_cleaned,           ## blank channels will be retrieved from metadata
    method = "butterworth", ## Butterworth digital filter is a common choice
    order = 2,              ## filter order number
    W = 0.02,               ## filter fractional critical frequency `[0, 1]`
    type = "low",           ## specify a "low-pass" filter
    na.rm = TRUE            ## explicitly ignore NAs
)

## we will add the non-filtered data back to the plot to compare
plot(data_filtered, time_labels = TRUE) +
    geom_line(
        data = data_cleaned, 
        aes(y = smo2_left, colour = "smo2_left"), alpha = 0.4
    ) +
    geom_line(
        data = data_cleaned, 
        aes(y = smo2_right, colour = "smo2_right"), alpha = 0.4
    )
```

<img src="man/figures/README-filter_mnirs-1.png" alt="" width="100%" />

### `shift_mnirs()` & `rescale_mnirs()`: Shift and rescale data

``` r
data_shifted <- shift_mnirs(
    data_filtered,
    group_channels = list(smo2_left, smo2_right), ## channels shifted separately
    to = 0,            ## NIRS values will be shifted to zero
    span = 120,        ## shift the *first* 120 sec of data to zero
    position = "first"
)

plot(data_shifted, time_labels = TRUE) +
    geom_hline(yintercept = 0, linetype = "dotted")
```

<img src="man/figures/README-shift_mnirs-1.png" alt="" width="100%" />

``` r
data_rescaled <- rescale_mnirs(
    data_filtered,
    group_channels = list(smo2_left, smo2_right), ## channels rescaled separately
    range = c(0, 100) ## rescale to a 0-100% functional exercise range
)

plot(data_rescaled, time_labels = TRUE) +
    geom_hline(yintercept = c(0, 100), linetype = "dotted")
```

<img src="man/figures/README-rescale_mnirs-1.png" alt="" width="100%" />

### Pipe-friendly functions

``` r
## global option to silence info & warning messages
options(mnirs.verbose = FALSE)

nirs_data <- read_mnirs(
    example_mnirs("train.red"),
    nirs_channels = c(
        smo2_left = "SmO2 unfiltered",
        smo2_right = "SmO2 unfiltered"
    ),
    time_channel = c(time = "Timestamp (seconds passed)"),
    zero_time = TRUE
) |>
    resample_mnirs(method = "linear") |> ## default settings will resample to the same `sample_rate`
    replace_mnirs(
        invalid_above = 73,
        outlier_cutoff = 3,
        span = 7
    ) |>
    filter_mnirs(
        method = "butterworth",
        order = 2,
        W = 0.02,
        na.rm = TRUE
    ) |>
    shift_mnirs(
        group_channels = list(smo2_left, smo2_right), ## 👈 channels grouped separately
        to = 0,
        span = 60,
        position = "first"
    ) |>
    rescale_mnirs(
        group_channels = list(c(smo2_left, smo2_right)), ## 👈 channels grouped together
        range = c(0, 100)
    )

plot(nirs_data, time_labels = TRUE)
```

<img src="man/figures/README-pipeline-1.png" alt="" width="100%" />

### `extract_intervals()`: detect events and extract intervals

``` r
## return each interval independently with `group_intervals = "distinct"`
distinct <- extract_intervals(
    nirs_data,                    ## channels blank for "distinct" grouping
    group_intervals = "distinct", ## return a list of data frames for each (2) event
    start = by_time(348, 1064),   ## manually identified interval start times
    end = by_time(458, 1174),     ## interval end time (start + 150 sec)
    span = c(0, 0),               ## no boundary modification
    zero_time = FALSE             ## return original time values
)

plot(distinct, time_labels = TRUE)
```

<img src="man/figures/README-extract_intervals_distinct-1.png" alt="" width="100%" />

``` r
## ensemble average both intervals with `group_intervals = "ensemble"`
ensemble <- extract_intervals(
    nirs_data,                    ## channels recycled to all intervals by default
    group_intervals = "ensemble", ## ensemble-average across two intervals
    start = by_time(368, 1084),   ## alternatively specify start times + span
    span = c(-20, 90),            ## span recycled to all intervals by default
    zero_time = TRUE              ## re-calculate common time to start from `0`
)

plot(ensemble, time_labels = TRUE) + 
    geom_vline(xintercept = 0, linetype = "dotted")
```

<img src="man/figures/README-extract_intervals_ensemble-1.png" alt="" width="100%" />

## Future *`mnirs`* development

- Process oxygenation kinetics

  - Monoexponential & sigmoidal non-linear curve fitting

  - Non-parametric response time & slope analysis

- Critical oxygenation breakpoint analysis

  - Manual selection and automation-assisted breakpoint detection
    (combine expert evaluation with robust probabilistic breakpoint
    detection)

- Oxidative capacity assessment

  - Repeated occlusion ensemble-averaging and model fitting

  - Blood volume correction

## mNIRS device compatibility

This package is designed to recognise file formats exported from common
wearable mNIRS devices. It should be flexible for use with other file
formats, and compatibility will improve with continued development.

Currently, it has been tested successfully with mNIRS data exported from
the following devices and apps:

- [Artinis](https://artinis.com/oxysoft) Oxysoft software (.csv and
  .xlsx)
- [Moxy](https://www.moxymonitor.com/) direct export (.csv)
- [PerfPro](https://perfprostudio.com/) PC software (.xlsx)
- [Train.Red](https://train.red/) app (.csv)
- [VO2 Master Manager](https://vo2master.com/features/) app (.xlsx)

------------------------------------------------------------------------

*Generative chatbots are used to assist with code optimisation. All code
is thoroughly reviewed and validated by the package author.*

<!-- README.md is generated from README.Rmd. Please edit that file -->

# mnirs <a href="https://jemarnold.github.io/mnirs/"><img src="man/figures/mnirs-hex.svg" align="right" height="139" alt="mnirs website" /></a>

<!-- badges: start -->

[![Lifecycle:
experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)
[![R-CMD-check](https://github.com/jemarnold/mnirs/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/jemarnold/mnirs/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/jemarnold/mnirs/graph/badge.svg)](https://app.codecov.io/gh/jemarnold/mnirs)
<!-- badges: end -->

*{mnirs}* contains standardised, reproducible methods for reading,
processing, and analysing data from muscle near-infrared spectroscopy
(mNIRS) devices. Intended for mNIRS researchers and practitioners in
exercise physiology, sports science, and clinical rehabilitation with
minimal coding experience required.

## Installation

``` r
install.packages("mnirs")
```

You can install the development version of *`mnirs`* from
[GitHub](https://github.com/jemarnold/mnirs) with:

``` r
# install.packages("pak")
pak::pak("jemarnold/mnirs")
```

## Citation

`<coming soon>`

## Online App

A very basic implementation of this package is hosted at
<https://jemarnold-mnirs-app.share.connect.posit.cloud/> and can
currently be used for reading and pre-processing mNIRS data.

[![mnirs processing shiny
app](https://raw.githubusercontent.com/jemarnold/mnirs/main/man/figures/README-mnirs-app.gif)](https://jemarnold-mnirs-app.share.connect.posit.cloud/)

## Usage

A more detailed vignette for common usage can be found on the package
website: [Reading and Cleaning Data with
{mnirs}](https://jemarnold.github.io/mnirs/articles/reading-mnirs-data.html)

*`mnirs`* is designed for mNIRS data, but it can be used to read, clean,
and process other time series datasets which require many of the same
processing steps. Enjoy!

### `read_mnirs()` Read data from file

``` r
library(ggplot2) ## for plotting
library(mnirs)

## {mnirs} includes sample files from a few mNIRS devices
example_mnirs()
#> [1] "artinis_intervals.xlsx"  "moxy_intervals.csv"     
#> [3] "moxy_ramp.xlsx"          "portamon-oxcap.xlsx"    
#> [5] "train.red_intervals.csv"

## rename channels in the format `renamed = "original_name"`
## where "original_name1" should match the file column name exactly
data_raw <- read_mnirs(
    file_path = example_mnirs("moxy_ramp"), ## call an example data file
    nirs_channels = c(
        smo2_left = "SmO2 Live",    ## identify and rename channels
        smo2_right = "SmO2 Live(2)"
    ),
    time_channel = c(time = "hh:mm:ss"), ## date-time format will be converted to numeric
    event_channel = NULL,           ## leave blank if unused
    sample_rate = NULL,             ## if blank, will be estimated from time_channel
    add_timestamp = FALSE,          ## omit a date-time timestamp column
    zero_time = TRUE,               ## recalculate time values from zero
    keep_all = FALSE,               ## return only the specified data channels
    verbose = TRUE                  ## show warnings & messages
)
#> ! Estimated `sample_rate` = 2 Hz.
#> ℹ Define `sample_rate` explicitly to override.
#> Warning: ! Duplicate or irregular `time_channel` samples detected.
#> ℹ `time` = 211.59 and 1183.6.
#> ℹ Re-sample with `mnirs::resample_mnirs()`.

## Note the above info message that sample_rate was estimated correctly at 2 Hz 👆
## ignore the warnings about irregular sampling for now, we will resample later

data_raw
#> # A tibble: 2,202 × 3
#>     time smo2_left smo2_right
#>    <dbl>     <dbl>      <dbl>
#>  1 0            54         68
#>  2 0.560        54         68
#>  3 1.11         54         66
#>  4 1.66         54         66
#>  5 2.21         54         66
#>  6 2.76         54         66
#>  7 3.31         57         67
#>  8 3.86         57         67
#>  9 4.41         57         67
#> 10 4.96         57         67
#> # ℹ 2,192 more rows

## note the `time_labels` plot argument to display time values as `h:mm:ss`
plot(
    data_raw,
    time_labels = TRUE,
    na.omit = FALSE
)
```

<img src="man/figures/README-read_mnirs-1.png" alt="" width="100%" />

### Metadata stored in `mnirs` data frames

``` r
## view metadata
attr(data_raw, "nirs_channels")
#> [1] "smo2_left"  "smo2_right"

attr(data_raw, "time_channel")
#> [1] "time"

attr(data_raw, "sample_rate")
#> [1] 2
```

### `resample_mnirs()`: Resample data

``` r
data_resampled <- resample_mnirs(
    data_raw,            ## blank channels will be retrieved from metadata
    time_channel = time, ## channels can be left blank or specified explicitly
    sample_rate = NULL,  ## blank by default will be retrieved from metadata
    resample_rate = 2,   ## blank by default will resample to `sample_rate`
    method = "linear"    ## linear interpolation across resampled indices
)
#> ℹ Output is resampled at 2 Hz.

## note the altered "time" values from the original data frame 👇
data_resampled
#> # A tibble: 2,419 × 3
#>     time smo2_left smo2_right
#>    <dbl>     <dbl>      <dbl>
#>  1   0        54         68  
#>  2   0.5      54         68  
#>  3   1        54         66.4
#>  4   1.5      54         66  
#>  5   2        54         66  
#>  6   2.5      54         66  
#>  7   3        55.3       66.4
#>  8   3.5      57         67  
#>  9   4        57         67  
#> 10   4.5      57         67  
#> # ℹ 2,409 more rows
```

### `replace_mnirs`: Replace local outliers, invalid values, and missing values

``` r
data_cleaned <- replace_mnirs(
    data_resampled,     ## blank channels will be retrieved from metadata
    invalid_values = 0, ## known invalid values in the data
    invalid_above = 90, ## remove data spikes above 90
    outlier_cutoff = 3, ## recommended default value
    width = 7,          ## window to detect and replace outliers/missing values
    method = "linear"   ## linear interpolation over `NA`s
)

plot(data_cleaned, time_labels = TRUE)
```

<img src="man/figures/README-replace_mnirs-1.png" alt="" width="100%" />

### `filter_mnirs()`: Digital filtering

``` r
data_filtered <- filter_mnirs(
    data_cleaned,           ## blank channels will be retrieved from metadata
    method = "butterworth", ## Butterworth digital filter is a common choice
    order = 2,              ## filter order number
    W = 0.02,               ## filter fractional critical frequency `[0, 1]`
    type = "low",           ## specify a "low-pass" filter
    na.rm = TRUE            ## explicitly ignore NAs
)

## we will add the non-filtered data back to the plot to compare
plot(data_filtered, time_labels = TRUE) +
    geom_line(
        data = data_cleaned, 
        aes(y = smo2_left, colour = "smo2_left"), alpha = 0.4
    ) +
    geom_line(
        data = data_cleaned, 
        aes(y = smo2_right, colour = "smo2_right"), alpha = 0.4
    )
```

<img src="man/figures/README-filter_mnirs-1.png" alt="" width="100%" />

### `shift_mnirs()` & `rescale_mnirs()`: Shift and rescale data

``` r
data_shifted <- shift_mnirs(
    data_filtered,     ## un-grouped nirs channels to shift separately 
    nirs_channels = list(smo2_left, smo2_right), 
    to = 0,            ## NIRS values will be shifted to zero
    span = 120,        ## shift the *first* 120 sec of data to zero
    position = "first"
)

plot(data_shifted, time_labels = TRUE) +
    geom_hline(yintercept = 0, linetype = "dotted")
```

<img src="man/figures/README-shift_mnirs-1.png" alt="" width="100%" />

``` r
data_rescaled <- rescale_mnirs(
    data_filtered,    ## un-grouped nirs channels to rescale separately 
    nirs_channels = list(smo2_left, smo2_right), 
    range = c(0, 100) ## rescale to a 0-100% functional exercise range
)

plot(data_rescaled, time_labels = TRUE) +
    geom_hline(yintercept = c(0, 100), linetype = "dotted")
```

<img src="man/figures/README-rescale_mnirs-1.png" alt="" width="100%" />

### Pipe-friendly functions

``` r
## global option to silence info & warning messages
options(mnirs.verbose = FALSE)

nirs_data <- read_mnirs(
    example_mnirs("train.red"),
    nirs_channels = c(
        smo2_left = "SmO2 unfiltered",
        smo2_right = "SmO2 unfiltered"
    ),
    time_channel = c(time = "Timestamp (seconds passed)"),
    zero_time = TRUE
) |>
    resample_mnirs(method = "linear") |> ## default settings will resample to the same `sample_rate`
    replace_mnirs(
        invalid_above = 73,
        outlier_cutoff = 3,
        span = 7
    ) |>
    filter_mnirs(
        method = "butterworth",
        order = 2,
        W = 0.02,
        na.rm = TRUE
    ) |>
    shift_mnirs(
        nirs_channels = list(smo2_left, smo2_right), ## 👈 channels grouped separately
        to = 0,
        span = 60,
        position = "first"
    ) |>
    rescale_mnirs(
        nirs_channels = list(c(smo2_left, smo2_right)), ## 👈 channels grouped together
        range = c(0, 100)
    )

plot(nirs_data, time_labels = TRUE)
```

<img src="man/figures/README-pipeline-1.png" alt="" width="100%" />

### `extract_intervals()`: detect events and extract intervals

``` r
## return each interval independently with `event_groups = "distinct"`
distinct <- extract_intervals(
    nirs_data,                  ## channels blank for "distinct" grouping
    start = by_time(348, 1064), ## manually identified interval start times
    end = by_time(458, 1174),   ## interval end time (start + 150 sec)
    event_groups = "distinct",  ## return a list of data frames for each (2) event
    span = c(0, 0),             ## no boundary modification
    zero_time = FALSE           ## return original time values
)

plot(distinct, time_labels = TRUE)
```

<img src="man/figures/README-extract_intervals_distinct-1.png" alt="" width="100%" />

``` r
## ensemble average both intervals with `event_groups = "ensemble"`
ensemble <- extract_intervals(
    nirs_data,                  ## channels recycled to all intervals by default
    nirs_channels = c(smo2_left, smo2_right),
    start = by_time(368, 1084), ## alternatively specify start times + span
    event_groups = "ensemble",  ## ensemble-average across two intervals
    span = c(-20, 90),          ## span recycled to all intervals by default
    zero_time = TRUE            ## re-calculate common time to start from `0`
)

plot(ensemble, time_labels = TRUE) + 
    geom_vline(xintercept = 0, linetype = "dotted")
```

<img src="man/figures/README-extract_intervals_ensemble-1.png" alt="" width="100%" />

## Future *`mnirs`* development

- Process oxygenation kinetics

  - Monoexponential & sigmoidal non-linear curve fitting

  - Non-parametric response time & slope analysis

- Critical oxygenation breakpoint analysis

  - Manual selection and automation-assisted breakpoint detection
    (combine expert evaluation with robust probabilistic breakpoint
    detection)

- Oxidative capacity assessment

  - Repeated occlusion ensemble-averaging and model fitting

  - Blood volume correction

## mNIRS device compatibility

This package is designed to recognise mNIRS data exported as *.csv* or
*.xls(x)* files. It should be flexible for use with many different NIRS
devices, and compatibility will improve with continued development.

Currently, it has been tested successfully with mNIRS data exported from
the following devices and apps:

- [Artinis](https://artinis.com/oxysoft) Oxysoft software (.csv and
  .xlsx)
- [Moxy](https://www.moxymonitor.com/) direct export (.csv)
- [PerfPro](https://perfprostudio.com/) PC software (.xlsx)
- [Train.Red](https://train.red/) app (.csv)
- [VO2 Master Manager](https://vo2master.com/features/) app (.xlsx)

------------------------------------------------------------------------

*Generative chatbots are used to assist with code optimisation. All code
is thoroughly reviewed and validated by the package author.*

## -----------------------------------------------------------------------------
#| label: setup
#| warning: false

# pak::pak("jemarnold/mnirs") ## install development version
library(ggplot2)   ## for plotting
library(mnirs)


## -----------------------------------------------------------------------------
nirs_channels = c(
    renamed1 = "original_name1", 
    renamed2 = "original_name2"
)


## -----------------------------------------------------------------------------
## {mnirs} includes sample files from a few NIRS devices
example_mnirs()

## partial matching will error if matches multiple
try(example_mnirs("moxy"))

data_raw <- read_mnirs(
    file_path = example_mnirs("moxy_ramp"), ## call an example data file
    nirs_channels = c(
        smo2_left = "SmO2 Live",    ## identify and rename channels
        smo2_right = "SmO2 Live(2)"
    ),
    time_channel = c(time = "hh:mm:ss"), ## date-time format will be converted to numeric
    event_channel = NULL,           ## leave blank if unused
    sample_rate = NULL,             ## if blank, will be estimated from time_channel
    add_timestamp = FALSE,          ## omit a date-time timestamp column
    zero_time = TRUE,               ## recalculate time values from zero
    keep_all = FALSE,               ## return only the specified data channels
    verbose = TRUE                  ## show warnings & messages
)

## Note the above info message that sample_rate was estimated correctly at 2 Hz ☝
## ignore the warnings about irregular sampling for now, we will resample later

data_raw


## -----------------------------------------------------------------------------
## note the `time_labels` plot argument to display time values as `h:mm:ss`
plot(
    data_raw,
    points = FALSE,
    time_labels = TRUE,
    na.omit = FALSE
)


## -----------------------------------------------------------------------------
## view metadata
attr(data_raw, "nirs_channels")

attr(data_raw, "time_channel")

attr(data_raw, "sample_rate")


## -----------------------------------------------------------------------------
data_resampled <- resample_mnirs(
    data_raw,            ## blank channels will be retrieved from metadata
    time_channel = time, ## channels can be left blank or specified explicitly
    sample_rate = NULL,  ## blank by default will be retrieved from metadata
    resample_rate = 2,   ## blank by default will resample to `sample_rate`
    method = "linear"    ## linear interpolation across resampled indices
)

## note the altered "time" values from the original data frame 👇
data_resampled


## -----------------------------------------------------------------------------
data_cleaned <- replace_mnirs(
    data_resampled,     ## blank channels will be retrieved from metadata
    invalid_values = 0, ## known invalid values in the data
    invalid_above = 90, ## remove data spikes above 90
    outlier_cutoff = 3, ## recommended default value
    width = 7,          ## window to detect and replace outliers/missing values
    method = "linear"   ## linear interpolation over `NA`s
)

plot(data_cleaned, time_labels = TRUE)


## -----------------------------------------------------------------------------
data_filtered <- filter_mnirs(
    data_cleaned,           ## blank channels will be retrieved from metadata
    method = "butterworth", ## Butterworth digital filter is a common choice
    order = 2,              ## filter order number
    W = 0.02,               ## filter fractional critical frequency `[0, 1]`
    type = "low",           ## specify a "low-pass" filter
    na.rm = TRUE            ## explicitly ignore NAs
)

## we will add the non-filtered data back to the plot to compare
plot(data_filtered, time_labels = TRUE) +
    geom_line(
        data = data_cleaned, 
        aes(y = smo2_left, colour = "smo2_left"), alpha = 0.4
    ) +
    geom_line(
        data = data_cleaned, 
        aes(y = smo2_right, colour = "smo2_right"), alpha = 0.4
    )


## -----------------------------------------------------------------------------
data_shifted <- shift_mnirs(
    data_filtered,     ## un-grouped nirs channels to shift separately 
    nirs_channels = list(smo2_left, smo2_right), ## 👈 channels grouped separately
    to = 0,            ## NIRS values will be shifted to zero
    span = 120,        ## shift the *first* 120 sec of data to zero
    position = "first"
)

plot(data_shifted, time_labels = TRUE) +
    geom_hline(yintercept = 0, linetype = "dotted")


## -----------------------------------------------------------------------------
data_rescaled <- rescale_mnirs(
    data_filtered,    ## un-grouped nirs channels to rescale separately 
    nirs_channels = list(smo2_left, smo2_right), ## 👈 channels grouped separately
    range = c(0, 100) ## rescale to a 0-100% functional exercise range
)

plot(data_rescaled, time_labels = TRUE) +
    geom_hline(yintercept = c(0, 100), linetype = "dotted")


## -----------------------------------------------------------------------------
options(mnirs.verbose = FALSE)

nirs_data <- read_mnirs(
    example_mnirs("train.red"),
    nirs_channels = c(
        smo2_left = "SmO2 unfiltered",
        smo2_right = "SmO2 unfiltered"
    ),
    time_channel = c(time = "Timestamp (seconds passed)"),
    zero_time = TRUE
) |>
    resample_mnirs(method = "linear") |> ## default settings will resample to the same `sample_rate`
    replace_mnirs(
        invalid_above = 73,
        outlier_cutoff = 3,
        span = 7
    ) |>
    filter_mnirs(
        method = "butterworth",
        order = 2,
        W = 0.02,
        na.rm = TRUE
    ) |>
    shift_mnirs(
        nirs_channels = list(smo2_left, smo2_right), ## 👈 channels grouped separately
        to = 0,
        span = 60,
        position = "first"
    ) |>
    rescale_mnirs(
        nirs_channels = list(c(smo2_left, smo2_right)), ## 👈 channels grouped together
        range = c(0, 100)
    )

plot(nirs_data, time_labels = TRUE)


## -----------------------------------------------------------------------------
#| fig.width: 8
#| fig.asp: 0.55

## return each interval independently with `event_groups = "distinct"`
distinct <- extract_intervals(
    nirs_data,                  ## channels blank for "distinct" grouping
    start = by_time(348, 1064), ## manually identified interval start times
    end = by_time(458, 1174),   ## interval end time (start + 150 sec)
    event_groups = "distinct",  ## return a list of data frames for each (2) event
    span = c(0, 0),             ## no boundary modification
    zero_time = FALSE           ## return original time values
)

plot(distinct, time_labels = TRUE)


## -----------------------------------------------------------------------------
## ensemble average both intervals with `event_groups = "ensemble"`
ensemble <- extract_intervals(
    nirs_data,                  ## channels recycled to all intervals by default
    nirs_channels = c(smo2_left, smo2_right),
    start = by_time(368, 1084), ## alternatively specify start times + span
    event_groups = "ensemble",  ## ensemble-average across two intervals
    span = c(-20, 90),          ## span recycled to all intervals by default
    zero_time = TRUE            ## re-calculate common time to start from `0`
)

plot(ensemble, time_labels = TRUE) + 
    geom_vline(xintercept = 0, linetype = "dotted")

## -----------------------------------------------------------------------------
#| label: setup
#| warning: false

# pak::pak("jemarnold/mnirs") ## install development version
library(ggplot2)   ## for plotting
library(mnirs)


## -----------------------------------------------------------------------------
nirs_channels = c(
    renamed1 = "original_name1", 
    renamed2 = "original_name2"
)


## -----------------------------------------------------------------------------
## {mnirs} includes sample files from a few NIRS devices
example_mnirs()

## partial matching will error if matches multiple
try(example_mnirs("moxy"))

data_raw <- read_mnirs(
    file_path = example_mnirs("moxy_ramp"), ## call an example data file
    nirs_channels = c(
        smo2_left = "SmO2 Live",    ## identify and rename channels
        smo2_right = "SmO2 Live(2)"
    ),
    time_channel = c(time = "hh:mm:ss"), ## date-time format will be converted to numeric
    event_channel = NULL,           ## leave blank if unused
    sample_rate = NULL,             ## if blank, will be estimated from time_channel
    add_timestamp = FALSE,          ## omit a date-time timestamp column
    zero_time = TRUE,               ## recalculate time values from zero
    keep_all = FALSE,               ## return only the specified data channels
    verbose = TRUE                  ## show warnings & messages
)

## Note the above info message that sample_rate was estimated correctly at 2 Hz ☝
## ignore the warnings about irregular sampling for now, we will resample later

data_raw


## -----------------------------------------------------------------------------
## note the `time_labels` plot argument to display time values as `h:mm:ss`
plot(
    data_raw,
    points = FALSE,
    time_labels = TRUE,
    na.omit = FALSE
)


## -----------------------------------------------------------------------------
## view metadata
attr(data_raw, "nirs_channels")

attr(data_raw, "time_channel")

attr(data_raw, "sample_rate")


## -----------------------------------------------------------------------------
data_resampled <- resample_mnirs(
    data_raw,            ## blank channels will be retrieved from metadata
    time_channel = time, ## channels can be left blank or specified explicitly
    sample_rate = NULL,  ## blank by default will be retrieved from metadata
    resample_rate = 2,   ## blank by default will resample to `sample_rate`
    method = "linear"    ## linear interpolation across resampled indices
)

## note the altered "time" values from the original data frame 👇
data_resampled


## -----------------------------------------------------------------------------
data_cleaned <- replace_mnirs(
    data_resampled,     ## blank channels will be retrieved from metadata
    invalid_values = 0, ## known invalid values in the data
    invalid_above = 90, ## remove data spikes above 90
    outlier_cutoff = 3, ## recommended default value
    width = 7,          ## window to detect and replace outliers/missing values
    method = "linear"   ## linear interpolation over `NA`s
)

plot(data_cleaned, time_labels = TRUE)


## -----------------------------------------------------------------------------
data_filtered <- filter_mnirs(
    data_cleaned,           ## blank channels will be retrieved from metadata
    method = "butterworth", ## Butterworth digital filter is a common choice
    order = 2,              ## filter order number
    W = 0.02,               ## filter fractional critical frequency `[0, 1]`
    type = "low",           ## specify a "low-pass" filter
    na.rm = TRUE            ## explicitly ignore NAs
)

## we will add the non-filtered data back to the plot to compare
plot(data_filtered, time_labels = TRUE) +
    geom_line(
        data = data_cleaned, 
        aes(y = smo2_left, colour = "smo2_left"), alpha = 0.4
    ) +
    geom_line(
        data = data_cleaned, 
        aes(y = smo2_right, colour = "smo2_right"), alpha = 0.4
    )


## -----------------------------------------------------------------------------
data_shifted <- shift_mnirs(
    data_filtered,
    group_channels = list(smo2_left, smo2_right), ## 👈 channels shifted separately
    to = 0,            ## NIRS values will be shifted to zero
    span = 120,        ## shift the *first* 120 sec of data to zero
    position = "first"
)

plot(data_shifted, time_labels = TRUE) +
    geom_hline(yintercept = 0, linetype = "dotted")


## -----------------------------------------------------------------------------
data_rescaled <- rescale_mnirs(
    data_filtered,
    group_channels = list(smo2_left, smo2_right), ## 👈 channels rescaled separately
    range = c(0, 100) ## rescale to a 0-100% functional exercise range
)

plot(data_rescaled, time_labels = TRUE) +
    geom_hline(yintercept = c(0, 100), linetype = "dotted")


## -----------------------------------------------------------------------------
options(mnirs.verbose = FALSE)

nirs_data <- read_mnirs(
    example_mnirs("train.red"),
    nirs_channels = c(
        smo2_left = "SmO2 unfiltered",
        smo2_right = "SmO2 unfiltered"
    ),
    time_channel = c(time = "Timestamp (seconds passed)"),
    zero_time = TRUE
) |>
    resample_mnirs(method = "linear") |> ## default settings will resample to the same `sample_rate`
    replace_mnirs(
        invalid_above = 73,
        outlier_cutoff = 3,
        span = 7
    ) |>
    filter_mnirs(
        method = "butterworth",
        order = 2,
        W = 0.02,
        na.rm = TRUE
    ) |>
    shift_mnirs(
        group_channels = list(smo2_left, smo2_right), ## 👈 channels grouped separately
        to = 0,
        span = 60,
        position = "first"
    ) |>
    rescale_mnirs(
        group_channels = list(c(smo2_left, smo2_right)), ## 👈 channels grouped together
        range = c(0, 100)
    )

plot(nirs_data, time_labels = TRUE)


## -----------------------------------------------------------------------------
#| fig.width: 8
#| fig.asp: 0.55

## return each interval independently with `group_intervals = "distinct"`
distinct <- extract_intervals(
    nirs_data,                    ## channels blank for "distinct" grouping
    group_intervals = "distinct", ## return a list of data frames for each (2) event
    start = by_time(348, 1064),   ## manually identified interval start times
    end = by_time(458, 1174),     ## interval end time (start + 150 sec)
    span = c(0, 0),               ## no boundary modification
    zero_time = FALSE             ## return original time values
)

plot(distinct, time_labels = TRUE)


## -----------------------------------------------------------------------------
## ensemble average both intervals with `group_intervals = "ensemble"`
ensemble <- extract_intervals(
    nirs_data,                    ## channels recycled to all intervals by default
    group_intervals = "ensemble", ## ensemble-average across two intervals
    start = by_time(368, 1084),   ## alternatively specify start times + span
    span = c(-20, 90),            ## span recycled to all intervals by default
    zero_time = TRUE              ## re-calculate common time to start from `0`
)

plot(ensemble, time_labels = TRUE) + 
    geom_vline(xintercept = 0, linetype = "dotted")

#' Plot *{mnirs}* objects
#'
#' Create a base plot for data frames or lists of data frames with class
#' *"mnirs"*.
#'
#' @param x Data frame or list of data frames of class *"mnirs"* (e.g. from
#'   [extract_intervals()]). List input produces a faceted plot with one
#'   panel per element.
#' @param points Logical. Default is `FALSE`. If `TRUE` displays
#'   `ggplot2::geom_points()`. Otherwise displays `ggplot2::geom_lines()`.
#' @param time_labels Logical. Default is `FALSE`. If `TRUE` displays x-axis
#'   time values formatted as *"hh:mm:ss"* using [format_hmmss()]. Otherwise,
#'   x-axis values are displayed as numeric.
#' @param n.breaks A numeric value specifying the number of breaks in both
#'   x- and y-axes. Default is `5`.
#' @param na.omit Logical. Default is `FALSE`. If `TRUE` omits missing (`NA`)
#'   and non-finite `c(Inf, -Inf, NaN)` from display.
#' @param ... Additional arguments.
#'
#' @details
#' When `x` is a named list of *"mnirs"* data frames, elements are bound into a
#' single data frame and displayed as faceted panels via
#' [ggplot2::facet_wrap()].
#'
#' Arguments in `...` are currently passed to [ggplot2::facet_wrap()]
#' formals, such as `nrow`, `ncol`, and `scales` for more precise control.
#'
#' @returns A [ggplot2][ggplot2::ggplot()] object.
#'
#' @examplesIf rlang::is_installed(c("ggplot2", "scales"))
#' data <- read_mnirs(
#'     example_mnirs("train.red"),
#'     nirs_channels = c(smo2 = "SmO2"),
#'     time_channel = c(time = "Timestamp (seconds passed)"),
#'     verbose = FALSE
#' )
#' 
#' ## plot time labels as "hh:mm:ss"
#' plot(data, time_labels = TRUE)
#' 
#' data_list <- extract_intervals(
#'     data,
#'     start = by_time(2452, 3168),
#'     span = c(-60, 120),
#'     verbose = FALSE
#' )
#' 
#' ## plot a list of mnirs data frames as faceted panels
#' plot(data_list, time_labels = TRUE)
#'
#' @export
plot.mnirs <- function(
    x,
    points = FALSE,
    time_labels = FALSE,
    n.breaks = 5,
    na.omit = FALSE,
    ...
) {
    check_installed(c("ggplot2"), reason = "to plot mNIRS data")

    args <- list(...)

    ## handle list of mnirs data frames
    if (is.list(x) && !is.data.frame(x)) {
        x <- as_plot_data(x)
    }

    nirs_channels <- attr(x, "nirs_channels")
    time_channel <- attr(x, "time_channel")

    ## pre-compute conditionals
    x_name <- if (time_labels) {
        paste(time_channel, "(h:mm:ss)")
    } else {
        ggplot2::waiver()
    }
    x_breaks <- if (time_labels) {
        breaks_timespan(n = n.breaks)
    } else if (rlang::is_installed("scales")) {
        scales::breaks_pretty(n = n.breaks)
    } else {
        ggplot2::waiver()
    }
    x_labels <- if (time_labels) {
        format_hmmss
    } else {
        ggplot2::waiver()
    }
    y_breaks <- if (rlang::is_installed("scales")) {
        scales::breaks_pretty(n = n.breaks)
    } else {
        ggplot2::waiver()
    }

    ## build base plot with axis configuration
    plot <- ggplot2::ggplot(x) +
        ggplot2::aes(x = .data[[time_channel]]) +
        theme_mnirs() +
        ggplot2::labs(x = x_name, y = "mNIRS") +
        ggplot2::scale_x_continuous(
            breaks = x_breaks,
            labels = x_labels,
            expand = ggplot2::expansion(mult = 0.01)
        ) +
        ggplot2::scale_y_continuous(
            breaks = y_breaks,
            expand = ggplot2::expansion(mult = 0.01)
        ) +
        scale_colour_mnirs(
            name = NULL,
            guide = ggplot2::guide_legend(
                override.aes = list(linewidth = 1)
            )
        )

    ## add one geom per channel
    layers <- lapply(nirs_channels, function(ch) {
        ch_data <- if (na.omit) x[is.finite(x[[ch]]), ] else x
        ch_aes <- ggplot2::aes(y = .data[[ch]], colour = ch)
        c(
            list(ggplot2::geom_line(ch_aes, data = ch_data)),
            if (points) list(
                ggplot2::geom_point(ch_aes, data = ch_data, size = 3)
            )
        )
    })

    ## facet when plotting multiple mnirs data frames
    if (".id" %in% names(x)) {
        facet_args <- intersect(
            names(args),
            names(formals(ggplot2::facet_wrap))
        )
        plot <- plot + do.call(
            ggplot2::facet_wrap,
            c(
                list(facets = ~.id, scales = args[["scales"]] %||% "free_x"),
                args[setdiff(facet_args, "scales")]
            )
        )
    }

    return(plot + layers)
}


#' Validate and bind a list of mnirs data frames for plotting
#' @keywords internal
as_plot_data <- function(x) {
    if (length(x) == 0L) {
        cli_abort(c(
            "x" = "{.fn plot.mnirs} must contain at least one \\
            {col_blue('\"mnirs\"')} data frame."
        ))
    }

    if (any(!vapply(x, is.data.frame, logical(1)))) {
        cli_abort(c(
            "x" = "{.fn plot.mnirs} must contain all {col_blue('\"mnirs\"')} \\
            data frames."
        ))
    }

    ## validate time_channel is consistent across elements
    time_channels <- vapply(x, \(.df) {
        attr(.df, "time_channel") %||% NA_character_
    }, character(1))
    
    if (anyNA(time_channels)) {
        cli_abort(c(
            "x" = "All elements of {.fn plot.mnirs} must have a \\
            {.field time_channel} attribute."
        ))
    }
    if (length(unique(time_channels)) > 1L) {
        cli_abort(c(
            "x" = "All elements of {.fn plot.mnirs} must share the same \\
            {.field time_channel}.",
            "i" = "Found: {.val {unique(time_channels)}}."
        ))
    }

    ## auto-name unnamed list elements
    if (is.null(names(x))) {
        names(x) <- paste0("interval_", seq_along(x))
    }

    ## length-1 list: unwrap to single data frame
    if (length(x) == 1L) {
        return(x[[1L]])
    }

    ## union of nirs_channels across all elements
    nirs_channels <- unique(unlist(
        lapply(x, attr, "nirs_channels"),
        use.names = FALSE
    ))

    ## pad each element with NA for any missing nirs_channels
    x <- lapply(x, \(.df) {
        .df[setdiff(nirs_channels, names(.df))] <- NA_real_
        .df
    })

    ## add .id column to each element, then row-bind
    x <- Map(\(.df, .nm) {
        .df[[".id"]] <- .nm
        .df
    }, x, names(x))
    plot_data <- do.call(rbind, unname(x))
    plot_data[[".id"]] <- factor(
        plot_data[[".id"]], levels = unique(plot_data[[".id"]])
    )
    attr(plot_data, "nirs_channels") <- nirs_channels
    attr(plot_data, "time_channel") <- time_channels[[1L]]

    return(plot_data)
}


#' Custom *{mnirs}* ggplot2 theme
#'
#' A `[ggplot2][ggplot2::ggplot2-package]` theme for display.
#'
#' @param base_size Base font size, given in pts.
#' @param base_family Base font family.
#' @param border Define either a *partial* or *full* border around plots.
#' @param ink Colour for text and lines. *Default* is *"black"*.
#' @param paper Background colour. *Default* is *"white"*.
#' @param accent Accent colour for highlights. *Default* is *"#0080ff"*.
#' @param ... Additional arguments to add to `[ggplot2::theme()]`.
#'
#' @details
#' - `axis.title = element_text(face = "bold")` by *default* Modify to *"plain"*.
#'
#' - `panel.grid.major` & `panel.grid.major` set to blank. Modify to
#'   `= element_line()` for visible grid lines.
#'
#' - `legend.position = "top"` by *default* Modify `"none"` to remove legend
#'   entirely.
#'
#' - `border = "partial"` uses `panel.border = element_blank()` and
#'   `axis.line = element_line()`.
#'
#' - `border = "full"` uses `panel.border = element_rect(colour = "black",`
#'   `linewidth = 1)` and `axis.line = element_line()`.
#'
#' - `base_family = "sans"` by *default*.
#'
#' @returns A [ggplot2][ggplot2::ggplot()] theme object.
#'
#' @seealso [palette_mnirs()], [scale_colour_mnirs()]
#'
#' @examplesIf rlang::is_installed(c("ggplot2", "scales"))
#' ## plot example data
#' read_mnirs(
#'     file_path = example_mnirs("moxy_ramp"),
#'     nirs_channels = c(smo2_left = "SmO2 Live",
#'                       smo2_right = "SmO2 Live(2)"),
#'     time_channel = c(time = "hh:mm:ss"),
#'     verbose = FALSE
#' ) |>
#'     plot(time_labels = TRUE)
#'
#' @export
theme_mnirs <- function(
    base_size = 14,
    base_family = "sans",
    border = c("partial", "full"),
    ink = "black",
    paper = "white",
    accent = "#0080ff",
    ...
) {
    check_installed("ggplot2", reason = "to plot mNIRS data")
    border <- match.arg(border)
    half_line <- base_size * 0.5

    if (border == "partial") {
        panel.border <- ggplot2::element_blank()
        axis.line <- ggplot2::element_line()
    } else {
        panel.border <- ggplot2::element_rect(colour = "black", linewidth = 1)
        axis.line <- ggplot2::element_blank()
    }

    ggplot2::theme_bw(
        base_size = base_size,
        base_family = base_family,
        ink = ink,
        paper = paper,
        accent = accent
    ) +
        ggplot2::theme(
            plot.title = ggplot2::element_text(
                size = ggplot2::rel(1.2),
                lineheight = 1.1
            ),
            plot.subtitle = ggplot2::element_text(lineheight = 1.1),
            plot.caption = ggplot2::element_text(colour = "grey50"),
            panel.border = panel.border,
            axis.line = axis.line,
            axis.title = ggplot2::element_text(face = "bold"),
            strip.background = ggplot2::element_rect(fill = "grey95"),
            strip.text = ggplot2::element_text(
                margin = ggplot2::margin_auto(t = half_line * 0.5)
            ),
            plot.margin = ggplot2::margin_part(r = base_size, unit = "pt"),
            panel.grid.major = ggplot2::element_blank(),
            panel.grid.minor = ggplot2::element_blank(),
            legend.position = "top",
            legend.justification = "right",
            legend.direction = "horizontal",
            legend.margin = ggplot2::margin_auto(t = 1),
            legend.box.spacing = ggplot2::unit(half_line * 0.5, "pt")
        ) +
        ggplot2::theme(...)
}


#' Custom *{mnirs}* colour palette
#'
#' @param ... Either a single numeric specifying the number of colours to
#'   return, or character strings specifying colour names. If empty, all
#'   colours are returned.
#'
#' @returns Named (when selecting by name) or unnamed character vector of
#'   hex colours.
#'
#' @seealso [theme_mnirs()], [scale_colour_mnirs()]
#'
#' @examplesIf rlang::is_installed("scales")
#' scales::show_col(palette_mnirs())
#' scales::show_col(palette_mnirs(2))
#' scales::show_col(palette_mnirs("red", "orange"))
#'
#' @export
palette_mnirs <- function(...) {
    # fmt: skip
    colours <- c(                         ## NIRS location codes
        `light blue`  = "#0080ff",      ## "VL"
        `dark red`    = "#ba2630",      ## "FCR"
        `light green` = "#5b8c52",      ## "BB" "#7dbf70" alt
        `pink`        = "#ff80ff",      ## "VM"
        `orange`      = "#ff7f00",      ## "SCM"
        `dark blue`   = "#00468Bff",    ## "TA"
        `light red`   = "#db5555",      ## "ECR"
        `green`       = "#42B540FF",    ## "DL"
        `purple`      = "#9f79ee",      ## "RF"
        `brown`       = "#8b4726",      ## "PS"
        `blue`        = "#0000ff",      ## "HHb"
        `red`         = "#ED0000FF"     ## "O2Hb"
    )

    dots <- list(...)

    if (length(dots) == 0L) {
        return(colours)
    }

    ## numeric -> subset by count
    if (length(dots) == 1L && is.numeric(dots[[1L]])) {
        n <- dots[[1L]]
        validate_numeric(n, 1, c(1, Inf), msg1 = "one-element positive")
        if (n <= length(colours)) {
            return(unname(colours[seq_len(n)]))
        }
        ## interpolate if more colours needed, but this probably won't look good!
        return(grDevices::colorRampPalette(colours)(n))
    }

    ## character args -> subset by name
    names <- unlist(dots)

    if (!is.character(names)) {
        ## covers condition of multiple numeric vals
        cli_abort(c(
            "x" = "{.fn palette_mnirs} expects a single numeric value \\
            for the number of colours to return, or character colour names."
        ))
    }
    names <- match.arg(names, choices = names(colours), several.ok = TRUE)
    return(colours[names])
}


#' Scales for custom *{mnirs}* palette
#'
#' @param ... Arguments passed to `ggplot2::discrete_scale()`.
#' @param aesthetics A character vector with aesthetic(s) passed to
#'   `ggplot2::discrete_scale()`. *Default* is `"colour"`.
#'
#' @returns A [ggplot2][ggplot2::ggplot()] scale object.
#'
#' @seealso [theme_mnirs()], [palette_mnirs()]
#'
#' @examplesIf rlang::is_installed(c("ggplot2", "scales"))
#' ## plot example data
#' data <- read_mnirs(
#'     file_path = example_mnirs("moxy_ramp"),
#'     nirs_channels = c(smo2_left = "SmO2 Live",
#'                       smo2_right = "SmO2 Live(2)"),
#'     time_channel = c(time = "hh:mm:ss"),
#'     verbose = FALSE
#' )
#'
#' ggplot2::ggplot(data, ggplot2::aes(x = time)) +
#'     theme_mnirs() +
#'     scale_colour_mnirs(name = NULL) +
#'     ggplot2::geom_line(ggplot2::aes(y = smo2_left, colour = "smo2_left")) +
#'     ggplot2::geom_line(ggplot2::aes(y = smo2_right, colour = "smo2_right"))
#'
#' @rdname scale_colour_mnirs
#' @export
scale_colour_mnirs <- function(..., aesthetics = "colour") {
    check_installed("ggplot2", reason = "to plot mNIRS data")

    ggplot2::discrete_scale(
        aesthetics = aesthetics,
        palette = palette_mnirs,
        na.value = "grey10",
        ...
    )
}

#' @rdname scale_colour_mnirs
#' @export
scale_color_mnirs <- scale_colour_mnirs

#' @rdname scale_colour_mnirs
#' @export
scale_fill_mnirs <- function(..., aesthetics = "fill") {
    check_installed("ggplot2", reason = "to plot mNIRS data")

    ggplot2::discrete_scale(
        aesthetics = aesthetics,
        palette = palette_mnirs,
        na.value = "grey10",
        ...
    )
}


#' Breaks for time span data
#'
#' Pretty time span breaks for plotting in units of 5, 15, 30, 60 sec, etc.
#' Modified from [scales::breaks_timespan()].
#'
#' @param unit The time unit used to interpret numeric data input (*defaults*
#'   to *"secs"*).
#' @param n Desired number of breaks. You may get slightly more or fewer breaks
#'   than requested.
#'
#' @returns Returns a function for generating breaks.
#'
#' @examplesIf rlang::is_installed("ggplot2")
#' x <- 0:120
#' y <- sin(2 * pi * x / 15) + rnorm(length(x), 0, 0.2)
#'
#' ggplot2::ggplot(data.frame(x, y), ggplot2::aes(x, y)) +
#'     theme_mnirs() +
#'     ggplot2::scale_x_continuous(breaks = breaks_timespan()) +
#'     ggplot2::geom_line()
#'
#' @export
breaks_timespan <- function(
    unit = c("secs", "mins", "hours", "days", "weeks"),
    n = 5
) {
    unit <- match.arg(unit)
    force(n)
    function(x) {
        x <- as.numeric(as.difftime(x, units = unit), units = "secs")
        range <- range(x, na.rm = TRUE)
        diff <- range[2L] - range[1L]

        ## scale of time range
        ## define nice steps for each unit
        if (diff <= 5 * 60) {
            scale <- 1 ## sec
            nice_steps <- c(1, 2, 5, 10, 15, 20, 30, 60, 120)
        } else if (diff <= 5 * 3600) {
            scale <- 60 ## min
            nice_steps <- c(1, 2, 5, 10, 15, 20, 30, 60, 120) * 60
        } else if (diff <= 5 * 86400) {
            scale <- 3600 ## hr
            nice_steps <- c(0.25, 0.5, 1, 2, 3, 4, 6, 8, 12, 24) * 3600
        } else {
            scale <- 86400 ## days
            nice_steps <- c(1, 7, 28) * 86400
        }

        ## scale to scale units
        range_scaled <- range / scale
        scaled_steps <- nice_steps / scale

        ## find optimal step size from nice_steps
        target_step <- diff(range_scaled) / n
        best_step <- scaled_steps[which.min(abs(scaled_steps - target_step))]

        ## generate breaks
        breaks_scaled <- seq(
            floor(range_scaled[1L] / best_step) * best_step,
            ceiling(range_scaled[2L] / best_step) * best_step,
            by = best_step
        )

        ## convert back to seconds
        round(as.numeric(as.difftime(breaks_scaled * scale, units = "secs")))
    }
}


#' Format time span data as h:mm:ss
#'
#' Convert numeric time span data to `h:mm:ss` format for pretty plotting.
#' Inspired by [ggplot2::scale_x_time()].
#'
#' @param x A numeric vector.
#'
#' @details
#' If all values are less than 3600 (1 hour), then format is returned as
#'   `mm:ss`. If any value is greater than 3600, format is returned as
#'   `h:mm:ss` with leading zeroes.
#'
#' @returns A character vector the same length as `x`.
#'
#' @examplesIf rlang::is_installed("ggplot2")
#' x <- 0:120
#' y <- sin(2 * pi * x / 15) + rnorm(length(x), 0, 0.2)
#'
#' ggplot2::ggplot(data.frame(x, y), ggplot2::aes(x, y)) +
#'     theme_mnirs() +
#'     ggplot2::scale_x_continuous(
#'         breaks = breaks_timespan(),
#'         labels = format_hmmss
#'     ) +
#'     ggplot2::geom_line()
#'
#' @export
format_hmmss <- function(x) {
    # validate_numeric(x)
    x <- as.numeric(x)
    ## logical whether to handle NAs
    handle_na <- anyNA(x)

    if (handle_na) {
        na_info <- preserve_na(x)
        x <- na_info$x_valid
    }

    sign <- ifelse(x < 0, "-", "")
    hrs <- as.integer(abs(x) %/% 3600)
    mins <- as.integer((abs(x) %% 3600) %/% 60)
    secs <- abs(x) %% 60

    hmmss_string <- if (any(hrs > 0, na.rm = TRUE)) {
        sprintf("%s%d:%02d:%02d", sign, hrs, mins, secs)
    } else {
        sprintf("%s%02d:%02d", sign, mins, secs)
    }

    ## return y to original x length with NAs if handled
    if (handle_na) {
        return(restore_na(hmmss_string, na_info))
    } else {
        return(hmmss_string)
    }
}

skip_if_not_installed("ggplot2")

## theme_mnirs() ============================================================
test_that("theme_mnirs returns a ggplot2 theme object", {
    theme_obj <- theme_mnirs()
    expect_s3_class(theme_obj, "theme")
    expect_s3_class(theme_obj, "gg")
})

test_that("theme_mnirs border argument works correctly", {
    partial <- theme_mnirs(border = "partial")
    full <- theme_mnirs(border = "full")
    expect_s3_class(partial$panel.border, "element_blank")
    expect_s3_class(full$panel.border, "element_rect")
})

test_that("theme_mnirs accepts custom colours", {
    custom <- theme_mnirs(ink = "red", paper = "blue", accent = "#ff0000")
    expect_s3_class(custom, "theme")
})

## palette_mnirs() ========================================
test_that("palettes returns correct colour vector", {
    all_colours <- palette_mnirs()
    expect_type(all_colours, "character")
    expect_length(all_colours, 12)
    expect_true(all(grepl("^#[0-9A-Fa-f]{6}", all_colours)))
})

test_that("palettes subset by number works", {
    expect_length(palette_mnirs(3), 3)
    expect_length(palette_mnirs(1), 1)
    expect_error(palette_mnirs(2:4), "valid.*numeric")
    expect_error(palette_mnirs(0), "valid.*numeric")
})

test_that("palettes subset by name works", {
    red <- palette_mnirs("red")
    expect_named(red, "red")
    expect_equal(red[["red"]], "#ED0000FF")
    expect_error(palette_mnirs("invalid"), "should be one of")

    multi <- palette_mnirs("red", "blue")
    expect_length(multi, 2)
    expect_named(multi, c("red", "blue"))
    expect_equal(palette_mnirs("red", "invalid"), red)

    ## mixed types error
    expect_error(palette_mnirs(TRUE), "expects")
})

test_that("palette_mnirs interpolates when n > 12", {
    many <- palette_mnirs(20)
    expect_length(many, 20)
})

## scale_colour_mnirs() ==================================================
test_that("scale_color_mnirs is an alias for scale_colour_mnirs", {
    expect_identical(scale_color_mnirs, scale_colour_mnirs)
})

test_that("scale_*_mnirs returns a ggplot2 Scale object", {
    expect_s3_class(scale_colour_mnirs(), "Scale")
    expect_s3_class(scale_colour_mnirs(), "ScaleDiscrete")

    expect_s3_class(scale_fill_mnirs(), "Scale")
    expect_s3_class(scale_fill_mnirs(), "ScaleDiscrete")
})

test_that("scale_colour_mnirs uses correct aesthetics", {
    expect_equal(scale_colour_mnirs()$aesthetics, "colour")
    expect_equal(scale_fill_mnirs()$aesthetics, "fill")
})

test_that("scale functions pass through additional arguments", {
    scale <- scale_colour_mnirs(name = "Test")
    expect_equal(scale$name, "Test")
})

test_that("scale functions use palette_mnirs", {
    # Extract the palette function and call it
    expect_equal(scale_colour_mnirs()$palette(5), palette_mnirs(5))
    expect_equal(scale_fill_mnirs()$palette(5), palette_mnirs(5))

    # Test with no argument (all colours)
    expect_equal(scale_colour_mnirs()$palette(), palette_mnirs())

    # Test with character argument
    expect_equal(
        scale_colour_mnirs()$palette("light blue", "dark red"),
        palette_mnirs("light blue", "dark red")
    )

    ## test with na.value
    expect_equal(scale_colour_mnirs()$na.value, "grey10")
    expect_equal(scale_fill_mnirs()$na.value, "grey10")
})

test_that("scale functions work in ggplot2 plots", {
    p <- ggplot2::ggplot(mtcars, ggplot2::aes(mpg, wt, colour = factor(cyl))) +
        ggplot2::geom_point() +
        scale_colour_mnirs()

    expect_s3_class(p, "gg")
    expect_s3_class(p$scales$get_scales("colour"), "ScaleDiscrete")
})

## breaks_timespan() ==================================================
test_that("breaks_timespan returns a function", {
    breaks_fn <- breaks_timespan()
    expect_type(breaks_fn, "closure")
})

test_that(" breaks_timespan corresponds to nice_steps for each scale level", {
    # Test scale = 1 (diff <= 5 * 60)
    nice_steps_sec <- c(1, 2, 5, 10, 15, 20, 30, 60, 120)
    x_sec <- c(0, 150) # 2.5 min range
    breaks_sec <- breaks_timespan("secs", n = 5)(x_sec)
    steps_sec <- unique(diff(breaks_sec))
    expect_true(all(steps_sec %in% nice_steps_sec))
    expect_type(breaks_sec, "double")
    expect_true(all(breaks_sec >= 0 & breaks_sec <= 150))
    expect_true(all.equal(length(breaks_sec), 5, tolerance = 2, scale = 1))

    # Test scale = 60 (5 * 60 < diff <= 5 * 3600)
    nice_steps_min <- c(1, 2, 5, 10, 15, 20, 30, 60, 120) * 60
    x_min <- c(0, 7200) # 2 hour range
    breaks_min <- breaks_timespan("secs", n = 5)(x_min)
    steps_min <- unique(diff(breaks_min))
    expect_true(all(steps_min %in% nice_steps_min))
    expect_type(breaks_min, "double")
    expect_true(all(breaks_min >= 0 & breaks_min <= 7200))
    expect_true(all.equal(length(breaks_min), 5, tolerance = 2, scale = 1))

    # Test scale = 3600 (5 * 3600 < diff <= 5 * 86400)
    nice_steps_hr <- c(0.25, 0.5, 1, 2, 3, 4, 6, 8, 12, 24) * 3600
    x_hr <- c(0, 86400) # 1 day range
    breaks_hr <- breaks_timespan("secs", n = 5)(x_hr)
    steps_hr <- unique(diff(breaks_hr))
    expect_true(all(steps_hr %in% nice_steps_hr))
    expect_type(breaks_hr, "double")
    expect_true(all(breaks_hr >= 0 & breaks_hr <= 86400))
    expect_true(all.equal(length(breaks_hr), 5, tolerance = 2, scale = 1))

    # Test scale = 86400 (diff > 5 * 86400)
    nice_steps_day <- c(1, 7, 28) * 86400
    x_day <- c(0, 86400 * 28) # 28 day range
    breaks_day <- breaks_timespan("secs", n = 5)(x_day)
    steps_day <- unique(diff(breaks_day))
    expect_true(all(steps_day %in% nice_steps_day))
    expect_type(breaks_day, "double")
    expect_true(all(breaks_day >= 0 & breaks_day <= 86400 * 28))
    expect_true(all.equal(length(breaks_day), 5, tolerance = 2, scale = 1))
})

## format_hmmss() ==================================================
test_that("format_hmmss formats time correctly", {
    ## seconds
    expect_equal(format_hmmss(0), "00:00")
    expect_equal(format_hmmss(30), "00:30")
    expect_equal(format_hmmss(90), "01:30")
    expect_equal(format_hmmss(3599), "59:59")

    ## hours
    expect_equal(format_hmmss(3600), "1:00:00")
    expect_equal(format_hmmss(3661), "1:01:01")
    expect_equal(format_hmmss(7325), "2:02:05")

    ## negative
    expect_equal(format_hmmss(-30), "-00:30")
    expect_equal(format_hmmss(-3661), "-1:01:01")
})

test_that("format_hmmss handles vectors", {
    result <- format_hmmss(c(0, 30, 90, 3600))
    expect_length(result, 4)
    expect_equal(result, c("0:00:00", "0:00:30", "0:01:30", "1:00:00"))
})

test_that("format_hmmss handles NA values", {
    result <- format_hmmss(c(30, NA, 90))
    expect_length(result, 3)
    expect_true(is.na(result[2]))
})


## as_plot_data() =============================================
# Helper to create mock mNIRS object
mock_mnirs <- function() {
    df <- data.frame(
        time = 1:10,
        HHb = c(1:6, NA, NA, 9:10),
        O2Hb = c(rep(2, 3), NA, NA, rep(2, 5))
    )
    structure(
        df,
        class = c("mnirs", "data.frame"),
        nirs_channels = c("HHb", "O2Hb"),
        time_channel = "time"
    )
}

test_that("as_plot_data errors on invalid lists", {
    ## empty list
    expect_error(as_plot_data(list()), "at least one")
    ## not a df
    expect_error(as_plot_data(list(mock_mnirs(), "not_a_df")), "must contain all")
})

test_that("as_plot_data errors when element missing time_channel attribute", {
    bad <- structure(
        data.frame(time = 1:3, HHb = 1:3),
        class = c("mnirs", "data.frame"),
        nirs_channels = "HHb"
    )
    expect_error(as_plot_data(list(bad)), "time_channel attribute")
})

test_that("as_plot_data errors when elements have differing time_channel", {
    a <- mock_mnirs()
    b <- structure(
        data.frame(t = 1:10, HHb = 1:10),
        class = c("mnirs", "data.frame"),
        nirs_channels = "HHb",
        time_channel = "t"
    )
    expect_error(as_plot_data(list(a, b)), "same.*time_channel")
})

test_that("as_plot_data unwraps single-element list", {
    x <- mock_mnirs()
    result <- as_plot_data(list(x))
    expect_identical(result, x)
})

test_that("as_plot_data row-binds named list with .id factor", {
    a <- mock_mnirs()
    b <- mock_mnirs()
    result <- as_plot_data(list(pre = a, post = b))
    expect_true(".id" %in% names(result))
    expect_s3_class(result[[".id"]], "factor")
    expect_equal(levels(result[[".id"]]), c("pre", "post"))
    expect_equal(nrow(result), nrow(a) + nrow(b))
    expect_equal(attr(result, "time_channel"), "time")
    expect_equal(attr(result, "nirs_channels"), c("HHb", "O2Hb"))
})

test_that("as_plot_data auto-names unnamed list with sequential integers", {
    a <- mock_mnirs()
    b <- mock_mnirs()
    result <- as_plot_data(list(a, b))
    expect_equal(levels(result[[".id"]]), c("interval_1", "interval_2"))
})

test_that("as_plot_data unions nirs_channels across elements", {
    a <- structure(
        data.frame(time = 1:3, HHb = 1:3),
        class = c("mnirs", "data.frame"),
        nirs_channels = "HHb",
        time_channel = "time"
    )
    b <- structure(
        data.frame(time = 1:3, O2Hb = 4:6),
        class = c("mnirs", "data.frame"),
        nirs_channels = "O2Hb",
        time_channel = "time"
    )
    result <- as_plot_data(x = list(a, b))

    # plot(result)
    expect_equal(attr(result, "nirs_channels"), c("HHb", "O2Hb"))
})

## plot.mnirs() ===============================================
test_that("na.omit removes rows with any NA in nirs_channels", {
    x <- mock_mnirs()

    ## na.omit = FALSE: all 10 rows retained per channel
    p1 <- plot(x, points = TRUE)
    expect_equal(nrow(p1$layers[[1L]]$data), 10L) ## HHb line
    expect_equal(nrow(p1$layers[[3L]]$data), 10L) ## O2Hb line

    p2 <- plot(x, points = TRUE, na.omit = TRUE)
    expect_equal(nrow(p2$layers[[1L]]$data), 8L) ## HHb line
    expect_equal(nrow(p2$layers[[3L]]$data), 8L) ## O2Hb line
})

test_that("time_labels controls x-axis name and formatting", {
    x <- mock_mnirs()

    # With time_labels = FALSE (default)
    p1 <- plot(x)
    expect_true(ggplot2::is_waiver(p1$scales$get_scales("x")$name))
    expect_true(ggplot2::is_waiver(p1$scales$get_scales("x")$labels))

    # With time_labels = TRUE
    p2 <- plot(x, time_labels = TRUE)
    expect_equal(p2$labels$x, "time (h:mm:ss)")
    expect_false(ggplot2::is_waiver(p2$scales$get_scales("x")$labels))
})

test_that("n.breaks controls number of breaks", {
    x <- mock_mnirs()

    # Extract breaks by building plot
    get_breaks <- function(p, axis = "x") {
        built <- ggplot2::ggplot_build(p)
        built$layout$panel_params[[1]]$x$breaks
    }

    p1 <- plot(x, n.breaks = 3)
    p2 <- plot(x, n.breaks = 10)

    breaks1 <- get_breaks(p1)
    breaks2 <- get_breaks(p2)

    # More n should generally produce more breaks
    expect_true(length(breaks2) >= length(breaks1))
})

test_that("plot.mnirs groups and facets", {
    x <- mock_mnirs()

    # Add a grouping column that we want to facet by
    x$group <- rep(c("A", "B"), each = 5)

    # Create plot
    p <- plot(x)

    # Check that the group column exists in plot data
    expect_true("group" %in% names(p$data))

    # Check that group values are correctly repeated
    expect_equal(p$data$group, x$group)

    # Verify faceting works without error
    expect_no_error(p + ggplot2::facet_wrap(~group))
})

test_that("plot.mnirs works with extract_intervals and faceting", {
    x <- mock_mnirs()

    # Simulate extract_intervals output with interval column
    x$interval <- factor(rep(1:2, each = 5))

    p <- plot(x)

    # Verify interval column preserved as factor
    expect_true("interval" %in% names(p$data))
    expect_s3_class(p$data$interval, "factor")

    # Test faceting works
    p_facet <- p + ggplot2::facet_wrap(~interval)
    expect_s3_class(p_facet, "ggplot")

    # Build plot to ensure no errors during rendering
    expect_no_error(ggplot2::ggplot_build(p_facet))
})

test_that("plot.mnirs uses waiver() for breaks when scales is unavailable", {
    x <- mock_mnirs()

    with_mocked_bindings(
        is_installed = function(pkg, ...) FALSE,
        .package = "rlang",
        {
            p <- plot(x, time_labels = FALSE)
            x_scale <- p$scales$get_scales("x")
            y_scale <- p$scales$get_scales("y")
            ## both break functions fall back to waiver()
            expect_true(ggplot2::is_waiver(x_scale$breaks))
            expect_true(ggplot2::is_waiver(y_scale$breaks))
        }
    )
})

test_that("plot.mnirs works on lists", {
    df_list <- read_mnirs(
        file_path = example_mnirs("train.red"),
        nirs_channels = c(
            smo2_left = "SmO2",
            smo2_right = "SmO2 unfiltered"
        ),
        time_channel = c(time = "Timestamp (seconds passed)"),
        event_channel = c(lap = "Lap/Event"),
        verbose = FALSE
    ) |>
        resample_mnirs(method = "linear", verbose = FALSE) |>
        extract_intervals(
            start = by_lap(3, 5),
            span = c(-30, 120),
            zero_time = TRUE,
            verbose = FALSE
        )

    expect_type(df_list, "list")
    expect_s3_class(df_list, "mnirs")
    expect_length(df_list, 2)

    ## visual check
    p <- plot(df_list)
    expect_s3_class(p, "ggplot")

    ## facet wrap present for multi-element list
    facet_layers <- Filter(\(l) inherits(l, "FacetWrap"), list(p$facet))
    expect_length(facet_layers, 1)

    ## renders without error
    expect_no_error(ggplot2::ggplot_build(p))
})

test_that("plot.mnirs() returns ggplot2 warnings for missing values", {
    a <- structure(
        data.frame(time = 1:3, HHb = 1:3),
        class = c("mnirs", "data.frame"),
        nirs_channels = "HHb",
        time_channel = "time"
    )
    b <- structure(
        data.frame(time = 1:3, O2Hb = 4:6),
        class = c("mnirs", "data.frame"),
        nirs_channels = "O2Hb",
        time_channel = "time"
    )
    result <- as_plot_data(x = list(a, b))

    w <- tryCatch(
        print(plot(result)),
        warning = \(w) conditionMessage(w)
    )
    expect_match(w, "Removed.*containing missing")
})

test_that("plot.mnirs moxy.perfpro works", {
    df <- read_mnirs(
        file_path = example_mnirs("moxy_ramp"),
        nirs_channels = c(smo2_left = "SmO2 Live", smo2_right = "SmO2 Live(2)"),
        time_channel = c(time = "hh:mm:ss"),
        verbose = FALSE
    )

    ## visual check
    plot <- plot(df, na.omit = TRUE, time_labels = TRUE, n.breaks = 8)
    expect_s3_class(plot, "ggplot")
})

#' Shift data range
#'
#' Move the range of data channels in a data frame up or down, while preserving
#' the absolute amplitude/dynamic range of each channel, and the relative
#' scaling across channels. e.g. shift the minimum data value to zero for all
#' positive values, or shift the mean of the first time span in a recording
#' to zero.
#'
#' @param to A numeric value in units of `nirs_channels` to which the data
#'   channels will be shifted, e.g. shift the minimum value to zero.
#' @param by A numeric value in units of `nirs_channels` by which the data
#'   channels will be shifted, e.g. shift all values up by 10 units.
#' @param position Indicates where the reference values will be shifted from.
#'   \describe{
#'      \item{`"min"`}{(The *default*) will shift the minimum value(s) `to`
#'      or `by` the specified value.}
#'      \item{`"max"`}{Will shift the maximum value(s) `to` or `by` the
#'      specified values.}
#'      \item{`"first"`}{Will shift first value(s) `to` or `by` the specified
#'      values.}
#'   }
#' @inheritParams validate_mnirs
#' @inheritParams replace_mnirs
#' @inheritParams rescale_mnirs
#'
#' @details
#' `nirs_channels = list()` can be used to group data channels (column names)
#'   to preserve absolute or relative scaling.
#'
#' - Channels grouped together in a vector (e.g. `list(c("A", "B"))`) will be
#'   shifted to a common value, and the relative scaling within that group
#'   will be preserved.
#'
#' - Channels in separate list vectors (e.g. `list("A", "B")`) will be
#'   shifted independently, and relative scaling between groups will be lost.
#'
#' - A single vector of channel names (e.g. `c("A", "B")`) will group
#'   channels together.
#'
#' - Channels (columns) in `data` not explicitly defined in `nirs_channels`
#'   will be passed through untouched to the output data frame.
#'
#' `nirs_channels` and `time_channel` can be retrieved automatically from
#'   `data` of class *"mnirs"* which has been processed with `{mnirs}`,
#'   if not defined explicitly. This will default to returning all
#'   `nirs_channels` grouped together, and should be defined explicitly
#'   for other grouping arrangements.
#'
#' Only one of either `to` or `by` and one of either `width` or `span` should
#'   be defined. If both of either pairing are defined, `to` will be preferred
#'   over `by`, and `width` will be preferred over `span`.
#'
#' @returns
#' A [tibble][tibble::tibble-package] of class *"mnirs"* with metadata
#'   available with `attributes()`.
#'
#' @examples
#' ## read example data
#' data <- read_mnirs(
#'     file_path = example_mnirs("moxy_ramp"),
#'     nirs_channels = c(smo2_left = "SmO2 Live",
#'                       smo2_right = "SmO2 Live(2)"),
#'     time_channel = c(time = "hh:mm:ss"),
#'     verbose = FALSE
#' ) |>
#'     shift_mnirs(        ## un-grouped nirs channels to shift separately 
#'         nirs_channels = list(smo2_left, smo2_right), 
#'         to = 0,         ## NIRS values will be shifted to zero
#'         span = 120,     ## shift the *first* 120 sec of data to zero
#'         position = "first"
#'     )
#'
#' data
#'
#' \donttest{
#'     if (requireNamespace("ggplot2", quietly = TRUE)) {
#'         plot(data, time_labels = TRUE) +
#'             ggplot2::geom_hline(yintercept = 0, linetype = "dotted")
#'     }
#' }  
#'
#' @export
shift_mnirs <- function(
    data,
    nirs_channels = list(),
    time_channel = NULL,
    to = NULL,
    by = NULL,
    width = NULL,
    span = NULL,
    position = c("min", "max", "first"),
    verbose = TRUE
) {
    ## validation =============================================
    ## do nothing condition
    if (is.null(c(to, by))) {
        cli_abort(c(
            "x" = "Shift value undefined",
            "i" = "One of {.arg to} or {.arg by} must be defined."
        ))
    }

    validate_mnirs_data(data)
    metadata <- attributes(data)
    if (missing(verbose)) {
        verbose <- getOption("mnirs.verbose", default = TRUE)
    }
    nirs_channels <- validate_nirs_channels(
        enquo(nirs_channels), data, verbose, as_list = TRUE
    )
    time_channel <- validate_time_channel(enquo(time_channel), data)
    validate_numeric(to, 1, msg1 = "one-element")
    validate_numeric(by, 1, msg1 = "one-element")
    if (!is.null(to) && !is.null(by)) {
        by <- NULL
        if (verbose) {
            cli_inform(c("i" = "{.arg to} = {.val {to}} overrides {.arg by}."))
        }
    }

    if (
        verbose &&
            is.null(attr(data, "nirs_channels")) &&
            !is.list(nirs_channels)
    ) {
        cli_inform(c(
            "!" = "{.fn shift_mnirs} accepts {.arg nirs_channels} = \\
            {col_blue('list()')} for channel grouping. See `?shift_mnirs`."
        ))
    }
    nirs_listed <- make_list(nirs_channels)
    nirs_unlisted <- unlist(nirs_listed, use.names = FALSE)

    ## Metadata =================================
    metadata$nirs_channels <- unique(nirs_unlisted)
    metadata$time_channel <- time_channel

    ## shift_by ====================================================
    ## shift_by does not require grouping or calculating positions
    if (!is.null(by) && is.null(to)) {
        data[nirs_unlisted] <- data[nirs_unlisted] + by
        return(create_mnirs_data(data, metadata))
    }

    ## calculate shift_to values ====================================
    ## validate
    position <- match.arg(position)
    validate_width_span(width, span, verbose, "for `shift_mnirs()`.")
    time_vec <- data[[time_channel]]

    if (position == "first") {
        ## for span, take data <= first time_channel value + span, assuming sorted
        width <- width %||% rev(which(time_vec <= (time_vec[1L] + span)))[1L]
        ## drop = FALSE to avoid reducing to vector with one `nirs_unlisted`
        shift_values <- colMeans(
            data[seq_len(width), nirs_unlisted, drop = FALSE],
            na.rm = TRUE
        )
    } else if (position %in% c("min", "max")) {
        ## find local windows within width/span centred around idx
        window_idx <- compute_local_windows(
            t = time_vec, width = width, span = span,
        )
        shift_fun <- match.fun(position)
        ## compute min or max along local means
        ## return named vec of min/max for each nirs_channel
        shift_values <- vapply(data[nirs_unlisted], \(.x) {
            shift_fun(
                compute_local_fun(.x, window_idx, mean, na.rm = TRUE), 
                na.rm = TRUE
            )
        }, numeric(1))
    }

    ## apply shifts ==============================================
    ## find the value to shift to per nirs_channel group
    group_shift_values <- vapply(nirs_listed, \(.cols) {
        switch(
            position,
            min = min(shift_values[.cols]),
            max = max(shift_values[.cols]),
            first = mean(shift_values[.cols])
        )
    }, numeric(1))

    ## expand grouped shift values for each nirs_channel
    channel_shifts <- rep(group_shift_values, lengths(nirs_listed))
    ## move the shift_value to the `to` value for each nirs_channel
    data[nirs_unlisted] <- lapply(seq_along(nirs_unlisted), \(.i) {
        data[[nirs_unlisted[.i]]] - channel_shifts[.i] + to
    })

    return(create_mnirs_data(data, metadata))
}

#' Methods for mnirs objects
#'
#' Generic methods for objects of class `"mnirs"`.
#'
#' @param x Object of class `"mnirs"`.
#' @param ... Additional arguments passed to [print()].
#'
#' @returns
#' \item{`print`}{Returns `x` without class attributes.}
#'
#' @examples
#' x <- read_mnirs(
#'     example_mnirs("train.red"),
#'     nirs_channels = c(smo2 = "SmO2"),
#'     time_channel = c(time = "Timestamp (seconds passed)"),
#'     verbose = FALSE
#' ) |>
#'     resample_mnirs(method = "linear", verbose = FALSE) |>
#'     extract_intervals(
#'         start = by_time(2452, 3168),
#'         span = c(-60, 120),
#'         verbose = FALSE
#'     )
#' 
#' print(x)
#'
#' @export
print.mnirs <- function(x, ...) {
    class(x) <- setdiff(class(x), "mnirs")
    print(x, ...)
    invisible(x)
}

#' Validate `{mnirs}` parameters
#'
#' Resolve and validate *{mnirs}* metadata and perform basic data quality
#' checks.
#'
#' @param data A data frame of class *"mnirs"* containing time series data and
#'   metadata.
#'
#' @param nirs_channels A character vector giving the names of mNIRS columns to
#'   operate on. Must match column names in `data` exactly.
#'   - If `NULL` (default), the `nirs_channels` metadata attribute of `data` is
#'     used.
#'
#' @param time_channel A character string naming the time or sample column. 
#'   Must match a column name in `data` exactly.
#'   - If `NULL` (default), the `time_channel` metadata attribute of `data` is
#'     used.
#'
#' @param event_channel A character string naming the event/lap column. Must 
#'   match a column name in `data` exactly.
#'   - If `NULL` (default), the `event_channel` metadata attribute of `data` is
#'     used.
#'
#' @param as_list Logical. Default is `FALSE`. If `nirs_channels` is specified
#'   as a list, it will be coerced to a flat character vector and an
#'   information message is displayed (when `verbose = TRUE`). If `TRUE`,  
#'   `nirs_channels` is returned as-is, i.e. as a list for callers which 
#'   require it. 
#'
#' @param required Logical. Default is `TRUE`. `event_channel` must be
#'   present or detected in metadata. If `FALSE`, `event_channel` may be `NULL`.
#'
#' @param x A numeric vector.
#'
#' @param sample_rate A numeric sample rate in Hz.
#'   - If `NULL` (default), the `sample_rate` metadata attribute of `data` will
#'     be used if detected, or the sample rate will be estimated from
#'     `time_channel`.
#'
#' @param elements An integer. Default is `Inf`. The number of numeric elements
#'   expected in `x`.
#'
#' @param range A two-element numeric vector giving the valid range for `x`.
#'
#' @param inclusive A character vector specifying which boundaries of `range`
#'   are included. Any of `"left"`, `"right"` (default is both). Use `FALSE` to
#'   exclude both endpoints.
#'
#' @param integer Logical. Default is `FALSE`. If `TRUE`, validate `x` as
#'   integer-like values using [rlang::is_integerish()]. Otherwise tested as a
#'   numeric value.
#' 
#' @param allow_na Logical. Default is `FALSE`. If `TRUE`, allows pass through
#'   of `NA` to the returned numeric/integer vector.
#'
#' @param msg1,msg2 A character string appended to the [cli::cli_abort()]
#'   message when numeric validation fails.
#' 
#' @inheritParams read_mnirs
#'
#' @details
#' `validate_mnirs()` is an internal documentation topic for a set of
#' validators used throughout the package. These validators:
#'
#' - Prefer explicit user-supplied arguments.
#' - Fall back to *"mnirs"* metadata attributes when available.
#' - Fail fast with informative [cli::cli_abort()] messages when values are
#'   missing or invalid.
#'
#' @returns Returns the validated object (e.g. a resolved `time_channel`
#'   string), or invisibly returns `NULL` for successful validations. On
#'   failure, an error is thrown via [cli::cli_abort()].
#'
#' @name validate_mnirs
#' @keywords internal
NULL

#' validate_numeric abort message construction
#' @keywords internal
abort_validation <- function(name, integer = FALSE, msg1 = "", msg2 = "") {
    type <- if (integer) "integer" else "numeric"

    cli_abort(c(
        "x" = paste0(
            "{.arg {name}} must be a valid ",
            msg1,
            " {.cls {type}} ",
            msg2
        )
    ))
}


#' @rdname validate_mnirs
validate_numeric <- function(
    x,
    elements = Inf,
    range = NULL,
    inclusive = c("left", "right"),
    integer = FALSE,
    allow_na = FALSE,
    msg1 = "",
    msg2 = ""
) {
    ## pass through NULL
    if (is.null(x)) {
        return(invisible(NULL))
    }

    name <- substitute(x)

    ## cheap early type check
    if (!is.numeric(x)) {
        abort_validation(name, integer, msg1, msg2)
    }

    ## valid elements length — skip NA scan when allow_na = TRUE
    if (!allow_na) {
        valid <- !is.na(x)
        n_valid <- sum(valid)
        if (n_valid == 0L) {
            abort_validation(name, integer, msg1, msg2)
        }
    } else {
        n_valid <- length(x)
    }

    ## elements check
    if (is.finite(elements) && n_valid != elements) {
        abort_validation(name, integer, msg1, msg2)
    }

    ## subset once for range/integer checks
    needs_subset <- !is.null(range) || integer
    if (needs_subset && !allow_na) {
        x_valid <- if (n_valid < length(x)) x[valid] else x
    }

    ## range check
    if (!is.null(range) && !all(within(x_valid, range, inclusive))) {
        abort_validation(name, integer, msg1, msg2)
    }
    ## expensive integer check
    if (integer && !rlang::is_integerish(x_valid)) {
        abort_validation(name, integer, msg1, msg2)
    }

    return(invisible())
}


#' @rdname validate_mnirs
validate_mnirs_data <- function(data, ncol = 2L) {
    ## validate is a data frame with at least two columns
    if (!is.data.frame(data) || length(data) < ncol) {
        cli_abort(c(
            "x" = "{.arg data} must be a data frame with at least \\
            {.val {ncol}} column{?s}."
        ))
    }

    return(invisible())
}

#' Parse channel expressions for NSE
#'
#' Converts quosures to character vectors, handling bare symbols, character
#' strings, lists, and tidyselect expressions.
#'
#' @param channel A quosure from `rlang::enquo()`.
#' @param data A data frame for tidyselect context.
#' @param env Environment for symbol evaluation.
#'
#' @returns A character vector, list of character vectors, or `NULL`.
#'
#' @keywords internal
parse_channel_name <- function(channel, data, env = rlang::caller_env()) {
    if (rlang::quo_is_null(channel)) {
        return(NULL)
    }

    channel_raw <- rlang::quo_get_expr(channel)

    ## already-evaluated list or character
    if (is.list(channel_raw) || is.character(channel_raw)) {
        return(channel_raw)
    }

    ## bare symbol: check if column name, otherwise evaluate
    if (rlang::quo_is_symbol(channel)) {
        sym_name <- rlang::as_name(channel)
        if (sym_name %in% names(data)) {
            return(sym_name)
        }

        ## external object: evaluate and return directly
        result <- tryCatch(
            rlang::eval_tidy(channel, env = env),
            error = \(e) NULL
        )
        if (is.list(result) || is.character(result)) {
            return(result)
        }
        return(sym_name)
    }

    ## list() call: recurse on each element
    if (rlang::is_call(channel_raw, "list")) {
        result <- lapply(rlang::call_args(channel_raw), \(.arg) {
            parse_channel_name(rlang::new_quosure(.arg, env = env), data, env)
        })
        return(unname(result))
    }

    ## evaluate: tidyselect first, then fallback to direct evaluation
    ## handles c(), tidyselect helpers, symbols, and external objects
    tryCatch(
        unname(names(tidyselect::eval_select(channel, data))),
        error = \(e) {
            result <- rlang::eval_tidy(channel, env = env)
            if (is.list(result) || is.character(result)) result else NULL
        }
    )
}


#' @rdname validate_mnirs
validate_nirs_channels <- function(
    nirs_channels,
    data,
    verbose = FALSE, ## only for functions requiring list()
    as_list = FALSE,
    env = rlang::caller_env()
) {
    ## parse tidy eval input
    if (rlang::is_quosure(nirs_channels)) {
        nirs_channels <- parse_channel_name(nirs_channels, data, env)
    }
    nirs_unlisted <- unlist(nirs_channels)

    ## if not defined, check metadata
    if (is.null(nirs_unlisted) || length(nirs_unlisted) == 0) {
        nirs_channels <- attr(data, "nirs_channels") ## should be vector
        nirs_unlisted <- nirs_channels
        if (verbose && as_list && !is.null(nirs_unlisted)) {
            cli_inform(c(
                "i" = "{.arg nirs_channels} = \\
                {col_blue({deparse(nirs_unlisted)})} \\
                grouped together from metadata."
            ))
        }
    }

    ## if still not defined, return error
    if (is.null(nirs_unlisted)) {
        cli_abort(c(
            "x" = "{.arg nirs_channels} not detected in metadata.",
            "i" = "Check your data attributes or define \\
            {.arg nirs_channels} explicitly."
        ))
    }

    ## validate exists in data
    if (!is.character(nirs_unlisted) || !all(nirs_unlisted %in% names(data))) {
        cli_abort(c(
            "x" = "{.arg nirs_channels} not detected in {.arg data}.",
            "i" = "Channel names are case-sensitive and must match exactly."
        ))
    }

    ## validate is numeric and has >=2 valid values
    invalid_channels <- vapply(data[nirs_unlisted], \(.x) {
        !is.numeric(.x) || sum(is.finite(.x)) < 2
    }, logical(1))

    if (sum(invalid_channels) > 0) {
        cli_abort(c(
            "x" = "{.arg nirs_channels} must contain valid {.cls numeric} data."
        ))
    }

    ## preserve list grouping for callers that need it
    if (as_list) {
        return(nirs_channels)
    }

    ## default: coerce to flat vector
    if (verbose && is.list(nirs_channels)) {
        cli_inform(c(
            "i" = "{.arg nirs_channels} = \\
            {col_blue({deparse(nirs_unlisted)})} passed through unlisted."
        ))
    }

    ## returns explicitly grouped nirs_channels
    ## or nirs_unlisted if retrieved from metadata
    return(nirs_unlisted)
}


#' @rdname validate_mnirs
validate_time_channel <- function(
    time_channel,
    data,
    env = rlang::caller_env()
) {
    ## parse tidy eval input
    if (rlang::is_quosure(time_channel)) {
        time_channel <- parse_channel_name(time_channel, data, env)
    }

    ## if not defined, check metadata
    if (is.null(time_channel)) {
        time_channel <- attr(data, "time_channel")
    }

    ## if still not defined, return error
    if (is.null(time_channel)) {
        cli_abort(c(
            "x" = "{.arg time_channel} not detected in metadata.",
            "i" = "Check your data attributes or define \\
            {.arg time_channel} explicitly."
        ))
    }

    ## validate exists in data
    if (!is.character(time_channel) || !time_channel %in% names(data)) {
        cli_abort(c(
            "x" = "{.arg time_channel} not detected in {.arg data}.",
            "i" = "Channel names are case-sensitive and must match exactly."
        ))
    }

    ## validate is numeric and has >=2 valid values
    if (
        !is.numeric(data[[time_channel]]) ||
            sum(is.finite(data[[time_channel]])) < 2
    ) {
        cli_abort(c(
            "x" = "{.arg time_channel} must contain valid {.cls numeric} data."
        ))
    }

    return(time_channel)
}


#' @rdname validate_mnirs
validate_event_channel <- function(
    event_channel,
    data,
    required = TRUE,
    env = rlang::caller_env()
) {
    ## parse tidy eval input
    if (rlang::is_quosure(event_channel)) {
        event_channel <- parse_channel_name(event_channel, data, env)
    }
    ## if not defined, check metadata
    if (is.null(event_channel)) {
        event_channel <- attr(data, "event_channel")
    }

    ## if still not defined, return error
    if (is.null(event_channel) && required) {
        cli_abort(c(
            "x" = "{.arg event_channel} not detected in metadata.",
            "i" = "Check your data attributes or define {.arg event_channel} \\
            explicitly."
        ))
    } else if (is.null(event_channel) && !required) {
        ## return event_channel = NULL if not required
        return(event_channel)
    }

    ## validate exists in data
    if (!is.character(event_channel) || !event_channel %in% names(data)) {
        cli_abort(c(
            "x" = "{.arg event_channel} not detected in {.arg data}.",
            "i" = "Channel names are case-sensitive and must match exactly."
        ))
    }

    ## validate column type: must be character or integerish
    col <- data[[event_channel]]
    if (!is.character(col) && !rlang::is_integerish(col)) {
        cli_abort(c(
            "x" = "{.arg event_channel} must contain valid {.cls character} \\
            event labels or {.cls integer} lap numbers."
        ))
    }

    ## check for empty column — character columns also check for empty strings
    valid_values <- if (is.character(col)) {
        !is.na(col) & nzchar(col)
    } else {
        !is.na(col)
    }
    if (sum(valid_values) == 0) {
        cli_abort(c(
            "x" = "{.arg event_channel} must contain valid {.cls character} \\
            event labels or {.cls integer} lap numbers."
        ))
    }

    return(event_channel)
}


#' @rdname validate_mnirs
estimate_sample_rate <- function(x) {
    ## estimate samples per second
    sample_rate_raw <- 1 / median(diff(x), na.rm = TRUE)
    
    if (!is.finite(sample_rate_raw) || sample_rate_raw == 0) {
        cli_abort(c(
            "x" = "Unable to estimate {.arg sample_rate}.",
            "i" = "Check that {.arg time_channel} values are consistent.",
            "i" = "Set {.arg sample_rate} = {.cls numeric}."
        ))
    }

    pretty_vals <- c(
        0.25, 0.5, 1, 2, 3, 4, 5, 10, 15, 20, 25, 30, 50, 60, 75, 100
    )
    rounded <- pretty_vals[which.min(abs(pretty_vals - sample_rate_raw))]
    return(rounded)
}


#' @rdname validate_mnirs
validate_sample_rate <- function(
    data,
    time_channel,
    sample_rate,
    verbose = TRUE
) {
    ## if not defined, check metadata
    if (is.null(sample_rate)) {
        sample_rate <- attr(data, "sample_rate")
    }

    ## estimate sample_rate from time_channel
    ## time_channel MUST be validated before this
    time_vec <- as.numeric(data[[time_channel]])
    ## will error on unable to estimate sample_rate
    sample_rate_est <- estimate_sample_rate(time_vec)

    ## if still not defined, use estimated sample_rate
    if (is.null(sample_rate)) {
        sample_rate <- sample_rate_est
        if (verbose) {
            cli_inform(c(
                "!" = "Estimated {.arg sample_rate} = {.val {sample_rate}} Hz.",
                "i" = "Define {.arg sample_rate} explicitly to override."
            ))
        }
    }

    ## validate has one numeric value
    validate_numeric(
        sample_rate, 1, c(0, Inf), FALSE, msg1 = "one-element positive"
    )

    ## if provided sample rate seems off and time_channel doesn't appear
    ## to be integer values, report warning
    if (
        verbose && !isTRUE(
            all.equal(1, sample_rate_est, tolerance = 0.001, scale = 1)
        ) & !isTRUE(
            all.equal(sample_rate_est, sample_rate, tolerance = 0.5, scale = 1)
        )
    ) {
        cli_warn(c(
            "!" = "`sample_rate = {.val {sample_rate}}` appears to be \\
            inconsistent with {.arg time_channel}. Estimated \\
            `sample_rate = {.val {sample_rate_est}}`.",
            "i" = "Check that your sample rate and {.arg time_channel} \\
            values are consistent."
        ))
    }

    return(sample_rate)
}

#' @rdname validate_mnirs
validate_width_span <- function(
    width = NULL,
    span = NULL,
    verbose = TRUE,
    msg = ""
) {
    if (is.null(c(width, span))) {
        cli_abort(c(
            "x" = "Window size undefined",
            "i" = paste(
                "One of {.arg width} or {.arg span} must be defined",
                msg
            )
        ))
    }
    validate_numeric(
        width, 1, c(1, Inf), integer = TRUE, msg1 = "one-element positive"
    )
    validate_numeric(span, 1, c(0, Inf), msg1 = "one-element positive")
    if (verbose && !is.null(width) && !is.null(span)) {
        cli_inform(c(
            "i" = "{.arg width} = {.val {width}} overrides {.arg span}."
        ))
    }
}


#' @rdname validate_mnirs
validate_x_t <- function(x, t, allow_na = FALSE) {
    ## exclude NULL by defaulting to allow_na character
    x <- x %||% character()
    t <- t %||% character()
    validate_numeric(x, allow_na = allow_na)
    validate_numeric(t, allow_na = allow_na)
    if (length(x) != length(t)) {
        cli_abort(c(
            "x" = "{.arg x} and {.arg t} must be {.cls numeric} vectors \\
            of equal length."
        ))
    }
}


#' Validate if an item is a list
#' @keywords internal
make_list <- function(x) {
    if (is.list(x)) {
        return(x)
    } else {
        return(list(x))
    }
}


#' Detect if numeric values fall within range of a vector
#'
#' Vectorised check for `x %in% vec`, inclusive or exclusive of left and right
#' boundary values, specified independently.
#'
#' @param x A numeric vector.
#' @param vec A numeric vector from which `left` and `right` boundary values
#'   for `x` will be taken.
#' @param inclusive A character vector to specify which of `left` and/or
#'   `right` boundary values should be included in the range, or both (the
#'   default), or excluded if `FALSE`.
#'
#' @details
#' `inclusive = FALSE` can be used to test for positive non-zero values:
#'   `within(x, c(0, Inf), inclusive = FALSE)`.
#'
#' @returns A logical vector the same length as `x`.
#'
#' @seealso [dplyr::between()]
#'
#' @keywords internal
within <- function(x, vec, inclusive = c("left", "right")) {
    if (!is.numeric(x)) {
        abort_validation(substitute(x))
    }
    if (!is.numeric(vec)) {
        abort_validation(substitute(vec))
    }
    inclusive <- match.arg(
        as.character(inclusive), ## force FALSE to character
        choices = c("left", "right", "FALSE"),
        several.ok = TRUE
    )

    ## extract bounds from vec
    left <- min(vec, na.rm = TRUE)
    right <- max(vec, na.rm = TRUE)

    if ("FALSE" %in% inclusive) {
        return(x > left & x < right)
    }

    left_op <- if ("left" %in% inclusive) `>=` else `>`
    right_op <- if ("right" %in% inclusive) `<=` else `<`

    return(left_op(x, left) & right_op(x, right))
}

## helper to create test data with metadata
create_test_data <- function(
    time_max = 10,
    sample_rate = 10,
    add_metadata = TRUE
) {
    time <- seq(0, time_max, by = 1 / sample_rate)
    nrow <- length(time)
    data <- tibble(
        time = time,
        nirs1 = rnorm(nrow, 50, 5),
        nirs2 = rnorm(nrow, 60, 5),
        nirs3 = rnorm(nrow, 80, 5),
        event = c(1, rep(NA, nrow - 2), 2),
    )
    class(data) <- c("mnirs", class(data))

    if (add_metadata) {
        attr(data, "time_channel") <- "time"
        attr(data, "nirs_channels") <- c("nirs1", "nirs2")
        attr(data, "event_channel") <- "event"
        attr(data, "sample_rate") <- sample_rate
    }

    return(data)
}


## validate_numeric() ==================================
test_that("validate_numeric accepts valid numeric values", {
    expect_silent(validate_numeric(5))
    expect_silent(validate_numeric(c(1, 2, 3)))
    expect_silent(validate_numeric(1.5))
})

test_that("validate_numeric checks element count", {
    x <- c(1, 2)
    expect_error(validate_numeric(x, elements = 1), "numeric")
    expect_silent(validate_numeric(5, elements = 1))
    expect_silent(validate_numeric(c(1, 2, 3), elements = 3))
    ## fewer elements than specified fails
    expect_error(validate_numeric(1, elements = 2), "numeric")
})

test_that("validate_numeric checks range with inclusive bounds", {
    expect_silent(validate_numeric(5, range = c(0, 10)))
    expect_silent(validate_numeric(0, range = c(0, 10)))
    expect_silent(validate_numeric(10, range = c(0, 10)))
    expect_error(validate_numeric(-1, range = c(0, 10)), "numeric")
    expect_error(validate_numeric(11, range = c(0, 10)), "numeric")
})

test_that("validate_numeric checks range with exclusive bounds", {
    expect_silent(validate_numeric(5, range = c(0, 10), inclusive = FALSE))
    expect_error(
        validate_numeric(0, range = c(0, 10), inclusive = FALSE),
        "numeric"
    )
    expect_error(
        validate_numeric(10, range = c(0, 10), inclusive = FALSE),
        "numeric"
    )
})

test_that("validate_numeric rejects non-numeric input", {
    expect_error(validate_numeric("text"), "numeric")
    expect_error(validate_numeric(TRUE), "numeric")
    expect_error(validate_numeric(list(1, 2))) ## "not implemented for type `list`"
})

test_that("validate_numeric handles NA, NaN, Inf", {
    expect_silent(validate_numeric(c(1, NA)))
    expect_error(validate_numeric(NA), "numeric")
    expect_error(validate_numeric(NaN), "numeric")
    expect_silent(validate_numeric(Inf))
    expect_error(validate_numeric(Inf, range = c(0, 1)), "numeric")
    expect_silent(validate_numeric(c(1, NA, 3), elements = 2))
})

test_that("validate_numeric handles NULL", {
    expect_silent(validate_numeric(NULL))
})

## validate_mnirs_data() ========================================
test_that("validate_mnirs_data() accepts valid data frames", {
    data <- create_test_data()
    expect_silent(validate_mnirs_data(data))
})

test_that("validate_mnirs_data() rejects non-data frames", {
    expect_error(
        validate_mnirs_data(list(a = 1, b = 2)),
        "must be a data frame"
    )
    expect_error(validate_mnirs_data(c(1, 2, 3)), "must be a data frame")
})

test_that("validate_mnirs_data() rejects data frames with < 2 columns", {
    expect_error(validate_mnirs_data(data.frame(x = 1:10)), "at least")
})


## parse_channel_name() ============================================
test_that("parse_channel_name() returns NULL for NULL input", {
    data <- data.frame(a = 1, b = 2)
    result <- parse_channel_name(rlang::quo(NULL), data)
    expect_null(result)
})

test_that("parse_channel_name() handles already-evaluated character", {
    data <- data.frame(a = 1, b = 2)
    channel <- "a"
    result <- parse_channel_name(channel = rlang::enquo(channel), data)
    expect_equal(result, "a")
})

test_that("parse_channel_name() handles already-evaluated character vector", {
    data <- data.frame(a = 1, b = 2, c = 3)
    ch <- c("a", "b")
    result <- parse_channel_name(rlang::enquo(ch), data)
    expect_equal(result, c("a", "b"))
})

test_that("parse_channel_name() handles already-evaluated list", {
    data <- data.frame(a = 1, b = 2, c = 3)
    ch <- list(c("a", "b"), "c")
    result <- parse_channel_name(rlang::enquo(ch), data)
    expect_equal(result, list(c("a", "b"), "c"))
})

test_that("parse_channel_name() handles bare symbol matching column", {
    data <- data.frame(smo2 = 1, thb = 2)
    env <- rlang::current_env()
    expr <- rlang::quo(smo2)
    result <- parse_channel_name(expr, data, env = env)
    expect_equal(result, "smo2")
})

test_that("parse_channel_name() handles bare symbol as external object", {
    data <- data.frame(a = 1, b = 2, c = 3)
    channels <- c("a", "b")
    result <- parse_channel_name(rlang::enquo(channels), data)
    expect_equal(result, c("a", "b"))
})

test_that("parse_channel_name() handles external list object", {
    data <- data.frame(a = 1, b = 2, c = 3)
    channels <- list(c("a", "b"), "c")
    result <- parse_channel_name(rlang::enquo(channels), data)
    expect_equal(result, list(c("a", "b"), "c"))
})

test_that("parse_channel_name() handles c() with bare symbols", {
    data <- data.frame(smo2 = 1, thb = 2, hhr = 3)
    env <- rlang::current_env()
    expr <- rlang::quo(c(smo2, thb))
    result <- parse_channel_name(expr, data, env = env)
    expect_equal(result, c("smo2", "thb"))
})

test_that("parse_channel_name() handles list() with bare symbols", {
    data <- data.frame(smo2 = 1, thb = 2, hhr = 3)
    env <- rlang::current_env()
    expr <- rlang::quo(list(c(smo2, thb), hhr))
    result <- parse_channel_name(expr, data, env = env)
    expect_equal(result, list(c("smo2", "thb"), "hhr"))
})

test_that("parse_channel_name() handles list() with mixed input", {
    data <- data.frame(a = 1, b = 2, c = 3)
    env <- rlang::current_env()
    expr <- rlang::quo(list(c("a", "b"), c))
    result <- parse_channel_name(expr, data, env = env)
    expect_equal(result, list(c("a", "b"), "c"))
})

test_that("parse_channel_name() handles tidyselect starts_with()", {
    data <- data.frame(smo2_left = 1, smo2_right = 2, thb = 3)
    env <- rlang::current_env()
    expr <- rlang::quo(tidyselect::starts_with("smo2"))
    result <- parse_channel_name(expr, data, env = env)
    expect_equal(result, c("smo2_left", "smo2_right"))
})

test_that("parse_channel_name() handles tidyselect ends_with()", {
    data <- data.frame(smo2_left = 1, thb_left = 2, smo2_right = 3)
    env <- rlang::current_env()
    expr <- rlang::quo(tidyselect::ends_with("left"))
    result <- parse_channel_name(expr, data, env = env)
    expect_equal(result, c("smo2_left", "thb_left"))
})

test_that("parse_channel_name() handles tidyselect matches()", {
    data <- data.frame(left_smo2 = 1, smo2_right = 2, thb = 3)
    env <- rlang::current_env()
    expr <- rlang::quo(tidyselect::matches("smo2"))
    result <- parse_channel_name(expr, data, env = env)
    expect_equal(result, c("left_smo2", "smo2_right"))
})

test_that("parse_channel_name() handles tidyselect in list()", {
    data <- data.frame(smo2_a = 1, smo2_b = 2, thb = 3)
    env <- rlang::current_env()
    expr <- rlang::quo(list(tidyselect::starts_with("smo2"), thb))
    result <- parse_channel_name(expr, data, env = env)
    expect_equal(result, list(c("smo2_a", "smo2_b"), "thb"))
})

test_that("parse_channel_name() returns non-existent colname symbol", {
    data <- data.frame(a = 1, b = 2)
    env <- rlang::new_environment()
    channel <- rlang::new_quosure(rlang::sym("nonexistent"), env = env)
    result <- parse_channel_name(channel, data, env = env)
    ## should pass through nonexistent colname to validate_nirs_channels
    expect_equal(result, "nonexistent")
})

test_that("parse_channel_name returns NULL on logical channel", {
    data <- data.frame(a = 1, b = 2)
    env <- rlang::new_environment()
    expect_null(parse_channel_name(rlang::quo(TRUE), data, env))
    list_null <- parse_channel_name(rlang::quo(list(TRUE)), data, env)
    expect_true(is.list(list_null))
    expect_null(unlist(list_null))
})



## validate_nirs_channels() ========================================
test_that("validate_nirs_channels() uses metadata when NULL", {
    data <- create_test_data()
    result <- validate_nirs_channels(NULL, data, verbose = FALSE)
    expect_equal(result, c("nirs1", "nirs2"))
})

test_that("validate_nirs_channels() uses explicit channels when provided", {
    data <- create_test_data()
    result <- validate_nirs_channels("nirs1", data)
    expect_equal(result, "nirs1")
})

test_that("validate_nirs_channels() works with nirs_channels = list()", {
    data <- create_test_data()
    nirs_vec <- c("nirs1", "nirs2")
    result <- validate_nirs_channels(nirs_vec, data)
    expect_equal(result, nirs_vec)

    result <- validate_nirs_channels(enquo(nirs_vec), data)
    expect_equal(result, nirs_vec)

    attr(data, "nirs_channels") <- nirs_vec
    expect_message(
        result <- validate_nirs_channels(
            NULL, data, verbose = TRUE, as_list = TRUE
        ),
        "`nirs_channels`.*grouped"
    )
    expect_equal(result, nirs_vec)

    nirs_list <- list(c("nirs1", "nirs2"), "nirs3")
    result <- validate_nirs_channels(nirs_list, data, as_list = TRUE)
    expect_equal(result, nirs_list)

    result <- validate_nirs_channels(enquo(nirs_list), data, as_list = TRUE)
    expect_equal(result, nirs_list)
})

test_that("validate_nirs_channels() errors when not in metadata or provided", {
    data <- create_test_data(add_metadata = FALSE)
    expect_error(validate_nirs_channels(NULL, data), "not detected in metadata")
})

test_that("validate_nirs_channels() errors when columns don't exist", {
    data <- create_test_data()
    expect_error(validate_nirs_channels("nonexistent", data), "match exactly")
})

test_that("validate_nirs_channels() errors when col symbols don't exist", {
    data <- data.frame(a = 1, b = 2)
    env <- rlang::new_environment()
    channel <- rlang::new_quosure(rlang::sym("nonexistent"), env = env)
    expect_error(validate_nirs_channels(channel, data), "match exactly")
})

test_that("validate_nirs_channels() errors for non-numeric channels", {
    data <- create_test_data()
    data$nirs1 <- as.character(data$nirs1)
    expect_error(
        validate_nirs_channels(c("nirs1", "nirs2"), data),
        "must contain valid.*numeric"
    )
})

test_that("validate_nirs_channels() errors when < 2 valid values", {
    data <- create_test_data()
    data$nirs1 <- c(1, rep(NA, nrow(data) - 1))
    expect_error(
        validate_nirs_channels(c("nirs1", "nirs2"), data),
        "must contain valid.*numeric"
    )
})

test_that("validate_nirs_channels() informs when list coerced to vector", {
    data <- create_test_data()
    nirs_list <- list(c("nirs1", "nirs2"), "nirs3")
    expect_message(
        result <- validate_nirs_channels(
            nirs_list, data, verbose = TRUE, as_list = FALSE
        ),
        "`nirs_channels`.*unlisted"
    )
    expect_equal(result, c("nirs1", "nirs2", "nirs3"))
})


## validate_time_channel() ========================================
test_that("validate_time_channel() uses metadata when NULL", {
    data <- create_test_data()
    result <- validate_time_channel(NULL, data)
    expect_equal(result, "time")
})

test_that("validate_time_channel() uses explicit channel when provided", {
    data <- create_test_data()
    data$time_new <- data$time
    result <- validate_time_channel("time_new", data)
    expect_equal(result, "time_new")
})

test_that("validate_time_channel() errors when not in metadata or provided", {
    data <- create_test_data(add_metadata = FALSE)
    expect_error(validate_time_channel(NULL, data), "not detected in metadata")
})

test_that("validate_time_channel() errors when column doesn't exist", {
    data <- create_test_data()
    expect_error(validate_time_channel("nonexistent", data), "match exactly")
})

test_that("validate_time_channel() errors for non-numeric channel", {
    data <- create_test_data()
    data$time <- as.character(data$time)
    expect_error(validate_time_channel("time", data), "must contain valid.*numeric")
})

test_that("validate_time_channel() errors when < 2 valid values", {
    data <- create_test_data()
    data$time <- c(1, rep(NA, nrow(data) - 1))
    expect_error(validate_time_channel("time", data), "must contain valid.*numeric")
})


## validate_event_channel() ========================================
test_that("validate_event_channel() uses metadata when NULL", {
    data <- create_test_data()
    result <- validate_event_channel(NULL, data)
    expect_equal(result, "event")
})

test_that("validate_event_channel() uses explicit channel when provided", {
    data <- create_test_data()
    data$lap <- data$event
    result <- validate_event_channel("lap", data)
    expect_equal(result, "lap")
})

test_that("validate_event_channel() errors when not in metadata or provided", {
    data <- create_test_data(add_metadata = FALSE)
    expect_error(validate_event_channel(NULL, data), "not detected in metadata")

    ## no metadata, required = FALSE
    expect_equal(validate_event_channel(NULL, data, required = FALSE), NULL)
})

test_that("validate_event_channel() errors when column doesn't exist", {
    data <- create_test_data()
    expect_error(validate_event_channel("nonexistent", data), "match exactly")
})

test_that("validate_event_channel() errors when all NA", {
    data <- create_test_data()
    data_one <- data[1:(nrow(data) - 1), ] ## one valid should work
    expect_equal(validate_event_channel("event", data_one), "event")
    data_na <- data[2:(nrow(data) - 1), ]
    expect_error(
        validate_event_channel("event", data_na),
        "must contain valid"
    )
})

test_that("validate_event_channel() works for quosures", {
    data <- create_test_data()
    result <- validate_event_channel(rlang::quo("event"), data)
    expect_equal(result, "event")
})

test_that("validate_event_channel() accepts integer column", {
    data <- create_test_data()
    ## create_test_data() uses c(1, rep(NA, nrow - 2), 2) — integerish
    data$lap <- c(1L, rep(NA_integer_, nrow(data) - 2L), 2L)
    result <- validate_event_channel("lap", data)
    expect_equal(result, "lap")
})

test_that("validate_event_channel() error conditions", {
    data <- create_test_data()

    ## errors for non-integerish numeric column
    data$bad_col <- c(1.5, rep(NA, nrow(data) - 2), 2.5)
    expect_error(validate_event_channel("bad_col", data), "integer")

    ## errors when integer column is all NA
    data$lap <- rep(NA_integer_, nrow(data))
    expect_error(validate_event_channel("lap", data), "must contain valid")
})



## within() ===============================================================
test_that("within() handles basic inclusive range (default)", {
    expect_equal(within(5, c(1, 10)), TRUE)
    expect_equal(within(1, c(1, 10)), TRUE)
    expect_equal(within(10, c(1, 10)), TRUE)
    expect_equal(within(0, c(1, 10)), FALSE)
    expect_equal(within(11, c(1, 10)), FALSE)
})

test_that("within() handles vectorised inputs", {
    expect_equal(
        within(c(0, 1, 5, 10, 11), c(1, 10)),
        c(FALSE, TRUE, TRUE, TRUE, FALSE)
    )
    expect_equal(
        within(1:10, c(3, 7)),
        c(FALSE, FALSE, TRUE, TRUE, TRUE, TRUE, TRUE, FALSE, FALSE, FALSE)
    )
})

test_that("within() handles non-numeric `vec` values", {
    expect_error(
        within(1:5, c("A", "B")),
        "valid.*numeric"
    )
})

test_that("within() handles inclusive/exclusive", {
    ## both exclusive
    expect_false(within(1, c(1, 10), inclusive = FALSE))
    expect_false(within(10, c(1, 10), inclusive = FALSE))
    expect_true(within(5, c(1, 10), inclusive = FALSE))
    expect_equal(
        within(c(1, 5, 10), c(1, 10), inclusive = FALSE),
        c(FALSE, TRUE, FALSE)
    )

    ## left inclusive
    expect_true(within(1, c(1, 10), inclusive = "left"))
    expect_false(within(10, c(1, 10), inclusive = "left"))
    expect_equal(
        within(c(1, 5, 10), c(1, 10), inclusive = "left"),
        c(TRUE, TRUE, FALSE)
    )

    ## right inclusive
    expect_false(within(1, c(1, 10), inclusive = "right"))
    expect_true(within(10, c(1, 10), inclusive = "right"))
    expect_equal(
        within(c(1, 5, 10), c(1, 10), inclusive = "right"),
        c(FALSE, TRUE, TRUE)
    )
})

test_that("within() detects positive non-zero values", {
    expect_true(within(0, c(0, Inf)))
    expect_true(within(Inf, c(0, Inf)))
    expect_false(within(0, c(0, Inf), inclusive = FALSE))
    expect_true(within(1, c(0, Inf), inclusive = FALSE))
    expect_false(within(-1, c(0, Inf)))
    expect_false(within(-1, c(0, Inf), inclusive = FALSE))
    expect_equal(
        within(c(-1, 0, 0.001, 1), c(0, Inf), inclusive = FALSE),
        c(FALSE, FALSE, TRUE, TRUE)
    )
})

test_that("within() handles NA values", {
    expect_error(within(NA, c(1, 10)), "valid.*numeric")
    expect_equal(within(NA_real_, c(1, 10)), NA)
    expect_equal(within(c(1, NA, 5), c(1, 10)), c(TRUE, NA, TRUE))
    expect_false(within(5, c(NA, 10)))
    expect_false(within(5, c(1, NA)))
})

test_that("within() handles infinite values", {
    expect_true(within(Inf, c(1, Inf)), TRUE)
    expect_false(within(Inf, c(1, Inf), inclusive = FALSE), FALSE)
    expect_true(within(-Inf, c(-Inf, 10)), TRUE)
    expect_false(within(-Inf, c(-Inf, 10), inclusive = FALSE), FALSE)
})

test_that("within() handles negative ranges", {
    expect_true(within(-5, c(-10, -1)))
    expect_true(within(-10, c(-10, -1)))
    expect_false(within(-11, c(-10, -1)))
})

test_that("within() handles degenerate ranges where left equals right", {
    expect_true(within(5, c(5, 5)), TRUE)
    expect_false(within(5, c(5, 5), inclusive = FALSE))
    expect_false(within(5, c(5, 5), inclusive = "left"))
    expect_false(within(5, c(5, 5), inclusive = "right"))
})

test_that("within is equivalent to dplyr::between()", {
    expect_equal(
        within(c(0, 5, 10, 15), c(1, 10)),
        dplyr::between(c(0, 5, 10, 15), 1, 10)
    )

    expect_equal(within(NA_real_, c(1, 10)), dplyr::between(NA_real_, 1, 10))
    expect_equal(
        within(c(1, NA, 5), c(1, 10)),
        dplyr::between(c(1, NA, 5), 1, 10)
    )
})


## estimate_sample_rate() ======================================
test_that("estimate_sample_rate works correctly", {
    # Regular 100 Hz data
    time_100hz <- seq(0, 1, by = 0.01)
    expect_equal(estimate_sample_rate(time_100hz), 100)

    # 50 Hz data
    time_50hz <- seq(0, 1, by = 0.02)
    expect_equal(estimate_sample_rate(time_50hz), 50)

    ## irregular sampling
    set.seed(13)
    x <- seq(0, 1, by = 0.02)
    x <- x + rnorm(length(x), 0, 0.001)
    expect_true(estimate_sample_rate(x) %in% c(49.5, 50, 50.5))
    expect_true(
        all.equal(estimate_sample_rate(x), 50, tolerance = 1, scale = 1)
    )
    expect_true(
        all.equal(estimate_sample_rate(x), 50, tolerance = 1, scale = 1)
    )

    # With NAs in diffs (not in x directly, as diff removes one element)
    time_with_gaps <- c(0, 0.01, NA, 0.03, 0.04)
    expect_type(estimate_sample_rate(time_with_gaps), "double")

    # Edge case: single diff value
    expect_equal(estimate_sample_rate(c(0, 0.01)), 100)
    expect_error(estimate_sample_rate(NA), "numeric")

    ## edge case sample rate undetectable returns NULL
    expect_error(
        estimate_sample_rate(c(1, 1, 1, 1, 1)),
        "Unable to estimate"
    )
    expect_error(
        estimate_sample_rate(Inf),
        "Unable to estimate"
    )
})


## validate_sample_rate() ========================================
test_that("validate_sample_rate() uses metadata when NULL", {
    data <- create_test_data(sample_rate = 15)
    result <- validate_sample_rate(data, "time", NULL)
    expect_equal(result, 15)
})

test_that("validate_sample_rate() uses explicit value with warning when provided", {
    data <- create_test_data(sample_rate = 10)
    expect_equal(validate_sample_rate(data, "time", 20), 20) |>
        expect_warning("appears to be inconsistent")
    ## uses explicit without warning
    expect_equal(validate_sample_rate(data, "time", 20, verbose = FALSE), 20) |>
        expect_silent()
})

test_that("validate_sample_rate() does not warn for integer time_channel", {
    data <- create_test_data(time_max = 100, sample_rate = 1)
    expect_silent(validate_sample_rate(data, "time", 5, FALSE))
})

test_that("validate_sample_rate() estimates from time_channel when NULL", {
    rate = 9.5
    data <- create_test_data(
        time_max = 10,
        sample_rate = rate,
        add_metadata = FALSE
    )
    result <- validate_sample_rate(data, "time", NULL, verbose = FALSE)
    expect_equal(result, round(rate))

    rate = 0.8
    data <- create_test_data(
        time_max = 10,
        sample_rate = rate,
        add_metadata = FALSE
    )
    result <- validate_sample_rate(data, "time", NULL, verbose = FALSE)
    expect_equal(result, round(rate))

    rate = 11
    data <- create_test_data(
        time_max = 10,
        sample_rate = rate,
        add_metadata = FALSE
    )
    result <- validate_sample_rate(data, "time", NULL, verbose = FALSE)
    expect_equal(result, 10)

    rate = 44
    data <- create_test_data(
        time_max = 10,
        sample_rate = rate,
        add_metadata = FALSE
    )
    result <- validate_sample_rate(data, "time", NULL, verbose = FALSE)
    expect_equal(result, 50)

    rate = 98
    data <- create_test_data(
        time_max = 10,
        sample_rate = rate,
        add_metadata = FALSE
    )
    result <- validate_sample_rate(data, "time", NULL, verbose = FALSE)
    expect_equal(result, 100)
})

test_that("validate_sample_rate() shows message when estimating", {
    data <- create_test_data(add_metadata = FALSE)
    expect_message(
        validate_sample_rate(data, "time", NULL, verbose = TRUE),
        "Estimated"
    )
})

test_that("validate_sample_rate() errors for non-single non-numeric non-positive", {
    data <- create_test_data()
    expect_error(
        validate_sample_rate(data, "time", "10", FALSE),
        "must be .*numeric"
    )
    expect_error(
        validate_sample_rate(data, "time", NA, FALSE),
        "must be .*numeric"
    )
    expect_error(
        validate_sample_rate(data, "time", 0, FALSE),
        "must be .*positive"
    )
    expect_error(
        validate_sample_rate(data, "time", -5, FALSE),
        "must be .*positive"
    )
    expect_error(
        validate_sample_rate(data, "time", c(10, 20), FALSE),
        "must be .*one-element"
    )
})

## validate_width_span ==============================
test_that("validate_width_span() validates inputs", {
    expect_error(
        validate_width_span(width = NULL, span = NULL),
        "width.*span.*must be defined"
    )

    expect_message(
        validate_width_span(width = 2, span = 1),
        "width.*overrides.*span"
    )

    expect_error(
        validate_width_span(width = -1),
        "width.*valid.*integer"
    )

    expect_error(
        validate_width_span(width = 1.5),
        "width.*valid.*integer"
    )

    expect_error(
        validate_width_span(span = -1),
        "span.*valid.*numeric"
    )
})

test_that("validate_width_span() reports message", {
    expect_error(
        validate_width_span(
            width = NULL, span = NULL, msg = "for this function."
        ),
        "width.*span.*must be defined.*for this function."
    )
})

## validate_x_t =================================
test_that("validate_x_t() validates inputs", {
    expect_error(validate_x_t(x = 1:10, t = 1:5), "numeric.*equal length")
    expect_error(validate_x_t(x = NULL, t = 1:10), "valid.*numeric")
    expect_error(validate_x_t(x = 1:10, t = NULL), "valid.*numeric")
    expect_error(validate_x_t(x = NA_real_, t = 1), "valid.*numeric")
    expect_silent(validate_x_t(x = NA_real_, t = 1, allow_na = TRUE))
})

#' @keywords internal
"_PACKAGE"

## usethis namespace: start
#' @importFrom cli cli_abort
#' @importFrom cli cli_inform
#' @importFrom cli cli_warn
#' @importFrom cli col_blue
#' @importFrom cli qty
#' @importFrom lifecycle deprecated
#' @importFrom rlang !!
#' @importFrom rlang !!!
#' @importFrom rlang %||%
#' @importFrom rlang .data
#' @importFrom rlang :=
#' @importFrom rlang check_installed
#' @importFrom rlang enquo
#' @importFrom stats coef
#' @importFrom stats median
#' @importFrom stats rnorm
#' @importFrom stats setNames
#' @importFrom tibble tibble
## usethis namespace: end
NULL

#' Filter a data frame
#'
#' Apply digital filtering/smoothing to numeric vector data within a data frame
#' using either:
#'   1. A cubic smoothing spline.
#'   2. A Butterworth digital filter.
#'   3. A simple moving average.
#'
#' @param method A character string indicating how to filter the data (see
#'   *Details*).
#'   \describe{
#'      \item{`"smooth_spline"`}{Fits a cubic smoothing spline.}
#'      \item{`"butterworth"`}{Uses a centred Butterworth digital filter.}
#'      \item{`"moving_average"`}{Uses a centred moving average filter.}
#'   }
#' @param na.rm Logical; default is `FALSE`, propagates any `NA`s to the 
#'   returned vector. If `TRUE`, ignores `NA`s and processes available valid 
#'   samples within the local window. May return errors or warnings. (see 
#'   *Details*).
#' @param ... Additional method-specific arguments must be specified 
#'   (see *Details*).
#' @inheritParams validate_mnirs
#'
#' @details
#' ## method = "smooth_spline"
#' 
#' Aliases: `method = c("smooth spline", "spline")`
#'
#' Applies a non-parametric cubic smoothing spline from
#' [stats::smooth.spline()]. Smoothing is defined by the parameter `spar`,
#' which can be left as `NULL` and automatically determined via penalised
#' log likelihood. This usually works well for responses occurring on the 
#' order of minutes or longer. `spar` can be specified typically, but not 
#' necessarily, in the range `spar = [0, 1]`.
#'
#' Additional arguments (`...`) accepted when `method = "smooth_spline"`:
#'
#' \describe{
#'   \item{`spar`}{A numeric smoothing parameter passed to
#'       [stats::smooth.spline()]. If `NULL` (*default*), automatically
#'       determined via penalised log likelihood.}
#' }
#'
#' ## method = "butterworth"
#' 
#' Aliases: `method = c("butter")`
#'
#' Applies a centred (two-pass symmetrical) Butterworth digital filter
#' from [signal::butter()] and [signal::filtfilt()].
#'
#' Filter `type` defines how the desired signal frequencies are either
#' passed or rejected from the output signal. *Low-pass* and *high-pass*
#' filters allow only frequencies *lower* or *higher* than the cutoff
#' frequency, respectively to be passed through to the output signal.
#' *Stop-band* defines a critical range of frequencies which are rejected
#' from the output signal. *Pass-band* defines a critical range of
#' frequencies which are passed through as the output signal.
#'
#' The filter order (number of passes) is defined by `order`, typically
#' in the range `order = [1, 10]`. Higher filter order tends to capture
#' more rapid changes in amplitude, but also causes more distortion
#' around those change points in the signal. General advice is to use
#' the lowest filter order which sufficiently captures the desired rapid
#' responses in the data.
#'
#' The critical (cutoff) frequency can be defined by `W`, a numeric value
#' for *low-pass* and *high-pass* filters, or a two-element vector
#' `c(low, high)` defining the lower and upper bands for *stop-band*
#' and *pass-band* filters. `W` represents the desired fractional cutoff
#' frequency in the range `W = [0, 1]`, where `1` is the Nyquist
#' frequency, i.e., half the `sample_rate` of the data in Hz.
#'
#' Alternatively, the cutoff frequency can be defined by `fc` and
#' `sample_rate` together. `fc` represents the desired cutoff frequency
#' directly in Hz, and `sample_rate` is the sample rate of the recorded data 
#' in Hz. Where `W = fc / (sample_rate / 2)`.
#'
#' Only one of either `W` or `fc` should be defined. If both are
#' defined, `W` will be preferred over `fc`.
#'
#' Additional arguments (`...`) accepted when `method = "butterworth"`:
#'
#' \describe{
#'   \item{`order`}{An integer for the filter order (*default* `2`).}
#'   \item{`W`}{A numeric fractional cutoff frequency within `[0, 1]`. One
#'       of either `W` or `fc` must be specified.}
#'   \item{`fc`}{A numeric absolute cutoff frequency in Hz. Used with
#'       `sample_rate` to compute `W`.}
#'   \item{`sample_rate`}{A numeric sample rate in Hz. Will be taken from
#'       metadata or estimated from `time_channel` if not defined.}
#'   \item{`type`}{A character string specifying filter type, one of:
#'       `c("low", "high", "stop", "pass")` (`"low"` is the default).}
#'   \item{`edges`}{A character string specifying the edge padding, one of:
#'       `c("rev", "rep1", "none")` (`"rev"` is the default). 
#'       See [filter_butter()].}
#' }
#'
#' ## method = "moving_average"
#' 
#' Aliases: `method = c("moving average", "ma")`
#'
#' Applies a centred (symmetrical) moving average filter in a local
#' window, defined by either `width` as the number of samples around
#' `idx` between `[idx - floor(width/2), idx + floor(width/2)]`. Or by
#' `span` as the timespan in units of `time_channel` between
#' `[t - span/2, t + span/2]`.
#'
#' Additional arguments (`...`) accepted when `method = "moving_average"`:
#'
#' \describe{
#'   \item{`width` or `span`}{Either an integer number of samples, or a
#'       numeric time duration in units of `time_channel` within the local 
#'       window. One of either `width` or `span` must be specified.}
#'   \item{`partial`}{Logical; `FALSE` by default, only returns values
#'       where a full window of valid (non-`NA`) samples are available. 
#'       If `TRUE`, ignores `NA` and allows calculation over partial windows 
#'       at the edges of the data.}
#' }
#' 
#' ## Missing values
#'
#' Missing values (`NA`) in `nirs_channels` will cause an error for
#' `method = "smooth_spline"` or `"butterworth"`, unless `na.rm = TRUE`.
#' Then `NA`s will be ignored and passed through to the returned data.
#' 
#' For `method = "moving_average"`, `na.rm` controls whether `NA`s within
#' each local window are either propagated to the returned vector when 
#' `na.rm = FALSE` (the default), or ignored before processing if 
#' `na.rm = TRUE`.
#'
#' @returns
#' A [tibble][tibble::tibble-package] of class *"mnirs"* with metadata
#'   available with `attributes()`.
#'
#' @examples
#' ## read example data and clean for outliers
#' data <- read_mnirs(
#'     file_path = example_mnirs("moxy_ramp"),
#'     nirs_channels = c(smo2 = "SmO2 Live"),
#'     time_channel = c(time = "hh:mm:ss"),
#'     verbose = FALSE
#' ) |>
#'     replace_mnirs(
#'         invalid_values = c(0, 100),
#'         outlier_cutoff = 3,
#'         width = 7,
#'         verbose = FALSE
#'     )
#' 
#' data
#'
#' data_filtered <- filter_mnirs(
#'     data,                   ## blank channels will be retrieved from metadata
#'     method = "butterworth", ## Butterworth digital filter is a common choice
#'     order = 2,              ## filter order number
#'     W = 0.02,               ## filter fractional critical frequency `[0, 1]`
#'     type = "low",           ## specify a "low-pass" filter
#'     na.rm = TRUE            ## explicitly ignore NAs
#' )
#' 
#' ## note the smoothed `smo2` values
#' data_filtered
#'
#' \donttest{
#'     if (requireNamespace("ggplot2", quietly = TRUE)) {
#'         ## plot filtered data and add the raw data back to the plot to compare
#'         plot(data_filtered, time_labels = TRUE) +
#'             ggplot2::geom_line(
#'                 data = data,
#'                 ggplot2::aes(y = smo2, colour = "smo2"), alpha = 0.4
#'             )
#'     }
#' }
#'
#' @rdname filter_mnirs
#' @export
filter_mnirs <- function(
    data,
    nirs_channels = NULL,
    time_channel = NULL,
    method = c("smooth_spline", "butterworth", "moving_average"),
    na.rm = FALSE,
    verbose = TRUE,
    ...
) {
    ## validation ====================================
    validate_mnirs_data(data)
    ## normalise method aliases before matching
    method <- gsub(
        "^(ma|moving[ _-]average)$",
        "moving_average",
        method,
        ignore.case = TRUE
    )
    method <- gsub(
        "^(spline|smooth[ _-]spline)$",
        "smooth_spline",
        method,
        ignore.case = TRUE
    )
    method <- match.arg(method)
    
    if (missing(verbose)) {
        verbose <- getOption("mnirs.verbose", default = TRUE)
    }

    UseMethod(
        "filter_mnirs", 
        structure(data, class = c(method, "mnirs_filtered"))
    )
}


#' @rdname filter_mnirs
#' @usage NULL
#' @export
filter_mnirs.smooth_spline <- function(
    data,
    nirs_channels = NULL,
    time_channel = NULL,
    method,
    na.rm = FALSE,
    verbose = TRUE,
    ...
) {
    ## validation ==========================================
    metadata <- attributes(data)
    nirs_channels <- validate_nirs_channels(enquo(nirs_channels), data, verbose)
    time_channel <- validate_time_channel(enquo(time_channel), data)
    spar <- list(...)$spar
    validate_numeric(spar, 1, c(0, Inf), FALSE, msg1 = "one-element positive")

    ## processing ==========================================
    time_vec <- data[[time_channel]]

    if (anyDuplicated(time_vec)) {
        cli_abort(c(
            "x" = "{.arg time_channel} has duplicated or irregular samples.",
            "i" = "Re-sample first with {.fn mnirs::resample_mnirs}."
        ))
    }

    data[nirs_channels] <- lapply(nirs_channels, \(.x) {
        x <- data[[.x]]
        ## handle NAs
        handle_na <- na.rm && anyNA(x)
        if (handle_na) {
            na_info <- preserve_na(x)
            x <- na_info$x_valid
            time_vec <- time_vec[!na_info$na_idx]
        } else if (anyNA(x)) {
            cli_abort(c(
                "x" = "{.arg nirs_channels} = {.val {.x}} contains internal \\
                {.val {NA}}'s.",
                "i" = "Set {.arg na.rm = TRUE} to ignore {.val {NA}}'s."
            ))
        }

        spline_model <- stats::smooth.spline(x = time_vec, y = x, spar = spar)

        if (is.null(spar) && verbose) {
            cli_inform(c(
                "i" = "{.arg nirs_channel} = {.val {.x}}: \\
                `smooth.spline(spar = {.val {round(spline_model$spar, 3)}})`"
            ))
        }

        if (handle_na) {
            restore_na(spline_model$y, na_info)
        } else {
            spline_model$y
        }
    })

    ## Metadata =================================
    metadata$nirs_channels <- unique(nirs_channels)

    return(create_mnirs_data(data, metadata))
}


#' @rdname filter_mnirs
#' @usage NULL
#' @export
filter_mnirs.butterworth <- function(
    data,
    nirs_channels = NULL,
    time_channel = NULL,
    method,
    na.rm = FALSE,
    verbose = TRUE,
    ...
) {
    ## validation ==========================================
    metadata <- attributes(data)
    nirs_channels <- validate_nirs_channels(enquo(nirs_channels), data, verbose)
    time_channel <- validate_time_channel(enquo(time_channel), data)
    args <- list(...)
    sample_rate <- args$sample_rate
    sample_rate <- validate_sample_rate(
        data, time_channel, sample_rate, verbose
    )
    order <- args$order %||% 2L
    W <- args$W
    fc <- args$fc
    type <- args$type %||% "low"
    type <- match.arg(type, c("low", "high", "stop", "pass"))
    edges <- args$edges %||% "rev"

    if (is.null(c(W, fc))) {
        cli_abort(c(
            "x" = "Cutoff frequency undefined.",
            "i" = "One of {.arg W} or {.arg fc} must be defined for a \\
            Butterworth filter."
        ))
    }

    fc_n <- if (type %in% c("low", "high")) 1 else 2
    ## order & W are validated in filter_butter
    validate_numeric(
        fc, fc_n, c(0, Inf), inclusive = FALSE,
        msg1 = paste0(fc_n, "-element positive")
    )

    if (!is.null(W) && !is.null(fc)) {
        fc <- NULL
        if (verbose) {
            cli_inform(c(
                "i" = "{.val Butterworth} parameter {.arg W} = \\
                {.val {W}} overrides {.arg fc}."
            ))
        }
    }

    if (is.null(W) && !is.null(fc) && !is.null(sample_rate)) {
        nq <- sample_rate * 0.5
        W <- fc / nq
        if (W > 1 | W <= 0) {
            cli_abort(c(
                "x" = "{.arg fc} must be between {.val {0}} and half \\
                the {.val sample_rate} ({.val {signif(nq, 3)}} Hz)"
            ))
        }
    }

    ## processing ==========================================
    data[nirs_channels] <- lapply(data[nirs_channels], \(.x) {
        filter_butter(.x, order, W, type, edges, na.rm)
    })

    ## Metadata =================================
    metadata$nirs_channels <- unique(nirs_channels)
    metadata$time_channel <- time_channel
    metadata$sample_rate <- sample_rate

    return(create_mnirs_data(data, metadata))
}


#' @rdname filter_mnirs
#' @usage NULL
#' @export
filter_mnirs.moving_average <- function(
    data,
    nirs_channels = NULL,
    time_channel = NULL,
    method,
    na.rm = FALSE,
    verbose = TRUE,
    ...
) {
    ## validation ==========================================
    metadata <- attributes(data)
    nirs_channels <- validate_nirs_channels(enquo(nirs_channels), data, verbose)
    time_channel <- validate_time_channel(enquo(time_channel), data)
    args <- list(...)
    width <- args$width
    span <- args$span
    partial <- args$partial %||% FALSE

    ## processing ==========================================
    time_vec <- data[[time_channel]]

    data[nirs_channels] <- lapply(data[nirs_channels], \(.x) {
        filter_ma(
            x = .x,
            t = time_vec,
            width = width,
            span = span,
            partial = partial,
            na.rm = na.rm,
            verbose = verbose,
            bypass_checks = TRUE
        )
    })

    ## Metadata =================================
    metadata$nirs_channels <- unique(nirs_channels)
    metadata$time_channel <- time_channel

    return(create_mnirs_data(data, metadata))
}


#' Apply a moving average filter
#'
#' Apply a simple moving average smoothing filter to vector data.
#' `filter_moving_average()` is an alias of `filter_ma()`.
#'
#' @param partial Logical; default is `FALSE`, requires local windows to have
#'   complete number of samples specified by `width` or `span`. If `TRUE`, 
#'   processes available samples within the local window. See *Details*.
#' @inheritParams replace_invalid
#' @inheritParams shift_mnirs
#' @inheritParams filter_mnirs
#'
#' @details
#' ## Rolling window
#' 
#' Applies a centred (symmetrical) moving average filter in a local
#' window, defined by either `width` as the number of samples around
#' `idx` between `[idx - floor(width/2), idx + floor(width/2)]`. Or by
#' `span` as the timespan in units of `time_channel` between
#' `[t - span/2, t + span/2]`.
#'
#' ## Partial windows
#'
#' The default `partial = FALSE` requires a complete number of samples
#' specified by `width` or `span` (estimated from the sample rate of `t` when
#' `span` is used). `NA` is returned if fewer samples are present in the
#' local window. 
#' 
#' Setting `partial = TRUE` allows computation with only a single valid sample, 
#' such as at edge conditions. But these values will be more sensitive to 
#' noise and should be used with caution.
#'
#' ## Missing values
#'
#' `na.rm` controls whether missing values (`NA`s) within each local window are 
#' either propagated to the returned vector when `na.rm = FALSE` (the default),
#' or ignored before processing if `na.rm = TRUE`.
#'
#' @returns A numeric vector the same length as `x`.
#'
#' @examples
#' x <- c(1, 3, 2, 5, 4, 6, 5, 7)
#' t <- c(0, 1, 2, 4, 5, 6, 7, 10)  ## irregular time with gaps
#'
#' ## width: centred window of 3 samples
#' filter_ma(x, width = 3)
#'
#' ## partial = TRUE fills edge values with a narrower window
#' filter_ma(x, width = 3, partial = TRUE)
#'
#' ## span: centred window of 2 time-units (accounts for irregular sampling)
#' filter_ma(x, t, span = 2)
#'
#' ## na.rm = FALSE (default): any NA in the window propagates to the result
#' x_na <- c(1, NA, 3, 4, 5, NA, 7, 8)
#' filter_ma(x_na, width = 3, na.rm = FALSE)
#'
#' ## na.rm = TRUE: skip NAs and return the local mean of local valid values
#' filter_ma(x_na, width = 3, partial = TRUE, na.rm = TRUE)
#'
#' @rdname filter_ma
#' @export
filter_ma <- function(
    x,
    t = seq_along(x),
    width = NULL,
    span = NULL,
    partial = FALSE,
    na.rm = FALSE,
    verbose = TRUE,
    ...
) {
    ## validation ===========================================
    if (!(list(...)$bypass_checks %||% FALSE)) {
        if (missing(verbose)) {
            verbose <- getOption("mnirs.verbose", default = TRUE)
        }
    }
    validate_x_t(x, t)
    validate_width_span(width, span, verbose, "for a moving average filter.")

    ## handle NAs
    if (verbose && !na.rm && anyNA(x)) {
        cli_warn(c(
            "!" = "{.arg x} contains internal {.val {NA}}'s.",
            "i" = "Set {.arg na.rm = TRUE} to ignore {.val {NA}}'s."
        ))
    }

    ## processing ==============================================
    window_idx <- compute_local_windows(t, width = width, span = span)

    if (!partial) {
        ## min_obs default to estimated width when span is specified
        ## less strict span_width - 2 to allow start & end buffer
        ## with irregular t values
        min_obs <- max(
            width %||% (floor(span * estimate_sample_rate(t)) - 2L),
            1L
        )

        ## error if fewer valid samples than min_obs
        if (sum(is.finite(x)) < min_obs) {
            cli_abort(c(
                "x" = "Insufficient valid samples detected.",
                "i" = "{.arg width} or {.arg span} must be smaller than \\
                the range of {.arg x} when {.arg partial} = {.val {FALSE}}."
            ))
        }

        which_partial <- lengths(window_idx) < min_obs
    }

    y <- vapply(window_idx, \(.idx) mean(x[.idx], na.rm = na.rm), numeric(1))

    if (!partial) {
        ## exclude incomplete windows (at edges)
        y[which_partial] <- NA_real_
    }
    ## NaN to NA
    y[!is.finite(y)] <- NA_real_
    return(y)
}


#' @rdname filter_ma
#' @export
filter_moving_average <- function(
    x,
    t = seq_along(x),
    width = NULL,
    span = NULL,
    partial = FALSE,
    na.rm = FALSE,
    verbose = TRUE,
    ...
) {
    filter_ma(
        x = x,
        t = t,
        width = width,
        span = span,
        partial = partial,
        na.rm = na.rm,
        verbose = verbose,
        ...
    )
}


#' Apply a Butterworth digital filter
#'
#' Apply a Butterworth digital filter to vector data with [signal::butter()]
#' and [signal::filtfilt()] which handles 'edges' better at the start and end
#' of the data.
#'
#' @param x A numeric vector.
#' @param order An integer defining the filter order (*default* `order = 2`).
#' @param W A one- or two-element numeric vector defining the filter cutoff
#'   frequency(ies) as a fraction of the Nyquist frequency (see *Details*).
#' @param type A character string indicating the digital filter type (see
#'   *Details*).
#'   \describe{
#'      \item{`"low"`}{For a *low-pass* filter (the *default*).}
#'      \item{`"high"`}{For a *high-pass* filter.}
#'      \item{`"stop"`}{For a *stop-band* (band-reject) filter.}
#'      \item{`"pass"`}{For a *pass-band* filter.}
#'   }
#' @param edges A character string indicating edge detection padding for `x`.
#'   \describe{
#'      \item{`"rev"`}{Will pad `x` with the preceding 5% data in reverse 
#'      sequence (*the default*).}
#'      \item{`"rep1"`}{Will pad `x` by repeating the last preceding value.}
#'      \item{`"none"`}{Will return the unpadded [signal::filtfilt()] output.}
#'   }
#' @inheritParams filter_mnirs
#'
#' @details
#' Applies a centred (two-pass symmetrical) Butterworth digital filter from
#'   [signal::butter()] and [signal::filtfilt()].
#'
#' Filter `type` defines how the desired signal frequencies are either
#'   passed or rejected from the output signal. *Low-pass* and *high-pass*
#'   filters allow only frequencies *lower* or *higher* than the cutoff
#'   frequency `W` to be passed through as the output signal, respectively.
#'   *Stop-band* defines a critical range of frequencies which are rejected
#'   from the output signal. *Pass-band* defines a critical range of
#'   frequencies which are passed through as the output signal.
#'
#' The filter order (number of passes) is defined by `order`, typically in
#'   the range `order = [1, 10]`. Higher filter order tends to capture more
#'   rapid changes in amplitude, but also causes more distortion around
#'   those change points in the signal. General advice is to use the
#'   lowest filter order which sufficiently captures the desired rapid
#'   responses in the data.
#'
#' The critical (cutoff) frequency is defined by `W`, a numeric value for
#'   *low-pass* and *high-pass* filters, or a two-element vector
#'   `c(low, high)` defining the lower and upper bands for *stop-band* and
#'   *pass-band* filters. `W` represents the desired fractional cutoff
#'   frequency in the range `W = [0, 1]`, where `1` is the Nyquist
#'   frequency, i.e., half the sample rate of the data in Hz.
#'
#' Missing values (`NA`) in `x` will cause an error unless `na.rm = TRUE`.
#'   Then `NA`s will be ignored and passed through to the returned vector.
#'
#' @returns A numeric vector the same length as `x`.
#'
#' @seealso [signal::filtfilt()], [signal::butter()]
#'
#' @examples
#' set.seed(13)
#' sin <- sin(2 * pi * 1:150 / 50) * 20 + 40
#' noise <- rnorm(150, mean = 0, sd = 6)
#' noisy_sin <- sin + noise
#' without_edge_detection <- filter_butter(
#'     x = noisy_sin,
#'     order = 2,
#'     W = 0.1,
#'     edges = "none"
#' )
#' with_edge_detection <- filter_butter(
#'     x = noisy_sin,
#'     order = 2,
#'     W = 0.1,
#'     edges = "rep1"
#' )
#'
#' @examplesIf rlang::is_installed("ggplot2")
#' ggplot2::ggplot(data.frame(), ggplot2::aes(x = seq_along(noise))) +
#'     theme_mnirs() +
#'     scale_colour_mnirs(name = NULL) +
#'     ggplot2::geom_line(ggplot2::aes(y = noisy_sin)) +
#'     ggplot2::geom_line(
#'         ggplot2::aes(y = without_edge_detection, colour = "without_edge_detection")
#'     ) +
#'     ggplot2::geom_line(
#'         ggplot2::aes(y = with_edge_detection, colour = "with_edge_detection")
#'     )
#'
#' @export
filter_butter <- function(
    x,
    order = 2L,
    W,
    type = c("low", "high", "stop", "pass"),
    edges = c("rev", "rep1", "none"),
    na.rm = FALSE,
    ...
) {
    ## validation ============================================
    check_installed("signal", "to use Butterworth digital filter")
    validate_numeric(x)
    validate_numeric(
        order, 1, c(1, Inf), integer = TRUE, msg1 = "one-element positive"
    )
    type <- match.arg(type)
    W_n <- if (type %in% c("low", "high")) 1 else 2
    validate_numeric(
        W, W_n, c(0, 1), inclusive = FALSE,
        msg1 = paste0(W_n, "-element positive"),
        msg2 = "between {col_blue('[0, 1]')}."
    )
    edges <- match.arg(edges)

    ## processing ==============================================
    ## handle NAs
    handle_na <- na.rm && anyNA(x)
    if (handle_na) {
        na_info <- preserve_na(x)
        x <- na_info$x_valid
    } else if (anyNA(x)) {
        cli_abort(c(
            "x" = "{.arg x} contains internal {.val {NA}}'s.",
            "i" = "Set {.arg na.rm = TRUE} to ignore {.val {NA}}'s."
        ))
    }

    if (edges == "none") {
        y <- signal::filtfilt(signal::butter(n = order, W, type), x = x)
    } else {
        x_n <- length(x)
        pad <- max(1L, x_n %/% 20) ## 5% padded length

        padded <- switch(
            edges,
            "rev" = c(x[pad:1], x, x[x_n:(x_n - pad + 1)]),
            "rep1" = c(rep(x[1], pad), x, rep(x[x_n], pad))
        )

        y_padded <- signal::filtfilt(
            signal::butter(n = order, W, type),
            x = padded
        )
        y <- y_padded[(pad + 1):(pad + x_n)]
    }

    ## return NAs to original positions in y
    if (handle_na) {
        return(restore_na(y, na_info))
    } else {
        return(y)
    }
}

## filter_ma() ===========================================
test_that("filter_ma() returns expected smoothed values", {
    x <- c(1, 2, 3, 4, 5)

    result <- filter_ma(x, width = 3, verbose = FALSE, na.rm = TRUE)
    expect_equal(result, c(NA, 2, 3, 4, NA))
    result <- filter_ma(x, width = 3, verbose = FALSE, partial = TRUE)
    expect_equal(result, c(1.5, 2, 3, 4, 4.5))

    # Width-based with single width (floor(1/2) = 0, so just x itself)
    result <- filter_ma(x, width = 1, verbose = FALSE)
    expect_equal(result, x)
    
    ## filter_moving_average alias for filter_ma
    result <- filter_moving_average(
        x,
        width = 3,
        verbose = FALSE,
        partial = FALSE
    )
    expect_equal(result, c(NA, 2, 3, 4, NA))
    result <- filter_moving_average(
        x,
        width = 3,
        verbose = FALSE,
        partial = TRUE
    )
    expect_equal(result, c(1.5, 2, 3, 4, 4.5))
})

test_that("filter_ma() handles custom time vectors", {
    x <- c(10, 20, 30, 40, 50, 60)
    t <- c(0, 1, 3, 6, 10, 11)

    # Span-based with time gaps
    result <- filter_ma(x, t, span = 2, verbose = FALSE)
    expect_equal(result, c(15, 15, 30, 40, 55, 55))
})

test_that("filter_ma() handles NA values correctly", {
    x <- c(1, NA, 3, 4, 5)

    result <- filter_ma(
        x,
        width = 3,
        partial = FALSE,
        na.rm = FALSE,
        verbose = FALSE
    )
    expect_equal(result, c(rep(NA, 3), 4, NA))

    result <- filter_ma(
        x,
        width = 3,
        partial = TRUE,
        na.rm = TRUE,
    )
    expect_equal(result, c(1, 2, 3.5, 4, 4.5))

    result <- filter_ma(
        x,
        width = 3,
        partial = FALSE, ## min 3 valid obs at edges
        na.rm = TRUE, ## mean excludes NA
    )
    expect_equal(result, c(NA, 2, 3.5, 4, NA))

    result <- filter_ma(
        x,
        width = 3,
        partial = TRUE,
        na.rm = FALSE,
        verbose = FALSE
    )
    expect_equal(result, c(rep(NA, 3), 4, 4.5))

    x <- c(1, NA, NA, NA, NA, 6, 7, 8)
    result <- filter_ma(x, width = 3, partial = TRUE, na.rm = TRUE)
    expect_equal(result, c(1, 1, NA, NA, 6, 6.5, 7, 7.5))
})

test_that("filter_ma() validates x as numeric", {
    expect_error(
        filter_ma(c("a", "b", "c"), width = 3, verbose = FALSE),
        "x.*numeric"
    )
})

test_that("filter_ma() validates t as numeric", {
    expect_error(
        filter_ma(
            1:5,
            t = letters[1:5],
            width = 3,
            verbose = FALSE
        ),
        "t.*numeric"
    )
})

test_that("filter_ma() validates x and t have same length", {
    expect_error(
        filter_ma(1:5, t = 1:3, width = 3, verbose = FALSE),
        "equal length"
    )
})

test_that("filter_ma() validates width & span constraints", {
    x <- 1:10

    ## width or span required
    expect_error(
        filter_ma(x, verbose = FALSE),
        "width.*or.*span.*must be defined"
    )

    # Width must be positive integer
    expect_error(
        filter_ma(x, width = -1, verbose = FALSE),
        "width.*integer"
    )

    ## width = 0
    expect_error(
        filter_ma(x, width = 0, verbose = FALSE),
        "width.*integer"
    )
    ## width > length(x)
    expect_true(all(
        filter_ma(x, width = 21, partial = TRUE) == mean(x)
    ))

    # Span must be positive
    expect_error(
        filter_ma(x, span = -1, verbose = FALSE),
        "span.*numeric"
    )

    ## span = 0
    expect_equal(filter_ma(x, span = 0, verbose = FALSE), x)
    ## span = 0.5 takes nearest 1 sample
    expect_equal(filter_ma(x, span = 0.5, verbose = FALSE), x)
    ## span > length(x)
    expect_true(all(
        filter_ma(x, span = 21, partial = TRUE) == mean(x)
    ))
})

test_that("filter_ma() warns when both width and span provided", {
    ## should produce warning
    expect_message(
        filter_ma(1:5, width = 3, span = 2),
        "width.*span"
    )

    old_verbose <- getOption("mnirs.verbose")
    on.exit(options(mnirs.verbose = old_verbose), add = TRUE)
    options(mnirs.verbose = FALSE)

    ## should not produce warning
    expect_silent(
        result <- filter_ma(1:5, width = 3, span = 2)
    )
})

test_that("filter_ma() handles edge cases", {
    # Single value
    expect_equal(
        filter_ma(5, width = 1, verbose = FALSE),
        5,
        ignore_attr = TRUE
    )

    # Two values
    expect_equal(
        filter_ma(c(1, 2), width = 1, verbose = FALSE),
        c(1, 2)
    )

    # All NA
    expect_error(
        filter_ma(c(NA, NA, NA), width = 3, verbose = FALSE),
        "x.*numeric"
    )
})

test_that("filter_ma with insufficient width returns NA with warning", {
    t <- 1:10
    x <- c(1, NA, 3, NA, 5, NA, 7, NA, 9, NA)

    expect_all_equal(filter_ma(x, t, width = 2), NA_real_) |>
        expect_warning("Set.*na.rm = TRUE")

    x <- c(NA, 1, NA)
    expect_error(
        filter_ma(x, width = 3, na.rm = TRUE),
        "Insufficient.*samples"
    )
})

## filter_butter() =========================================
test_that("filter_butter validates inputs correctly", {
    expect_error(
        filter_butter(x = "not_numeric", order = 1, W = 0.1),
        "x.*numeric"
    )

    expect_error(
        filter_butter(x = 1:10, order = 0, W = 0.1),
        "n.*integer"
    )

    expect_error(
        filter_butter(x = 1:10, order = -1, W = 0.1),
        "n.*integer"
    )

    expect_error(
        filter_butter(x = 1:10, order = 1.5, W = 0.1),
        "n.*integer"
    )

    expect_error(
        filter_butter(x = 1:10, order = 1, W = 1.5),
        "W.*numeric"
    )

    expect_error(
        filter_butter(x = 1:10, order = 1, W = 0),
        "W.*numeric"
    )

    expect_error(
        filter_butter(x = 1:10, order = 1, W = c(0.1, 0.3), type = "low"),
        "W.*numeric"
    )
})

test_that("filter_butter returns correct output structure", {
    x <- rnorm(100)
    result <- filter_butter(x, order = 2, W = 0.1)

    expect_type(result, "double")
    expect_length(result, length(x))
    expect_false(anyNA(result))
})

test_that("filter_butter handles different edge options", {
    set.seed(42)
    x <- sin(2 * pi * 1:50 / 10) + rnorm(50, 0, 0.5)

    result_none <- filter_butter(x, order = 2, W = 0.1, edges = "none")
    result_rev <- filter_butter(x, order = 2, W = 0.1, edges = "rev")
    result_rep1 <- filter_butter(x, order = 2, W = 0.1, edges = "rep1")

    expect_length(result_none, length(x))
    expect_length(result_rev, length(x))
    expect_length(result_rep1, length(x))

    # Results should differ at edges
    expect_false(identical(result_none, result_rev))
    expect_false(identical(result_none, result_rep1))
    expect_false(identical(result_rev, result_rep1))
})

test_that("filter_butter handles NA values", {
    x_with_na <- c(1:5, NA, 7:10)

    # NA propagates through filtering
    expect_error(
        filter_butter(x_with_na, order = 1, W = 0.1, na.rm = FALSE),
        "x.*NA"
    )

    result <- filter_butter(x_with_na, order = 1, W = 0.1, na.rm = TRUE)
    expect_true(is.na(result[6]))

    # All NA input
    expect_error(
        filter_butter(rep(NA_real_, 5), order = 1, W = 0.1, na.rm = TRUE),
        "x.*valid.*numeric"
    )

    # Leading/trailing NAs
    x_edge_na <- c(NA, NA, 3:8, NA, NA)
    result_edge <- filter_butter(x_edge_na, order = 1, W = 0.1, na.rm = TRUE)
    expect_true(all(is.na(result_edge[c(1:2, 9:10)])))
    expect_length(result_edge, length(x_edge_na))
})

test_that("filter_butter handles different filter types", {
    set.seed(123)
    x <- rnorm(100)

    low <- filter_butter(x, order = 2, W = 0.1, type = "low")
    high <- filter_butter(x, order = 2, W = 0.9, type = "high")
    stop <- filter_butter(x, order = 2, W = c(0.3, 0.7), type = "stop")
    pass <- filter_butter(x, order = 2, W = c(0.3, 0.7), type = "pass")

    expect_length(low, length(x))
    expect_length(high, length(x))
    expect_length(stop, length(x))
    expect_length(pass, length(x))

    ## wrong W elements
    expect_error(
        filter_butter(x, order = 2, W = c(0.3, 0.7), type = "low"),
        "W.*1-element.*numeric"
    )
    expect_error(
        filter_butter(x, order = 2, W = 0.9, type = "pass"),
        "W.*2-element.*numeric"
    )
})

test_that("filter_butter handles edge cases", {
    # Very short vector
    x_short <- c(1, 2, 3)
    result <- filter_butter(x_short, order = 1, W = 0.5)
    expect_length(result, 3)

    # Single value
    x_single <- 5
    result_single <- filter_butter(x_single, order = 1, W = 0.5)
    expect_length(result_single, 1)

    # Constant signal
    x_const <- rep(10, 50)
    result_const <- filter_butter(x_const, order = 1, W = 0.5)
    expect_true(all(abs(result_const - 10) < 1e-10))
})

test_that("filter_butter smooths noisy signal", {
    set.seed(999)
    sin_wave <- sin(2 * pi * 1:100 / 20)
    x <- seq_along(sin_wave)
    noisy <- sin_wave + rnorm(100, 0, 0.3)
    filtered <- filter_butter(noisy, order = 2, W = 0.2)

    # Filtered signal should be smoother (lower variance)
    expect_lt(var(diff(filtered)), var(diff(noisy)))
})

test_that("filter_butter works visually", {
    skip("visual check for filter_butter")
    set.seed(999)
    sin_wave <- sin(2 * pi * 1:100 / 20)
    x <- seq_along(sin_wave)
    noisy <- sin_wave + rnorm(100, 0, 0.3)
    filtered <- filter_butter(noisy, order = 2, W = 0.2)

    ggplot2::ggplot(tibble::tibble()) +
        ggplot2::aes(x = x) +
        ggplot2::geom_line(ggplot2::aes(y = sin_wave, colour = "sin")) +
        ggplot2::geom_line(ggplot2::aes(y = noisy, colour = "noisy")) +
        ggplot2::geom_line(ggplot2::aes(y = filtered, colour = "filtered"))
})


## moxy_data for filter_mnirs() ==============================================
moxy_data <- read_mnirs(
    file_path = example_mnirs("moxy_ramp.xlsx"),
    nirs_channels = c(smo2_left = "SmO2 Live", smo2_right = "SmO2 Live(2)"),
    time_channel = c(time = "hh:mm:ss"),
    verbose = FALSE
) |>
    resample_mnirs(method = "linear", verbose = FALSE)

## Input validation tests =======================================
test_that("filter_mnirs validates input data", {
    expect_error(
        filter_mnirs(data.frame(x = 1:10)),
        "`data`.*data.*frame"
    )
})

test_that("filter_mnirs method aliases work", {
    result_ma <- filter_mnirs(
        moxy_data, method = "moving_average", width = 5, verbose = FALSE
    )
    result_spl <- filter_mnirs(
        moxy_data, method = "smooth_spline", verbose = FALSE
    )

    expect_equal(
        filter_mnirs(
            moxy_data, method = "moving average", width = 5, verbose = FALSE
        ),
        result_ma
    )
    expect_equal(
        filter_mnirs(
            moxy_data, method = "Moving Average", width = 5, verbose = FALSE
        ),
        result_ma
    )
    expect_equal(
        filter_mnirs(
            moxy_data, method = "MA", width = 5, verbose = FALSE
        ),
        result_ma
    )
    expect_equal(
        filter_mnirs(moxy_data, verbose = FALSE),
        result_spl
    )
    expect_equal(
        filter_mnirs(moxy_data, method = "spline", verbose = FALSE),
        result_spl
    )
    expect_equal(
        filter_mnirs(moxy_data, method = "smooth spline", verbose = FALSE),
        result_spl
    )
    expect_equal(
        filter_mnirs(moxy_data, method = "Smooth Spline", verbose = FALSE),
        result_spl
    )
})

test_that("filter_mnirs validates method argument", {
    expect_error(
        filter_mnirs(moxy_data, method = "invalid", verbose = FALSE),
        "should be one of"
    )
})

## Smooth spline tests ==============================================
test_that("smooth_spline filters data correctly", {
    result <- filter_mnirs(moxy_data, method = "smooth_spline", verbose = FALSE)

    expect_s3_class(result, "mnirs")
    expect_equal(nrow(result), nrow(moxy_data))
    expect_true(all(c("time", "smo2_left", "smo2_right") %in% names(result)))

    # Filtered data should be smoother (lower variance)
    expect_lt(var(result$smo2_left), var(moxy_data$smo2_left))
    expect_lt(var(result$smo2_right), var(moxy_data$smo2_right))
})

test_that("smooth_spline respects spar parameter", {
    result_low <- filter_mnirs(
        moxy_data,
        method = "smooth_spline",
        spar = 0.1,
        verbose = FALSE
    )
    result_high <- filter_mnirs(
        moxy_data,
        method = "smooth_spline",
        spar = 0.9,
        verbose = FALSE
    )

    # Higher spar = more smoothing
    expect_lt(var(result_high$smo2_left), var(result_low$smo2_left))
    expect_lt(var(result_high$smo2_right), var(result_low$smo2_right))
})

test_that("smooth_spline handles NAs", {
    moxy_data <- read_mnirs(
        file_path = example_mnirs("moxy_ramp.xlsx"),
        nirs_channels = c(smo2_left = "SmO2 Live", smo2_right = "SmO2 Live(2)"),
        time_channel = c(time = "hh:mm:ss"),
        verbose = FALSE
    ) |>
        resample_mnirs(method = "none", verbose = FALSE)

    result <- filter_mnirs(
        moxy_data,
        method = "smooth_spline",
        na.rm = TRUE,
        verbose = FALSE
    )

    expect_true(anyNA(result$smo2_left))
    expect_true(anyNA(result$smo2_right))
    expect_equal(
        which(is.na(result$smo2_right)),
        which(is.na(moxy_data$smo2_right))
    )

    expect_error(
        filter_mnirs(
            moxy_data, method = "smooth_spline", na.rm = FALSE,
            verbose = FALSE
        ),
        "NA"
    )
})

test_that("smooth_spline errors with irregular samples", {
    moxy_data <- read_mnirs(
        file_path = example_mnirs("moxy_ramp.xlsx"),
        nirs_channels = c(smo2_left = "SmO2 Live", smo2_right = "SmO2 Live(2)"),
        time_channel = c(time = "hh:mm:ss"),
        verbose = FALSE
    )

    expect_error(
        filter_mnirs(
            moxy_data, method = "smooth_spline", verbose = FALSE
        ),
        "irregular samples"
    )
})

test_that("smooth_spline validates spar parameter", {
    expect_error(
        filter_mnirs(
            moxy_data, method = "smooth_spline", spar = -1,
            verbose = FALSE
        ),
        "spar.*numeric"
    )

    expect_error(
        filter_mnirs(
            moxy_data, method = "smooth_spline", spar = c(0.1, 0.5),
            verbose = FALSE
        ),
        "spar.*numeric"
    )
})

## Butterworth tests ==============================================
test_that("butterworth low-pass filter works", {
    result <- filter_mnirs(
        moxy_data,
        method = "butterworth",
        type = "low",
        order = 2,
        W = 0.1,
        verbose = FALSE
    )

    expect_s3_class(result, "mnirs")
    expect_equal(nrow(result), nrow(moxy_data))

    # Low-pass should reduce high-frequency variance
    expect_lt(var(diff(result$smo2_left)), var(diff(moxy_data$smo2_left)))
    expect_lt(var(diff(result$smo2_right)), var(diff(moxy_data$smo2_right)))
})

test_that("butterworth accepts fc instead of W", {
    sr <- attr(moxy_data, "sample_rate")

    result_W <- filter_mnirs(
        moxy_data,
        method = "butterworth",
        type = "low",
        order = 2,
        W = 0.1,
        verbose = FALSE
    )

    result_fc <- filter_mnirs(
        moxy_data,
        method = "butterworth",
        type = "low",
        order = 2,
        fc = 0.1 * sr * 0.5,
        verbose = FALSE
    )

    expect_equal(result_W$smo2_left, result_fc$smo2_left)
})

test_that("butterworth prefers W when both W and fc specified", {
    expect_message(
        filter_mnirs(
            moxy_data,
            method = "butterworth",
            type = "low",
            order = 2,
            W = 0.1,
            fc = 5
        ),
        ".*W.*fc"
    )
})

test_that("butterworth errors without W or fc", {
    expect_error(
        filter_mnirs(
            moxy_data,
            method = "butterworth",
            type = "low",
            order = 2,
            verbose = FALSE
        ),
        "W.*fc.*must be defined"
    )
})

test_that("butterworth errors for invalid fc", {
    expect_error(
        filter_mnirs(
            moxy_data,
            method = "butterworth",
            type = "low",
            order = 2,
            fc = 100,
            verbose = FALSE
        ),
        "must be between.*0.*half"
    )
})

test_that("butterworth validates filter type", {
    expect_error(
        filter_mnirs(
            moxy_data,
            method = "butterworth",
            type = "invalid",
            W = 0.1,
            verbose = FALSE
        ),
        "should be one of"
    )
})

test_that("butterworth handles different filter types", {
    result_low <- filter_mnirs(
        moxy_data,
        method = "butterworth",
        type = "low",
        order = 2,
        W = 0.1,
        verbose = FALSE
    )

    result_high <- filter_mnirs(
        moxy_data,
        method = "butterworth",
        type = "high",
        order = 2,
        W = 0.1,
        verbose = FALSE
    )

    result_stop <- filter_mnirs(
        moxy_data,
        method = "butterworth",
        type = "stop",
        order = 2,
        W = c(0.05, 0.15),
        verbose = FALSE
    )

    result_pass <- filter_mnirs(
        moxy_data,
        method = "butterworth",
        type = "pass",
        order = 2,
        W = c(0.05, 0.15),
        verbose = FALSE
    )

    expect_s3_class(result_low, "mnirs")
    expect_s3_class(result_high, "mnirs")
    expect_s3_class(result_stop, "mnirs")
    expect_s3_class(result_pass, "mnirs")
})

test_that("butterworth validates W length for stop/pass filters", {
    expect_error(
        filter_mnirs(
            moxy_data,
            method = "butterworth",
            type = "stop",
            order = 2,
            W = 0.5,
            verbose = FALSE
        ),
        "W.*numeric"
    )

    expect_error(
        filter_mnirs(
            moxy_data,
            method = "butterworth",
            type = "pass",
            order = 2,
            fc = 5,
            verbose = FALSE
        ),
        "fc.*numeric"
    )
})

test_that("butterworth handles NAs with na.rm = TRUE", {
    moxy_data <- read_mnirs(
        file_path = example_mnirs("moxy_ramp.xlsx"),
        nirs_channels = c(smo2_left = "SmO2 Live", smo2_right = "SmO2 Live(2)"),
        time_channel = c(time = "hh:mm:ss"),
        verbose = FALSE
    ) |>
        resample_mnirs(method = "none", verbose = FALSE)

    result <- filter_mnirs(
        moxy_data,
        method = "butterworth",
        type = "low",
        order = 2,
        W = 0.1,
        na.rm = TRUE,
        verbose = FALSE
    )

    expect_true(anyNA(result$smo2_left))
    expect_true(anyNA(result$smo2_right))
    expect_equal(
        which(is.na(result$smo2_right)),
        which(is.na(moxy_data$smo2_right))
    )

    expect_error(
        filter_mnirs(
            moxy_data,
            method = "butterworth",
            type = "low",
            order = 2,
            W = 0.1,
            na.rm = FALSE,
            verbose = FALSE
        ),
        "NA"
    )
})

## Moving average tests ==============================================
test_that("moving_average with width works", {
    result <- filter_mnirs(
        moxy_data,
        method = "moving_average",
        width = 10,
        verbose = FALSE
    )

    expect_s3_class(result, "mnirs")
    expect_equal(nrow(result), nrow(moxy_data))

    # Should smooth the data
    expect_lt(
        var(diff(result$smo2_left), na.rm = TRUE),
        var(diff(moxy_data$smo2_left), na.rm = TRUE)
    )
    expect_lt(
        var(diff(result$smo2_right), na.rm = TRUE),
        var(diff(moxy_data$smo2_right), na.rm = TRUE)
    )
})

test_that("moving_average with span works", {
    result <- filter_mnirs(
        moxy_data,
        method = "moving_average",
        span = 1,
        verbose = FALSE
    )

    expect_s3_class(result, "mnirs")
    expect_equal(nrow(result), nrow(moxy_data))
})

test_that("moving_average with larger width = more smoothing", {
    result_narrow <- filter_mnirs(
        moxy_data,
        method = "moving_average",
        width = 5,
        verbose = FALSE
    )

    result_wide <- filter_mnirs(
        moxy_data,
        method = "moving_average",
        width = 20,
        verbose = FALSE
    )

    expect_lt(
        var(result_wide$smo2_left, na.rm = TRUE),
        var(result_narrow$smo2_left, na.rm = TRUE)
    )
})

## Channel selection tests ==============================================
test_that("filter_mnirs respects nirs_channels argument", {
    result <- filter_mnirs(
        moxy_data,
        nirs_channels = "smo2_right",
        method = "smooth_spline",
        spar = 0.5,
        verbose = FALSE
    )

    # channel1 should be filtered
    expect_false(identical(result$smo2_right, moxy_data$smo2_right))
    expect_lt(var(diff(result$smo2_right)), var(diff(moxy_data$smo2_right)))

    expect_true(identical(result$smo2_left, moxy_data$smo2_left))
    expect_equal(var(diff(result$smo2_left)), var(diff(moxy_data$smo2_left)))

    expect_equal(attr(result, "nirs_channels"), "smo2_right")

    expect_false(isTRUE(all.equal(
        attr(result, "nirs_channels"),
        attr(moxy_data, "nirs_channels")
    )))
})

## Metadata preservation tests ==============================================
test_that("filter_mnirs preserves and updates metadata", {
    original_meta <- attributes(moxy_data)
    result <- filter_mnirs(moxy_data, method = "smooth_spline", verbose = FALSE)
    result_meta <- attributes(result)

    expect_equal(result_meta$nirs_channels, original_meta$nirs_channels)
    expect_equal(result_meta$time_channel, original_meta$time_channel)
})

test_that("butterworth updates sample_rate in metadata", {
    result <- filter_mnirs(
        moxy_data,
        method = "butterworth",
        type = "low",
        order = 2,
        W = 0.1,
        sample_rate = 10,
        verbose = FALSE
    )

    expect_equal(attr(result, "sample_rate"), 10)
    expect_false(attr(result, "sample_rate") == attr(moxy_data, "sample_rate"))
})

## verbose output tests ==============================================
test_that("verbose output works", {
    expect_message(
        filter_mnirs(moxy_data, method = "smooth_spline", verbose = TRUE),
        "spar ="
    ) |>
        expect_message("spar =")

    expect_silent(
        filter_mnirs(moxy_data, method = "smooth_spline", verbose = FALSE)
    )

    expect_message(
        filter_mnirs(
            moxy_data,
            method = "butterworth",
            type = "low",
            order = 2,
            W = 0.1,
            fc = 0.1,
            verbose = TRUE
        ),
        "W.*overrides.*fc"
    )

    expect_silent(
        filter_mnirs(
            moxy_data,
            method = "butterworth",
            type = "low",
            order = 2,
            W = 0.1,
            fc = 0.1,
            verbose = FALSE
        )
    )
})

## Integration tests ==============================================
test_that("filter_mnirs works with pipe", {
    result <- moxy_data |>
        filter_mnirs(method = "smooth_spline", verbose = FALSE)

    expect_s3_class(result, "mnirs")
    expect_equal(nrow(result), nrow(moxy_data))
    expect_lt(var(diff(result$smo2_left)), var(diff(moxy_data$smo2_left)))
})

test_that("multiple filtering operations can be chained", {
    result <- moxy_data |>
        filter_mnirs(
            method = "butterworth",
            type = "high",
            order = 1,
            W = 0.01,
            verbose = FALSE
        ) |>
        filter_mnirs(method = "smooth_spline", spar = 0.3, verbose = FALSE)

    expect_s3_class(result, "mnirs")
    expect_equal(nrow(result), nrow(moxy_data))
})

test_that("filter_mnirs works visually on Moxy data", {
    skip("visual check for filter_mnirs() on Moxy data")

    data <- read_mnirs(
        file_path = example_mnirs("moxy_ramp.xlsx"),
        nirs_channels = c(smo2 = "SmO2 Live(2)"),
        time_channel = c(time = "hh:mm:ss"),
        verbose = FALSE
    ) |>
        resample_mnirs(method = "linear", verbose = FALSE) |>
        dplyr::mutate(
            dplyr::across(
                smo2,
                \(.x) filter_butter(.x, 2, 0.05),
                .names = "{.col}_filtfilt"
            )
        )

    data_filt <- filter_mnirs(
        data,
        nirs_channels = "smo2",
        time_channel = NULL,
        sample_rate = NULL,
        method = "moving_average", #c("smooth_spline", "butterworth", "moving_average"),
        # spar = 0.5,
        # type = c("low"),
        # order = 2,
        # W = 0.05,
        # fc = NULL,
        width = NULL,
        span = 30,
        # na.rm = FALSE,
        verbose = TRUE
    )

    ## visual check
    plot(data) +
        ggplot2::scale_colour_manual(
            breaks = c("smo2", "filt_vec", "filt_mnirs"),
            values = palette_mnirs()
        ) +
        ggplot2::geom_line(
            ggplot2::aes(y = smo2_filtfilt, colour = "filt_vec"),
            linewidth = 1
        ) +
        ggplot2::geom_line(
            data = data_filt,
            ggplot2::aes(y = smo2, colour = "filt_mnirs")
        )
})

#' Specify interval boundaries by time, label, lap, or sample
#'
#' Helper functions to define interval start or end boundaries for
#' [extract_intervals()].
#'
#' @param ... Specify start or end boundaries.
#'   \describe{
#'     \item{`by_time(...)`}{Numeric time values in units of `time_channel`.}
#'     \item{`by_label(...)`}{Character strings to match in `event_channel`.
#'     All matching occurrences are returned.}
#'     \item{`by_lap(...)`}{Integer lap numbers to match in `event_channel`.
#'     For `start`, resolves to the first sample of each lap. For `end`,
#'     resolves to the last sample.}
#'     \item{`by_sample(...)`}{Integer sample indices (row numbers).}
#'   }
#'
#' @details
#' These helpers can be used explicitly for arguments `start`/`end`, or raw
#' values can be passed directly:
#'   - Numeric -> [by_time()]
#'   - Character -> [by_label()],
#'   - Explicit integer (e.g. `2L`) -> [by_lap()].
#'   - Use [by_sample()] explicitly for sample indices.
#'
#' @returns An object of class `"mnirs_interval"` for use with the `start`
#'   and `end` arguments of [extract_intervals()].
#'
#' @examples
#' ## read example data
#' data <- read_mnirs(
#'     example_mnirs("train.red"),
#'     nirs_channels = c(
#'         smo2_left = "SmO2 unfiltered",
#'         smo2_right = "SmO2 unfiltered"
#'     ),
#'     time_channel = c(time = "Timestamp (seconds passed)"),
#'     event_channel = c(lap = "Lap/Event"),
#'     zero_time = TRUE,
#'     verbose = FALSE
#' )
#'
#' ## start and end by time
#' extract_intervals(data, start = by_time(66), end = by_time(357))
#'
#' ## start by lap
#' extract_intervals(data, start = by_lap(2, 4), span = 0)
#'
#' ## introduce event_channel with "start" string
#' data$event <- NA_character_
#' data$event[1000] <- "start"
#' data <- create_mnirs_data(data, event_channel = "event")
#'
#' ## start by label, end by time
#' extract_intervals(data, start = by_label("start"), end = by_time(1500))
#'
#' ## multiple intervals by sample index
#' extract_intervals(data, start = by_sample(1000, 1500), end = by_sample(2000, 2600))
#'
#' @export
by_time <- function(...) {
    by_time <- c(...)
    validate_numeric(by_time)
    structure(
        list(type = "time", by_time = by_time),
        class = "mnirs_interval"
    )
}


#' @rdname by_time
#' @export
by_label <- function(...) {
    by_label <- c(...)
    if (!is.character(by_label) || length(by_label) == 0L) {
        cli_abort("{.fn by_label} must be a valid {.cls character} vector.")
    }
    structure(
        list(type = "label", by_label = by_label),
        class = "mnirs_interval"
    )
}


#' @rdname by_time
#' @export
by_lap <- function(...) {
    by_lap <- c(...)
    validate_numeric(by_lap, range = c(1, Inf), integer = TRUE)
    structure(
        list(type = "lap", by_lap = as.integer(by_lap)),
        class = "mnirs_interval"
    )
}


#' @rdname by_time
#' @export
by_sample <- function(...) {
    by_sample <- c(...)
    validate_numeric(by_sample, range = c(1, Inf), integer = TRUE)
    structure(
        list(type = "sample", by_sample = as.integer(by_sample)),
        class = "mnirs_interval"
    )
}


#' coerce raw values to mnirs_interval objects
#' @param x A raw value or mnirs_interval object.
#' @param arg Name of the argument for error messages.
#' @keywords internal
as_mnirs_interval <- function(x, arg = "start") {
    if (is.null(x) || inherits(x, "mnirs_interval")) {
        return(x)
    }
    ## integer before numeric — integers are also numeric in R
    if (is.integer(x)) {
        return(by_lap(x))
    }
    if (is.numeric(x)) {
        return(by_time(x))
    }
    if (is.character(x)) {
        return(by_label(x))
    }
    cli_abort(c(
        "x" = "{.arg {arg}} must be {.cls numeric}, {.cls integer}, \\
        {.cls character}, or a {.fn by_time}, {.fn by_sample}, \\
        {.fn by_label}, {.fn by_lap} specification."
    ))
}


#' recycle a single-element span to c(before, after)
#' positive -> c(0, x), negative -> c(x, 0)
#' @keywords internal
recycle_span <- function(span) {
    if (is.numeric(span) && length(span) == 2L) {
        return(span)
    }
    if (!is.numeric(span) || length(span) != 1L) {
        cli_abort(c(
            "x" = "{.arg span} must be a one- or two-element \\
            {.cls numeric} vector."
        ))
    }
    ## positive shifts end, negative shifts start
    return(if (span >= 0) c(0, span) else c(span, 0))
}


#' resolve a single mnirs_interval object to time values
#' @keywords internal
find_interval_time <- function(
    interval,
    time_vec,
    event_vec = NULL,
    position = c("first", "last")
) {
    switch(
        interval$type,
        time = interval$by_time,
        sample = time_vec[interval$by_sample],
        label = {
            pattern <- paste(interval$by_label, collapse = "|")
            matches <- which(grepl(pattern, event_vec))
            if (length(matches) == 0L) {
                cli_abort(c(
                    "x" = "No events detected matching \\
                    {.val {interval$by_label}}.",
                    "i" = "Must match contents of {.arg event_channel} exactly."
                ))
            }
            time_vec[matches]
        },
        lap = {
            ## convert once outside loop
            event_int <- as.integer(event_vec)
            vapply(interval$by_lap, \(lap_val) {
                matches <- which(event_int == lap_val)
                if (length(matches) == 0L) {
                    cli_abort(c(
                        "x" = "No samples found for lap {.val {lap_val}}.",
                        "i" = "Check that {.arg event_channel} contains \\
                        lap numbers."
                    ))
                }
                idx <- if (position == "first") 1L else length(matches)
                time_vec[matches[idx]]
            }, numeric(1))
        }
    )
}


#' resolve start/end into time value vectors (no span applied)
#' @keywords internal
resolve_interval <- function(
    start,
    end,
    time_vec,
    event_vec = NULL
) {
    has_start <- !is.null(start)
    has_end <- !is.null(end)
    interval <- start %||% end

    ## single-boundary non-lap: reference point for span
    if (xor(has_start, has_end)) {
        start_time <- find_interval_time(
            interval,
            time_vec,
            event_vec,
            position = if (has_start) "first" else "last"
        )
        return(list(
            start_time = start_time,
            end_time = NULL,
            has_start = has_start,
            has_end = has_end
        ))
    }

    ## both boundaries specified
    start_time <- find_interval_time(start, time_vec, event_vec, "first")
    end_time <- find_interval_time(end, time_vec, event_vec, "last")

    ## warn and truncate if lengths differ
    n_start <- length(start_time)
    n_end <- length(end_time)
    if (n_start != n_end) {
        n_intervals <- min(n_start, n_end)
        cli_warn(c(
            "!" = "Unequal lengths for {.arg start} ({col_blue(n_start)}) \\
            and {.arg end} ({col_blue(n_end)}).",
            "i" = "Returning {col_blue(n_intervals)} paired interval{?s}."
        ))
        start_time <- start_time[seq_len(n_intervals)]
        end_time <- end_time[seq_len(n_intervals)]
    }

    return(list(
        start_time = start_time,
        end_time = end_time,
        has_start = TRUE,
        has_end = TRUE
    ))
}


#' Recycle parameter list to target length
#' @keywords internal
recycle_to_length <- function(
    param,
    n,
    name = c("event", "group"),
    verbose = TRUE
) {
    n_param <- length(param)

    if (n_param == n) {
        return(param)
    }

    if (n_param > n) {
        if (verbose) {
            cli_inform(c(
                "!" = "{.arg {substitute(param)}} exceeds the number of \\
                {name}s by {.val {n_param - n}}.",
                "i" = "Extra values are ignored."
            ))
        }
        return(param[seq_len(n)])
    }

    ## n_param < n:  recycle last element forward
    if (verbose && n_param > 1L) {
        cli_inform(c(
            "i" = "{.arg {substitute(param)}} recycled to meet \\
            {.val {n - n_param}} unspecified {name}{qty(n - n_param)}{?s}."
        ))
    }
    return(param[c(seq_len(n_param), rep(n_param, n - n_param))])
}


#' Recycle parameter to match number of events
#'
#' Recycle an argument vector to a list or repeat the last list item to match
#' the number of events.
#'
#' @keywords internal
recycle_param <- function(
    param,
    n_events,
    event_groups,
    verbose = TRUE
) {
    ## flatten nested lists to single-depth list
    param <- if (is.list(param)) {
        lapply(param, \(.x) if (is.list(.x)) unlist(.x) else .x)
    } else {
        list(param)
    }

    ## custom grouping: recycle per group, then map to event order
    if (is.numeric(event_groups[[1L]])) {
        n_groups <- length(event_groups)
        groups_unlisted <- unlist(event_groups)

        ## recycle param to number of groups
        param <- recycle_to_length(param, n_groups, "group", verbose)

        ## create mapping:  event_id -> group_id
        ## rep(1:n_groups, lengths(event_groups)) gives group index per event in event_groups
        group_for_event <- rep(seq_len(n_groups), lengths(event_groups))

        ## build lookup:  position i holds group index for event i (NA if ungrouped)
        event_to_group <- integer(n_events)
        valid_events <- groups_unlisted[groups_unlisted <= n_events]
        valid_groups <- group_for_event[groups_unlisted <= n_events]
        event_to_group[valid_events] <- valid_groups

        ## fill ungrouped (zero) positions with last group index
        event_to_group[event_to_group == 0L] <- n_groups

        ## index into param by group assignment
        return(param[event_to_group])
    }

    ## standard recycling for "distinct" or "ensemble"
    return(recycle_to_length(param, n_events, "event", verbose))
}


#' apply span to resolved times and build interval_spec data frame
#' @keywords internal
apply_span <- function(
    interval_list,
    time_vec,
    span,
    verbose = TRUE
) {
    ## extract span values (recycled by recycle_param)
    span_before <- vapply(span, `[`, numeric(1), 1L)
    span_after <- vapply(span, `[`, numeric(1), 2L)
    event_times <- interval_list$start_time

    if (interval_list$has_start && interval_list$has_end) {
        ## span[1] shifts starts, span[2] shifts ends
        start_times <- event_times + span_before
        end_times <- interval_list$end_time + span_after
        ## original boundary times for metadata
        interval_times <- Map(c, event_times, interval_list$end_time)
    } else {
        ## single-boundary: both span values on reference
        start_times <- event_times + span_before
        end_times <- event_times + span_after
        interval_times <- as.list(event_times)
    }

    ## data time bounds
    t_min <- time_vec[1L]
    t_max <- time_vec[length(time_vec)]

    ## check for entirely out of bounds intervals
    entirely_oob <- end_times < t_min | start_times > t_max
    if (any(entirely_oob)) {
        oob_ids <- which(entirely_oob)
        n_oob <- qty(length(oob_ids))
        cli_abort(c(
            "x" = "{n_oob} Interval{?s} {.val {oob_ids}} {n_oob} \\
            {?is/are} entirely outside data bounds.",
            "i" = "Intervals must be specified within existing data bounds."
        ))
    }

    ## clamp partial out of bounds to data range
    partial_oob <- start_times < t_min | end_times > t_max
    if (any(partial_oob)) {
        start_times <- pmax(t_min, start_times)
        end_times <- pmin(t_max, end_times)

        if (verbose) {
            oob_ids <- which(partial_oob)
            n_oob <- qty(length(oob_ids))
            cli_warn(c(
                "!" = "{n_oob} Interval{?s} {.val {oob_ids}} {n_oob} \\
                {?is/are} partially outside data bounds.",
                "i" = "Returning available data only."
            ))
        }
    }

    result <- data.frame(
        span_before = span_before,
        span_after = span_after,
        start_times = start_times,
        end_times = end_times,
        stringsAsFactors = FALSE
    )
    result$interval_times <- interval_times
    return(result)
}


#' Extract interval data by time range
#' @keywords internal
extract_df_list <- function(
    data,
    time_vec,
    interval_spec,
    nirs_channels ## as list
) {
    n_vec <- seq_len(nrow(interval_spec))
    df_list <- lapply(n_vec, \(.i) {
        time_range <- time_vec >= interval_spec$start_times[.i] &
            time_vec <= interval_spec$end_times[.i]
        interval_data <- data[time_range, , drop = FALSE]

        ## return interval_data with metadata
        create_mnirs_data(
            interval_data,
            nirs_channels = nirs_channels[[.i]], ## overwrite for interval data
            interval_times = interval_spec$interval_times[[.i]],
            interval_span = c(
                interval_spec$span_before[.i],
                interval_spec$span_after[.i]
            )
        )
    })

    names(df_list) <- sprintf("interval_%d", n_vec)
    return(df_list)
}


#' Recalculate time_channel values with zero offset at event time (t0)
#' @keywords internal
zero_offset_data <- function(data, time_channel, t0) {
    ## zero time channel to start_time `t0`
    data[[time_channel]] <- data[[time_channel]] - t0
    return(data)
}


#' Ensemble average multiple intervals
#' @keywords internal
ensemble_intervals <- function(
    df_list,
    nirs_channels,
    metadata,
    verbose = TRUE
) {
    time_channel <- metadata$time_channel
    sample_rate <- metadata$sample_rate
    ## extract data & metadata from df_list
    interval_data <- lapply(df_list, \(.df) {
        interval_times <- attr(.df, "interval_times")
        time_channel <- attr(.df, "time_channel")
        ## zero-offset to start time (first element of interval_times)
        t0 <- interval_times[[1L]]
        ## return data & metadata from each interval
        list(
            ## ensemble-average time values makes no sense, so return zero-offset
            data = zero_offset_data(.df, time_channel, t0),
            interval_times = interval_times - t0, ## zero_time recalc
            interval_span = attr(.df, "interval_span")
        )
    })

    ## stack interval data frames
    df_long <- do.call(rbind, lapply(interval_data, `[[`, "data"))
    ## resample times to nearest estimated sample rate for binned ensembling
    time_resampled <- round(df_long[[time_channel]] * sample_rate) / sample_rate
    ## split row indices by unique time
    time_groups <- split(seq_len(nrow(df_long)), time_resampled)
    unique_times <- as.numeric(names(time_groups))

    ## warn if any time samples have only one value, implying irregular
    ## samples and may result in alternating samples instead of ensemble-means
    if (verbose && min(lengths(time_groups), na.rm = TRUE) < 2) {
        cli_warn(c(
            "!" = "Duplicate or irregular {.arg time_channel} samples \\
            detected after ensemble-averaging.",
            "i" = "Check your resulting data for inconsistent results.",
            "i" = "Re-sample with {.fn mnirs::resample_mnirs}."
        ))
    }

    col_n <- length(nirs_channels)
    ## nirs_channel-wise means per unique time matrix operation
    ## nirs_channel must be vectorised and exist in the interval_data
    channel_matrix <- as.matrix(df_long[, nirs_channels, drop = FALSE])
    result_matrix <- vapply(time_groups, \(.idx) {
        colMeans(channel_matrix[.idx, , drop = FALSE], na.rm = TRUE)
    }, numeric(col_n))

    ## vapply returns channels x times (or vector if 1 channel)
    ## coerce to times x channels data frame
    if (col_n == 1L) {
        result_df <- data.frame(setNames(list(result_matrix), nirs_channels))
    } else {
        result_df <- as.data.frame(t(result_matrix))
        names(result_df) <- nirs_channels
    }
    result <- data.frame(setNames(list(unique_times), time_channel), result_df)

    ## return with metadata
    return(
        create_mnirs_data(
            result,
            nirs_device = metadata$nirs_device,
            nirs_channels = unique(nirs_channels),
            time_channel = time_channel,
            event_channel = metadata$event_channel,
            sample_rate = sample_rate,
            start_timestamp = metadata$start_timestamp,
            interval_times = lapply(interval_data, `[[`, "interval_times"),
            interval_span = lapply(interval_data, `[[`, "interval_span")
        )
    )
}

#' Zero-offset time values and add metadata
#' @keywords internal
preserve_metadata <- function(data, metadata, zero_time = FALSE) {
    if (zero_time) {
        ## Zero-offset a single interval's time and interval_times attr
        interval_times <- attr(data, "interval_times")
        t0 <- interval_times[[1L]]
        data <- zero_offset_data(data, metadata$time_channel, t0)
        attr(data, "interval_times") <- interval_times - t0
        data
    }

    return(
        create_mnirs_data(
            data,
            nirs_device = metadata$nirs_device,
            nirs_channels = unique(attr(data, "nirs_channels")),
            time_channel = metadata$time_channel,
            event_channel = metadata$event_channel,
            sample_rate = metadata$sample_rate,
            start_timestamp = metadata$start_timestamp,
            interval_times = attr(data, "interval_times"),
            interval_span = attr(data, "interval_span")
        )
    )
}


#' Apply grouping to intervals
#' @keywords internal
group_intervals <- function(
    df_list,
    nirs_channels,
    metadata,
    event_groups,
    zero_time = FALSE,
    verbose = TRUE
) {
    time_channel <- metadata$time_channel
    n_intervals <- length(df_list)

    ## return distinct intervals
    if (n_intervals == 1L || event_groups[[1L]][1L] == "distinct") {
        return(lapply(df_list, \(.x) {
            preserve_metadata(.x, metadata, zero_time)
        }))
    }

    ## return ensembled intervals
    if (event_groups[[1L]][1L] == "ensemble") {
        return(list(
            ensemble = ensemble_intervals(
                df_list,
                unique(unlist(nirs_channels)),
                metadata,
                verbose
            )
        ))
    }

    ## custom grouping ===================================
    grouped_ids <- unlist(event_groups)
    ungrouped_ids <- setdiff(seq_len(n_intervals), grouped_ids)

    ## add ungrouped ids as individual groups, sort by position
    if (length(ungrouped_ids) > 0) {
        event_groups <- c(event_groups, as.list(ungrouped_ids))
        event_groups <- event_groups[
            order(vapply(event_groups, median, numeric(1), na.rm = TRUE))
        ]
        if (verbose) {
            cli_inform(c(
                "!" = "Intervals not specified by {.arg event_groups}.",
                "i" = "Ungrouped intervals included as discrete."
            ))
        }
    }

    ## warn for duplicated intervals across groups
    dupes <- grouped_ids[duplicated(grouped_ids)]
    if (verbose && length(dupes) > 0) {
        cli_warn(c(
            "!" = "Duplicates detected of {qty(length(dupes))} \\
            interval{?s} {.val {dupes}}.",
            "i" = "Re-specify {.arg event_groups} to remove duplicates."
        ))
    }

    ## process each group
    result <- lapply(event_groups, \(.g) {
        if (length(.g) == 1L) {
            return(preserve_metadata(df_list[[.g]], metadata, zero_time))
        }
        ## multi-interval group: ensemble-average
        group_nirs <- unique(unlist(nirs_channels[.g]))
        ensemble_intervals(df_list[.g], group_nirs, metadata, verbose)
    })

    names(result) <- vapply(event_groups, \(.g) {
        paste0("interval_", paste(.g, collapse = "_"))
    }, character(1))

    return(result)
}

# Mock mnirs data with attributes
create_mock_mnirs <- function(n = 100, sample_rate = 10) {
    df <- tibble::tibble(
        time = seq(0, (n - 1) / sample_rate, length.out = n),
        smo2_left = sin(time * 0.5) * 10 + 50,
        smo2_right = cos(time * 0.5) * 10 + 50,
        event = c(rep("", 10), "start", rep("", n - 21), "end", rep("", 9))
    )
    structure(
        df,
        class = c("mnirs", class(df)),
        nirs_channels = c("smo2_left", "smo2_right"),
        time_channel = "time",
        event_channel = "event",
        sample_rate = sample_rate,
        nirs_device = "test_device"
    )
}

# Mock interval data (as returned by extract_df_list)
create_mock_interval <- function(
    time_start = 0,
    n = 50,
    sample_rate = 10,
    event_time = 0,
    span = c(-1, 4)
) {
    time_vec <- seq(
        time_start,
        time_start + (n - 1) / sample_rate,
        length.out = n
    )
    df <- tibble::tibble(
        time = time_vec,
        smo2_left = sin(time_vec) * 5 + 50,
        smo2_right = cos(time_vec) * 5 + 50
    )
    structure(
        df,
        class = c("mnirs", class(df)),
        nirs_channels = c("smo2_left", "smo2_right"),
        time_channel = "time",
        interval_times = event_time,
        interval_span = span,
        nirs_device = "test_device",
        event_channel = NULL
    )
}

## by_time(), by_sample(), by_label() constructors =======================
test_that("by_time creates mnirs_interval with correct structure", {
    result <- by_time(2, 5, 8)
    expect_s3_class(result, "mnirs_interval")
    expect_equal(result$type, "time")
    expect_equal(result$by_time, c(2, 5, 8))
})

test_that("by_sample creates mnirs_interval with correct structure", {
    result <- by_sample(10, 30, 70)
    expect_s3_class(result, "mnirs_interval")
    expect_equal(result$type, "sample")
    expect_equal(result$by_sample, c(10L, 30L, 70L))
})

test_that("by_sample validates input", {
    expect_error(by_sample(0), "valid.*integer")
    expect_error(by_sample(-1), "valid.*integer")
    expect_error(by_sample(1.5), "valid.*integer")
})

test_that("by_label creates mnirs_interval with correct structure", {
    result <- by_label("start", "end")
    expect_s3_class(result, "mnirs_interval")
    expect_equal(result$type, "label")
    expect_equal(result$by_label, c("start", "end"))
})

test_that("by_label validates input", {
    expect_error(by_label(123), "valid.*character")
})

test_that("by_lap creates mnirs_interval with correct structure", {
    result <- by_lap(1, 3, 5)
    expect_s3_class(result, "mnirs_interval")
    expect_equal(result$type, "lap")
    expect_equal(result$by_lap, c(1L, 3L, 5L))
})

test_that("by_lap validates input", {
    expect_error(by_lap(0), "valid.*integer")
    expect_error(by_lap(-1), "valid.*integer")
    expect_error(by_lap(1.5), "valid.*integer")
})


## as_mnirs_interval() =====================================================
test_that("as_mnirs_interval passes through NULL", {
    expect_null(as_mnirs_interval(NULL))
})

test_that("as_mnirs_interval passes through mnirs_interval", {
    interval <- by_time(5)
    result <- as_mnirs_interval(interval)
    expect_identical(result, interval)
})

test_that("as_mnirs_interval coerces numeric to by_time", {
    result <- as_mnirs_interval(c(2, 5, 8))
    expect_s3_class(result, "mnirs_interval")
    expect_equal(result$type, "time")
    expect_equal(result$by_time, c(2, 5, 8))
})

test_that("as_mnirs_interval coerces character to by_label", {
    result <- as_mnirs_interval(c("start", "end"))
    expect_s3_class(result, "mnirs_interval")
    expect_equal(result$type, "label")
    expect_equal(result$by_label, c("start", "end"))
})

test_that("as_mnirs_interval coerces integer to by_lap", {
    result <- as_mnirs_interval(c(1L, 3L))
    expect_s3_class(result, "mnirs_interval")
    expect_equal(result$type, "lap")
    expect_equal(result$by_lap, c(1L, 3L))
})

test_that("as_mnirs_interval errors on unsupported type", {
    expect_error(as_mnirs_interval(TRUE, "start"), "start.*must be")
    expect_error(as_mnirs_interval(list(1), "end"), "end.*must be")
})


## recycle_span() =========================================================
test_that("recycle_span works", {
    ## recycle_span expands positive scalar to c(0, x)
    expect_equal(recycle_span(60), c(0, 60))
    ## recycle_span expands negative scalar to c(x, 0)
    expect_equal(recycle_span(-60), c(-60, 0))
    ## recycle_span treats zero as positive
    expect_equal(recycle_span(0), c(0, 0))
    ## recycle_span passes through two-element vector
    expect_equal(recycle_span(c(-5, 10)), c(-5, 10))
})

test_that("recycle_span validates span", {
    ## length > 2 errors
    expect_error(recycle_span(c(1, 2, 3)), "span.*must be")
    ## length 0 errors
    expect_error(recycle_span(numeric(0)), "span.*must be")
    ## non-numeric errors
    expect_error(recycle_span("a"), "span.*must be")
})


## find_interval_time() ====================================================
test_that("find_interval_time resolves time values directly", {
    time_vec <- seq(0.1, 10, by = 0.1)

    result <- find_interval_time(by_time(2, 5, 8), time_vec)
    expect_equal(result, c(2, 5, 8))
})

test_that("find_interval_time resolves samples to times", {
    time_vec <- seq(0.1, 10, by = 0.1)

    result <- find_interval_time(
        by_sample(10, 30, 70), time_vec
    )
    expect_equal(result, time_vec[c(10, 30, 70)])
})

test_that("find_interval_time resolves labels to times", {
    time_vec <- seq(0.1, 10, by = 0.1)
    event_vec <- c(
        "start", rep("", 4), "mid", rep("", 4), "end"
    )

    result <- find_interval_time(
        by_label("start", "mid"),
        time_vec,
        event_vec
    )
    expect_equal(result, time_vec[c(1, 6)])
})

test_that("find_interval_time errors when no labels match", {
    event_vec <- c(rep("", 50), "marker", rep("", 50))

    expect_error(
        find_interval_time(
            by_label("invalid"),
            time_vec = NULL,
            event_vec
        ),
        "No events detected"
    )
})

test_that("find_interval_time resolves laps with position = first", {
    time_vec <- seq(0, 0.8, by = 0.1)
    event_vec <- c(1L, 1L, 1L, 2L, 2L, 2L, 3L, 3L, 3L)

    result <- find_interval_time(
        by_lap(1, 3),
        time_vec,
        event_vec,
        position = "first"
    )
    expect_equal(result, time_vec[c(1, 7)])
})

test_that("find_interval_time resolves laps with position = last", {
    time_vec <- seq(0, 0.8, by = 0.1)
    event_vec <- c(1L, 1L, 1L, 2L, 2L, 2L, 3L, 3L, 3L)

    result <- find_interval_time(
        by_lap(1, 3),
        time_vec,
        event_vec,
        position = "last"
    )
    expect_equal(result, time_vec[c(3, 9)])
})

test_that("find_interval_time errors when lap not found", {
    event_vec <- c(1L, 1L, 2L, 2L)

    expect_error(
        find_interval_time(
            by_lap(5),
            time_vec = NULL,
            event_vec,
            position = "first"
        ),
        "No samples found for lap"
    )
})

## resolve_interval() ===============================================
test_that("resolve_interval returns start-only times", {
    time_vec <- seq(0, 10, by = 0.1)

    result <- resolve_interval(
        start = by_time(2, 5),
        end = NULL,
        time_vec = time_vec
    )

    expect_true(result$has_start)
    expect_false(result$has_end)
    expect_equal(result$start_time, c(2, 5))
    expect_null(result$end_time)
})

test_that("resolve_interval returns paired start+end times", {
    time_vec <- seq(0, 10, by = 0.1)

    result <- resolve_interval(
        start = by_time(2, 5),
        end = by_time(4, 8),
        time_vec = time_vec
    )

    expect_true(result$has_start)
    expect_true(result$has_end)
    expect_equal(result$start_time, c(2, 5))
    expect_equal(result$end_time, c(4, 8))
})

test_that("resolve_interval warns and truncates unequal lengths", {
    time_vec <- seq(0, 10, by = 0.1)

    expect_warning(
        result <- resolve_interval(
            start = by_time(2, 5, 8),
            end = by_time(4, 7),
            time_vec = time_vec
        ),
        "Unequal lengths"
    )

    ## truncated to 2 paired intervals
    expect_equal(length(result$start_time), 2)
    expect_equal(length(result$end_time), 2)
})

test_that("resolve_interval resolves lap start-only to lap start", {
    event_vec <- c(1L, 1L, 1L, 2L, 2L, 2L, 3L, 3L, 3L)
    time_vec <- seq(0, 0.8, by = 0.1)

    result <- resolve_interval(
        start = by_lap(2),
        end = NULL,
        time_vec = time_vec,
        event_vec = event_vec
    )

    ## lap 2 starts row 4; time 0.3
    expect_equal(result$start_time, time_vec[4])
    expect_null(result$end_time)
    expect_true(result$has_start)
    expect_false(result$has_end)
})

test_that("resolve_interval resolves lap end-only to lap end", {
    event_vec <- c(1L, 1L, 1L, 2L, 2L, 2L, 3L, 3L, 3L)
    time_vec <- seq(0, 0.8, by = 0.1)

    result <- resolve_interval(
        start = NULL,
        end = by_lap(3),
        time_vec = time_vec,
        event_vec = event_vec
    )

    ## lap 3 ends row 9; time 0.8
    expect_equal(result$start_time, time_vec[9])
    expect_null(result$end_time)
    expect_false(result$has_start)
    expect_true(result$has_end)
})

test_that("resolve_interval lap single-boundary supports multiple laps", {
    event_vec <- c(1L, 1L, 2L, 2L, 3L, 3L)
    time_vec <- seq(0, 0.5, by = 0.1)

    result <- resolve_interval(
        start = by_lap(1, 3),
        end = NULL,
        time_vec = time_vec,
        event_vec = event_vec
    )

    ## lap 1: row 1; lap 3: row 5
    expect_equal(result$start_time, time_vec[c(1, 5)])
    expect_null(result$end_time)
    expect_true(result$has_start)
    expect_false(result$has_end)
})

## recycle_to_length() ==============================================
test_that("recycle_to_length returns unchanged when lengths match", {
    input <- list(c(-1, 1), c(-2, 2), c(-3, 3))
    result <- recycle_to_length(input, n = 3, verbose = FALSE)

    expect_equal(result, input)
})

test_that("recycle_to_length recycles last element when n_param < n", {
    result <- recycle_to_length(
        list(c(-1, 1), c(-2, 2)),
        n = 4,
        verbose = FALSE
    )

    expect_length(result, 4)
    expect_equal(result[[1]], c(-1, 1))
    expect_equal(result[[2]], c(-2, 2))
    expect_equal(result[[3]], c(-2, 2))
    expect_equal(result[[4]], c(-2, 2))
})

test_that("recycle_to_length truncates when n_param > n", {
    result <- recycle_to_length(
        list(c(-1, 1), c(-2, 2), c(-3, 3), c(-4, 4)),
        n = 2,
        verbose = FALSE
    )

    expect_length(result, 2)
    expect_equal(result[[1]], c(-1, 1))
    expect_equal(result[[2]], c(-2, 2))
})

test_that("recycle_to_length handles single element recycling", {
    result <- recycle_to_length(
        list(c(-1, 1)),
        n = 3,
        verbose = FALSE
    )

    expect_length(result, 3)
    expect_equal(result[[1]], c(-1, 1))
    expect_equal(result[[2]], c(-1, 1))
    expect_equal(result[[3]], c(-1, 1))
})

test_that("recycle_to_length messages when recycling with verbose", {
    expect_message(
        recycle_to_length(
            list(c(-1, 1), c(-2, 2), c(-3, 3)),
            n = 1,
            verbose = TRUE
        ),
        regexp = "exceeds.*ignored"
    )

    expect_message(
        recycle_to_length(
            list(c(-1, 1), c(-2, 2)),
            n = 4,
            verbose = TRUE
        ),
        regexp = "recycled.*unspecified"
    )

    ## no message when single param recycled (common case)
    expect_no_message(
        recycle_to_length(
            list(c(-1, 1)),
            n = 4,
            verbose = TRUE
        )
    )
})

## recycle_param() ==============================================
test_that("recycle_param converts non-list to list", {
    result <- recycle_param(
        c(1, 2),
        n_events = 1,
        event_groups = "distinct",
        verbose = FALSE
    )

    expect_true(is.list(result))
    expect_equal(result, list(c(1, 2)))
})

test_that("recycle_param returns unchanged when lengths match", {
    input <- list(c(-1, 1), c(-2, 2), c(-3, 3))
    result <- recycle_param(
        input,
        n_events = 3,
        event_groups = "distinct",
        verbose = FALSE
    )

    expect_equal(result, input)
})

test_that("recycle_param recycles last element when param_length < n_events", {
    result <- recycle_param(
        list(c(-1, 1), c(-2, 2)),
        n_events = 4,
        event_groups = "distinct",
        verbose = FALSE
    )

    expect_length(result, 4)
    expect_equal(result[[3]], c(-2, 2))
    expect_equal(result[[4]], c(-2, 2))
})

test_that("recycle_param truncates when param_length > n_events", {
    result <- recycle_param(
        list(c(-1, 1), c(-2, 2), c(-3, 3), c(-4, 4)),
        n_events = 2,
        event_groups = "distinct",
        verbose = FALSE
    )

    expect_length(result, 2)
    expect_equal(result[[1]], c(-1, 1))
    expect_equal(result[[2]], c(-2, 2))
})

test_that("recycle_param flattens nested lists", {
    result <- recycle_param(
        list(list(c(-1, 1)), list(c(-2, 2))),
        n_events = 2,
        event_groups = "distinct",
        verbose = FALSE
    )

    expect_equal(result[[1]], c(-1, 1))
    expect_equal(result[[2]], c(-2, 2))
})

test_that("recycle_param warns when recycling & truncating", {
    expect_message(
        recycle_param(
            list(c(-1, 1), c(-2, 2)),
            n_events = 4,
            event_groups = "distinct",
            verbose = TRUE
        ),
        regexp = "recycled.*unspecified"
    )

    expect_message(
        recycle_param(
            list(c(-1, 1), c(-2, 2), c(-3, 3)),
            n_events = 1,
            event_groups = "distinct",
            verbose = TRUE
        ),
        regexp = "exceeds.*ignored"
    )
})

test_that("recycle_param expands per group and reorders to event order", {
    ## 4 events, 2 groups:  group 1 = events 1,3; group 2 = events 2,4
    ## 2 span values should map:  span1 -> events 1,3; span2 -> events 2,4
    result <- recycle_param(
        list(c(-0.3, 0.3), c(-0.5, 0.5)),
        n_events = 5,
        event_groups = list(c(1, 3, 5), c(2, 4)),
        verbose = FALSE
    )

    expect_length(result, 5)
    expect_equal(result[[1]], c(-0.3, 0.3))
    expect_equal(result[[2]], c(-0.5, 0.5))
    expect_equal(result[[3]], c(-0.3, 0.3))
    expect_equal(result[[4]], c(-0.5, 0.5))
    expect_equal(result[[5]], c(-0.3, 0.3))
})

test_that("recycle_param recycles params to match group count", {
    ## 4 events, 3 groups but only 1 param: recycle to all groups
    result <- recycle_param(
        list(c(-1, 1)),
        n_events = 4,
        event_groups = list(c(1, 2), c(3), c(4)),
        verbose = FALSE
    )

    expect_length(result, 4)
    expect_equal(result[[1]], c(-1, 1))
    expect_equal(result[[2]], c(-1, 1))
    expect_equal(result[[3]], c(-1, 1))
    expect_equal(result[[4]], c(-1, 1))

    ## 2 params for 3 groups: last param recycled to group 3
    result <- recycle_param(
        list(c(-0.3, 0.3), c(-0.5, 0.5)),
        n_events = 4,
        event_groups = list(c(4, 3), c(2), c(1)),
        verbose = FALSE
    )

    expect_length(result, 4)
    expect_equal(result[[4]], c(-0.3, 0.3)) # group 1
    expect_equal(result[[3]], c(-0.3, 0.3)) # group 1
    expect_equal(result[[2]], c(-0.5, 0.5)) # group 2
    expect_equal(result[[1]], c(-0.5, 0.5)) # group 3 (recycled from group 2)
})

test_that("recycle_param handles ungrouped events with custom grouping", {
    ## 5 events but only 4 grouped: event 5 ungrouped
    result <- recycle_param(
        param = list(c(-0.3, 0.3), c(-0.5, 0.5)),
        n_events = 5,
        event_groups = list(c(1, 3), c(2, 4)),
        verbose = FALSE
    )

    expect_length(result, 5)
    expect_equal(result[[1]], c(-0.3, 0.3))
    expect_equal(result[[2]], c(-0.5, 0.5))
    expect_equal(result[[3]], c(-0.3, 0.3))
    expect_equal(result[[4]], c(-0.5, 0.5))
    expect_equal(result[[5]], c(-0.5, 0.5)) # ungrouped:  last param recycled
})

test_that("recycle_param truncates when more grouped events than actual events", {
    ## event_groups specifies events 1-6 but only 4 actual events
    result <- recycle_param(
        param = list(c(-0.3, 0.3), c(-0.5, 0.5)),
        n_events = 4,
        event_groups = list(c(1, 3, 5), c(2, 4, 6)),
        verbose = FALSE
    )

    expect_length(result, 4)
    expect_equal(result[[1]], c(-0.3, 0.3))
    expect_equal(result[[2]], c(-0.5, 0.5))
    expect_equal(result[[3]], c(-0.3, 0.3))
    expect_equal(result[[4]], c(-0.5, 0.5))
})

test_that("recycle_param handles non-contiguous group indices", {
    ## groups reference events 1, 4, 2, 5 (non-sequential)
    result <- recycle_param(
        param = list(c(-0.3, 0.3), c(-0.5, 0.5)),
        n_events = 5,
        event_groups = list(c(1, 4), c(2, 5)),
        verbose = FALSE
    )

    expect_length(result, 5)
    expect_equal(result[[1]], c(-0.3, 0.3)) # group 1
    expect_equal(result[[2]], c(-0.5, 0.5)) # group 2
    expect_equal(result[[3]], c(-0.5, 0.5)) # ungrouped (event 3 not in groups)
    expect_equal(result[[4]], c(-0.3, 0.3)) # group 1
    expect_equal(result[[5]], c(-0.5, 0.5)) # group 2
})

test_that("recycle_param truncates excess params for custom grouping", {
    ## 3 params but only 2 groups
    expect_message(
        result <- recycle_param(
            list(c(-0.3, 0.3), c(-0.5, 0.5), c(-1, 1)),
            n_events = 4,
            event_groups = list(c(1, 3), c(2, 4)),
            verbose = TRUE
        ),
        regexp = "exceeds"
    )

    expect_length(result, 4)
    expect_equal(result[[1]], c(-0.3, 0.3))
    expect_equal(result[[2]], c(-0.5, 0.5))
    expect_equal(result[[3]], c(-0.3, 0.3))
    expect_equal(result[[4]], c(-0.5, 0.5))
})

test_that("recycle_param messages when recycling groups with verbose", {
    expect_message(
        recycle_param(
            list(c(-1, 1), c(-2, 2), c(-3, 3)),
            n_events = 1,
            event_groups = list(1),
            verbose = TRUE
        ),
        regexp = "exceeds.*ignored"
    )

    expect_message(
        recycle_param(
            list(c(-0.3, 0.3), c(-0.5, 0.5)),
            n_events = 6,
            event_groups = list(c(1, 2), c(3, 4), c(5, 6)),
            verbose = TRUE
        ),
        regexp = "recycled.*unspecified"
    )

    ## no message when single param recycled (common case)
    expect_no_message(
        recycle_param(
            list(c(-0.3, 0.3)),
            n_events = 6,
            event_groups = list(c(1, 2), c(3, 4), c(5, 6)),
            verbose = TRUE
        )
    )
})


## apply_span() ============================================================
test_that("apply_span creates correct interval specification", {
    ## floating point precision issues
    time_vec <- seq(0, 10, by = 0.1)
    interval_list <- list(
        start_time = time_vec[c(20, 50, 80)], ## c(1.9, 4.9, 7.9)
        end_time = NULL,
        has_start = TRUE,
        has_end = FALSE
    )
    span <- list(c(-1, 1), c(-1, 1), c(-1, 1))

    result <- apply_span(
        interval_list,
        time_vec,
        span,
        verbose = FALSE
    )

    expect_s3_class(result, "data.frame")
    expect_equal(nrow(result), 3)
    expect_equal(result$start_times, time_vec[c(20, 50, 80)] - 1)
    expect_equal(result$end_times, time_vec[c(20, 50, 80)] + 1)
    expect_equal(result$interval_times, as.list(time_vec[c(20, 50, 80)]))
})

test_that("apply_span creates correct specification with start, end", {
    ## floating point precision issues
    time_vec <- seq(0, 10, by = 0.1)
    start_times <- time_vec[c(10, 40, 70)]
    end_times <- time_vec[c(30, 60, 90)]
    interval_list <- list(
        start_time = start_times,
        end_time = end_times,
        has_start = TRUE,
        has_end = TRUE
    )

    result <- apply_span(
        interval_list,
        time_vec,
        span = list(c(0, 1), c(0, 1), c(0, 1)),
        verbose = FALSE
    )

    expect_s3_class(result, "data.frame")
    expect_equal(nrow(result), 3)
    expect_equal(result$interval_times, Map(c, start_times, end_times))
})

test_that("apply_span calculates correct start/end indices", {
    time_vec <- seq(0, 10, by = 0.1)
    interval_list <- list(
        start_time = 5.0,
        end_time = NULL,
        has_start = TRUE,
        has_end = FALSE
    )

    result <- apply_span(
        interval_list,
        time_vec,
        span = list(c(-1, 2)),
        verbose = FALSE
    )

    ## span = c(-1, 2) around time = 5 means [4, 7]
    expect_equal(result$start_times, 4)
    expect_equal(result$end_times, 7)
})

test_that("apply_span clips partial out-of-bounds intervals", {
    time_vec <- seq(0, 10, by = 0.1)
    interval_list <- list(
        start_time = time_vec[6],
        end_time = NULL,
        has_start = TRUE,
        has_end = FALSE
    )

    result <- apply_span(
        interval_list,
        time_vec,
        span = list(c(-2, 2)),
        verbose = FALSE
    )

    expect_equal(result$start_times, 0) # clamped to t_min

    ## returns warning with verbose = TRUE
    expect_warning(
        result <- apply_span(
            interval_list,
            time_vec,
            span = list(c(-2, 10)),
            verbose = TRUE
        ),
        "partially outside"
    )

    expect_equal(result$start_times, 0) # clamped to t_min
    expect_equal(result$end_times, 10) # clamped to t_max
})

test_that("apply_span errors for entirely out-of-bounds", {
    time_vec <- seq(0, 10, by = 0.1)
    interval_list <- list(
        start_time = time_vec[50],
        end_time = NULL,
        has_start = TRUE,
        has_end = FALSE
    )

    expect_error(
        apply_span(
            interval_list,
            time_vec,
            span = list(c(100, 200)),
            verbose = FALSE
        ),
        regexp = "entirely outside"
    )
})

test_that("apply_span applies span correctly with start+end", {
    time_vec <- seq(0, 10, by = 0.1)
    interval_list <- list(
        start_time = 2.0,
        end_time = 6.0,
        has_start = TRUE,
        has_end = TRUE
    )

    ## span[1] shifts start, span[2] shifts end
    result <- apply_span(
        interval_list,
        time_vec,
        span = list(c(-1, 2)),
        verbose = FALSE
    )

    ## start: time 2 + (-1) = 1; end: time 6 + 2 = 8
    expect_equal(result$start_times, 1)
    expect_equal(result$end_times, 8)
})


## extract_df_list() ==============================================
test_that("extract_df_list returns correct number of intervals", {
    data <- create_mock_mnirs(n = 100, sample_rate = 10)
    time_vec <- data$time
    
    interval_spec <- data.frame(
        span_before = c(-1, -1),
        span_after = c(2, 2),
        start_times = time_vec[c(10, 50)],
        end_times = time_vec[c(30, 70)]
    )
    interval_spec$interval_times <- list(1, 5) ## two start_times, no end_times

    result <- extract_df_list(
        data,
        time_vec,
        interval_spec,
        nirs_channels = list(
            c("smo2_left", "smo2_right"),
            c("smo2_left", "smo2_right")
        )
    )

    expect_length(result, 2)
    expect_named(result, c("interval_1", "interval_2"))
})

test_that("extract_df_list extracts correct row ranges", {
    data <- create_mock_mnirs(n = 100, sample_rate = 10)
    time_vec <- data$time

    interval_spec <- data.frame(
        span_before = -1,
        span_after = 1,
        start_times = time_vec[20],
        end_times = time_vec[40]
    )
    interval_spec$interval_times <- list(3) ## one start_times, no end_times

    result <- extract_df_list(
        data,
        time_vec,
        interval_spec,
        nirs_channels = list(c("smo2_left", "smo2_right"))
    )

    expect_equal(nrow(result[[1L]]), 21) # rows 20 to 40 inclusive
})

test_that("extract_df_list preserves metadata attributes", {
    data <- create_mock_mnirs(n = 100, sample_rate = 10)
    time_vec <- data$time

    interval_spec <- data.frame(
        span_before = -0.5,
        span_after = 1.5,
        start_times = time_vec[10],
        end_times = time_vec[30]
    )
    interval_spec$interval_times <- list(c(1.5, 3.0)) ## one start_times, one end_times

    result <- extract_df_list(
        data,
        time_vec, 
        interval_spec,
        nirs_channels = list(c("smo2_left"))
    )

    expect_equal(attr(result[[1L]], "interval_times"), c(1.5, 3.0))
    expect_equal(attr(result[[1L]], "interval_span"), c(-0.5, 1.5))
    expect_equal(attr(result[[1L]], "nirs_channels"), "smo2_left")
})


## zero_offset_data() ===============================================
test_that("zero_offset_data shifts time channel by event time", {
    df <- tibble::tibble(time = c(5, 6, 7, 8, 9), value = 1:5)
    result <- zero_offset_data(df, time_channel = "time", t0 = 7)
    expect_equal(result$time, c(-2, -1, 0, 1, 2))
    expect_equal(result$value, 1:5) # other columns unchanged

    ## negative event times
    df <- tibble::tibble(time = c(-5, -4, -3, -2, -1), value = 1:5)
    result <- zero_offset_data(df, time_channel = "time", t0 = -3)
    expect_equal(result$time, c(-2, -1, 0, 1, 2))
})


## ensemble_intervals() ====================================================
test_that("ensemble_intervals averages across intervals correctly", {
    # Create two intervals with known values
    interval1 <- create_mock_interval(time_start = 10, n = 11, event_time = 10)
    interval1$smo2_left <- rep(40, 11)
    interval1$smo2_right <- rep(60, 11)

    interval2 <- create_mock_interval(time_start = 20, n = 11, event_time = 20)
    interval2$smo2_left <- rep(60, 11)
    interval2$smo2_right <- rep(40, 11)

    df_list <- list(interval_1 = interval1, interval_2 = interval2)
    metadata <- list(time_channel = "time", sample_rate = 10)

    result <- ensemble_intervals(
        df_list = df_list,
        nirs_channels = c("smo2_left", "smo2_right"),
        metadata = metadata,
        verbose = FALSE
    )

    # Ensemble mean of 40 and 60 should be 50
    expect_true(all(abs(result$smo2_left - 50) < 1e-10))
    expect_true(all(abs(result$smo2_right - 50) < 1e-10))

    # Times should be zero-offset (start at 0)
    expect_equal(min(result$time), 0)
})

test_that("ensemble_intervals preserves metadata", {
    interval1 <- create_mock_interval(time_start = 10, n = 11, event_time = 10)
    interval2 <- create_mock_interval(time_start = 20, n = 11, event_time = 20)
    df_list <- list(interval_1 = interval1, interval_2 = interval2)
    metadata <- list(time_channel = "time", sample_rate = 10)

    result <- ensemble_intervals(
        df_list = df_list,
        nirs_channels = c("smo2_left", "smo2_right"),
        metadata = metadata,
        verbose = FALSE
    )

    expect_equal(attr(result, "time_channel"), "time")
    expect_equal(attr(result, "sample_rate"), 10)
    expect_true(is.list(attr(result, "interval_times")))
    expect_setequal(unlist(attr(result, "interval_times")), c(0, 0))
    expect_true(is.list(attr(result, "interval_span")))
    expect_setequal(lengths(attr(result, "interval_span")), 2)
})

test_that("ensemble_intervals warns on irregular samples with verbose", {
    interval1 <- create_mock_interval(time_start = 0, n = 5, event_time = 0)
    interval2 <- create_mock_interval(
        time_start = 0.05,
        n = 5,
        event_time = 0.05
    )
    interval2[2, ] <- NA
    df_list <- list(interval_1 = interval1, interval_2 = interval2)
    metadata <- list(time_channel = "time", sample_rate = 10)

    expect_warning(
        ensemble_intervals(
            df_list = df_list,
            nirs_channels = c("smo2_left", "smo2_right"),
            metadata = metadata,
            verbose = TRUE
        ),
        regexp = "irregular.*samples"
    )
})

test_that("ensemble_intervals returns the right number of dims", {
    interval1 <- create_mock_interval(time_start = 0, n = 5, event_time = 0)
    interval2 <- create_mock_interval(
        time_start = 0.05,
        n = 5,
        event_time = 0.05
    )
    df_list <- list(interval_1 = interval1, interval_2 = interval2)
    metadata <- list(time_channel = "time", sample_rate = 10)
    nirs_channels = c("smo2_left")

    result <- ensemble_intervals(
        df_list = df_list,
        nirs_channels = nirs_channels,
        metadata = metadata,
        verbose = FALSE
    )

    expect_equal(ncol(result), length(nirs_channels) + 1)

    nirs_channels = c("smo2_left", "smo2_right")

    result <- ensemble_intervals(
        df_list = df_list,
        nirs_channels = nirs_channels,
        metadata = metadata,
        verbose = FALSE
    )

    expect_equal(ncol(result), length(nirs_channels) + 1)
})

test_that("ensemble_intervals preserves all metadata attributes", {
    interval1 <- create_mock_interval(time_start = 10, n = 11, event_time = 10)
    interval2 <- create_mock_interval(time_start = 20, n = 11, event_time = 20)
    df_list <- list(interval_1 = interval1, interval_2 = interval2)
    metadata <- list(
        time_channel = "time",
        sample_rate = 10,
        nirs_device = "MockDevice",
        event_channel = "event",
        start_timestamp = as.POSIXct("2024-01-01")
    )

    result <- ensemble_intervals(
        df_list = df_list,
        nirs_channels = c("smo2_left", "smo2_right"),
        metadata = metadata,
        verbose = FALSE
    )

    expect_equal(attr(result, "nirs_device"), "MockDevice")
    expect_equal(attr(result, "nirs_channels"), c("smo2_left", "smo2_right"))
    expect_equal(attr(result, "event_channel"), "event")
    expect_equal(
        attr(result, "start_timestamp"),
        as.POSIXct("2024-01-01")
    )
    ## class is preserved
    expect_true(inherits(result, "mnirs"))
})

test_that("ensemble_intervals deduplicates nirs_channels attr", {
    interval1 <- create_mock_interval(time_start = 0, n = 11, event_time = 0)
    df_list <- list(i1 = interval1, i2 = interval1)
    metadata <- list(time_channel = "time", sample_rate = 10)

    ## duplicated channel name supplied; attr must be unique
    result <- ensemble_intervals(
        df_list = df_list,
        nirs_channels = c("smo2_left", "smo2_left"),
        metadata = metadata,
        verbose = FALSE
    )

    expect_equal(attr(result, "nirs_channels"), "smo2_left")
})


## group_intervals() ==================================================
test_that("group_intervals returns distinct intervals unchanged", {
    interval1 <- create_mock_interval(time_start = 0, n = 11, event_time = 0)
    interval2 <- create_mock_interval(time_start = 10, n = 11, event_time = 10)
    df_list <- list(interval_1 = interval1, interval_2 = interval2)
    metadata <- list(time_channel = "time", sample_rate = 10)

    result <- group_intervals(
        df_list = df_list,
        nirs_channels = list(
            c("smo2_left", "smo2_right"),
            c("smo2_left", "smo2_right")
        ),
        metadata = metadata,
        event_groups = "distinct",
        zero_time = TRUE,
        verbose = FALSE
    )

    expect_length(result, 2)
    expect_named(result, c("interval_1", "interval_2"))
    ## zero offset explicitly for "distinct"
    expect_equal(min(result[[1]]$time), 0)
})

test_that("group_intervals ensembles all intervals with 'ensemble'", {
    interval1 <- create_mock_interval(time_start = 0, n = 11, event_time = 0)
    interval2 <- create_mock_interval(time_start = 10, n = 11, event_time = 10)
    df_list <- list(interval_1 = interval1, interval_2 = interval2)
    metadata <- list(time_channel = "time", sample_rate = 10)

    result <- group_intervals(
        df_list = df_list,
        nirs_channels = list(
            c("smo2_left", "smo2_right"),
            c("smo2_left", "smo2_right")
        ),
        metadata = metadata,
        event_groups = "ensemble",
        zero_time = FALSE,
        verbose = FALSE
    )

    expect_length(result, 1)
    expect_named(result, "ensemble")
    ## zero offset regardless for "ensemble"
    expect_equal(min(result[[1]]$time), 0)
})

test_that("group_intervals handles custom grouping", {
    interval1 <- create_mock_interval(time_start = 0, n = 11, event_time = 0)
    interval2 <- create_mock_interval(time_start = 10, n = 11, event_time = 10)
    interval3 <- create_mock_interval(time_start = 20, n = 11, event_time = 20)
    interval4 <- create_mock_interval(time_start = 30, n = 11, event_time = 30)
    df_list <- list(
        interval_1 = interval1,
        interval_2 = interval2,
        interval_3 = interval3,
        interval_4 = interval4
    )
    metadata <- list(time_channel = "time", sample_rate = 10)

    result <- group_intervals(
        df_list = df_list,
        nirs_channels = rep(list(c("smo2_left", "smo2_right")), 4),
        metadata = metadata,
        event_groups = list(c(1, 2), c(3, 4)),
        zero_time = TRUE,
        verbose = FALSE
    )

    expect_length(result, 2)
    expect_named(result, c("interval_1_2", "interval_3_4"))

    # Group only intervals 1, 2 & 3, leaving 4 ungrouped
    expect_message(
        result <- group_intervals(
            df_list = df_list,
            nirs_channels = rep(list(c("smo2_left", "smo2_right")), 3),
            metadata = metadata,
            event_groups = list(c(1, 2), 4),
            zero_time = TRUE,
            verbose = TRUE
        ),
        "Ungrouped.*discrete"
    )

    expect_length(result, 3)
    expect_named(result, c("interval_1_2", "interval_3", "interval_4"))

    ## group the same interval multiple times throws warning
    expect_warning(
        result <- group_intervals(
            df_list = df_list,
            nirs_channels = rep(list(c("smo2_left", "smo2_right")), 3),
            metadata = metadata,
            event_groups = list(c(1, 2, 3), c(2, 4)),
            zero_time = TRUE,
            verbose = TRUE
        ),
        "Duplicates detected"
    )
})

test_that("group_intervals returns single interval as distinct regardless", {
    interval1 <- create_mock_interval(time_start = 0, n = 11, event_time = 0)
    df_list <- list(interval_1 = interval1)
    metadata <- list(time_channel = "time", sample_rate = 10)

    result <- group_intervals(
        df_list = df_list,
        nirs_channels = list(c("smo2_left", "smo2_right")),
        metadata = metadata,
        event_groups = "ensemble", # request ensemble but only 1 interval
        zero_time = FALSE,
        verbose = FALSE
    )

    expect_length(result, 1)
})

test_that("group_intervals (distinct) preserves all metadata on each interval", {
    interval1 <- create_mock_interval(time_start = 0, n = 11, event_time = 0)
    interval2 <- create_mock_interval(time_start = 10, n = 11, event_time = 10)
    df_list <- list(interval_1 = interval1, interval_2 = interval2)
    metadata <- list(
        time_channel = "time",
        sample_rate = 10,
        nirs_device = "MockDevice",
        event_channel = "event",
        start_timestamp = as.POSIXct("2024-01-01")
    )

    result <- group_intervals(
        df_list = df_list,
        nirs_channels = list(
            c("smo2_left", "smo2_right"),
            c("smo2_left", "smo2_right")
        ),
        metadata = metadata,
        event_groups = "distinct",
        zero_time = FALSE,
        verbose = FALSE
    )

    for (iv in result) {
        expect_equal(attr(iv, "nirs_device"), "MockDevice")
        expect_equal(attr(iv, "nirs_channels"), c("smo2_left", "smo2_right"))
        expect_equal(attr(iv, "time_channel"), "time")
        expect_equal(attr(iv, "event_channel"), "event")
        expect_equal(attr(iv, "sample_rate"), 10)
        expect_equal(
            attr(iv, "start_timestamp"),
            as.POSIXct("2024-01-01")
        )
    }
    ## interval_times and interval_span forwarded from original interval attrs
    expect_true(inherits(result[[1]], "mnirs"))
    expect_equal(attr(result[[1]], "interval_times"), 0)
    expect_equal(attr(result[[1]], "interval_span"), c(-1, 4))
    expect_true(inherits(result[[2]], "mnirs"))
    expect_equal(attr(result[[2]], "interval_times"), 10)
    expect_equal(attr(result[[2]], "interval_span"), c(-1, 4))
})


test_that("group_intervals custom multi-interval groups preserve metadata", {
    interval1 <- create_mock_interval(time_start = 0, n = 11, event_time = 0)
    interval2 <- create_mock_interval(time_start = 10, n = 11, event_time = 10)
    interval3 <- create_mock_interval(time_start = 20, n = 11, event_time = 20)
    interval4 <- create_mock_interval(time_start = 30, n = 11, event_time = 30)
    df_list <- list(
        interval_1 = interval1,
        interval_2 = interval2,
        interval_3 = interval3,
        interval_4 = interval4
    )
    metadata <- list(
        time_channel = "time",
        sample_rate = 10,
        nirs_device = "MockDevice",
        event_channel = "event"
    )

    result <- group_intervals(
        df_list = df_list,
        nirs_channels = rep(list(c("smo2_left", "smo2_right")), 4),
        metadata = metadata,
        event_groups = list(c(1, 2), c(3, 4)),
        # zero_time = FALSE, ## ensemble auto zeroes
        verbose = FALSE
    )

    for (iv in result) {
        expect_equal(attr(iv, "nirs_device"), "MockDevice")
        expect_equal(attr(iv, "sample_rate"), 10)
        expect_equal(attr(iv, "event_channel"), "event")
        expect_true(inherits(iv, "mnirs"))
    }
    ## ensemble sub-groups collect interval_times as a list
    expect_length(attr(result[[1]], "interval_times"), 2)
    expect_equal(
        attr(result[[1]], "interval_times"),
        list(0, 0), ## start times for two grouped intervals
        ignore_attr = TRUE
    )
    expect_length(attr(result[[2]], "interval_times"), 2)
    expect_equal(
        attr(result[[2]], "interval_times"),
        list(0, 0), ## adheres to `zero_time` to represent output data frame
        ignore_attr = TRUE
    )
})

test_that("group_intervals custom single-interval group retains original attrs", {
    interval1 <- create_mock_interval(
        time_start = 0,
        n = 11,
        event_time = 0,
        span = c(-2, 5)
    )
    interval2 <- create_mock_interval(
        time_start = 10,
        n = 11,
        event_time = 10,
        span = c(-1, 4)
    )
    interval3 <- create_mock_interval(
        time_start = 20,
        n = 11,
        event_time = 20,
        span = c(-1, 4)
    )
    df_list <- list(
        interval_1 = interval1,
        interval_2 = interval2,
        interval_3 = interval3
    )
    metadata <- list(time_channel = "time", sample_rate = 10)

    ## intervals 1+2 ensembled; interval 3 returned as lone group (raw)
    result <- group_intervals(
        df_list = df_list,
        nirs_channels = rep(list(c("smo2_left", "smo2_right")), 3),
        metadata = metadata,
        event_groups = list(c(1, 2), 3),
        zero_time = FALSE,
        verbose = FALSE
    )

    lone <- result[["interval_3"]]
    ## original attrs are preserved on the lone interval
    expect_equal(attr(lone, "nirs_channels"), c("smo2_left", "smo2_right"))
    expect_equal(attr(lone, "time_channel"), "time")
    expect_equal(attr(lone, "interval_times"), 20)
    expect_equal(attr(lone, "interval_span"), c(-1, 4))

    ## with `zero_time = TRUE`
    result <- group_intervals(
        df_list = df_list,
        nirs_channels = rep(list(c("smo2_left", "smo2_right")), 3),
        metadata = metadata,
        event_groups = list(c(1, 2), 3),
        zero_time = TRUE,
        verbose = FALSE
    )

    lone <- result[["interval_3"]]
    expect_equal(attr(lone, "interval_times"), 0) ## adheres to `zero_time`
})


## extract_intervals() ===================================================
test_that("extract_intervals validates start/end args", {
    data <- create_mock_mnirs(n = 100, sample_rate = 10)

    old <- options(mnirs.verbose = FALSE)
    on.exit(options(old), add = TRUE)

    ## unsupported types still error via as_mnirs_interval
    expect_error(
        extract_intervals(
            data,
            start = TRUE,
            span = c(-1, 1)
        ),
        "start.*must be"
    )

    expect_error(
        extract_intervals(
            data,
            end = list(1),
            span = c(-1, 1)
        ),
        "end.*must be"
    )
})

test_that("extract_intervals returns list of tibbles", {
    data <- create_mock_mnirs(n = 100, sample_rate = 10)

    result <- extract_intervals(
        data = data,
        start = by_time(2, 5),
        event_groups = "distinct",
        span = c(-1, 1),
        verbose = FALSE
    )

    expect_type(result, "list")
    expect_true(all(vapply(result, tibble::is_tibble, logical(1))))
    expect_equal(result[[1]]$time[1], 2 - 1)
    expect_equal(rev(result[[1]]$time)[1], 2 + 1)
    expect_equal(result[[2]]$time[1], 5 - 1)
    expect_equal(rev(result[[2]]$time)[1], 5 + 1)
})

test_that("extract_intervals works with start and end", {
    data <- create_mock_mnirs(n = 100, sample_rate = 10)

    result <- extract_intervals(
        data = data,
        start = by_time(2, 5),
        end = by_time(4, 8),
        event_groups = "distinct",
        span = c(0, 0),
        verbose = FALSE
    )

    expect_length(result, 2)
    expect_equal(result[[1]]$time[1], 2)
    expect_equal(rev(result[[1]]$time)[1], 4)
    expect_equal(result[[2]]$time[1], 5)
    expect_equal(rev(result[[2]]$time)[1], 8)
    ## interval_times is c(start, end) when both boundaries defined
    expect_equal(attr(result[[1]], "interval_times"), c(2, 4))
    expect_equal(attr(result[[2]], "interval_times"), c(5, 8))
})

test_that("extract_intervals works with by_sample", {
    data <- create_mock_mnirs(n = 100, sample_rate = 10)

    result <- extract_intervals(
        data = data,
        start = by_sample(21, 51),
        event_groups = "distinct",
        span = c(-1, 1),
        verbose = FALSE
    )

    expect_length(result, 2)
    expect_equal(result[[1]]$time[1], 2 - 1)
    expect_equal(rev(result[[1]]$time)[1], 2 + 1)
    expect_equal(result[[2]]$time[1], 5 - 1)
    expect_equal(rev(result[[2]]$time)[1], 5 + 1)
})

test_that("extract_intervals works with by_label", {
    data <- create_mock_mnirs(n = 100, sample_rate = 10)
    data$event[51] <- "marker"

    result <- extract_intervals(
        data = data,
        event_channel = "event",
        start = by_label("marker"),
        event_groups = "distinct",
        span = c(-1, 1),
        verbose = FALSE
    )

    expect_length(result, 1)
    expect_equal(result[[1]]$time[1], 5 - 1)
    expect_equal(rev(result[[1]]$time)[1], 5 + 1)
})

test_that("extract_intervals works with by_lap start only", {
    data <- create_mock_mnirs(n = 100, sample_rate = 10)
    ## replace character event with integer laps
    data$event <- rep(1:10, each = 10)

    ## single boundary: full lap returned (first to last sample)
    result <- extract_intervals(
        data = data,
        event_channel = "event",
        start = by_lap(3),
        event_groups = "distinct",
        span = c(0, 0),
        verbose = FALSE
    )

    expect_length(result, 1)
    ## lap 3: starts row 21, times 2.0
    expect_equal(result[[1]]$time, 2.0)
    expect_equal(nrow(result[[1]]), 1)
    ## interval_times reflects both boundaries
    expect_equal(attr(result[[1]], "interval_times"), 2.0)

    ## span shifts boundaries around the full lap
    result <- extract_intervals(
        data = data,
        event_channel = "event",
        start = by_lap(3),
        event_groups = "distinct",
        span = c(-0.5, 0.5),
        verbose = FALSE
    )

    ## lap 3 starts at 2.0; span[-0.5, 0.5] -> [1.5, 2.5]
    expect_equal(result[[1]]$time[1], 1.5)
    expect_equal(rev(result[[1]]$time)[1], 2.5)
})

test_that("extract_intervals works with by_lap end only", {
    data <- create_mock_mnirs(n = 100, sample_rate = 10)
    data$event <- rep(1:10, each = 10)

    result <- extract_intervals(
        data = data,
        event_channel = "event",
        end = by_lap(5),
        event_groups = "distinct",
        span = c(0, 0),
        verbose = FALSE
    )

    expect_length(result, 1)
    ## lap 5: ends row 50, time 4.9
    expect_equal(result[[1]]$time, 4.9)
    expect_equal(nrow(result[[1]]), 1)
    expect_equal(attr(result[[1]], "interval_times"), 4.9)
})

test_that("extract_intervals works with by_lap start and end", {
    data <- create_mock_mnirs(n = 100, sample_rate = 10)
    data$event <- rep(1:10, each = 10)

    result <- extract_intervals(
        data = data,
        event_channel = "event",
        start = by_lap(2),
        end = by_lap(4),
        event_groups = "distinct",
        span = c(0, 0),
        verbose = FALSE
    )

    expect_length(result, 1)
    ## lap 2 first sample: row 11 (time = 1.0)
    ## lap 4 last sample: row 40 (time = 3.9)
    expect_equal(result[[1]]$time[1], 1.0)
    expect_equal(rev(result[[1]]$time)[1], 3.9)
    expect_equal(nrow(result[[1]]), 30) ## rows 11 to 40
})

test_that("extract_intervals works with multiple by_lap pairs", {
    data <- create_mock_mnirs(n = 100, sample_rate = 10)
    data$event <- rep(1:10, each = 10)

    result <- extract_intervals(
        data = data,
        event_channel = "event",
        start = by_lap(1, 5),
        end = by_lap(2, 7),
        event_groups = "distinct",
        span = c(0, 0),
        verbose = FALSE
    )

    expect_length(result, 2)
    ## lap 1 first sample: row 1 (time = 0.0)
    ## lap 2 last sample: row 20 (time = 1.9)
    expect_equal(result[[1]]$time[1], 0.0)
    expect_equal(rev(result[[1]]$time)[1], 1.9)
    ## interval 1: lap 1 first (row 1) to lap 2 last (row 20)
    expect_equal(nrow(result[[1]]), 20)
    ## lap 5 first sample: row 50 (time = 4.0)
    ## lap 7 last sample: row 70 (time = 6.9)
    expect_equal(result[[2]]$time[1], 4.0)
    expect_equal(rev(result[[2]]$time)[1], 6.9)
    ## interval 2: lap 5 first (row 41) to lap 7 last (row 70)
    expect_equal(nrow(result[[2]]), 30)
})

test_that("extract_intervals errors when by_lap used without event_channel", {
    data <- create_mock_mnirs(n = 100, sample_rate = 10)
    ## remove event_channel from metadata
    attr(data, "event_channel") <- NULL
    data$event <- NULL

    expect_error(
        extract_intervals(
            data = data,
            start = by_lap(1),
            span = c(0, 1),
            verbose = FALSE
        ),
        "event_channel.*required"
    )
})

test_that("extract_intervals coerces raw numeric to by_time", {
    data <- create_mock_mnirs(n = 100, sample_rate = 10)

    result <- extract_intervals(
        data = data,
        start = 2,
        event_groups = "distinct",
        span = c(-1, 1),
        verbose = FALSE
    )

    expect_length(result, 1)
    expect_equal(result[[1]]$time[1], 2 - 1)
    expect_equal(rev(result[[1]]$time)[1], 2 + 1)


    obj <- 2
    result <- extract_intervals(
        data = data,
        start = obj,
        event_groups = "distinct",
        span = c(-0.5, 0.5),
        verbose = FALSE
    )

    expect_length(result, 1)
    expect_equal(result[[1]]$time[1], 2 - 0.5)
    expect_equal(rev(result[[1]]$time)[1], 2 + 0.5)
})

test_that("extract_intervals coerces raw character to by_label", {
    data <- create_mock_mnirs(n = 100, sample_rate = 10)
    data$event[51] <- "marker"

    result <- extract_intervals(
        data = data,
        event_channel = "event",
        start = "marker",
        event_groups = "distinct",
        span = c(-1, 1),
        verbose = FALSE
    )

    expect_length(result, 1)
    expect_equal(result[[1]]$time[1], 5 - 1)
    expect_equal(rev(result[[1]]$time)[1], 5 + 1)
})

test_that("extract_intervals coerces raw integer to by_lap", {
    data <- create_mock_mnirs(n = 100, sample_rate = 10)
    data$event <- rep(1:10, each = 10)

    result <- extract_intervals(
        data = data,
        event_channel = "event",
        start = 2L,
        end = 4L,
        event_groups = "distinct",
        span = c(0, 0),
        verbose = FALSE
    )

    expect_length(result, 1)
    ## same as by_lap(2) / by_lap(4)
    expect_equal(result[[1]]$time[1], 1.0)
    expect_equal(rev(result[[1]]$time)[1], 3.9)
    expect_equal(nrow(result[[1]]), 30)
})

test_that("extract_intervals recycles positive span scalar", {
    data <- create_mock_mnirs(n = 100, sample_rate = 10)

    result <- extract_intervals(
        data = data,
        start = by_time(2),
        end = by_time(5),
        event_groups = "distinct",
        span = 1,
        verbose = FALSE
    )

    ## span = 1 → c(0, 1): start unchanged, end shifted +1
    expect_equal(result[[1]]$time[1], 2)
    expect_equal(rev(result[[1]]$time)[1], 6)
})

test_that("extract_intervals recycles negative span scalar", {
    data <- create_mock_mnirs(n = 100, sample_rate = 10)

    result <- extract_intervals(
        data = data,
        start = by_time(2),
        end = by_time(5),
        event_groups = "distinct",
        span = -1,
        verbose = FALSE
    )

    ## span = -1 → c(-1, 0): start shifted -1, end unchanged
    expect_equal(result[[1]]$time[1], 1)
    expect_equal(rev(result[[1]]$time)[1], 5)
})

test_that("extract_intervals applies zero_time correctly", {
    data <- create_mock_mnirs(n = 100, sample_rate = 10)

    result <- extract_intervals(
        data = data,
        start = by_time(5),
        event_groups = "distinct",
        span = c(-1, 1),
        zero_time = TRUE,
        verbose = FALSE
    )

    # Time should start at -1 (span before) after zero offset
    expect_equal(min(result[[1]]$time), -1)
})

test_that("extract_intervals handles grouping", {
    data <- create_mock_mnirs(n = 100, sample_rate = 10)

    result <- extract_intervals(
        data = data,
        start = by_time(2, 5, 8),
        event_groups = "ensemble",
        span = c(-0.5, 0.5), ## single span recycled to all events
        verbose = FALSE
    )

    expect_length(result, 1)
    expect_named(result, "ensemble")
    ## check interval span
    expect_setequal(range(result$ensemble$time), c(-0.5, 0.5))

    result <- extract_intervals(
        data = data,
        start = by_time(2, 4, 6, 8),
        event_groups = list(c(1, 3), c(2, 4)),
        span = list(c(-0.3, 0.3), c(-0.5, 0.5)),
        verbose = FALSE
    )

    expect_length(result, 2)
    expect_named(result, c("interval_1_3", "interval_2_4"))
    expect_setequal(range(result$interval_1_3$time), c(-0.3, 0.3))
    expect_setequal(range(result$interval_2_4$time), c(-0.5, 0.5))
})

test_that("extract_intervals handles different spans per event", {
    data <- create_mock_mnirs(n = 100, sample_rate = 10)

    result <- extract_intervals(
        data = data,
        start = by_time(2, 5),
        event_groups = "distinct",
        span = list(c(-0.5, 0.5), c(-1, 1)),
        verbose = FALSE
    )

    expect_length(result, 2)
    # Second interval should be larger due to wider span
    expect_true(nrow(result[[2]]) > nrow(result[[1]]))
})

test_that("extract_intervals errors & messages", {
    data <- create_mock_mnirs(n = 100, sample_rate = 10)

    ## no interval specified
    expect_error(
        extract_intervals(
            data = data,
            span = c(-1, 1),
            verbose = FALSE
        ),
        regexp = "No interval specification"
    )

    ## edge case: event at data boundary
    expect_warning(
        result <- extract_intervals(
            data = data,
            start = by_time(0.5),
            event_groups = "distinct",
            span = c(-1, 1),
            verbose = TRUE
        ),
        regexp = "partially outside"
    ) |>
        expect_message("nirs_channels.*grouped together")

    expect_length(result, 1)
    ## start value bounded by time = zero
    expect_setequal(range(result$interval_1$time), c(0, 1.5))
})

test_that("extract_intervals respects nirs_channels metadata", {
    data <- create_mock_mnirs(n = 100, sample_rate = 10)
    all_channels <- attr(data, "nirs_channels")

    result <- extract_intervals(
        data = data,
        nirs_channels = "smo2_left",
        start = by_time(1, 5),
        event_groups = "distinct",
        span = c(-1, 1),
        verbose = FALSE
    )

    expect_equal(attr(result[[1]], "nirs_channels"), "smo2_left")
    expect_equal(attr(result[[2]], "nirs_channels"), "smo2_left")

    result <- extract_intervals(
        data = data,
        nirs_channels = "smo2_left",
        time_channel = "time",
        start = by_time(1, 5),
        event_groups = "ensemble",
        span = c(-1, 1),
        verbose = TRUE
    )

    expect_equal(attr(result[[1]], "nirs_channels"), "smo2_left")
})


test_that("extract_intervals informs when nirs_channels is not a list()", {
    data <- create_mock_mnirs(n = 60, sample_rate = 10)
    attr(data, "nirs_channels") <- NULL  ## remove metadata

    ## fires: verbose=TRUE, ensemble, no metadata, non-list channels
    expect_message(
        extract_intervals(
            data,
            nirs_channels = "smo2_left",
            time_channel = "time",
            start = by_time(1, 4),
            span = c(-0.5, 0.5),
            event_groups = "ensemble",
            verbose = TRUE
        ),
        "list\\(\\).*channel grouping"
    )

    ## silent: verbose=FALSE
    expect_no_message(
        extract_intervals(
            data,
            nirs_channels = "smo2_left",
            time_channel = "time",
            start = by_time(1, 4),
            span = c(-0.5, 0.5),
            event_groups = "ensemble",
            verbose = FALSE
        )
    )

    ## silent: nirs_channels already a list()
    expect_no_message(
        extract_intervals(
            data,
            nirs_channels = list("smo2_left"),
            time_channel = "time",
            start = by_time(1, 4),
            span = c(-0.5, 0.5),
            event_groups = "ensemble",
            verbose = TRUE
        )
    )

    ## silent: event_groups = "distinct"
    expect_no_message(
        extract_intervals(
            data,
            nirs_channels = "smo2_left",
            time_channel = "time",
            start = by_time(1, 4),
            span = c(-0.5, 0.5),
            event_groups = "distinct",
            verbose = TRUE
        )
    )

    ## silent: data has nirs_channels metadata
    attr(data, "nirs_channels") <- "smo2_left"
    expect_no_message(
        extract_intervals(
            data,
            nirs_channels = "smo2_left",
            time_channel = "time",
            start = by_time(1, 4),
            span = c(-0.5, 0.5),
            event_groups = "ensemble",
            verbose = TRUE
        )
    )
})

test_that("extract_intervals returns a list of class mnirs", {
    data <- create_mock_mnirs(n = 100, sample_rate = 10)

    result <- extract_intervals(
        data = data,
        nirs_channels = "smo2_left",
        start = by_time(1, 5),
        event_groups = "distinct",
        span = c(-1, 1),
        verbose = FALSE
    )

    expect_equal(class(result), c("mnirs", "list"))

    ## print.mnirs should not show attr(,"class") trailer
    output <- capture.output(print(result))
    expect_false(any(grepl('attr\\(,"class"\\)', output)))
    
    ## calling `result` directly should not show attr(,"class") trailer
    output <- capture.output(result)
    expect_false(any(grepl('attr\\(,"class"\\)', output)))
})

## integration tests ===================================

test_that("extract_intervals works on Moxy data", {
    data <- read_mnirs(
        example_mnirs("moxy_ramp"),
        nirs_channels = c(smo2_left = "SmO2 Live", smo2_right = "SmO2 Live(2)"),
        verbose = FALSE
    )

    result <- extract_intervals(
        data,
        nirs_channels = c("smo2_left", "smo2_right"),
        start = by_time(878),
        span = list(c(-30, 180)),
        zero_time = FALSE,
        verbose = FALSE
    )

    ## structure
    expect_length(result, 1)
    expect_length(result[[1L]], 3)
    expect_named(result[[1L]], c("hh:mm:ss", "smo2_left", "smo2_right"))
    ## range of time_channel
    expect_gte(min(result[[1L]][[1]]), 878 - 30)
    expect_true(
        all.equal(min(result[[1L]][[1]]), 878 - 30, tolerance = 1, scale = 1)
    )
    expect_lte(max(result[[1L]][[1]]), 878 + 180)
    expect_true(
        all.equal(max(result[[1L]][[1]]), 878 + 180, tolerance = 1, scale = 1)
    )
    ## equivalent to intake df
    expect_equal(
        result[[1L]],
        data[within(data$`hh:mm:ss`, c(878 - 30, 878 + 180)), ],
        ignore_attr = TRUE
    )
})

test_that("extract_intervals works on train.red data", {
    data <- read_mnirs(
        example_mnirs("train.red"),
        nirs_channels = c(
            smo2_left = "SmO2 unfiltered",
            smo2_right = "SmO2 unfiltered"
        ),
        time_channel = c(time = "Timestamp (seconds passed)"),
        verbose = FALSE
    ) |>
        resample_mnirs(method = "linear", verbose = FALSE)

    result <- extract_intervals(
        data,
        nirs_channels = c("smo2_left", "smo2_right"),
        start = by_time(2150, 3168),
        event_groups = "ensemble",
        span = list(c(-30, 180)),
        zero_time = FALSE,
        verbose = FALSE
    )

    ## visual check
    # plot(result[[1L]], time_labels = TRUE)

    ## structure
    expect_length(result, 1)
    expect_length(result[[1L]], 3)
    expect_named(result[[1L]], c("time", "smo2_left", "smo2_right"))
    ## range of time_channel
    expect_gte(min(result[[1L]][[1]]), -30)
    expect_true(
        all.equal(min(result[[1L]][[1]]), -30, tolerance = 0.1, scale = 1)
    )
    expect_lte(max(result[[1L]][[1]]), 180)
    expect_true(
        all.equal(max(result[[1L]][[1]]), 180, tolerance = 0.1, scale = 1)
    )

    result <- extract_intervals(
        data,
        nirs_channels = c("smo2_left", "smo2_right"),
        start = by_time(2150, 3168),
        event_groups = "distinct",
        span = list(c(-30, 180)),
        zero_time = FALSE,
        verbose = FALSE
    )

    ## visual check
    # plot(result)

    ## structure
    expect_length(result, 2)
    expect_length(result[[1L]], 3)
    expect_length(result[[2L]], 3)
    expect_named(result[[1L]], c("time", "smo2_left", "smo2_right"))
    expect_named(result[[2L]], c("time", "smo2_left", "smo2_right"))
    ## range of time_channel
    expect_lte(min(result[[1L]][[1]]), 2150 - 30)
    expect_true(all.equal(
        min(result[[1L]][[1]]), 2150 - 30, tolerance = 0.1, scale = 1
    ))
    expect_lte(max(result[[1L]][[1]]), 2150 + 180)
    expect_true(all.equal(
        max(result[[1L]][[1]]), 2150 + 180, tolerance = 0.1, scale = 1
    ))
    expect_lte(min(result[[2L]][[1]]), 3168 - 30)
    expect_true(all.equal(
        min(result[[2L]][[1]]), 3168 - 30, tolerance = 0.1, scale = 1
    ))
    expect_lte(max(result[[2L]][[1]]), 3168 + 180)
    expect_true(all.equal(
        max(result[[2L]][[1]]), data$time[nrow(data)], tolerance = 0.1, scale = 1
    ))
})

## benchmark ===========================================================
test_that("extract_intervals benchmark", {
    ## baselne established from documented example on initial run;
    ## fails if itr/sec regresses by >10%
    skip("benchmark baseline test")

    data_list <- read_mnirs(
        example_mnirs("train.red"),
        nirs_channels = c(
            smo2_left = "SmO2 unfiltered",
            smo2_right = "SmO2 unfiltered"
        ),
        time_channel = c(time = "Timestamp (seconds passed)"),
        zero_time = TRUE,
        verbose = FALSE
    ) |>
        resample_mnirs(method = "linear", verbose = FALSE)

    # for (i in seq_len(3)) {
    #     bm <- bench::mark(
    #         extract_intervals = extract_intervals(
    #             data_list,
    #             start = by_time(368, 1084),
    #             event_groups = "distinct",
    #             span = c(-20, 90),
    #             zero_time = TRUE,
    #             verbose = FALSE
    #         ),
    #         iterations = 50,
    #         check = FALSE
    #     )
    #     print(bm)
    # }

    itr_per_sec <- bm$`itr/sec`

    ## baseline: update this value when optimising (seconds)
    ## run test interactively to calibrate:
    ##   itr_per_sec will be printed on first failure
    baseline <- 330
    threshold <- baseline * 1.10 ## 10% regression budget

    expect_lte(
        itr_per_sec,
        threshold,
        label = sprintf(
            "%.3f itr/sec exceeds %.0f%% of baseline %.3fs (limit %.3fs)",
            itr_per_sec,
            110,
            baseline,
            threshold
        )
    )
})

#' Replace outliers, invalid, and missing values in *{mnirs}* data
#'
#' Detect and replace local outliers, specified invalid values, and missing
#' `NA` values across `nirs_channels` within an *"mnirs"* data frame.
#' `replace_mnirs()` operates on a data frame, extending the vectorised 
#' functions:.
#'
#' @param invalid_values A numeric vector of invalid values to be replaced,
#'   e.g. `invalid_values = c(0, 100, 102.3)`. Default `NULL` will not replace
#'   invalid values.
#'
#' @param invalid_above,invalid_below Numeric values each specifying cutoff
#'   values, above or below which (respectively) will be replaced, *inclusive*
#'   of the specified cutoff values.
#'
#' @param outlier_cutoff A numeric value for the local outlier threshold, as
#'   the number of standard deviations from the local median.
#'   - Default `NULL` will not replace outliers.
#'   - Lower values are more sensitive and flag more outliers; higher values
#'     are more conservative.
#'   - `outlier_cutoff = 3` Pearson's 3 sigma edit rule.
#'     `outlier_cutoff = 2` approximates a Tukey-style 1.5*IQR rule.
#'     `outlier_cutoff = 0` Tukey's median filter.
#'
#' @param width An integer defining the local window in number of samples
#'   centred on `idx`, between
#'   `[idx - floor(width/2), idx + floor(width/2)]`.
#'
#' @param span A numeric value defining the local window time span around
#'   `idx` in units of `time_channel` or `t`, between
#'   `[t - span/2, t + span/2]`.
#'
#' @param method A character string indicating how to handle `NA`
#'   replacement (see *Details*):
#'   \describe{
#'     \item{`"linear"`}{Replaces `NA`s via linear interpolation (the
#'     *default*) using [stats::approx()].}
#'     \item{`"median"`}{Replaces `NA`s with the local median of valid
#'     values within a centred window defined by `width` or `span`.}
#'     \item{`"locf"`}{*"Last observation carried forward"*. Replaces
#'     `NA`s with the most recent valid value to the left for trailing
#'     `NA`s or to the right for leading `NA`s, using
#'     [stats::approx()].}
#'     \item{`"none"`}{Returns `NA`s without replacement.}
#'   }
#'
#' @inheritParams validate_mnirs
#'
#' @details
#' ## Automatic channel detection
#'
#' `nirs_channels` and `time_channel` are retrieved automatically from
#' *"mnirs"* metadata if not specified explicitly. Columns in `data` not
#' listed in `nirs_channels` are passed through unprocessed.
#'
#' ## The rolling window
#'
#' `replace_outliers()` and `replace_missing()` (when `method = "median"`)
#' operate over a local rolling window for outlier detection and median
#' interpolation. The window is specified by either `width` as the number
#' of samples, or `span` as the time span in units of `time_channel`.
#' A partial window is calculated at the edges of the data.
#'
#' @returns `replace_mnirs()` return a [tibble][tibble::tibble-package] of 
#' class `"mnirs"` with metadata available via `attributes()`.
#'
#' @examples
#' ## vectorised operations
#' x <- c(1, 999, 3, 4, 999, 6)
#' replace_invalid(x, invalid_values = 999, width = 3, method = "median")
#'
#' (x_na <- replace_outliers(x, outlier_cutoff = 3, width = 3, method = "none"))
#'
#' replace_missing(x_na, method = "linear")
#'
#' ## read example data
#' data <- read_mnirs(
#'     file_path = example_mnirs("moxy_ramp"),
#'     nirs_channels = c(smo2 = "SmO2 Live"),
#'     time_channel = c(time = "hh:mm:ss"),
#'     verbose = FALSE
#' )
#'
#' ## clean data
#' data_clean <- replace_mnirs(
#'     data,                  ## channels retrieved from metadata
#'     invalid_values = 0,    ## known invalid values in the data
#'     invalid_above = 90,    ## remove data spikes above 90
#'     outlier_cutoff = 3,    ## Pearson's 3 sigma edit rule
#'     width = 7,            ## window for outlier detection and interpolation
#'     method = "linear"      ## linear interpolation over NAs
#' )
#'
#' \donttest{
#'   if (requireNamespace("ggplot2", quietly = TRUE)) {
#'     ## plot original and show where values have been replaced
#'     ## ignore warning about replacing the existing colour scale
#'     plot(data, time_labels = TRUE) +
#'       ggplot2::scale_colour_manual(
#'         name = NULL,
#'         breaks = c("smo2", "replaced"),
#'         values = palette_mnirs(2)
#'       ) +
#'       ggplot2::geom_point(
#'         data = data[data_clean$smo2 != data$smo2, ],
#'         ggplot2::aes(y = smo2, colour = "replaced"),
#'         na.rm = TRUE
#'       ) +
#'       ggplot2::geom_line(
#'         data = {
#'           data_clean[!is.na(data$smo2), "smo2"] <- NA
#'           data_clean
#'         },
#'         ggplot2::aes(y = smo2, colour = "replaced"),
#'         linewidth = 1, na.rm = TRUE
#'       )
#'   }
#' }
#'
#' @rdname replace_mnirs
#' @order 1
#' @export
replace_mnirs <- function(
    data,
    nirs_channels = NULL,
    time_channel = NULL,
    invalid_values = NULL,
    invalid_above = NULL,
    invalid_below = NULL,
    outlier_cutoff = NULL,
    width = NULL,
    span = NULL,
    method = c("linear", "median", "locf", "none"),
    verbose = TRUE
) {
    ## validation ====================================
    method <- match.arg(method)
    check_conditions <- c(
        !is.null(c(invalid_values, invalid_above, invalid_below)),
        !is.null(outlier_cutoff),
        method != "none"
    )
    ## do nothing condition
    if (!any(check_conditions)) {
        cli_abort(c(
            "x" = "No replacement criteria specified",
            "i" = "At least one of {.arg invalid_values}, \\
            {.arg invalid_above}, {.arg invalid_below}, \\
            {.arg outlier_cutoff}, or {.arg method} must be specified."
        ))
    }
    if (missing(verbose)) {
        verbose <- getOption("mnirs.verbose", default = TRUE)
    }

    validate_mnirs_data(data)
    metadata <- attributes(data)
    nirs_channels <- validate_nirs_channels(enquo(nirs_channels), data, verbose)
    time_channel <- validate_time_channel(enquo(time_channel), data)
    time_vec <- data[[time_channel]]

    if (check_conditions[2L] || method == "median") {
        validate_width_span(width, span, verbose, "for `replace_mnirs()`.")
    }

    ## remove invalid, outliers, and NA ==============================
    data[nirs_channels] <- lapply(data[nirs_channels], \(.x) {
        if (check_conditions[1L]) {
            .x <- replace_invalid(
                x = .x,
                t = time_vec,
                invalid_values = invalid_values,
                invalid_above = invalid_above,
                invalid_below = invalid_below,
                method = "none",
                bypass_checks = TRUE
            )
        }
        if (check_conditions[2L]) {
            .x <- replace_outliers(
                x = .x,
                t = time_vec,
                width = width,
                span = span,
                method = "none",
                outlier_cutoff = outlier_cutoff,
                bypass_checks = TRUE
            )
        }
        if (check_conditions[3L]) {
            .x <- replace_missing(
                x = .x,
                t = time_vec,
                width = width,
                span = span,
                method = method,
                bypass_checks = TRUE
            )
        }
        .x
    })

    ## Metadata =================================
    metadata$nirs_channels <- unique(nirs_channels)
    metadata$time_channel <- time_channel

    return(create_mnirs_data(data, metadata))
}


#' Replace invalid values
#'
#' `replace_invalid()` detects specified invalid values or range cutoffs in a 
#' numeric vector and replace them with the local median value or `NA`.
#'
#' @param x A numeric vector of the response variable.
#' @param t An *optional* numeric vector of the predictor variable (e.g. time). 
#'   Default is `seq_along(x)`.
#' @inheritParams replace_mnirs
#'
#' @details
#' ## Replace invalid values with with replace_invalid()
#' 
#' Specific `invalid_values` can be replaced, such as `c(0, 100, 102.3)`.
#' Data ranges can be replaced with cutoff values specified by `invalid_above` 
#' and `invalid_below`, where any values higher or lower than the specified 
#' cutoff values (respectively) will be replaced, *inclusive* of the cutoff 
#' values themselves.
#'
#' @returns `replace_invalid()` returns a numeric vector the same length as 
#' `x` with invalid values replaced.
#'
#' @rdname replace_mnirs
#' @order 2
#' @export
replace_invalid <- function(
    x,
    t = seq_along(x),
    invalid_values = NULL,
    invalid_above = NULL,
    invalid_below = NULL,
    width = NULL,
    span = NULL,
    method = c("median", "none"),
    verbose = TRUE,
    ...
) {
    ## validate ===============================================
    args <- list(...)
    if (is.null(c(invalid_values, invalid_above, invalid_below))) {
        cli_abort(c(
            "x" = "No replacement criteria specified",
            "i" = "At least one of {.arg invalid_values}, \\
            {.arg invalid_above}, or {.arg invalid_below} must be specified."
        ))
    }
    if (!(args$bypass_checks %||% FALSE)) {
        validate_x_t(x, t)
        if (missing(verbose)) {
            verbose <- getOption("mnirs.verbose", default = TRUE)
        }
    }
    validate_numeric(invalid_values)
    validate_numeric(invalid_above, 1, msg1 = "one-element")
    validate_numeric(invalid_below, 1, msg1 = "one-element")
    method <- match.arg(method)

    ## process ========================================================
    ## fill invalid indices with NA
    invalid_idx <- c(
        which(x %in% invalid_values),
        which(x >= invalid_above),
        which(x <= invalid_below)
    )

    if (length(invalid_idx) == 0) {
        return(x)
    }

    y <- x
    y[invalid_idx] <- NA_real_

    if (method == "median") {
        if (!(args$bypass_checks %||% FALSE)) {
            validate_width_span(width, span, verbose, "for median replacement.")
        }

        window_idx <- compute_local_windows(t, invalid_idx, width, span)
        local_medians <- compute_local_fun(y, window_idx, median, na.rm = TRUE)
        ## if method = "median"
        ## invalid_values removed to NA first,
        ## so returns local median excluding idx
        y[invalid_idx] <- local_medians
    }

    return(y)
}


#' Replace local outliers
#'
#' `replace_outliers()` detects local outliers in a numeric vector using a Hampel filter and replaces with the local median value or `NA`.
#'
#' @inheritParams replace_invalid
#'
#' @details
#' ## Outlier detection with replace_outliers()
#'
#' Rolling local medians are computed across `x` within a window defined
#' by `width` (number of samples) or `span` (time span in units of `t`).
#'
#' Outliers are detected with robust median absolute deviation (MAD),
#' adapted from `pracma::hampel()`. Deviations equal to or less than the
#' smallest absolute time series difference in `x` are excluded, to avoid
#' flagging negligible differences where local data have minimal or zero
#' variation.
#'
#' ## Replacement behaviour
#'
#' Values of `x` outside the local bounds defined by `outlier_cutoff` are
#' identified as outliers and either replaced with the local median
#' (`method = "median"`, the *default*) or set to `NA` (`method = "none"`).
#'
#' Existing `NA` values in `x` are *not* replaced. They are passed
#' through to the returned vector. See [replace_missing()].
#'
#' ## Choosing outlier_cutoff
#'
#' `outlier_cutoff` is the number of (MAD-normalised) standard deviations
#' from the local median. Higher values are more conservative; lower
#' values flag more outliers.
#'   - `outlier_cutoff = 3` -- Pearson's 3 sigma edit rule (default).
#'   - `outlier_cutoff = 2` -- approximately Tukey-style 1.5*IQR rule.
#'   - `outlier_cutoff = 0` -- Tukey's median filter (every point
#'     replaced by local median).
#'
#' @returns `replace_outliers()` returns a numeric vector the same length as 
#' `x` with outliers replaced.
#'
#' @rdname replace_mnirs
#' @order 3
#' @export
replace_outliers <- function(
    x,
    t = seq_along(x),
    outlier_cutoff = 3,
    width = NULL,
    span = NULL,
    method = c("median", "none"),
    verbose = TRUE,
    ...
) {
    ## validate ===============================================
    args <- list(...)
    if (!(args$bypass_checks %||% FALSE)) {
        if (missing(verbose)) {
            verbose <- getOption("mnirs.verbose", default = TRUE)
        }
        validate_x_t(x, t)
        validate_width_span(width, span, verbose, "for `replace_outliers()`.")
    }
    validate_numeric(
        outlier_cutoff, 1, c(0, Inf), msg1 = "one-element positive"
    )
    method <- match.arg(method)

    ## process =====================================================
    window_idx <- compute_local_windows(t, width = width, span = span)
    outlier_stats <- compute_outliers(x, window_idx, outlier_cutoff)
    local_medians <- outlier_stats$local_medians
    is_outlier <- outlier_stats$is_outlier

    ## fill outliers with median or NA
    y <- x
    y[is_outlier] <- if (method == "median") {
        local_medians[is_outlier]
    } else {
        NA_real_
    }
    
    return(y)
}

#' Replace missing values
#'
#' `replace_missing()` detects missing (`NA`) values in a numeric vector and
#' replaces via interpolation.
#'
#' @param ... Additional arguments.
#'
#' @inheritParams replace_invalid
#'
#' @details
#' ## Interpolation with replace_missing()
#'
#' `method = "linear"` and `method = "locf"` use [stats::approx()] with
#' `rule = 2`, so leading `NA`s are filled by *"nocb"*
#' (*"next observation carried backward"*) and trailing `NA`s by *"locf"*.
#'
#' `method = "median"` calculates the local median of valid (non-`NA`)
#' values to either side of `NA`s, within a window defined by `width`
#' (number of samples) or `span` (time span in units of `t`). Sequential
#' `NA`s are all replaced by the same median value.
#'
#' ## Edge behaviour for method = "median"
#'
#' If there are no valid values within `span` to one side of the `NA`,
#' the median of the other side is used (i.e. for leading and trailing
#' `NA`s). If there are no valid values within either side, the first
#' valid sample on either side is used (equivalent to
#' `replace_missing(x, width = 1)`).
#'
#' @returns `replace_missing()` returns a numeric vector the same length as
#' `x` with missing values replaced.
#'
#' @rdname replace_mnirs
#' @order 4
#' @export
replace_missing <- function(
    x,
    t = seq_along(x),
    width = NULL,
    span = NULL,
    method = c("linear", "median", "locf"),
    verbose = TRUE,
    ...
) {
    ## validate ===============================================
    args <- list(...)
    if (!(args$bypass_checks %||% FALSE)) {
        validate_x_t(x, t)
    }
    method <- match.arg(method)
    if (method == "locf") {
        method <- "constant" ## swap for approx method arg
    }

    ## process ==============================================
    if (method %in% c("linear", "constant")) {
        y <- stats::approx(
            x = t,
            y = x,
            xout = args$xout %||% t, ## = t unless explicitly specified by hidden option
            method = method, ## c("linear", "constant")
            rule = 2, ## fill leading and trailing `NA`s
            f = 0, ## locf if method = "constant"
            ties = list("ordered", mean) ## assume ordered, take mean of ties
        )$y
    } else if (method == "median") {
        if (!(args$bypass_checks %||% FALSE)) {
            if (missing(verbose)) {
                verbose <- getOption("mnirs.verbose", default = TRUE)
            }
            validate_width_span(width, span, verbose, "for median replacement.")
        }
        ## median of width or span VALID values to either side of sequential NAs
        y <- x
        na_idx <- which(is.na(x))
        window_idx <- compute_valid_neighbours(x, t, width, span, verbose)
        local_medians <- compute_local_fun(x, window_idx, median, na.rm = TRUE)
        y[na_idx] <- local_medians
    }

    return(y)
}

## preserve_na() -------------------------------------------------------
test_that("preserve_na() correctly stores NA positions and clean values", {
    x <- c(1, NA, 3, NA, 5)
    na_info <- preserve_na(x)

    expect_type(na_info, "list")
    expect_named(na_info, c("x_valid", "x_length", "na_idx"))
    expect_equal(na_info$x_valid, x[!is.na(x)])
    expect_equal(na_info$x_length, 5)
    expect_equal(na_info$na_idx, is.na(x))
})

test_that("preserve_na() handles vector with no NA", {
    x <- 1:5
    na_info <- preserve_na(x)

    expect_equal(na_info$x_valid, x)
    expect_equal(na_info$x_length, 5)
    expect_equal(na_info$na_idx, is.na(x))
})

test_that("preserve_na() handles vector with all NA", {
    x <- rep(NA_real_, 5)
    na_info <- preserve_na(x)

    expect_length(na_info$x_valid, 0)
    expect_equal(na_info$x_length, 5)
    expect_equal(na_info$na_idx, is.na(x))
})


## restore_na() --------------------------------------------------------
test_that("restore_na() correctly restores NA positions", {
    x <- c(1, NA, 3, NA, 5)
    na_info <- preserve_na(x)
    y <- na_info$x_valid * 2 ## y_valid = c(2, 6, 10)
    result <- restore_na(y, na_info)

    expect_length(result, 5)
    expect_equal(result, x * 2)
})

test_that("restore_na() handles no NA in original vector", {
    x <- 1:5
    na_info <- preserve_na(x)
    y <- na_info$x_valid * 2
    result <- restore_na(y, na_info)

    expect_equal(result, x * 2)
})

test_that("restore_na() handles all NA in original vector", {
    x <- rep(NA_real_, 5)
    na_info <- preserve_na(x)
    y <- na_info$x_valid # no clean values to process
    result <- restore_na(y, na_info)

    expect_length(result, 5)
    expect_true(all(is.na(result)))
})

test_that("preserve_na() and restore_na() work together correctly", {
    x <- c(1, NA, 3, NA, 5, 6, NA, 8)
    na_info <- preserve_na(x)

    # Simulate processing
    y_processed <- na_info$x_valid + 10
    result <- restore_na(y_processed, na_info)

    expect_length(result, length(x))
    expect_equal(result[!is.na(x)], x[!is.na(x)] + 10)
    expect_true(is.na(result[2]))
    expect_true(is.na(result[4]))
    expect_true(is.na(result[7]))
})


## replace_outliers() --------------------------------------------------
test_that("replace_outliers() returns unchanged vector with no outliers", {
    x <- 1:20
    result <- replace_outliers(x, width = 5)
    expect_equal(result, x)

    ## span
    result <- replace_outliers(x, span = 3)
    expect_equal(result, x)
})

test_that("replace_outliers() detects and replaces outliers with median", {
    x <- c(1:4, 100, 6:10) # 100 is clear outlier
    result <- replace_outliers(x, width = 5)

    expect_type(result, "double")
    expect_length(result, length(x))
    expect_lt(result[5], 100) # outlier replaced
    expect_equal(result[5], median(x[c(3:7)])) # outlier replaced
    expect_false(anyNA(result))

    ## span
    result <- replace_outliers(x, span = 3)

    expect_type(result, "double")
    expect_length(result, length(x))
    expect_lt(result[5], 100) # outlier replaced
    expect_equal(result[5], median(x[c(3:7)])) # outlier replaced
    expect_false(anyNA(result))
})

test_that("replace_outliers() detects and replaces outliers with NA", {
    x <- c(1:10, 100, 11:20)
    result <- replace_outliers(x, method = "none", width = 5)

    expect_length(result, length(x))
    expect_true(is.na(result[11])) # outlier replaced with NA

    ## span
    result <- replace_outliers(x, method = "none", span = 3)

    expect_length(result, length(x))
    expect_true(is.na(result[11])) # outlier replaced with NA
})

test_that("replace_outliers() handles NA values in input", {
    x <- c(1:9, NA, 100, 12:15, NA, 17:20)
    result <- replace_outliers(x, width = 5)

    median(c(8, 9, NA), na.rm = TRUE)

    expect_length(result, length(x))
    expect_true(result[11] != 100) # outlier replaced
    expect_true(all(is.na(result[c(10, 16)]))) # original NA preserved

    ## span
    result <- replace_outliers(x, span = 4)

    expect_length(result, length(x))
    expect_true(result[11] != 100) # outlier replaced
    expect_true(all(is.na(result[c(10, 16)]))) # original NA preserved
})

test_that("replace_outliers() respects outlier_cutoff threshold", {
    x <- c(1:10, 15, 11:20) # mild outlier

    strict <- replace_outliers(x, width = 5, outlier_cutoff = 1)
    lenient <- replace_outliers(x, width = 5, outlier_cutoff = 5)

    expect_true(strict[11] != 15) # detected with strict threshold
    expect_equal(lenient[11], 15) # not detected with lenient threshold

    ## span
    strict <- replace_outliers(x, span = 3, outlier_cutoff = 1)
    lenient <- replace_outliers(x, span = 3, outlier_cutoff = 5)

    expect_true(strict[11] != 15) # detected with strict threshold
    expect_equal(lenient[11], 15) # not detected with lenient threshold

    ## numeric outlier_cutoff
    expect_equal(replace_outliers(x, span = 3, outlier_cutoff = 2.5)[11], 11)

    ## Tukey's median filter.
    x <- c(1:10, 15, 11:20, 20.1) ## reduce min diff to 0.1 to avoid modification
    result <- replace_outliers(x, width = 5, outlier_cutoff = 0)
    medians <- vapply(seq_along(x), \(.idx) {
        median(x[pmax(1, .idx - 2):pmin(length(x), .idx + 2)])
    }, numeric(1))
    ## added tolerance to accomodate the robust variance threshold
    expect_true(all.equal(result, medians, tolerance = 0.11, scale = 1))
})

test_that("replace_outliers() validates inputs correctly", {
    x <- 1:10

    expect_error(replace_outliers(x = "text", width = 5), "x.*?numeric") # non-numeric x
    expect_error(
        replace_outliers(x, method = "none", width = -1),
        "width.*?integer"
    ) # negative width
    expect_error(
        replace_outliers(x, method = "none", width = 5, outlier_cutoff = -1),
        "outlier_cutoff.*?positive"
    ) # negative outlier_cutoff

    ## halfes all NA
    x <- rep(NA_real_, 10)
    expect_error(replace_outliers(x, width = 5), "x.*?numeric") ## x is all NA
    ## handles all same values
    expect_equal(replace_outliers(c(1, 1, 1), width = 1), rep(1, 3))

    ## span
    x <- 1:10
    expect_error(replace_outliers("text", span = 3), "x.*?numeric") # non-numeric x
    expect_error(
        replace_outliers(x, method = "none", span = -1),
        "span.*?numeric"
    ) # negative span

    ## haldes all NA
    x <- rep(NA_real_, 10)
    expect_error(replace_outliers(x, span = 3), "x.*?numeric") ## x is all NA
})

test_that("replace_outliers() errors when x and t have different lengths", {
    expect_error(
        replace_outliers(c(1, 2, 3), t = c(1, 2), outlier_cutoff = 3, span = 1),
        "equal length"
    )
})

## replace_invalid() ==================================================
test_that("replace_invalid() returns expected structure", {
    x <- c(1, 999, 3, 4, 999, 6)
    result <- replace_invalid(x, invalid_values = 999, width = 1)
    expect_length(result, length(x))
    expect_type(result, "double")
})

test_that("replace_invalid() replaces invalid values with NA when method = 'NA'", {
    x <- c(1, 999, 3, 4, 999, 6)
    result <- replace_invalid(x, invalid_values = 999, method = "none")
    expect_equal(result, c(1, NA, 3, 4, NA, 6))
})

test_that("replace_invalid() replaces invalid values with local median", {
    x <- c(1, 2, 999, 4, 999, 6)
    result <- replace_invalid(x, invalid_values = 999, width = 3)
    expect_equal(result, c(1, 2, 3, 4, 5, 6))
})

test_that("replace_invalid() handles multiple invalid values", {
    x <- c(1, 999, 3, -1, 5, 999)

    result <- replace_invalid(
        x,
        invalid_values = c(999, -1),
        width = 3
    )

    expect_equal(result, c(1, 2, 3, 4, 5, 5))
})

test_that("replace_invalid() uses custom time vector", {
    x <- c(1, 999, 3, 4)
    t <- c(0, 1, 10, 11)
    result <- replace_invalid(x, t, invalid_values = 999, span = 2)
    expect_equal(result, c(1, 1, 3, 4))
})

test_that("replace_invalid() handles non-integer span argument", {
    x <- c(1, 999, 3, 4, 999, 6)
    t <- c(0, 1, 2, 10, 11, 12)
    result <- replace_invalid(x, t, invalid_values = 999, span = 3)
    expect_equal(result, c(1, 2, 3, 4, 5, 6))
})

test_that("replace_invalid() errors when x and t have different lengths", {
    expect_error(
        replace_invalid(c(1, 2, 3), t = c(1, 2), invalid_values = 999),
        "equal length"
    )
})

test_that("replace_invalid() errors when args are not numeric", {
    expect_error(
        replace_invalid(c("a", "b"), invalid_values = 999),
        "`x`.*numeric"
    )

    expect_error(
        replace_invalid(c(1, 2), t = c("a", "b"), invalid_values = 999),
        "`t`.*numeric"
    )

    expect_error(
        replace_invalid(c(1, 2), invalid_values = "a"),
        "invalid.*numeric"
    )
})

test_that("replace_invalid() handles edge cases", {
    ## no invalid values present
    x <- c(1, 2, 3, 4)
    expect_equal(
        replace_invalid(x, invalid_values = 999, width = 1),
        x,
        ignore_attr = TRUE
    )
    expect_equal(
        replace_invalid(x, invalid_values = 999, width = 1, method = "median"),
        x,
        ignore_attr = TRUE
    )

    ## all invalid values
    x <- c(999, 999, 999)
    expect_equal(
        replace_invalid(x, invalid_values = 999, width = 1),
        rep(NA_real_, 3)
    )
    expect_equal(
        replace_invalid(x, invalid_values = 999, width = 1, method = "median"),
        rep(NA_real_, 3)
    )
})


test_that("replace_invalid handles range filtering", {
    x <- c(0, 5, 10, 20, 25, 50.5, 100, 150, 35, 40)

    # Test invalid_below only (inclusive)
    result <- replace_invalid(x, invalid_below = 5, method = "none")
    expect_setequal(result[1:2], NA_real_)
    expect_equal(result[3:10], x[3:10])

    # Test invalid_above only (inclusive)
    result <- replace_invalid(x, invalid_above = 100, method = "none")
    expect_setequal(result[7:8], NA_real_)
    expect_equal(result[c(1:5, 9:10)], x[c(1:5, 9:10)])

    # Test both invalid_below and invalid_above
    result <- replace_invalid(
        x,
        invalid_below = 5,
        invalid_above = 100,
        method = "none"
    )
    expect_setequal(result[c(1:2, 7:8)], NA_real_)
    expect_equal(result[c(3:5, 9:10)], x[c(3:5, 9:10)])

    ## combines exact values and ranges
    result <- replace_invalid(
        x,
        invalid_values = 50.5,
        invalid_below = 5,
        invalid_above = 100,
        method = "none"
    )
    expect_setequal(result[c(1:2, 6, 7:8)], NA_real_)
    expect_equal(result[c(3:5, 9:10)], x[c(3:5, 9:10)])
})

test_that("replace_invalid errors when no invalid criteria specified", {
    x <- c(1, 2, 3)

    expect_error(
        replace_invalid(
            x,
            invalid_values = NULL,
            invalid_above = NULL,
            invalid_below = NULL
        ),
        "At least one of.*must be specified"
    )
})

test_that("replace_invalid handles overlapping conditions", {
    x <- c(-999, 0, 5, -999, 10)

    # Value meets both invalid_values and invalid_below criteria
    result <- replace_invalid(
        x,
        invalid_values = -999,
        invalid_below = 5,
        method = "none"
    )
    expect_setequal(is.na(result[1:4]), TRUE)
})


test_that("replace_invalid() works correctly", {
    x <- c(1, 2, 3, 16, 5, 6, 7)
    expect_equal(
        replace_invalid(x, invalid_values = 16, method = "none")[4],
        NA_real_
    )
    expect_equal(
        replace_invalid(x, invalid_values = 16, width = 5, method = "median")[
            4
        ],
        median(x[c(1:3, 5:7)])
    )

    ## no invalid
    x_valid <- 1:7
    result_clean <- replace_invalid(x_valid, invalid_values = 16, width = 5)
    expect_equal(result_clean, x_valid)

    ## edge cases
    expect_equal(replace_invalid(c(1), invalid_values = 16, method = "none"), 1)
    expect_equal(
        sum(replace_invalid(rep(1, 7), invalid_values = 16, method = "none")),
        7
    )
    expect_error(
        replace_invalid(rep(NA, 7), invalid_values = 16),
        "`x` must be a"
    )

    ## NA handling
    x_na <- c(1, 2, 3, NA, 5, 6, 7)
    expect_equal(
        replace_invalid(x_na, invalid_values = 35, method = "none"),
        x_na
    )
    expect_equal(
        replace_invalid(
            x_na,
            invalid_values = 35,
            width = 3,
            method = "median"
        ),
        x_na
    )
    expect_true(
        is.na(replace_invalid(x_na, invalid_values = 35, method = "none")[4])
    )

    x_na <- c(1, 2, NA, 35, 5, 6, 7)
    expect_true(
        all(is.na(replace_invalid(x_na, invalid_values = 35, method = "none")[
            3:4
        ]))
    )

    expect_equal(
        replace_invalid(
            x_na,
            invalid_values = 35,
            width = 3,
            method = "median"
        )[4],
        median(x_na[c(2:3, 5:6)], na.rm = TRUE)
    )
})


## replace_missing() =================================================
test_that("replace_missing() validates inputs", {
    # x must be numeric
    expect_error(
        replace_missing(c("a", NA, "c")),
        "x.*numeric"
    )

    # t must be numeric
    expect_error(
        replace_missing(c(1, NA, 3), t = c("a", "b", "c")),
        "t.*numeric"
    )

    # x and t must be equal length
    expect_error(
        replace_missing(c(1, NA, 3), t = c(1, 2)),
        "x.*t.*numeric.*equal length"
    )

    ## missing width/span
    expect_error(
        replace_missing(c(1, NA, 3), method = "median"),
        "width.*span.*must be defined"
    )
})

test_that("replace_missing() returns unchanged vector when no NAs", {
    x <- c(1, 2, 3, 4, 5)
    expect_equal(replace_missing(x), x)
    expect_equal(replace_missing(x, method = "locf"), x)
    expect_equal(replace_missing(x, method = "median", width = 1), x)
})

test_that("replace_missing() replaces NAs with linear method", {
    x <- c(1, 2, NA, 4, 5)
    result <- replace_missing(x, method = "linear")
    expect_equal(result, c(1, 2, 3, 4, 5))

    # Multiple consecutive NAs
    x <- c(1, NA, NA, 4)
    result <- replace_missing(x, method = "linear")
    expect_equal(result, c(1, 2, 3, 4))

    # Leading NA left extrapolation
    x <- c(NA, NA, 2, 3, 4)
    result <- replace_missing(x, method = "linear")
    expect_equal(result, c(2, 2, 2, 3, 4))

    # Trailing NA right extrapolation = "locf"
    x <- c(1, 2, 3, NA, NA)
    result <- replace_missing(x, method = "linear")
    expect_equal(result, c(1, 2, 3, 3, 3))

    # Both leading and trailing NAs
    x <- c(NA, 2, NA, 4, NA)
    result <- replace_missing(x, method = "linear")
    expect_equal(result, c(2, 2, 3, 4, 4))
})

test_that("replace_missing() replaces NAs with locf method", {
    x <- c(1, 2, NA, NA, 5)
    result <- replace_missing(x, method = "locf")
    expect_equal(result, c(1, 2, 2, 2, 5))

    # Leading NA (next observation carried backward)
    x <- c(NA, NA, 3, 4)
    result <- replace_missing(x, method = "locf")
    expect_equal(result, c(3, 3, 3, 4))

    # Trailing NA (last observation carried forward)
    x <- c(1, 2, NA, NA)
    result <- replace_missing(x, method = "locf")
    expect_equal(result, c(1, 2, 2, 2))
})

test_that("replace_missing() replaces NAs with median method", {
    x <- c(1, 2, NA, 4, 5)
    result <- replace_missing(x, method = "median", width = 1)
    expect_equal(result, c(1, 2, 3, 4, 5)) # median(2, 4) = 3

    # width = 3
    x <- c(1, 2, 3, NA, 5, 6, 7)
    result <- replace_missing(x, method = "median", width = 3)
    expect_equal(result, c(1, 2, 3, 4, 5, 6, 7)) # median(2,3,5,6) = 4

    # Multiple NAs
    x <- c(1, NA, 3, NA, 5)
    result <- replace_missing(x, method = "median", width = 1)
    expect_equal(result, c(1, 2, 3, 4, 5))
})

test_that("replace_missing() replaces NAs with span method", {
    x <- c(1, 2, NA, 4, 5)
    t <- c(0, 1, 2, 3, 4)
    expect_equal(
        replace_missing(x, t = t, method = "median", span = 1),
        c(1, 2, 3, 4, 5)
    )
    expect_equal(
        replace_missing(x, t = t, method = "median", span = 0.5),
        c(1, 2, 3, 4, 5)
    )
})

test_that("replace_missing() replaces NaN", {
    x <- c(1, NaN, 3, NaN, 5)
    result <- replace_missing(x, method = "linear")
    expect_equal(result, c(1, 2, 3, 4, 5))
    expect_false(any(is.nan(result)))
})

test_that("replace_missing() works with custom time vector", {
    x <- c(10, NA, 30)
    t <- c(0, 5, 10)
    result <- replace_missing(x, t = t, method = "linear")
    expect_equal(result, c(10, 20, 30))
})

test_that("replace_missing() handles all NAs", {
    x <- c(NA, NA, NA)

    expect_error(replace_missing(x, method = "linear"), "x.*numeric")
})


## replace_mnirs() =================================================
test_that("replace_mnirs outlier removal skipped when outlier_cutoff = NULL", {
    data <- data.frame(
        time = 1:5,
        ch1 = c(50, 51, 200, 53, 54)
    )
    class(data) <- c("mnirs", "data.frame")
    attr(data, "nirs_channels") <- "ch1"
    attr(data, "time_channel") <- "time"

    ## Store original to verify no outlier processing occurred
    original_ch1 <- data$ch1

    result <- replace_mnirs(
        data,
        invalid_values = c(999),
        method = "none",
        verbose = FALSE
    )

    ## Data should be unchanged except for invalid value processing
    expect_equal(result$ch1, original_ch1)
})

test_that("replace_mnirs outlier removal processes when outlier_cutoff provided", {
    data <- data.frame(
        time = 1:10,
        ch1 = c(50, 51, 52, 200, 54, 55, 56, 57, 58, 59),
        ch2 = c(60, 61, 62, 63, 64, 65, 250, 67, 68, 69)
    )
    class(data) <- c("mnirs", "data.frame")
    attr(data, "nirs_channels") <- c("ch1", "ch2")
    attr(data, "time_channel") <- "time"

    result <- replace_mnirs(
        data,
        outlier_cutoff = 3,
        width = 5,
        method = "none",
        verbose = FALSE
    )

    ## Outlier (200) should be replaced with NA
    expect_true(is.na(result$ch1[4]))
    expect_equal(result$ch1[-4], data$ch1[-4])
    expect_true(is.na(result$ch2[7]))
    expect_equal(result$ch2[-7], data$ch2[-7])
})

test_that("replace_mnirs do nothing condition throws error appropriately", {
    data <- data.frame(
        time = 1:5,
        ch1 = 50:54
    )
    class(data) <- c("mnirs", "data.frame")
    attr(data, "nirs_channels") <- "ch1"
    attr(data, "time_channel") <- "time"

    expect_error(
        replace_mnirs(
            data,
            invalid_values = NULL,
            width = NULL,
            span = NULL,
            method = "none",
            verbose = FALSE
        ),
        "must be specified"
    )
})

test_that("replace_mnirs updates metadata correctly", {
    data <- read_mnirs(
        file_path = example_mnirs("moxy_ramp"),
        nirs_channels = c(smo2_left = "SmO2 Live", smo2_right = "SmO2 Live(2)"),
        time_channel = c(time = "hh:mm:ss"),
        verbose = FALSE
    ) |>
        replace_mnirs(
            invalid_values = c(0, 100),
            span = 7,
            method = "linear",
            verbose = FALSE
        )
    expect_equal(attr(data, "nirs_channels"), c("smo2_left", "smo2_right"))
    expect_equal(attr(data, "time_channel"), "time")
    expect_equal(attr(data, "sample_rate"), 2)

    ## overwrites nirs_channels
    data <- read_mnirs(
        file_path = example_mnirs("moxy_ramp"),
        nirs_channels = c(smo2_left = "SmO2 Live", smo2_right = "SmO2 Live(2)"),
        time_channel = c(time = "hh:mm:ss"),
        verbose = FALSE
    ) |>
        replace_mnirs(
            nirs_channels = "smo2_left",
            invalid_values = c(0, 100),
            span = 7,
            method = "linear",
            verbose = FALSE
        )
    expect_equal(attr(data, "nirs_channels"), "smo2_left")
})

test_that("replace_mnirs global verbose works", {
    expect_warning(
        read_mnirs(
            file_path = example_mnirs("moxy_ramp.xlsx"),
            nirs_channels = c(smo2 = "SmO2 Live"),
            time_channel = c(time = "hh:mm:ss")
        ) |> 
            replace_mnirs(),
        "irregular.*detected"
    ) |>
        expect_message("Estimated.*sample_rate.*2")

    old_verbose <- getOption("mnirs.verbose")
    on.exit(options(mnirs.verbose = old_verbose), add = TRUE)
    options(mnirs.verbose = FALSE)

    expect_silent(
        read_mnirs(
            file_path = example_mnirs("moxy_ramp.xlsx"),
            nirs_channels = c(smo2 = "SmO2 Live"),
            time_channel = c(time = "hh:mm:ss")
        ) |>
            replace_mnirs()
    )
})

test_that("replace_mnirs works visually on moxy data", {
    skip_if_not_installed("ggplot2")
    skip("visual check of ggplots")
    data <- read_mnirs(
        file_path = example_mnirs("moxy_ramp.xlsx"),
        nirs_channels = c(smo2 = "SmO2 Live"),
        time_channel = c(time = "hh:mm:ss"),
        verbose = FALSE
    )

    plot(data) +
        ggplot2::ylim(0, 100)

    replace_mnirs(
        data,
        invalid_values = c(0, 100),
        span = 7,
        method = "none"
    ) |>
        plot() +
        ggplot2::ylim(0, 100)

    replace_mnirs(
        data,
        invalid_values = c(0, 100),
        span = 7,
        method = "median"
    ) |>
        plot() +
        ggplot2::ylim(0, 100)

    replace_mnirs(
        data,
        invalid_values = c(0, 100),
        span = 7,
        method = "locf"
    ) |>
        plot() +
        ggplot2::ylim(0, 100)

    replace_mnirs(
        data,
        invalid_values = c(0, 100),
        span = 7,
        method = "linear"
    ) |>
        plot() +
        ggplot2::ylim(0, 100)
})

#' @name moxy_ramp.xlsx
#'
#' @title 2 Hz PerfPro export of Moxy data
#'
#' @description Exported from PerfPro Studio software, recorded
#'   at 0.5 Hz no smoothing and exported at 2 Hz. Containing a
#'   ramp incremental cycling protocol, placed on bilateral
#'   vastus lateralis muscle sites. Intentional data errors
#'   (outliers, invalid values, and missing `NA` values) have
#'   been introduced to demonstrate *{mnirs}* cleaning
#'   functions.
#'
#' @docType data
#'
#' @format .xlsx file with five columns and 2202 rows:
#'   \describe{
#'     \item{mm-dd}{Recording date (dd-MMM format).}
#'     \item{hh:mm:ss}{Time of day (hh:mm:ss format).}
#'     \item{SmO2 Live}{Muscle oxygen saturation, left leg (%). Contains
#'       simulated erroneous and missing samples.}
#'     \item{SmO2 Live(2)}{Muscle oxygen saturation, right leg (%).}
#'     \item{Lap}{Lap marker (integer).}
#'   }
#'
#'   Channel mapping for [read_mnirs()]:
#'   - `nirs_channels = c("SmO2 Live", "SmO2 Live(2)")`
#'   - `time_channel = c("hh:mm:ss")`
#'   - `event_channel = c("Lap")`
#'   - `interval_times = list(start = c(204, 878))` (start and end of exercise)
#'
#' @source Moxy Monitor (Fortiori Design LLC), exported via PerfPro Studio
#'   desktop software (https://perfprostudio.com/).
#'
#' @seealso [read_mnirs()], [example_mnirs()]
#'
#' @examples
#' example_mnirs("moxy_ramp")
#'
NULL


#' @name moxy_intervals.csv
#'
#' @title 0.5 Hz Moxy onboard export
#'
#' @description Exported from Moxy onboard recording at 0.5 Hz
#'   no smoothing. Containing four 4-minute cycling work intervals,
#'   placed on the vastus lateralis muscle site.
#'
#' @docType data
#'
#' @format .csv file with seven columns and 936 rows:
#'   \describe{
#'     \item{mm-dd}{Recording date (dd-MMM format).}
#'     \item{hh:mm:ss}{Recording time of day (hh:mm:ss format).}
#'     \item{SmO2 Live}{Muscle oxygen saturation, raw signal (%).}
#'     \item{SmO2 Averaged}{Muscle oxygen saturation, rolling average (%).}
#'     \item{THb}{Total haemoglobin (arbitrary units).}
#'     \item{Lap}{Lap marker (integer). Not typically in use.}
#'     \item{Session Ct}{Session count of recordings.}
#'   }
#'
#'   Channel mapping for [read_mnirs()]:
#'   - `nirs_channels = c("SmO2 Live", "SmO2 Averaged", "THb")`
#'   - `time_channel = c("hh:mm:ss")`
#'   - `interval_times = list(start = c(124, 486, 848, 1210), end = c(364, 726, 1088, 1450))`
#'
#' @source Moxy Monitor (Fortiori Design LLC), exported via Moxy Portal App.
#'   (https://www.moxymonitor.com/)
#'
#' @seealso [read_mnirs()], [example_mnirs()]
#'
#' @examples
#' example_mnirs("moxy_intervals")
#'
NULL


#' @name train.red_intervals.csv
#'
#' @title 10 Hz Train.Red App export
#'
#' @description Exported from Train.Red app, recorded at 10 Hz.
#'   Containing two 5-minute cycling work intervals, placed on
#'   bilateral vastus lateralis muscle sites. Some data channels
#'   have been omitted to reduce file size.
#'
#' @docType data
#'
#' @format .csv file with header metadata and 10 columns and 11995 rows:
#'   \describe{
#'     \item{Timestamp (seconds passed)}{Elapsed time (s).}
#'     \item{Lap/Event}{Lap number (numeric).}
#'     \item{SmO2}{Muscle oxygen saturation, filtered (%). Two channels have
#'       duplicated names. If both are called, the second will be renamed to
#'       `SmO2_1`.}
#'     \item{SmO2 unfiltered}{Muscle oxygen saturation, raw signal (%). Two
#'       channels have duplicated names. If both are called, the second will
#'       be renamed to `SmO2 unfiltered_1`.}
#'     \item{O2HB unfiltered}{Oxyhaemoglobin concentration, raw signal
#'       (arbitrary units). Two channels have duplicated names. If both are
#'       called, the second will be renamed to `O2HB unfiltered_1`.}
#'     \item{HHB unfiltered}{Deoxyhaemoglobin concentration, raw signal
#'       (arbitrary units). Two channels have duplicated names. If both are
#'       called, the second will be renamed to `HHb unfiltered_1`.}
#'   }
#'
#'   Channel mapping for [read_mnirs()]:
#'   - `nirs_channels = c("SmO2", "SmO2 unfiltered", "O2HB unfiltered", "HHb unfiltered")`
#'   - `time_channel = c("Timestamp (seconds passed)")`
#'   - `event_channel = c("Lap/Event")`
#'   - `interval_times = list(start = c(2150.09, 2872.28), end = c(2452.26, 3167.98))`
#'   - `interval_times = list(start = c(65.94, 788.13), end = c(368.11, 1083.83))` from zero_time
#'
#' @source Train.Red (Train.Red B.V.), exported via Train.Red app
#'   (https://train.red/)
#'
#' @seealso [read_mnirs()], [example_mnirs()]
#'
#' @examples
#' example_mnirs("train.red")
#'
NULL


#' @name artinis_intervals.xlsx
#'
#' @title 10 Hz Artinis Oxysoft export recorded with Oxymon MKIII
#'
#' @description Exported from Artinis Oxysoft, recorded on Oxymon MKIII at
#'   50 Hz and exported at 10 Hz. Containing two 5-minute cycling work
#'   intervals and an ischaemic occlusion, placed on the vastus lateralis
#'   muscle site.
#'
#' @docType data
#'
#' @format .xlsx file with header metadata and five columns and 20919 rows:
#'   \describe{
#'     \item{Column 1}{Sample index (divide by sample rate for seconds).}
#'     \item{Column 2}{O2Hb: oxyhaemoglobin concentration change (\eqn{\mu}M).}
#'     \item{Column 3}{HHb: deoxyhaemoglobin concentration change (\eqn{\mu}M).}
#'     \item{Column 4}{Event marker (character).}
#'     \item{Column 5}{Unmarked event label (character).}
#'   }
#'
#'   Channel mapping for [read_mnirs()]:
#'   - `nirs_channels = c(O2Hb = 2, HHb = 3)`
#'   - `time_channel = c(sample = 1)`
#'   - `event_channel = c(event = 4, label = "col_5")`
#'   - `interval_times = list(start = c(158, 999, 1750) end = c(493, 1333, 1961))` two intervals and post-exercise occlusion
#'
#' @source Artinis Medical Systems. Oxymon MKIII, exported via Oxysoft desktop
#'   software (https://artinis.com/)
#'
#' @seealso [read_mnirs()], [example_mnirs()]
#'
#' @examples
#' example_mnirs("artinis_intervals")
#'
NULL


#' @name portamon-oxcap.xlsx
#'
#' @title 10 Hz Artinis Oxysoft export recorded with Portamon
#'
#' @description Exported from Artinis Oxysoft, recorded on
#'   Portamon at 10 Hz on the vastus lateralis muscle. Containing
#'   two trials of repeated occlusion oxidative capacity testing,
#'   each with 17 occlusions.
#'
#' @docType data
#'
#' @format .xlsx file with header metadata and six columns and 7943 rows:
#'   \describe{
#'     \item{Column 1}{Sample index (divide by sample rate for seconds).}
#'     \item{Column 2}{tHb: total haemoglobin concentration change
#'       (\eqn{\mu}M).}
#'     \item{Column 3}{HHb: deoxyhaemoglobin concentration change (\eqn{\mu}M).}
#'     \item{Column 4}{O2Hb: oxyhaemoglobin concentration change (\eqn{\mu}M).}
#'     \item{Column 5}{Event marker (character).}
#'     \item{Column 6}{Unmarked event label (character).}
#'   }
#'
#'   Channel mapping for [read_mnirs()]:
#'   - `nirs_channels = c(THb = 2, HHb = 3, O2Hb = 4)`
#'   - `time_channel = c(sample = 1)`
#'   - `event_channel = c(event = 5, label = "col_6")`
#'
#' @source Artinis Medical Systems. Portamon, exported via Oxysoft desktop
#'   software (https://artinis.com/)
#'
#' @seealso [read_mnirs()], [example_mnirs()]
#'
#' @examples
#' example_mnirs("portamon")
#'
NULL

#' Read raw data frame from file path
#' @keywords internal
read_file <- function(file_path) {
    ## validation: check file exists
    if (!file.exists(file_path)) {
        cli_abort(c(
            "x" = "File not found. Check that file exists.",
            "i" = "{.arg file_path} = {.path {file_path}}"
        ))
    }

    ## import data_raw from either excel or csv
    if (grepl("\\.csv$", file_path, ignore.case = TRUE)) {
        ## sample lines for separator and column count detection
        lines <- readLines(file_path, warn = FALSE)
        nrows <- length(lines)

        ## detect separator: comma vs tab from first 10 lines
        head_lines <- lines[seq_len(min(10L, nrows))]
        sep <- if (any(grepl("\t", head_lines))) "\t" else ","

        ## find the max number of separators from the end of the data file
        tail_lines <- lines[seq(to = nrows, by = 1L, len = min(50, nrows))]
        n_seps <- max(lengths(gregexpr(sep, tail_lines, fixed = TRUE)))

        ## pad the first line so fread infers the correct column count
        lines <- c(strrep(sep, n_seps), lines)

        ## read with explicit sep and column count to handle
        ## irregular header rows with fewer columns than data
        data_raw <- data.frame(
            data.table::fread(
                text = lines,
                header = FALSE,
                fill = Inf,
                sep = sep,
                colClasses = "character",
            )[-1, ]
        )
    } else if (grepl("\\.xls(x)?$", file_path, ignore.case = TRUE)) {
        ## report error when file is open and cannot be accessed by readxl
        data_raw <- tryCatch(
            readxl::read_excel(
                path = file_path,
                col_names = FALSE,
                col_types = "text",
                .name_repair = "minimal"
            ),
            error = \(e) {
                if (grepl("cannot be opened", e$message)) {
                    cli_abort(c(
                        "{e}",
                        "x" = "File cannot be opened.",
                        "i" = "Check the file is not in use by another \\
                        application."
                    ))
                } else {
                    cli_abort(e$message)
                }
            }
        )
        data_raw <- data.frame(data_raw)
    } else {
        ## validation: check file types
        cli_abort(c(
            "i" = "{.arg file_path} = {.path {file_path}}",
            "x" = "Unsupported file type.",
            "i" = "Only {.var .csv} or {.var .xls(x)} supported."
        ))
    }

    return(data_raw)
}


#' Known channel names and detection patterns for supported mNIRS devices
#' @keywords internal
device_patterns <- list(
    Artinis = list(
        ## keep `time_channel = NULL` to force `detect_time_channel` message
        time_channel = NULL,
        nirs_channels = c("2"),
        pattern = "^(\\d+ )+(\\d+|NA) ?$",
        fixed = FALSE
    ),
    Train.Red = list(
        time_channel = c("Timestamp (seconds passed)"),
        nirs_channels = c("SmO2"),
        pattern = c("Timestamp (seconds passed)", "SmO2"),
        fixed = TRUE
    ),
    Moxy = list(
        time_channel = c("hh:mm:ss"),
        nirs_channels = c("SmO2 Live"),
        pattern = c("hh:mm:ss", "SmO2 Live"),
        fixed = TRUE
    ),
    VO2master = list(
        time_channel = c("Time[s]"),
        nirs_channels = c("SmO2[%]"),
        pattern = c("Time[s]", "SmO2[%]"),
        fixed = TRUE
    ),
    PerfPro = list(
        time_channel = c("Time"),
        nirs_channels = NULL,
        pattern = c("Time.*SmO2"),
        fixed = FALSE
    )
)


#' Datetime format strings for POSIXct parsing
#' @keywords internal
datetime_formats <- c(
    "%H:%M:%OS",
    "%Y-%m-%dT%H:%M:%OS",
    "%Y-%m-%dT%H:%M:%OS%z",
    "%Y-%m-%d %H:%M:%OS",
    "%Y/%m/%d %H:%M:%OS",
    "%d-%m-%Y %H:%M:%OS",
    "%d/%m/%Y %H:%M:%OS"
)


#' Detect mnirs device from file metadata
#' @keywords internal
detect_mnirs_device <- function(data) {
    ## find the first row in `string` where all `patterns` match
    find_row <- function(string, patterns, fixed = TRUE) {
        Find(\(.i) {
            all(vapply(patterns, \(.x) {
                    grepl(.x, string[.i], fixed = fixed)
                }, logical(1L)))
        }, seq_along(string))
    }

    ## collapse each row to a single string for pattern matching
    data_strings <- apply(data, 1L, paste, collapse = " ")

    ## find first row of `data_strings` which matches any of `device_patterns`
    matched_row <- Find(\(.i) {
        any(vapply(device_patterns, \(.d) {
            !is.null(find_row(data_strings[.i], .d$pattern, .d$fixed))
        }, logical(1L)))
    }, seq_along(data_strings))

    if (is.null(matched_row)) {
        return(list(nirs_device = NULL, header_row = 1L))
    }

    ## return the first device name which matches the row of `data_strings`
    device_name <- Find(\(.nm) {
        .d <- device_patterns[[.nm]]
        !is.null(find_row(data_strings[matched_row], .d$pattern, .d$fixed))
    }, names(device_patterns))

    return(list(nirs_device = device_name, header_row = matched_row))
}


#' Detect known channels for a device
#' @keywords internal
detect_device_channels <- function(
    data,
    header_row = 1L,
    nirs_device = NULL,
    nirs_channels = NULL,
    time_channel = NULL,
    keep_all = FALSE,
    verbose = TRUE
) {
    ## user-specified channels always take priority
    if (!is.null(nirs_channels)) {
        return(list(
            ## if `time_channel = NULL` defined at `detect_time_channel`
            time_channel = time_channel,
            nirs_channels = nirs_channels,
            keep_all = keep_all
        ))
    }

    ## need device detection when is.null(nirs_channel)
    if (is.null(nirs_device)) {
        cli_abort(c(
            "x" = "{.arg nirs_channels} cannot be determined automatically.",
            "i" = "Define {.arg nirs_channels} explicitly."
        ))
    }

    ## successfully detected `nirs_device` with `nirs_channels = NULL`
    ch_list <- device_patterns[[nirs_device]]
    
    if (nirs_device == "PerfPro") {
        ch_list$nirs_channels <- Find(\(.x) {
            startsWith(.x, "SmO2")
        }, data[header_row, ])
    }
    
    ch_list <- list(
        ## user-specified `time_channel` takes priority here
        time_channel = time_channel %||% ch_list$time_channel,
        nirs_channels = ch_list$nirs_channels,
        keep_all = TRUE ## return all cols to view potential nirs_channels
    )

    if (verbose) {
        cli_inform(c(
            "!" = "{.val {nirs_device}} file format detected. \\
            {.arg nirs_channels} set to {.val {ch_list$nirs_channels}}.",
            "i" = "Override by specifying {.arg nirs_channels} explicitly."
        ))
    }

    return(ch_list)
}


#' Read data table from raw data
#' @keywords internal
read_data_table <- function(
    data,
    nirs_channels = NULL,
    header_row = 1L
) {
    nrows <- nrow(data)
    ## find the first row where ALL nirs_channels match
    ## start with `header_row` passed from `detect_mnirs_device()`
    header_row <- Find(\(.i) {
        all(nirs_channels %in% data[.i, ])
    }, c(header_row, seq_len(nrows)))

    ## validation: all channels must be detected to extract the data frame
    ## return error if channels string is detected at multiple rows
    if (length(header_row) == 0) {
        cli_abort(c(
            "x" = "Channel names not detected.",
            "i" = "Column names are case sensitive and must match exactly."
        ))
    }

    ## extract the data_table, and name by header row
    table_rows <- (header_row + 1L):nrows
    data_table <- setNames(data[table_rows, ], data[header_row, ])
    file_header <- data[seq_len(header_row), ]

    return(list(
        file_header = file_header,
        data_table = data_table
    ))
}


#' Extract earliest POSIXct value from file header metadata
#' @keywords internal
extract_start_timestamp <- function(file_header) {
    header_values <- unlist(file_header, use.names = FALSE)
    header_values <- header_values[!is_empty(header_values)]

    ## search for POSIXct values, return the earliest time value
    ## vulnerable to invalid timestamps
    parsed <- which(!is.na(vapply(header_values, \(.x) {
        as.POSIXct(.x, tryFormats = datetime_formats, optional = TRUE)
    }, numeric(1L))))

    if (length(parsed) == 0L) {
        return(NULL)
    }

    ## return the earliest character string timestamp, assuming start time
    return(min(header_values[parsed]))
}


#' Detect time_channel from header row
#' @keywords internal
detect_time_channel <- function(
    data,
    time_channel = NULL,
    nirs_device = NULL,
    verbose = TRUE
) {
    if (!is.null(time_channel)) {
        return(time_channel)
    }

    ## return default sample column for Artinis Oxysoft
    ## when `nirs_channels` defined but `time_channel = NULL`
    if (!is.null(nirs_device) && nirs_device == "Artinis") {
        if (verbose) {
            cli_inform(c(
                "!" = "Oxysoft {.val sample} column detected."
            ))
        }
        return(c(sample = "1"))
    }

    col_names <- names(data)

    ## match column names to possible time column names
    time_regex <- "time|duration|hms|h+:m+:s+"
    time_idx <- grep(time_regex, col_names, ignore.case = TRUE)[1L]

    ## find name of POSIXct column
    if (is.na(time_idx)) {
        time_idx <- Position(\(.col) inherits(.col, "POSIXct"), data)
    }

    ## find name of character column with time format strings
    if (is.na(time_idx)) {
        time_idx <- Position(\(.col) {
            is.character(.col) && {
                val <- .col[which(!is.na(.col))[1L]]
                !is.na(val) && grepl("^\\d{1,2}:\\d{2}(:\\d{2})?", val)
            }
        }, data)
    }

    if (!is.na(time_idx)) {
        if (verbose) {
            cli_inform(c(
                "!" = "Detected {.arg time_channel} = \\
                {col_blue(col_names[time_idx])}."
            ))
        }
        return(col_names[time_idx])
    }

    cli_abort(c(
        "x" = "{.arg time_channel} not detected.",
        "i" = "Define {.arg time_channel} explicitly."
    ))
}


#' Detect empty or NA strings
#' @keywords internal
is_empty <- function(x) {
    is.na(x) | x == ""
}


#' Rename duplicate strings in a vector with `make.unique()`
#' @keywords internal
rename_duplicates <- function(x) {
    if (is.null(x)) {
        return(x)
    }
    ## rename blank strings
    empty <- which(is_empty(x))
    x[empty] <- paste0("col_", empty)

    return(make.unique(x, sep = "_"))
}


#' Force names on character strings
#' @keywords internal
name_channels <- function(x) {
    ## if channels not named, names from object
    names <- names(x) %||% character(length(x))
    empty_names <- is_empty(names)
    names[empty_names] <- as.character(x)[empty_names]
    return(setNames(x, names))
}


#' Select data table columns and rename from channels, handling duplicates
#' @keywords internal
select_rename_data <- function(
    data,
    nirs_channels,
    time_channel,
    event_channel = NULL,
    keep_all = FALSE,
    verbose = TRUE
) {
    ## ensure all channel inputs are named (name = original_col_name)
    ch_list <- list(
        time_channel = time_channel,
        event_channel = event_channel,
        nirs_channels = nirs_channels
    ) |>
        lapply(\(.x) if (is.null(.x)) .x else name_channels(.x))

    ## original column names (values) mapped to user names (names)
    ## rename_duplicates makes user-facing names unique
    orig_names <- lapply(ch_list, \(.x) {
        if (is.null(.x)) NULL else rename_duplicates(as.character(.x))
    })
    user_names <- lapply(ch_list, \(.x) {
        if (is.null(.x)) NULL else rename_duplicates(names(.x))
    })

    ## flat vectors for column matching
    orig_vec <- unlist(orig_names, use.names = FALSE)
    user_vec <- unlist(user_names, use.names = FALSE)

    ## de-duplicate data column names
    data_names <- rename_duplicates(names(data))

    ## check channels exist in data
    missing <- setdiff(orig_vec, data_names)
    if (length(missing) > 0L) {
        cli_abort(c(
            "x" = "Channel names not detected.",
            "i" = "Column names are case sensitive and must match exactly."
        ))
    }

    ## keep all columns or only specified channels
    selected_cols <- if (keep_all) {
        c(orig_vec, setdiff(data_names, orig_vec))
    } else {
        orig_vec
    }

    ## select and rename: channel columns get user names,
    ## remaining columns keep de-duplicated data names
    result <- setNames(data, data_names)[, selected_cols, drop = FALSE]
    channel_idx <- match(orig_vec, names(result))

    ## resolve clashes: user names take priority over data names
    all_names <- rename_duplicates(c(user_vec, names(result)))
    names(result) <- all_names[!all_names %in% user_vec]
    names(result)[channel_idx] <- user_vec

    ## warn if any channels were renamed from their input names
    renamed <- user_vec != unlist(lapply(ch_list, names), use.names = FALSE)
    if (verbose && any(renamed)) {
        old <- unlist(lapply(ch_list, names), use.names = FALSE)[renamed]
        new <- user_vec[renamed]
        cli_warn(c(
            "!" = "Duplicate channel names detected.",
            "i" = "Renamed: {.val {paste(old, new, sep = ' = ')}}",
            "i" = "Unique channel names can be defined explicitly."
        ))
    }

    return(list(
        data = result,
        nirs_channel = user_names$nirs_channels,
        time_channel = user_names$time_channel,
        event_channel = user_names$event_channel
    ))
}


#' Standardise comma decimals to periods in character columns
#' @keywords internal
convert_type <- function(
    data,
    time_channel,
    event_channel = NULL,
    verbose = TRUE
) {
    colnames <- names(data)
    ## convert decimal "," to "."
    char_cols <- setdiff(
        colnames[vapply(data, is.character, logical(1L))],
        time_channel
    )
    for (col in char_cols) {
        data.table::set(
            data, j = col, value = gsub(",", ".", data[[col]], fixed = TRUE)
        )
    }

    ## convert column types
    data <- utils::type.convert(
        data, na.strings = c("NA", ""), dec = ".", as.is = TRUE
    )

    ## coerce integer columns to numeric (except event_channel)
    int_cols <- setdiff(
        colnames[vapply(data, is.integer, logical(1L))],
        event_channel
    )
    data[int_cols] <- lapply(data[int_cols], as.numeric)

    ## standardise Inf/NaN/empty to NA
    data[] <- lapply(data, \(.x) {
        if (is.character(.x)) {
            .x[.x %in% c("", "NA")] <- NA_character_
        } else if (is.integer(.x)) {
            .x[!is.finite(.x)] <- NA_integer_
        } else if (is.numeric(.x)) {
            .x[!is.finite(.x)] <- NA_real_
        }
        .x
    })

    return(data)
}


#' Remove Empty Rows and Columns
#' @keywords internal
remove_empty_rows_cols <- function(data) {
    data <- data[rowSums(!is_empty(data)) > 0, , drop = FALSE]
    return(data[, colSums(!is_empty(data)) > 0, drop = FALSE])
}


#' Parse time_channel character or dttm to numeric
#' @keywords internal
parse_time_channel <- function(
    data,
    time_channel,
    start_timestamp = NULL,
    add_timestamp = FALSE,
    zero_time = FALSE
) {
    time_vec <- data[[time_channel]]

    ## fractional unix time to POSIXct coerced to local time zone
    if (is.numeric(time_vec) && all(time_vec <= 1, na.rm = TRUE)) {
        time_vec <- as.POSIXct(
            as.character(as.POSIXct(Sys.Date(), "UTC")),
            tz = Sys.timezone()
        ) + time_vec * 86400
    }

    ## recalculate numeric time to start from zero
    if (zero_time && is.numeric(time_vec)) {
        time_vec <- time_vec - time_vec[1L]
    }

    ## character time to POSIXct
    if (is.character(time_vec)) {
        time_vec <- as.POSIXct(
            time_vec, tryFormats = datetime_formats, optional = TRUE
        )
    }

    ## preserve POSIXct timestamp and convert to numeric seconds
    timestamp_vec <- NULL
    if (inherits(time_vec, "POSIXct")) {
        timestamp_vec <- time_vec
        time_vec <- as.numeric(difftime(time_vec, time_vec[1L], units = "secs"))
    }

    data[[time_channel]] <- time_vec

    ## add_timestamp preserves or adds POSIXct/dttm column
    if (add_timestamp) {
        ## add "timestamp" col after `time_channel` position
        col_names <- names(data)
        time_idx <- match(time_channel, col_names)
        data_names <- append(col_names, "timestamp", time_idx)
        data$timestamp <- NA_real_
        data <- data[data_names]

        ## if neither header start_timestamp or timestamp_vec exist
        ## then return NULL and don't append column
        if (!is.null(start_timestamp)) {
            start_time <- as.POSIXct(
                start_timestamp, tryFormats = datetime_formats, optional = TRUE
            )
            data$timestamp <- start_time + time_vec
        } else if (!is.null(timestamp_vec)) {
            data$timestamp <- timestamp_vec
        } else {
            ## column removed if no timestamp detected
            data$timestamp <- NULL
        }
    }

    ## extract earliest POSIXct value as start_timestamp metadata
    ## if not already passed from file header
    if (is.null(start_timestamp) && !is.null(timestamp_vec)) {
        start_timestamp <- min(timestamp_vec, na.rm = TRUE)
    }

    return(list(data = data, start_timestamp = start_timestamp))
}


#' Validate and Estimate Sample Rate
#' @keywords internal
parse_sample_rate <- function(
    data,
    file_header,
    time_channel,
    sample_rate = NULL,
    nirs_device = NULL,
    verbose = TRUE
) {
    ## if Oxysoft, sample_rate will be detected = 1
    ## extract and overwrite with exported sample_rate
    ## create new "time" column at col_idx behind `time_channel`
    if (!is.null(nirs_device) && nirs_device == "Artinis") {
        pos <- which(file_header == "Export sample rate", arr.ind = TRUE)
        sample_rate <- as.numeric(file_header[pos[1L], pos[2L] + 1L])

        col_names <- names(data)
        time_vec <- data[[time_channel]] / sample_rate
        time_idx <- match(time_channel, col_names)
        data_names <- append(col_names, "time", time_idx)
        data_names <- rename_duplicates(data_names)
        time_channel <- setdiff(data_names, col_names)
        data[[time_channel]] <- time_vec
        data <- data[data_names]

        if (verbose) {
            cli_inform(c(
                "!" = "Oxysoft {.arg sample_rate} = {.val {sample_rate}} Hz.",
                "i" = "{.arg time_channel} = {.val {time_channel}} added to \\
                the data frame, in {.cls seconds}."
            ))
        }
    }

    ## validate priority user input sample_rate
    ## metadata check will be skipped
    ## will estimate from time_channel (time_channel)
    ## will error on unable to estimate sample_rate
    sample_rate <- validate_sample_rate(
        data, time_channel, sample_rate, verbose
    )

    return(list(
        data = data,
        time_channel = time_channel,
        sample_rate = sample_rate
    ))
}


#' Report warnings for unbalanced time_channel samples
#' @keywords internal
detect_irregular_samples <- function(
    x,
    time_channel,
    verbose = TRUE
) {
    if (!verbose) {
        return(invisible())
    }

    ## find duplicated, unordered, or big gaps in samples
    diffs <- diff(x)
    irregular_idx <- c(
        which(duplicated(x)), which(diffs < 0), which(diffs >= 3600)
    )

    ## silence if no irregular samples
    if (length(irregular_idx) == 0) {
        return(invisible())
    }

    irregular_vec <- round(unique(x[irregular_idx]), 6)

    info_msg <- if (length(irregular_vec) > 5L) {
        ## if more than 5 irregular samples, print the first three
        irregular_display <- irregular_vec[seq_len(3L)]

        "Investigate at {.arg {time_channel}} = {.val {irregular_display}} \\
        and {length(irregular_vec) - 3L} more."
    } else {
        ## if 5 or fewer irregular samples, print all of them
        "Investigate at {.arg {time_channel}} = {.val {irregular_vec}}."
    }

    cli_warn(c(
        "!" = "Duplicate or irregular {.arg time_channel} samples detected.",
        "i" = info_msg,
        "i" = "Re-sample with {.fn mnirs::resample_mnirs}."
    ))

    return(invisible())
}

## example_mnirs() ====================================================
test_that("example_mnirs() returns all files when file = NULL", {
    files <- example_mnirs()
    expect_type(files, "character")
    expect_true(length(files) > 0)
})

test_that("example_mnirs() returns valid path for exact match", {
    # Assumes at least one file exists in extdata
    all_files <- example_mnirs()
    skip_if(length(all_files) == 0, "No example files available")

    path <- example_mnirs(all_files[1])
    expect_type(path, "character")
    expect_true(file.exists(path))
    expect_match(path, all_files[1], fixed = TRUE)
})

test_that("example_mnirs() returns valid path for partial match", {
    skip_if_not(
        any(grepl("moxy_ramp", example_mnirs(), fixed = TRUE)),
        "moxy_ramp.xlsx not available"
    )

    path <- example_mnirs("moxy_ramp")
    expect_true(file.exists(path))
    expect_match(path, "moxy_ramp", fixed = TRUE)
})

test_that("example_mnirs() errors on multiple partial matches", {
    skip_if_not(
        sum(grepl("moxy", example_mnirs(), fixed = TRUE)) > 1,
        "Multiple moxy files not available"
    )

    expect_error(example_mnirs("moxy"), "Multiple files match")
})

test_that("example_mnirs() errors on non-existent file", {
    expect_error(
        example_mnirs("nonexistent_file_xyz"),
        "'arg' should be one of"
    )
})

test_that("example_mnirs() does not show files with `~`", {
    # Create test directory structure
    test_dir <- file.path(tempdir(), "epl_test")
    dir.create(test_dir, recursive = TRUE, showWarnings = FALSE)

    # Create test files
    file.create(file.path(test_dir, "data.csv"))
    file.create(file.path(test_dir, "~temp.csv"))
    file.create(file.path(test_dir, "results.xlsx"))

    result <- list.files(test_dir)
    expect_length(result, 3)
    expect_true(any(grepl("^~", result)))

    result <- list.files(test_dir, pattern = "^[^~]")

    expect_length(result, 2)
    expect_false(any(grepl("^~", result)))
    expect_true(all(c("data.csv", "results.xlsx") %in% result))
})


## read_file() ==================================================

test_that("read_file() reads moxy (perfpro) xlsx files correctly", {
    file_path <- example_mnirs("moxy_ramp")
    skip_if(!grepl("\\.xls(x)?$", file_path, ignore.case = TRUE))

    result <- read_file(file_path)

    expect_s3_class(result, "data.frame")
    expect_equal(ncol(result), 5)
    expect_gt(nrow(result), 0)
    expect_all_true(unlist(lapply(result, is.character)))
})

test_that("read_file() reads moxy .csv files correctly", {
    file_path <- example_mnirs("moxy_intervals")
    skip_if(!grepl("\\.csv$", file_path, ignore.case = TRUE))

    result <- read_file(file_path)

    expect_s3_class(result, "data.frame")
    expect_equal(ncol(result), 7)
    expect_gt(nrow(result), 0)
    expect_all_true(unlist(lapply(result, is.character)))
})

test_that("read_file() reads explicit comma train.red csv correctly", {
    file_path <- example_mnirs("train.red")
    skip_if(!grepl("\\.csv$", file_path, ignore.case = TRUE))

    result <- read_file(file_path)

    expect_s3_class(result, "data.frame")
    expect_equal(ncol(result), 12)
    expect_gt(nrow(result), 0)
    expect_all_true(unlist(lapply(result, is.character)))
})

test_that("read_file() reads train.red files correctly", {
    file_path <- test_path("testdata/train.red-mre.csv")
    skip_if_not(file.exists(file_path), "testdata not available")

    result <- read_file(file_path)

    expect_s3_class(result, "data.frame")
    expect_equal(ncol(result), 23)
    expect_gt(nrow(result), 0)
    expect_all_true(unlist(lapply(result, is.character)))
})

test_that("read_file() reads vo2master files correctly", {
    file_path <- test_path("testdata/vo2master.csv")
    skip_if_not(file.exists(file_path), "testdata not available")

    result <- read_file(file_path)

    expect_s3_class(result, "data.frame")
    expect_gt(ncol(result), 12)
    expect_gt(nrow(result), 0)
    expect_all_true(unlist(lapply(result, is.character)))
})

test_that("read_file() errors", {
    expect_error(
        read_file("nonexistent_file.xlsx"),
        "File not found"
    )

    temp_file <- tempfile(fileext = ".txt")
    writeLines("test", temp_file)
    on.exit(unlink(temp_file))

    expect_error(
        read_file(temp_file),
        "Unsupported file type"
    )
})

test_that("read_file() errors when Excel file cannot be opened", {
    file_path <- example_mnirs("moxy_ramp")
    skip_if(!grepl("\\.xls(x)?$", file_path, ignore.case = TRUE))

    local_mocked_bindings(
        read_excel = function(...) stop("cannot be opened"),
        .package = "readxl"
    )

    expect_error(
        read_file(file_path),
        "File cannot be opened"
    )
})

test_that("read_file() re-throws other Excel read errors", {
    file_path <- example_mnirs("moxy_ramp")
    skip_if(!grepl("\\.xls(x)?$", file_path, ignore.case = TRUE))

    local_mocked_bindings(
        read_excel = function(...) stop("some other error"),
        .package = "readxl"
    )

    expect_error(
        read_file(file_path),
        "some other error"
    )
})


## detect_mnirs_device() ===============================================
test_that("detect_mnirs_device works on example files", {
    ## xlsx files
    expect_equal(
        read_file(example_mnirs("moxy_ramp")) |>
            detect_mnirs_device(),
        list(
            nirs_device = "Moxy",
            header_row = 6
        )
    )

    expect_equal(
        read_file(example_mnirs("artinis_intervals")) |>
            detect_mnirs_device(),
        list(
            nirs_device = "Artinis",
            header_row = 38
        )
    )

    ## csv files
    expect_equal(
        read_file(example_mnirs("train.red")) |>
            detect_mnirs_device(),
        list(
            nirs_device = "Train.Red",
            header_row = 41
        )
    )
})

test_that("detect_mnirs_device works on internal example files", {
    skip_on_ci()
    skip_on_covr()
    file_path <- test_path("testdata/train.red-mre.csv")
    skip_if_not(file.exists(file_path), "testdata not available")

    expect_equal(
        read_file(file_path) |>
            detect_mnirs_device(),
        list(
            nirs_device = "Train.Red",
            header_row = 521
        )
    )

    file_path <- test_path("testdata/vo2master.csv")
    skip_if_not(file.exists(file_path), "testdata not available")

    expect_equal(
        read_file(file_path) |>
            detect_mnirs_device(),
        list(
            nirs_device = "VO2master",
            header_row = 1
        )
    )
})

test_that("detect_mnirs_device() returns NULL when no match", {
    data <- data.frame(
        V1 = c("Unknown", "device", "data"),
        V2 = c("header", "col1", "val1")
    )

    expect_equal(
        detect_mnirs_device(data),
        list(nirs_device = NULL, header_row = 1)
    )
})


## detect_device_channels() ============================================
test_that("detect_device_channels() returns user channels when provided", {
    result <- detect_device_channels(
        nirs_device = "Moxy",
        nirs_channels = c(smo2 = "SmO2 Live"),
        time_channel = c(time = "hh:mm:ss"),
        verbose = FALSE
    )

    expect_equal(result$nirs_channels, c(smo2 = "SmO2 Live"))
    expect_equal(result$time_channel, c(time = "hh:mm:ss"))
})

test_that("detect_device_channels() returns user channels even with NULL device", {
    result <- detect_device_channels(
        nirs_device = NULL,
        nirs_channels = "O2Hb",
        time_channel = "Time",
        verbose = FALSE
    )

    expect_equal(result$nirs_channels, "O2Hb")
    expect_equal(result$time_channel, "Time")
})

test_that("detect_device_channels() detects known channels for device", {
    result <- detect_device_channels(
        nirs_device = "Moxy",
        nirs_channels = NULL,
        time_channel = NULL,
        verbose = FALSE
    )

    expect_equal(result$nirs_channels, device_patterns$Moxy$nirs_channels)
    expect_equal(result$time_channel, device_patterns$Moxy$time_channel)
})

test_that("detect_device_channels() detects known channels for PerfPro", {
    file_path <- test_path("testdata/perfpro-mre.xlsx")
    skip_if_not(file.exists(file_path), "testdata not available")

    data <- read_file(file_path)
    detected_list <- detect_mnirs_device(data)
    nirs_device <- detected_list$nirs_device
    header_row <- detected_list$header_row

    expect_match(nirs_device, "PerfPro")
    expect_equal(header_row, 3)

    result <- detect_device_channels(
        data,
        header_row,
        nirs_device,
        nirs_channels = NULL,
        time_channel = NULL,
        verbose = FALSE
    )

    expect_equal(result$nirs_channels, "SmO2 (1614)")
    expect_equal(result$time_channel, "Time")
})

test_that("detect_device_channels() user time_channel overrides device default", {
    result <- detect_device_channels(
        nirs_device = "Moxy",
        nirs_channels = NULL,
        time_channel = c(time = "custom_time"),
        verbose = FALSE
    )

    expect_equal(result$nirs_channels, device_patterns$Moxy$nirs_channels)
    expect_equal(result$time_channel, c(time = "custom_time"))
})

test_that("detect_device_channels() errors when device is NULL and no channels", {
    expect_error(
        detect_device_channels(
            nirs_device = NULL,
            nirs_channels = NULL,
            verbose = FALSE
        ),
        "cannot be determined"
    )
})

test_that("detect_device_channels() verbose messages for detection", {
    expect_message(
        detect_device_channels(
            nirs_device = "Moxy",
            nirs_channels = NULL,
            verbose = TRUE
        ),
        "Moxy.*detected"
    )

    ## no message when user provides channels
    expect_no_message(
        detect_device_channels(
            nirs_device = "Moxy",
            nirs_channels = c("SmO2 Live"),
            verbose = TRUE
        )
    )
})

test_that("detect_device_channels() returns appropriate keep_all", {
    result <- detect_device_channels(
        nirs_device = "Moxy",
        nirs_channels = NULL,
        time_channel = c(time = "custom_time"),
        verbose = FALSE
    )

    expect_true(result$keep_all)

    result <- detect_device_channels(
        nirs_device = "Moxy",
        nirs_channels = "SmO2",
        time_channel = c(time = "custom_time"),
        verbose = FALSE
    )

    expect_false(result$keep_all)
})


## read_data_table() ===================================================
test_that("read_data_table() extracts data with valid channels", {
    data <- data.frame(
        V1 = c("meta1", "meta2", "O2Hb", "10", "20"),
        V2 = c("meta1", "meta2", "HHb", "5", "15"),
        V3 = c("meta1", "meta2", "Time", "0.1", "0.2"),
        stringsAsFactors = FALSE
    )

    result <- read_data_table(data, nirs_channels = c("O2Hb", "HHb"))

    expect_type(result, "list")
    expect_named(result, c("file_header", "data_table"))
    expect_s3_class(result$data_table, "data.frame")
    expect_s3_class(result$file_header, "data.frame")

    expect_equal(nrow(result$data_table), 2)
    expect_equal(ncol(result$data_table), 3)
    expect_equal(names(result$data_table), c("O2Hb", "HHb", "Time"))
    expect_true(all(result$data_table == data[4:5, ]))

    expect_equal(nrow(result$file_header), 3)
    expect_equal(ncol(result$file_header), 3)
    expect_true(all(result$file_header == data[1:3, ]))
})

test_that("read_data_table() errors when channels not found", {
    data <- data.frame(
        V1 = c("header", "WrongChannel", "10"),
        V2 = c("header", "Time", "0.1"),
        stringsAsFactors = FALSE
    )

    expect_error(
        read_data_table(data, "O2Hb"),
        "Channel names not detected"
    )
})

test_that("read_data_table() is case sensitive", {
    data <- data.frame(
        V1 = c("meta", "o2hb", "10"),
        V2 = c("meta", "time", "0.1"),
        stringsAsFactors = FALSE
    )

    expect_error(
        read_data_table(data, "O2Hb"),
        "case sensitive"
    )
})


## detect_time_channel() ==============================================
test_that("detect_time_channel returns provided time_channel", {
    df <- data.frame(x = 1:5, y = 6:10)
    expect_equal(
        detect_time_channel(df, time_channel = "custom", verbose = FALSE),
        "custom"
    )
})

test_that("detect_time_channel finds time column by name", {
    df <- data.frame(time = 1:5, value = 6:10)
    expect_equal(
        detect_time_channel(df, verbose = FALSE),
        "time"
    )

    df <- data.frame(Time = 1:5, value = 6:10)
    expect_equal(
        detect_time_channel(df, verbose = FALSE),
        "Time"
    )

    df <- tibble::tibble("hh:mm:ss" = 1:5, value = 6:10)
    expect_equal(
        detect_time_channel(df, verbose = FALSE),
        "hh:mm:ss"
    )

    df <- data.frame("hms" = 1:5, value = 6:10)
    expect_equal(
        detect_time_channel(df, verbose = FALSE),
        "hms"
    )

    df <- data.frame(duration = 1:5, value = 6:10)
    expect_equal(
        detect_time_channel(df, verbose = FALSE),
        "duration"
    )
})

test_that("detect_time_channel finds POSIXct column", {
    df <- data.frame(
        value = 1:5,
        posixct_col = as.POSIXct("2024-01-01 12:00:00") + 1:5
    )
    expect_equal(
        detect_time_channel(df, verbose = FALSE),
        "posixct_col"
    )
})

test_that("detect_time_channel finds character time format", {
    df <- data.frame(
        value = 1:5,
        string_col = c(
            "12:30:45",
            "12:30:46",
            "12:30:47",
            "12:30:48",
            "12:30:49"
        )
    )
    expect_equal(
        detect_time_channel(df, verbose = FALSE),
        "string_col"
    )

    # Test with H:MM format
    df <- data.frame(
        value = 1:5,
        string_col = c("1:30", "1:31", "1:32", "1:33", "1:34")
    )
    expect_equal(
        detect_time_channel(df, verbose = FALSE),
        "string_col"
    )
})

test_that("detect_time_channel handles NA values in character column", {
    df <- data.frame(
        value = 1:5,
        time_str = c(NA, "12:30:45", "12:30:46", "12:30:47", "12:30:48")
    )
    expect_equal(
        detect_time_channel(df, verbose = FALSE),
        "time_str"
    )
})

test_that("detect_time_channel errors when no time column found", {
    df <- data.frame(x = 1:5, y = 6:10)
    expect_error(
        detect_time_channel(df, verbose = FALSE),
        "time_channel.*not detected"
    )
})

test_that("detect_time_channel prioritises time_channel argument", {
    df <- data.frame(
        time = 1:5,
        custom = 6:10,
        timestamp = as.POSIXct("2024-01-01 12:00:00") + 1:5
    )
    expect_equal(
        detect_time_channel(df, time_channel = "custom", verbose = FALSE),
        "custom"
    )
})

test_that("detect_time_channel verbose messages work", {
    df <- data.frame(time = 1:5, value = 6:10)
    expect_message(
        detect_time_channel(df, verbose = TRUE),
        "Detected.*time_channel"
    )
})


## rename_duplicates() ================================================
test_that("rename_duplicates() handles duplicate strings", {
    x <- c("O2Hb", "HHb", "O2Hb", "Time")
    result <- rename_duplicates(x)

    expect_equal(result, c("O2Hb", "HHb", "O2Hb_1", "Time"))
    expect_true(all(!duplicated(result)))
})

test_that("rename_duplicates() handles empty strings", {
    x <- c("O2Hb", "", "HHb", "")
    result <- rename_duplicates(x)

    expect_equal(result, c("O2Hb", "col_2", "HHb", "col_4"))
})

test_that("rename_duplicates() handles NA values", {
    x <- c("O2Hb", NA, "HHb", NA)
    result <- rename_duplicates(x)

    expect_equal(result, c("O2Hb", "col_2", "HHb", "col_4"))
})

test_that("rename_duplicates() handles NULL input", {
    expect_null(rename_duplicates(NULL))
})

test_that("rename_duplicates() handles all unique values", {
    x <- c("O2Hb", "HHb", "Time")
    result <- rename_duplicates(x)

    expect_equal(result, x)
})

test_that("rename_duplicates() handles multiple duplicates", {
    x <- c("O2Hb", "O2Hb", "O2Hb")
    result <- rename_duplicates(x)

    expect_equal(result, c("O2Hb", "O2Hb_1", "O2Hb_2"))
})

## name_channels() ====================================================
test_that("name_channels() names unnamed vector", {
    x <- c("O2Hb", "HHb", "Time")
    result <- name_channels(x)

    expect_equal(names(result), c("O2Hb", "HHb", "Time"))
    expect_equal(as.character(result), x)
})

test_that("name_channels() preserves existing names", {
    x <- c(oxy = "O2Hb", deoxy = "HHb", time = "Time")
    result <- name_channels(x)

    expect_equal(names(result), c("oxy", "deoxy", "time"))
    expect_equal(as.character(result), c("O2Hb", "HHb", "Time"))
})

test_that("name_channels() fills in missing names", {
    x <- c(oxy = "O2Hb", "HHb", time = "Time")
    result <- name_channels(x)

    expect_equal(names(result), c("oxy", "HHb", "time"))
})

test_that("name_channels() handles NA names", {
    x <- c("O2Hb", "HHb")
    names(x) <- c(NA, "deoxy")
    result <- name_channels(x)

    expect_equal(names(result), c("O2Hb", "deoxy"))
})

test_that("name_channels() handles all empty names", {
    x <- c("O2Hb", "HHb", "Time")
    names(x) <- c("", "", "")
    result <- name_channels(x)

    expect_equal(names(result), c("O2Hb", "HHb", "Time"))
})


## select_rename_data() ===========================================
test_that("select_rename_data() selects and renames channels in order", {
    data <- data.frame(
        O2Hb = c("10", "20"),
        HHb = c("5", "15"),
        Time = c("0.1", "0.2"),
        stringsAsFactors = FALSE
    )

    result <- select_rename_data(
        data,
        nirs_channels = c(oxy = "O2Hb", deoxy = "HHb"),
        time_channel = c(time = "Time"),
        verbose = FALSE
    )

    expect_equal(names(result$data), c("time", "oxy", "deoxy"))
    expect_equal(result$nirs_channel, c("oxy", "deoxy"))
    expect_equal(result$time_channel, c("time"))
})

test_that("select_rename_data() works with unnamed channels", {
    data <- data.frame(
        O2Hb = c("10"),
        Time = c("0.1"),
        stringsAsFactors = FALSE
    )

    result <- select_rename_data(
        data,
        nirs_channels = "O2Hb",
        time_channel = "Time",
        verbose = FALSE
    )

    expect_equal(names(result$data), c("Time", "O2Hb"))
    expect_equal(result$nirs_channel, "O2Hb")
    expect_equal(result$time_channel, "Time")
})

test_that("select_rename_data() includes event channel", {
    data <- data.frame(
        O2Hb = c("10"),
        Time = c("0.1"),
        Event = c("Start"),
        stringsAsFactors = FALSE
    )

    result <- select_rename_data(
        data,
        nirs_channels = "O2Hb",
        time_channel = "Time",
        event_channel = "Event",
        verbose = FALSE
    )

    expect_equal(names(result$data), c("Time", "Event", "O2Hb"))
    expect_equal(result$event_channel, "Event")
})

test_that("select_rename_data() handles duplicate channel names", {
    skip('currently returns "Channel names not detected error"')

    data <- data.frame(
        O2Hb = c("10"),
        Time = c("0.1"),
        stringsAsFactors = FALSE
    )

    expect_warning(
        result <- select_rename_data(
            data,
            nirs_channels = c("O2Hb", "O2Hb"),
            time_channel = "Time",
            verbose = TRUE
        ),
        "Duplicated channel names"
    )

    expect_equal(result$nirs_channel, c("O2Hb", "O2Hb_1"))
})

test_that("select_rename_data() handles duplicate data columns", {
    data <- data.frame(
        O2Hb = c("10"),
        O2Hb = c("20"),
        Time = c("0.1"),
        check.names = FALSE,
        stringsAsFactors = FALSE
    )

    result <- select_rename_data(
        data,
        nirs_channels = c(oxy1 = "O2Hb", oxy2 = "O2Hb"),
        time_channel = "Time",
        verbose = FALSE
    )

    expect_equal(names(result$data), c("Time", "oxy1", "oxy2"))
})

test_that("select_rename_data() keeps all columns with keep_all", {
    data <- data.frame(
        O2Hb = c("10"),
        HHb = c("5"),
        Time = c("0.1"),
        Extra = c("x"),
        stringsAsFactors = FALSE
    )

    result <- select_rename_data(
        data,
        nirs_channels = c(o2hb = "O2Hb", hhb = "HHb"),
        time_channel = c(time = "Time"),
        keep_all = TRUE,
        verbose = FALSE
    )

    expect_equal(ncol(result$data), 4)
    expect_true(all(
        c("o2hb", "hhb", "time", "Extra") %in% names(result$data)
    ))
})

test_that("select_rename_data() drops extra columns with default keep_all = FALSE", {
    data <- data.frame(
        O2Hb = c("10"),
        HHb = c("5"),
        Time = c("0.1"),
        Extra = c("x"),
        stringsAsFactors = FALSE
    )

    result <- select_rename_data(
        data,
        nirs_channels = "O2Hb",
        time_channel = "Time",
        # keep_all = FALSE,
        verbose = FALSE
    )

    expect_equal(ncol(result$data), 2)
    expect_false("Extra" %in% names(result$data))
})

test_that("select_rename_data() errors when channel not found", {
    data <- data.frame(
        O2Hb = c("10"),
        Time = c("0.1"),
        stringsAsFactors = FALSE
    )

    expect_error(
        select_rename_data(
            data,
            nirs_channels = "HHb",
            time_channel = "Time"
        ),
        "Channel names not detected"
    )
})

test_that("select_rename_data() suppresses warnings with verbose", {
    data <- data.frame(
        O2Hb = c("10"),
        O2Hb = c("20"),
        Time = c("0.1"),
        check.names = FALSE,
        stringsAsFactors = FALSE
    )

    expect_silent(
        select_rename_data(
            data,
            nirs_channels = c(o2hb = "O2Hb", o2hb = "O2Hb"),
            time_channel = "Time",
            verbose = FALSE
        )
    )

    expect_warning(
        select_rename_data(
            data,
            nirs_channels = c(o2hb = "O2Hb", o2hb = "O2Hb"),
            time_channel = "Time",
            verbose = TRUE
        ),
        "o2hb = o2hb_1"
    )
})

test_that("select_rename_data() prioritises custom names over data", {
    data <- data.frame(
        O2Hb = c("10"),
        Time = c("0.1"),
        custom = c("x"),
        stringsAsFactors = FALSE
    )

    result <- select_rename_data(
        data,
        nirs_channels = c(custom = "O2Hb"),
        time_channel = "Time",
        keep_all = TRUE,
        verbose = FALSE
    )

    expect_true(all(c("custom", "custom_1") %in% names(result$data)))
    expect_equal(result$data$custom, "10")
})

## convert_type() ================================================
test_that("convert_type() coerces integer columns apart from event_channel", {
    data <- data.frame(
        time = c(1, 2, 3),
        lap = c(1.0, NA_real_, 2.0),
        B = c(10, 20, 30),
        x = c(10.5, 11.0, 11.5),
        stringsAsFactors = FALSE
    )

    result <- convert_type(data, "time", event_channel = "lap")

    expect_type(result$time, "double")
    expect_type(result$B, "double")
    expect_type(result$lap, "integer")
    expect_equal(result$lap, c(1L, NA_integer_, 2L))
    ## other columns unaffected
    expect_type(result$x, "double")
})


test_that("convert_type() standardises empty and 'NA' strings to NA", {
    data <- data.frame(
        time = c("1", "2", "3"),
        A = c("a", "", "b"),
        B = c("x", "NA", "y"),
        stringsAsFactors = FALSE
    )

    result <- convert_type(data, "time")

    expect_equal(result$A, c("a", NA_character_, "b"))
    expect_equal(result$B, c("x", NA_character_, "y"))
})

test_that("convert_type() standardises Inf/NaN to NA in numeric cols", {
    data <- data.frame(
        time = c("1", "2", "3", "4"),
        A = c("1.5", "Inf", "-Inf", "NaN"),
        stringsAsFactors = FALSE
    )

    result <- convert_type(data, "time")

    expect_type(result$A, "double")
    expect_equal(result$A, c(1.5, NA_real_, NA_real_, NA_real_))
})

test_that("convert_type() standardises non-finite integers to NA", {
    ## event_channel kept as integer; other int cols coerced to numeric
    data <- data.frame(
        time = c("1", "2", "3"),
        lap = c("1", NA, "2"),
        B = c("10", "20", "30"),
        stringsAsFactors = FALSE
    )

    result <- convert_type(data, "time", event_channel = "lap")

    expect_type(result$lap, "integer")
    expect_equal(result$lap, c(1L, NA_integer_, 2L))
    expect_type(result$B, "double")
})

test_that("convert_type() preserves valid numeric values", {
    data <- data.frame(
        time = c("1", "2"),
        A = c("0", "-0"),
        stringsAsFactors = FALSE
    )

    result <- convert_type(data, "time")
    expect_equal(result$A, c(0, 0))
})

## remove_empty_rows_cols() ===========================================
test_that("remove_empty_rows_cols() removes empty rows & cols", {
    data <- data.frame(
        A = c("1", "", "3"),
        B = c("x", "", "z"),
        stringsAsFactors = FALSE
    )

    result <- remove_empty_rows_cols(data)
    expect_equal(nrow(result), 2)
    expect_equal(result$A, c("1", "3"))

    data <- data.frame(
        A = c("1", "2"),
        B = c("", ""),
        C = c("x", "y"),
        stringsAsFactors = FALSE
    )

    result <- remove_empty_rows_cols(data)

    expect_equal(ncol(result), 2)
    expect_equal(names(result), c("A", "C"))

    data <- data.frame(
        A = c("1", "", "3"),
        B = c("", "", ""),
        C = c("x", "", "z"),
        stringsAsFactors = FALSE
    )

    result <- remove_empty_rows_cols(data)
    expect_equal(nrow(result), 2)
    expect_equal(result$A, c("1", "3"))
    expect_equal(ncol(result), 2)
    expect_equal(names(result), c("A", "C"))
})

test_that("remove_empty_rows_cols() handles NA values", {
    data <- data.frame(
        A = c("1", NA, "3"),
        B = c(NA, NA, NA),
        stringsAsFactors = FALSE
    )

    result <- remove_empty_rows_cols(data)

    expect_equal(nrow(result), 2)
    expect_equal(ncol(result), 1)
})

test_that("remove_empty_rows_cols() preserves non-empty data", {
    data <- data.frame(
        A = c("1", "2", "3"),
        B = c("x", "y", "z"),
        stringsAsFactors = FALSE
    )

    result <- remove_empty_rows_cols(data)

    expect_equal(result, data)
})

test_that("remove_empty_rows_cols() handles all empty data", {
    data <- data.frame(
        A = c("", ""),
        B = c("", ""),
        stringsAsFactors = FALSE
    )

    result <- remove_empty_rows_cols(data)

    expect_equal(nrow(result), 0)
    expect_equal(ncol(result), 0)
})

## extract_start_timestamp() ==========================================
test_that("extract_start_timestamp() returns NULL when no timestamps present", {
    data <- data.frame(
        V1 = c("device", "sensor"),
        V2 = c("model_x", "ch1"),
        stringsAsFactors = FALSE
    )
    expect_null(extract_start_timestamp(data))
})

test_that("extract_start_timestamp() detects ISO 8601 timestamp", {
    data <- data.frame(
        V1 = c("Start", "2025-03-01T08:30:00"),
        V2 = c("Device", "Moxy"),
        stringsAsFactors = FALSE
    )
    result <- extract_start_timestamp(data)
    expect_type(result, "character")
    expect_match(result, "2025-03-01T08:30:00")
})

test_that("extract_start_timestamp() detects yyyy-mm-dd HH:MM:SS format", {
    data <- data.frame(
        V1 = c("Start", "2025-06-15 14:22:10"),
        V2 = c("Device", "Moxy"),
        stringsAsFactors = FALSE
    )
    result <- extract_start_timestamp(data)
    expect_type(result, "character")
    expect_match(result, "2025-06-15 14:22:10")
})

test_that("extract_start_timestamp() returns earliest timestamp when multiple present", {
    data <- data.frame(
        V1 = c("2025-01-01 10:00:00", "2025-01-01 11:00:00"),
        V2 = c("2025-01-01 09:00:00", "metadata"),
        stringsAsFactors = FALSE
    )
    result <- extract_start_timestamp(data)
    expect_match(result, "2025-01-01 09:00:00")
})

test_that("extract_start_timestamp() ignores NA and empty strings", {
    data <- data.frame(
        V1 = c(NA, ""),
        V2 = c("2025-05-10T07:00:00", NA),
        stringsAsFactors = FALSE
    )
    result <- extract_start_timestamp(data)
    expect_type(result, "character")
    expect_match(result, "2025-05-10T07:00:00")
})

test_that("extract_start_timestamp() works with real example file header", {
    file_header <- read_file(example_mnirs("moxy_ramp.xlsx"))[1:20, ]
    result <- extract_start_timestamp(file_header)
    expect_false(is.null(result))
    expect_type(result, "character")
})


## parse_time_channel() ================================================
test_that("parse_time_channel() returns a list with $data and $start_timestamp", {
    data <- data.frame(time = c(0, 1, 2), value = c(1, 2, 3))
    result <- parse_time_channel(data, "time")

    expect_type(result, "list")
    expect_named(result, c("data", "start_timestamp"))
    expect_s3_class(result$data, "data.frame")
})

test_that("parse_time_channel() preserves numeric time (zero_time = FALSE)", {
    data <- data.frame(
        time = c(10.5, 20.5, 30.5),
        value = c(1, 2, 3)
    )

    result <- parse_time_channel(data, "time", zero_time = FALSE)

    expect_equal(result$data$time, data$time)
    expect_null(result$start_timestamp)
})

test_that("parse_time_channel() recalculates numeric time from zero", {
    data <- data.frame(
        time = c(10, 20, 30),
        value = c(1, 2, 3)
    )

    result <- parse_time_channel(data, "time", zero_time = TRUE)

    expect_equal(result$data$time, c(0, 10, 20))
})

test_that("parse_time_channel() parses ISO 8601 character timestamps to numeric", {
    data <- data.frame(
        time = c("2025-01-01T10:00:00", "2025-01-01T10:00:01"),
        value = c(1, 2),
        stringsAsFactors = FALSE
    )

    result <- parse_time_channel(data, "time")

    expect_type(result$data$time, "double")
    expect_equal(result$data$time, c(0, 1))
})

test_that("parse_time_channel() parses date-time formats to numeric", {
    formats <- list(
        c("2025-01-01 10:00:00", "2025-01-01 10:00:01"),
        c("2025/01/01 10:00:00", "2025/01/01 10:00:01"),
        c("01-01-2025 10:00:00", "01-01-2025 10:00:01"),
        c("01/01/2025 10:00:00", "01/01/2025 10:00:01")
    )

    for (fmt in formats) {
        data <- data.frame(
            time = fmt,
            value = c(1, 2),
            stringsAsFactors = FALSE
        )
        result <- parse_time_channel(data, "time")
        expect_equal(result$data$time[1L], 0)
    }
})

test_that("parse_time_channel() parses time-only H:MM:SS character format", {
    data <- data.frame(
        time = c("10:00:00", "10:00:01"),
        value = c(1, 2),
        stringsAsFactors = FALSE
    )

    result <- parse_time_channel(data, "time")

    expect_equal(result$data$time, c(0, 1))
})

test_that("parse_time_channel() handles milliseconds in timestamps", {
    data <- data.frame(
        time = c("2025-01-01T10:00:00.123", "2025-01-01T10:00:01.456"),
        value = c(1, 2),
        stringsAsFactors = FALSE
    )

    result <- parse_time_channel(data, "time")

    expect_equal(result$data$time[1], 0)
    expect_true(result$data$time[2] > 1 & result$data$time[2] < 2)
})

test_that("parse_time_channel() converts POSIXct to numeric seconds from zero", {
    data <- data.frame(
        time = as.POSIXct(c("2025-01-01 10:00:00", "2025-01-01 10:00:01")),
        value = c(1, 2)
    )

    result <- parse_time_channel(data, "time", add_timestamp = FALSE)

    expect_type(result$data$time, "double")
    expect_false("timestamp" %in% names(result$data))
    expect_equal(result$data$time, c(0, 1))
})

test_that("parse_time_channel() always zeros POSIXct regardless of zero_time", {
    data <- data.frame(
        time = as.POSIXct(c("2025-01-01 10:00:00", "2025-01-01 10:00:01")),
        value = c(1, 2)
    )

    result <- parse_time_channel(data, "time", zero_time = FALSE)

    expect_type(result$data$time, "double")
    ## POSIXct is always relative — always starts from 0
    expect_equal(result$data$time, c(0, 1))
})

test_that("parse_time_channel() returns start_timestamp from POSIXct time_channel", {
    t0 <- as.POSIXct("2025-03-15 08:00:00")
    data <- data.frame(
        time = t0 + c(0, 1, 2),
        value = c(1, 2, 3)
    )

    result <- parse_time_channel(data, "time")

    ## start_timestamp is extracted from the POSIXct column when not in header
    expect_false(is.null(result$start_timestamp))
    expect_equal(result$start_timestamp, t0, ignore_attr = TRUE)
})

test_that("parse_time_channel() add_timestamp=TRUE adds POSIXct column after time_channel", {
    t0 <- as.POSIXct("2025-01-01 10:00:00")
    data <- data.frame(
        time = t0 + c(0, 1),
        value = c(1, 2)
    )

    result <- parse_time_channel(data, "time", add_timestamp = TRUE)

    expect_true("timestamp" %in% names(result$data))
    expect_s3_class(result$data$timestamp, "POSIXct")
    expect_type(result$data$time, "double")
    ## timestamp column inserted immediately after time_channel
    expect_equal(which(names(result$data) == "timestamp"), 2)
    expect_equal(
        result$data$timestamp[1L],
        result$start_timestamp,
        ignore_attr = TRUE
    )
    expect_equal(as.numeric(result$data$timestamp), as.numeric(t0 + c(0, 1)))
})


test_that("parse_time_channel() add_timestamp=TRUE with start_timestamp reconstructs absolute timestamps", {
    data <- data.frame(
        time = c(0, 1, 2),
        value = c(1, 2, 3)
    )
    start_ts <- "2025-06-01T09:00:00"

    result <- parse_time_channel(
        data,
        "time",
        start_timestamp = start_ts,
        add_timestamp = TRUE
    )

    expect_true("timestamp" %in% names(result$data))
    expect_s3_class(result$data$timestamp, "POSIXct")

    expected_t0 <- as.POSIXct(start_ts, format = "%Y-%m-%dT%H:%M:%OS")
    expect_equal(
        as.numeric(result$data$timestamp),
        as.numeric(expected_t0 + c(0, 1, 2))
    )
    ## start_timestamp is passed through unchanged
    expect_equal(result$start_timestamp, start_ts)
})

test_that("parse_time_channel() add_timestamp=TRUE with no timestamps skips column silently", {
    data <- data.frame(
        time = c(10, 20, 30),
        value = c(1, 2, 3)
    )

    result <- parse_time_channel(
        data,
        time_channel = "time",
        start_timestamp = NULL,
        add_timestamp = TRUE
    )

    ## no timestamp available — column not added
    expect_false("timestamp" %in% names(result$data))
    expect_null(result$start_timestamp)
})

test_that("parse_time_channel() works on fractional unix time", {
    ## Moxy.csv saved as excel will coerce date-time
    ## to numeric fractional Unix time.
    file_path <- test_path("testdata/moxy-occlusion.xlsx")
    skip_if_not(file.exists(file_path), "testdata not available")

    data <- suppressMessages(readxl::read_excel(file_path)[-(1:2), 1:2])
    names(data)[2L] <- "time"
    data$time <- as.numeric(data$time)

    result <- parse_time_channel(
        data,
        time_channel = "time",
        start_timestamp = NULL,
        add_timestamp = TRUE
    )

    expect_equal(class(result$data$time), "numeric")
    expect_equal(result$data$time[1L], 0)
    expect_equal(median(diff(result$data$time)), 2)
    expect_equal(class(result$data$timestamp), c("POSIXct", "POSIXt"))
    expect_equal(class(result$start_timestamp), c("POSIXct", "POSIXt"))
    ## should return today's date, local time zone, precise timestamp
    expect_equal(as.Date(result$start_timestamp), Sys.Date())
    expect_equal(format(result$start_timestamp, "%Z"), format(Sys.time(), "%Z"))
    expect_equal(format(result$start_timestamp, "%H:%M:%OS"), "13:52:59")
})

test_that("parse_time_channel() returns local time zonel", {
    perfpro <- test_path("testdata/perfpro-mre.xlsx")
    moxy_occl <- test_path("testdata/moxy-occlusion.xlsx")
    vo2master <- test_path("testdata/vo2master.csv")
    skip_if_not(file.exists(perfpro), "testdata not available")
    skip("run in local time PDT/PST zone")
    
    file_list <- c(
        perfpro, ## today's date, 0:00:00
        moxy_occl, ## today's date, 13:52:59
        example_mnirs("moxy_intervals"), ## today's date, 13:17:13
        example_mnirs("moxy_ramp") ## today's date 0:29:00.41
    )

    timestamp_list <- list()
    for (.file in file_list) {
        # .file = file_list[3]
        data <- read_file(.file)
        detected_list <- detect_mnirs_device(data)
        nirs_device <- detected_list$nirs_device
        header_row <- detected_list$header_row

        channels <- detect_device_channels(
            data,
            header_row,
            nirs_device,
            nirs_channels = NULL,
            time_channel = NULL,
            keep_all = FALSE,
            verbose = FALSE
        )
        nirs_channels <- channels$nirs_channels
        time_channel <- channels$time_channel
        keep_all <- channels$keep_all

        table_list <- read_data_table(data, nirs_channels, header_row)
        data <- table_list$data_table
        file_header <- table_list$file_header

        start_timestamp <- extract_start_timestamp(file_header)

        time_channel <- detect_time_channel(
            data,
            time_channel,
            nirs_device,
            verbose = FALSE
        )

        renamed_list <- select_rename_data(
            data,
            nirs_channels,
            time_channel,
            event_channel = NULL,
            keep_all,
            verbose = FALSE
        )
        data <- renamed_list$data
        nirs_renamed <- renamed_list$nirs_channel
        time_renamed <- renamed_list$time_channel
        event_renamed <- renamed_list$event_channel

        data <- remove_empty_rows_cols(data)
        data <- convert_type(data, time_renamed, event_renamed, verbose = FALSE)
        time_list <- parse_time_channel(
            data,
            time_renamed,
            start_timestamp,
            add_timestamp = TRUE,
            zero_time = TRUE
        )

        expect_true(
            format(time_list$start_timestamp, "%Z") %in% c("PDT", "PST")
        )
        timestamp_list <- c(timestamp_list, list(time_list$start_timestamp))
    }
    timestamp_list <- do.call(c, timestamp_list)
    expect_equal(as.Date(timestamp_list), rep(Sys.Date(), 4))
    
    expect_equal(
        format(timestamp_list, "%H:%M:%OS2"),
        c("00:00:00.00", "13:52:59.00", "13:17:13.00", "00:29:00.41")
    )
})

## parse_sample_rate() ================================================
test_that("parse_sample_rate returns correct structure", {
    data <- data.frame(
        time = seq(0, 10, by = 0.1),
        value = rnorm(101, 10, 1)
    )
    file_header <- matrix(NA, nrow = 5, ncol = 5)

    result <- parse_sample_rate(
        data = data,
        file_header = file_header,
        time_channel = "time",
        sample_rate = 10,
        nirs_device = NULL,
        verbose = FALSE
    )

    expect_type(result, "list")
    expect_named(result, c("data", "time_channel", "sample_rate"))
    expect_s3_class(result$data, "data.frame")
    expect_type(result$time_channel, "character")
    expect_type(result$sample_rate, "double")
    expect_equal(result$data$value, data$value)
    expect_equal(nrow(result$data), nrow(data))
    expect_equal(result$sample_rate, 10)
})

test_that("parse_sample_rate handles Artinis device", {
    file_header <- read_file(example_mnirs("artinis_intervals"))
    data <- read_mnirs(
        example_mnirs("artinis_intervals"),
        nirs_channels = c(HHb = 2, O2Hb = 3),
        time_channel = c(sample = 1),
        event_channel = NULL,
        verbose = FALSE
    ) |>
        dplyr::select(-time)

    result <- parse_sample_rate(
        data = data,
        file_header = file_header,
        time_channel = "sample",
        sample_rate = NULL,
        nirs_device = "Artinis",
        verbose = FALSE
    )

    expect_equal(result$sample_rate, 10)
    expect_true("time" %in% names(result$data))
    expect_equal(result$time_channel, "time")
    expect_equal(ncol(result$data), 4)
    expect_equal(result$data$time, data$sample / 10)
})

test_that("parse_sample_rate errors when rate indeterminable", {
    data <- data.frame(x = rep(1, 10))
    file_header <- matrix(NA, nrow = 5, ncol = 5)

    expect_error(
        parse_sample_rate(
            data = data,
            file_header = file_header,
            time_channel = "x",
            sample_rate = NULL,
            verbose = FALSE
        ),
        "Unable to estimate.*sample_rate"
    )
})

test_that("parse_sample_rate verbose output for Artinis", {
    file_header <- read_file(example_mnirs("artinis_intervals"))
    data <- read_mnirs(
        example_mnirs("artinis_intervals"),
        nirs_channels = c(HHb = 2, O2Hb = 3),
        time_channel = c(sample = 1),
        event_channel = NULL,
        verbose = FALSE
    ) |>
        dplyr::select(-time)

    expect_message(
        result <- parse_sample_rate(
            data = data,
            file_header = file_header,
            time_channel = "sample",
            sample_rate = NULL,
            nirs_device = "Artinis",
            verbose = TRUE
        ),
        "Oxysoft.*sample_rate"
    )
})

## detect_irregular_samples() =========================================
test_that("detect_irregular_samples returns invisibly with no irregularities", {
    x <- seq(0, 100, by = 1)
    expect_invisible(detect_irregular_samples(x, "time", verbose = TRUE))
    expect_invisible(detect_irregular_samples(x, "time", verbose = FALSE))
})

test_that("detect_irregular_samples detects duplicated samples", {
    x <- c(0, 1, 2, 2, 3, 4)
    expect_warning(
        detect_irregular_samples(x, "time"),
        "irregular.*detected"
    )
    expect_warning(
        detect_irregular_samples(x, "time"),
        "time.*=.*2"
    )
})

test_that("detect_irregular_samples detects unordered samples", {
    x <- c(0, 1, 3, 2, 4, 5)
    expect_warning(
        detect_irregular_samples(x, "time"),
        "irregular.*detected"
    )
    expect_warning(
        detect_irregular_samples(x, "time"),
        "time.*=.*3"
    )
})

test_that("detect_irregular_samples detects large gaps (>= 3600)", {
    x <- c(0, 1, 2, 3602, 3603)
    expect_warning(
        detect_irregular_samples(x, "time"),
        "irregular.*detected"
    )
    expect_warning(
        detect_irregular_samples(x, "time"),
        "time.*=.*2"
    )
})

test_that("detect_irregular_samples shows first 3 when > 5 irregularities", {
    x <- c(0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6)
    expect_warning(
        detect_irregular_samples(x, "time"),
        "and 3 more"
    )
})

test_that("detect_irregular_samples shows all when <= 5 irregularities", {
    x <- c(0, 1, 1, 2, 2, 3)
    w <- expect_warning(detect_irregular_samples(x, "time"))
    expect_false(grepl("more", w$message))
})

test_that("detect_irregular_samples respects verbose = FALSE", {
    x <- c(0, 1, 1, 2, 2)
    expect_invisible(detect_irregular_samples(x, "time", verbose = FALSE))
    expect_no_warning(detect_irregular_samples(x, "time", verbose = FALSE))
})

test_that("detect_irregular_samples handles multiple irregularity types", {
    x <- c(0, 1, 1, 3, 2, 3606)
    expect_warning(
        detect_irregular_samples(x, "time"),
        "irregular.*detected"
    )
})

test_that("detect_irregular_samples uses correct time_channel name", {
    x <- c(0, 1, 1, 2)
    expect_warning(
        detect_irregular_samples(x, "my_time_col"),
        "my_time_col"
    )
})

## read_mnirs() =======================================================
## read_mnirs() auto-detection =========================================
test_that("read_mnirs auto-detects Moxy channels when nirs_channels = NULL", {
    file_path <- example_mnirs("moxy_ramp")

    expect_message(
        df <- read_mnirs(
            file_path = file_path,
            nirs_channels = NULL,
            time_channel = NULL,
            verbose = TRUE
        ),
        "Moxy.*detected"
    ) |>
        expect_warning("irregular") |>
        expect_message("Estimated.*sample_rate.*2")

    expect_s3_class(df, "mnirs")
    expect_equal(attr(df, "nirs_device"), "Moxy")
    expect_equal(attr(df, "nirs_channels"), device_patterns$Moxy$nirs_channels)
    expect_equal(attr(df, "time_channel"), device_patterns$Moxy$time_channel)
    ## auto-detected channels should keep original names (not renamed)
    expect_true(all(
        device_patterns$Moxy$nirs_channels %in% names(df)
    ))
})

test_that("read_mnirs auto-detects Train.Red channels when nirs_channels = NULL", {
    file_path <- example_mnirs("train.red")

    expect_message(
        df <- read_mnirs(
            file_path = file_path,
            nirs_channels = NULL,
            time_channel = NULL,
            verbose = TRUE
        ),
        "Train.Red.*detected"
    ) |>
        expect_warning("irregular") |>
        expect_message("Estimated.*sample_rate.*10")

    expect_s3_class(df, "mnirs")
    expect_equal(attr(df, "nirs_device"), "Train.Red")
    expect_equal(
        attr(df, "nirs_channels"),
        device_patterns$Train.Red$nirs_channels
    )
    expect_equal(
        attr(df, "time_channel"),
        device_patterns$Train.Red$time_channel
    )
    expect_true(all(
        device_patterns$Train.Red$nirs_channels %in% names(df)
    ))
})

test_that("read_mnirs auto-detects Artinis channels when nirs_channels = NULL", {
    file_path <- example_mnirs("artinis_intervals")

    expect_message(
        df <- read_mnirs(
            file_path = file_path,
            nirs_channels = NULL,
            time_channel = NULL,
            verbose = TRUE
        ),
        "Artinis.*detected"
    ) |>
        expect_message("Oxysoft.*sample.*column") |> 
        expect_message("Oxysoft.*sample_rate.*10")

    expect_s3_class(df, "mnirs")
    expect_equal(attr(df, "nirs_device"), "Artinis")
    expect_equal(
        attr(df, "nirs_channels"),
        device_patterns$Artinis$nirs_channels
    )
    expect_equal(attr(df, "time_channel"), "time")
    expect_true(all(
        device_patterns$Artinis$nirs_channels %in% names(df)
    ))
})

test_that("read_mnirs keep_all = FALSE returns only specified columns by default", {
    file_path <- example_mnirs("moxy_ramp")

    df <- read_mnirs(
        file_path = file_path,
        nirs_channels = c(smo2 = "SmO2 Live"),
        time_channel = c(time = "hh:mm:ss"),
        # keep_all = FALSE,
        verbose = FALSE
    )

    ## default keep_all = FALSE returns only smo2 + time
    expect_equal(ncol(df), 2)
    expect_true("smo2" %in% names(df))
    expect_true("time" %in% names(df))
})

test_that("read_mnirs keep_all = TRUE returns all columns", {
    file_path <- example_mnirs("moxy_ramp")

    df <- read_mnirs(
        file_path = file_path,
        nirs_channels = c(smo2 = "SmO2 Live"),
        time_channel = c(time = "hh:mm:ss"),
        keep_all = TRUE,
        verbose = FALSE
    )

    ## default keep_all = TRUE returns more columns than just smo2 + time
    expect_gt(ncol(df), 2)
    expect_true("smo2" %in% names(df))
    expect_true("time" %in% names(df))
})

test_that("read_mnirs returns all columns when auto-detecting nirs_channels", {
    file_path <- example_mnirs("moxy_ramp")

    df <- read_mnirs(
        file_path = file_path,
        nirs_channels = NULL,
        time_channel = NULL,
        keep_all = FALSE, ## explicitly FALSE should be overridden
        verbose = FALSE
    )

    ## default keep_all = TRUE returns more columns than just smo2 + time
    expect_gt(ncol(df), 2)
    expect_true("SmO2 Live" %in% names(df))
    expect_true("hh:mm:ss" %in% names(df))
})


## moxy ===============================================================
test_that("read_mnirs moxy .xlsx works with timestamp", {
    expect_warning(
        df <- read_mnirs(
            file_path = example_mnirs("moxy_ramp.xlsx"),
            nirs_channels = c(
                smo2_left = "SmO2 Live",
                smo2_right = "SmO2 Live(2)"
            ),
            time_channel = c(time = "hh:mm:ss"),
            add_timestamp = TRUE,
            verbose = TRUE
        ),
        "irregular"
    ) |>
        expect_message("Estimated.*sample_rate.*2")

    expect_s3_class(df, "mnirs")
    expect_s3_class(df, "data.frame")
    expect_true(all(
        c("time", "timestamp", "smo2_left", "smo2_right") %in% names(df)
    ))
    expect_equal(df$time[1], 0)
    expect_equal(class(df$time), "numeric")
    expect_true(any(class(df$timestamp) %in% "POSIXct"))

    ## check that time diffs should be 0 < Δ < 1 with proper POSIXct import
    expect_gt(sum(diff(df$time[1:100]) < 1 & diff(df$time[1:100]) > 0), 0)
    expect_lt(sum(diff(df$time[1:100]) %in% c(0, 1)), 99)

    expect_true(all(
        c("nirs_device", "nirs_channels", "time_channel", "sample_rate") %in%
            names(attributes(df))
    ))

    expect_equal(attr(df, "nirs_device"), "Moxy")
    expect_equal(attr(df, "sample_rate"), 2)
})

test_that("read_mnirs moxy .csv works converting time to numeric", {
    expect_message(
        df <- read_mnirs(
            file_path = example_mnirs("moxy_intervals"),
            nirs_channels = c(smo2_left = "SmO2 Live", thb = "THb"),
            time_channel = c(time = "hh:mm:ss"),
            add_timestamp = FALSE,
            verbose = TRUE
        ),
        "Estimated.*sample_rate.*0.5"
    )

    expect_equal(class(df$time), "numeric")
    expect_false(class(df$time) %in% "POSIXct")
    expect_false(c("timestamp") %in% names(df))

    ## check that time diffs should be Δ >= 2 with proper POSIXct import
    expect_gt(sum(diff(df$time[1:100]) >= 2), 0)

    expect_true(all(
        c("nirs_device", "nirs_channels", "time_channel", "sample_rate") %in%
            names(attributes(df))
    ))

    expect_equal(attr(df, "nirs_device"), "Moxy")
    expect_equal(attr(df, "sample_rate"), 0.5)
})


test_that("read_mnirs moxy invalid channel names", {
    file_path <- example_mnirs("moxy_ramp.xlsx")

    ## invalid channel names
    expect_error(
        read_mnirs(
            file_path = file_path,
            nirs_channels = c(""),
            time_channel = c(time = "hh:mm:ss"),
            verbose = FALSE
        ),
        "not detected"
    )

    expect_error(
        read_mnirs(
            file_path = file_path,
            nirs_channels = c(smo2_left = "smo2_doesnt_exist"),
            time_channel = c(time = "hh:mm:ss"),
            verbose = FALSE
        ),
        "not detected"
    )

    expect_message(
        df <- read_mnirs(
            file_path = file_path,
            nirs_channels = c(
                smo2_left = "SmO2 Live",
                smo2_right = "SmO2 Live(2)"
            ),
            time_channel = NULL,
            verbose = TRUE
        ),
        "Estimated.*sample_rate.*2"
    ) |>
        expect_warning("irregular") |> 
        expect_message("Detected.*time_channel")

    expect_equal(attr(df, "time_channel"), "hh:mm:ss")

    ## duplicate input names are renamed
    expect_warning(
        df <- read_mnirs(
            file_path = file_path,
            nirs_channels = c(smo2 = "SmO2 Live", smo2 = "SmO2 Live(2)"),
            time_channel = c(time = "hh:mm:ss"),
            verbose = TRUE
        ),
        "Duplicate"
    ) |>
        expect_warning("irregular") |>
        expect_message("Estimated.*sample_rate.*2")

    expect_true(all(c("smo2", "smo2_1") %in% names(df)))
})


## train.red ========================================================
test_that("read_mnirs train.red works", {
    file_path <- example_mnirs("train.red_intervals.csv")

    expect_no_message(
        read_mnirs(
            file_path = file_path,
            nirs_channels = c(
                smo2_left = "SmO2 unfiltered",
                smo2_right = "SmO2 unfiltered"
            ),
            time_channel = c(time = "Timestamp (seconds passed)"),
            verbose = FALSE
        )
    )

    expect_warning(
        df <- read_mnirs(
            file_path = file_path,
            nirs_channels = c(
                smo2_left = "SmO2 unfiltered",
                smo2_right = "SmO2 unfiltered"
            ),
            time_channel = c(time = "Timestamp (seconds passed)"),
            verbose = TRUE
        ),
        "irregular"
    ) |>
        expect_message("Estimated.*sample_rate.*10")

    expect_s3_class(df, "mnirs")
    expect_s3_class(df, "data.frame")
    expect_true(all(
        c("time", "smo2_left", "smo2_right") %in% names(df)
    ))
    expect_equal(class(df$time), "numeric")
    expect_gte(df$time[1], 0)

    ## check that time diffs should be 0 < Δ < 1 with proper POSIXct import
    expect_gt(sum(diff(df$time[1:100]) < 1 & diff(df$time[1:100]) > 0), 0)
    expect_lt(sum(diff(df$time[1:100]) %in% c(0, 1)), 99)

    expect_true(all(
        c("nirs_device", "nirs_channels", "time_channel", "sample_rate") %in%
            names(attributes(df))
    ))

    expect_equal(attr(df, "nirs_device"), "Train.Red")
    expect_equal(attr(df, "sample_rate"), 10)
})

test_that("read_mnirs external train.red mre works", {
    file_path <- test_path("testdata/train.red-mre.csv")
    skip_if_not(file.exists(file_path), "testdata not available")

    expect_no_message(
        read_mnirs(
            file_path = file_path,
            nirs_channels = c(
                smo2_left = "SmO2 unfiltered",
                smo2_right = "SmO2 unfiltered"
            ),
            time_channel = c(time = "Timestamp (seconds passed)"),
            verbose = FALSE
        )
    )

    expect_warning(
        df <- read_mnirs(
            file_path = file_path,
            nirs_channels = c(
                smo2_left = "SmO2 unfiltered",
                smo2_right = "SmO2 unfiltered"
            ),
            time_channel = c(time = "Timestamp (seconds passed)"),
            verbose = TRUE
        ),
        "irregular"
    ) |>
        expect_message("Estimated.*sample_rate.*10")

    expect_s3_class(df, "mnirs")
    expect_s3_class(df, "data.frame")
    expect_true(all(
        c("time", "smo2_left", "smo2_right") %in% names(df)
    ))
    expect_equal(class(df$time), "numeric")
    expect_gte(df$time[1], 0)

    ## check that time diffs should be 0 < Δ < 1 with proper POSIXct import
    expect_gt(sum(diff(df$time[1:100]) < 1 & diff(df$time[1:100]) > 0), 0)
    expect_lt(sum(diff(df$time[1:100]) %in% c(0, 1)), 99)

    expect_true(all(
        c("nirs_device", "nirs_channels", "time_channel", "sample_rate") %in%
            names(attributes(df))
    ))

    expect_equal(attr(df, "nirs_device"), "Train.Red")
    expect_equal(attr(df, "sample_rate"), 10)
})

test_that("read_mnirs coerces integerish event_channel to integer", {
    ## train.red "Lap/Event" column contains lap numbers (doubles from CSV)
    data <- read_mnirs(
        file_path = example_mnirs("train.red_intervals.csv"),
        nirs_channels = c(smo2_left = "SmO2 unfiltered"),
        time_channel = c(time = "Timestamp (seconds passed)"),
        event_channel = c(lap = "Lap/Event"),
        verbose = FALSE
    )
    # rlang::is_integerish(df$lap)

    expect_type(data$lap, "integer")
})

test_that("read_mnirs train.red works with zero_time", {
    file_path <- example_mnirs("train.red_intervals.csv")

    expect_equal(
        read_mnirs(
            file_path = file_path,
            nirs_channels = c(
                smo2_left = "SmO2 unfiltered",
                smo2_right = "SmO2 unfiltered"
            ),
            time_channel = c(time = "Timestamp (seconds passed)"),
            zero_time = TRUE,
            verbose = FALSE
        )$time[1],
        0
    )
})

test_that("read_mnirs train.red invalid channel names", {
    file_path <- example_mnirs("train.red_intervals.csv")

    ## invalid channel names
    expect_error(
        read_mnirs(
            file_path = file_path,
            nirs_channels = c(" "),
            time_channel = c(time = "Timestamp (seconds passed)"),
        ),
        "not detected"
    )

    expect_error(
        read_mnirs(
            file_path = file_path,
            nirs_channels = c(smo2_left = "smo2_doesnt_exist"),
            time_channel = c(time = "Timestamp (seconds passed)"),
        ),
        "not detected"
    )

    expect_message(
        df <- read_mnirs(
            file_path = file_path,
            nirs_channels = c(
                smo2_left = "SmO2 unfiltered",
                smo2_right = "SmO2 unfiltered"
            ),
            time_channel = NULL,
            verbose = TRUE
        ),
        "Estimated.*sample_rate.*10"
    ) |>
        expect_message("Detected.*time_channel") |> 
        expect_warning("irregular")

    expect_equal(attr(df, "time_channel"), "Timestamp (seconds passed)")

    ## duplicate input names are renamed
    expect_warning(
        df <- read_mnirs(
            file_path = file_path,
            nirs_channels = c(
                smo2 = "SmO2 unfiltered",
                smo2 = "SmO2 unfiltered"
            ),
            time_channel = c(time = "Timestamp (seconds passed)"),
            verbose = TRUE
        ),
        "Duplicate"
    ) |>
        expect_warning("irregular") |>
        expect_message("Estimated.*sample_rate.*10")

    expect_true(all(c("smo2", "smo2_1") %in% names(df)))
})


## oxysoft =======================================================
test_that("read_mnirs oxysoft works", {
    file_path <- example_mnirs("artinis_intervals")

    expect_message(
        df <- read_mnirs(
            file_path = file_path,
            nirs_channels = c(HHb = 2, O2Hb = 3),
            time_channel = c(sample = 1),
            verbose = TRUE
        ),
        "Oxysoft.*sample_rate.*10"
    )

    expect_s3_class(df, "mnirs")
    expect_s3_class(df, "data.frame")
    expect_true(all(
        c("time", "HHb", "O2Hb") %in% names(df)
    ))
    expect_equal(class(df$time), "numeric")
    expect_gte(df$time[1], 0)
    expect_equal(df$sample[1:10] / 10, df$time[1:10])

    expect_true(all.equal(diff(df$time[1:100]), rep(0.1, 99)))

    expect_true(all(
        c("nirs_device", "nirs_channels", "time_channel", "sample_rate") %in%
            names(attributes(df))
    ))

    expect_equal(attr(df, "nirs_device"), "Artinis")
    expect_equal(attr(df, "sample_rate"), 10)
    expect_equal(attr(df, "time_channel"), "time")
})

test_that("read_mnirs Oxysoft Portamon works", {
    file_path <- example_mnirs("portamon-oxcap")

    expect_equal(
        read_file(file_path) |>
            detect_mnirs_device(),
        list(
            nirs_device = "Artinis",
            header_row = 42
        )
    )

    df <- read_mnirs(
        file_path = example_mnirs("portamon-oxcap.xlsx"),
        nirs_channels = c(thb = 2, hhb = 3, o2hb = 4),
        time_channel = NULL,
        event_channel = c(event = "col_6"),
        verbose = FALSE
    )

    expect_true(all(
        c("sample", "time", "event", "thb", "hhb", "o2hb") %in% names(df)
    ))

    ## auto detect channels
    df2 <- read_mnirs(
        file_path,
        nirs_channels = NULL,
        time_channel = NULL,
        verbose = FALSE
    )

    expect_true(all(
        c("sample", "time", "2", "3", "4", "5", "col_6") %in% names(df2)
    ))

    for (d in list(df, df2)) {
        expect_s3_class(d, "mnirs")
        expect_s3_class(d, "data.frame")

        expect_equal(class(d$time), "numeric")
        expect_equal(d$time[1], 0)
        expect_equal(d$sample[1:10] / 10, d$time[1:10])

        expect_true(all.equal(diff(d$time[1:100]), rep(0.1, 99)))

        expect_true(all(
            c(
                "nirs_device",
                "nirs_channels",
                "time_channel",
                "sample_rate"
            ) %in% names(attributes(d))
        ))

        expect_equal(attr(d, "nirs_device"), "Artinis")
        expect_equal(attr(d, "sample_rate"), 10)
        expect_equal(attr(d, "time_channel"), "time")
    }
})

test_that("read_mnirs Oxysoft edge case channel names", {
    file_path <- example_mnirs("artinis_intervals")

    old_verbose <- getOption("mnirs.verbose")
    on.exit(options(mnirs.verbose = old_verbose), add = TRUE)
    options(mnirs.verbose = FALSE)

    ## invalid channel names
    expect_error(
        read_mnirs(
            file_path = file_path,
            nirs_channels = c(""),
            time_channel = c(sample = 1),
        ),
        "not detected"
    )

    expect_error(
        read_mnirs(
            file_path = file_path,
            nirs_channels = c(smo2_left = "smo2_doesnt_exist"),
            time_channel = c(sample = 1),
        ),
        "not detected"
    )

    expect_message(
        df <- read_mnirs(
            file_path = file_path,
            nirs_channels = c(HHb = 2, O2Hb = 3),
            time_channel = NULL,
            verbose = TRUE
        ),
        "Oxysoft.*sample"
    ) |>
        expect_message("Oxysoft.*sample_rate.*10")

    ## detected as "sample" then updated to "time" automatically
    expect_equal(attr(df, "time_channel"), "time")

    ## duplicate input names are renamed
    expect_message(
        df <- read_mnirs(
            file_path = file_path,
            nirs_channels = c(HHb = 2, HHb = 3),
            time_channel = c(sample = 1),
            verbose = TRUE
        ),
        "Oxysoft.*sample_rate.*10"
    ) |>
        expect_warning("Duplicate")

    expect_true(all(c("HHb", "HHb_1") %in% names(df)))
})

## VO2master app ========================================================
test_that("read_mnirs VO2master with ',' decimals returns numeric", {
    file_path <- test_path("testdata/vo2master.csv")
    skip_if_not(file.exists(file_path), "testdata not available")

    nirs_channels <- c(
        smo2_1 = "SmO2[%]",
        smo2_2 = "SmO2 -  2[%]",
        smo2_3 = "SmO2 -  3[%]"
    )
    time_channel <- c(time = "Time[s]")

    df_raw <- tibble::as_tibble(
        data.table::fread(
            file_path,
            header = TRUE,
            colClasses = "character"
        )
    )

    expect_all_true(vapply(df_raw, is.character, logical(1L)))

    ## should convert decimal "," to numeric
    df <- convert_type(df_raw, time_channel)
    expect_all_true(vapply(df[, -c(1:2)], is.numeric, logical(1L)))

    ## integrated test
    expect_message(
        df <- read_mnirs(
            file_path = file_path,
            nirs_channels = c(
                smo2_1 = "SmO2[%]",
                smo2_2 = "SmO2 -  2[%]",
                smo2_3 = "SmO2 -  3[%]"
            ),
            time_channel = c(time = "Time[s]"),
            verbose = TRUE
        ),
        "Estimated.*sample_rate.*1"
    )

    expect_type(df$time, "double")
    expect_equal(sum(diff(df$time[1:100]) == 1), 99)
    ## smo2 should be numeric from "27,90"
    expect_type(df$smo2_1, "double")
    expect_type(df$smo2_2, "double")
    expect_type(df$smo2_3, "double")

    expect_true(all(
        c("nirs_channels", "time_channel", "sample_rate") %in%
            names(attributes(df))
    ))
    expect_equal(attr(df, "sample_rate"), 1)
})

## PerfPro ========================================================
test_that("read_mnirs PerfPro", {
    file_path <- test_path("testdata/perfpro-mre.xlsx")
    skip_if_not(file.exists(file_path), "testdata not available")

    data <- read_file(file_path)
    detected_list <- detect_mnirs_device(data)
    nirs_device <- detected_list$nirs_device
    header_row <- detected_list$header_row

    expect_equal(nirs_device, "PerfPro")
    expect_equal(header_row, 3)

    expect_message(
        channels <- detect_device_channels(
            data,
            header_row,
            nirs_device,
            nirs_channels = NULL,
            time_channel = NULL,
            keep_all = FALSE,
            verbose = TRUE
        ),
        "PerfPro"
    )

    expect_equal(channels$time_channel, "Time")
    expect_equal(channels$nirs_channels, "SmO2 (1614)")
    expect_true(channels$keep_all)

    ## integrated test
    expect_message(
        df <- read_mnirs(
            file_path = file_path,
            nirs_channels = c(smo2 = "SmO2 (1614)"),
            time_channel = c(time = "Time"),
            add_timestamp = TRUE
        ),
        "Estimated.*sample_rate.*2"
    )

    expect_type(df$time, "double")
    expect_true(inherits(df$timestamp, "POSIXct"))
    expect_equal(df$time[1L], 0)
    expect_true(all(diff(df$time[1:25]) < 1))
    ## smo2 should be numeric from "27,90"
    expect_type(df$smo2, "double")

    expect_true(all(
        c("nirs_channels", "time_channel", "sample_rate") %in%
            names(attributes(df))
    ))
    expect_equal(attr(df, "sample_rate"), 2)
})

## create_mnirs_data() ==================================================

test_that("create_mnirs_data edge cases", {
    ## error when data isn't a dataframe
    vec <- c(1, 2)
    expect_error(create_mnirs_data(vec), "must be a data frame")

    ## attach metadata unlisted
    df <- tibble(A = 1:2, B = letters[1:2])
    df_meta <- create_mnirs_data(
        df,
        nirs_channels = c("A", "B"),
        sample_rate = 1
    )
    expect_null(attr(df, "sample_rate"))
    expect_equal(attr(df_meta, "sample_rate"), 1)
    expect_equal(attr(df_meta, "nirs_channels"), c("A", "B"))
})

test_that("create_mnirs_data accepts NSE for *_channels", {
    df <- tibble(A = 1:2, B = 3:4, C = 5:6, lap = c("a", "b"))

    ## bare symbols
    df_sym <- create_mnirs_data(
        df,
        nirs_channels = B,
        time_channel = A,
        event_channel = lap
    )
    expect_equal(attr(df_sym, "nirs_channels"), "B")
    expect_equal(attr(df_sym, "time_channel"), "A")
    expect_equal(attr(df_sym, "event_channel"), "lap")

    ## c() expression
    df_c <- create_mnirs_data(df, nirs_channels = c(B, C))
    expect_equal(attr(df_c, "nirs_channels"), c("B", "C"))

    ## tidyselect helper
    df_sel <- create_mnirs_data(
        df, nirs_channels = tidyselect::starts_with("B")
    )
    expect_equal(attr(df_sel, "nirs_channels"), "B")

    ## external character vector
    chans <- c("B", "C")
    df_ext <- create_mnirs_data(df, nirs_channels = chans)
    expect_equal(attr(df_ext, "nirs_channels"), c("B", "C"))

    ## list form still works (single list argument)
    meta <- list(nirs_channels = c("B", "C"), sample_rate = 2)
    df_list <- create_mnirs_data(df, meta)
    expect_equal(attr(df_list, "nirs_channels"), c("B", "C"))
    expect_equal(attr(df_list, "sample_rate"), 2)
})

#' Re-sample an *{mnirs}* data frame
#'
#' Up- or down-sample an *"mnirs"* data frame to a new sample rate, filling
#' new samples via nearest-neighbour matching or interpolation.
#'
#' @param resample_rate An *optional* sample rate (Hz) for the output data
#'   frame. If `NULL` (*default*) resamples to the existing `sample_rate`,
#'  which regularises any irregular samples without changing the rate.
#'
#' @param method A character string specifying how new samples are filled.
#'   Default is *"none"*. Filling must be opted into explicitly (see *Details*):
#'   \describe{
#'     \item{`"none"`}{Matches each new sample to the nearest original
#'     `time_channel` value without any interpolation, to within tolerance of
#'     half a sample-interval. New samples are returned as `NA`.}
#'     \item{`"locf"`}{(*"Last observation carried forward"*). Fills new and
#'     missing samples with the most recent valid non-`NA` value to the left,
#'     or the nearest valid value to the right for leading `NA`s. Safe for
#'     numeric, integer, and character columns.}
#'     \item{`"linear"`}{Fills new and missing samples via linear interpolation
#'     using [stats::approx()]. Suitable for numeric columns only; non-numeric
#'     columns will fall back to `"locf"` behaviour.}
#'   }
#'
#' @inheritParams validate_mnirs
#'
#' @details
#' This function uses [replace_missing()] (based on [stats::approx()]) to
#' interpolate across new samples in the resampled data range.
#'
#' ## Sample rate and time channel
#'
#' `time_channel` and `sample_rate` are retrieved automatically from `data`
#' of class *"mnirs"*, if not defined explicitly.
#'
#' Otherwise, `sample_rate` will be estimated from the values in `time_channel`.
#' However, this may return unexpected values, and it is safer to define
#' `sample_rate` explicitly or retrieve it from *"mnirs"* metadata.
#'
#' ## Default behaviour
#'
#' When `resample_rate` is omitted, the output has the same `sample_rate` as
#' the input but with a regular, evenly-spaced `time_channel`. This is useful
#' for regularising data that contains missing or repeated samples without
#' changing the nominal rate.
#'
#' ## Column handling
#'
#' Numeric columns are interpolated according to `method` (see
#' `?replace_missing`). Non-numeric columns (e.g. character event labels,
#' integer lap numbers) are always filled by last-observation-carried-forward,
#' regardless of `method`:
#'
#' - For `method = "none"`, existing rows are matched to the nearest original
#'   values of `time_channel` without interpolation or filling, meaning newly
#'   created samples and any `NA`s in the original data are returned as `NA`.
#' - When down-sampling, numeric columns use time-weighted averaging. 
#'   Non-numeric columns use the first valid value in each output bin.
#'
#' @returns A [tibble][tibble::tibble-package] of class `"mnirs"`. Metadata are
#'   stored as attributes and can be accessed with `attributes(data)`.
#'
#' @examples
#' ## read example data
#' data <- read_mnirs(
#'     file_path = example_mnirs("moxy_ramp"),
#'     nirs_channels = c(smo2 = "SmO2 Live"),
#'     time_channel = c(time = "hh:mm:ss"),
#'     verbose = TRUE
#' )
#'
#' ## note warning about irregular sampling
#' data
#'
#' data_resampled <- resample_mnirs(
#'     data,               ## blank channels will be retrieved from metadata
#'     resample_rate = 2,  ## blank by default will resample to `sample_rate`
#'     method = "linear",  ## linear interpolation across resampled indices
#'     verbose = TRUE      
#' )
#'
#' ## note the altered `time` values resolving the above warning
#' data_resampled
#'
#' @export
resample_mnirs <- function(
    data,
    time_channel = NULL,
    sample_rate = NULL,
    resample_rate = sample_rate, ## placeholder indicating default condition
    method = c("none", "linear", "locf"),
    verbose = TRUE
) {
    ## validation ====================================
    validate_mnirs_data(data)
    metadata <- attributes(data)
    if (missing(verbose)) {
        verbose <- getOption("mnirs.verbose", default = TRUE)
    }
    time_channel <- validate_time_channel(enquo(time_channel), data)
    sample_rate <- validate_sample_rate(
        data, time_channel, sample_rate, verbose
    )
    validate_numeric(
        resample_rate, 1, c(0, Inf), FALSE, msg1 = "one-element positive"
    )
    ## assign default resample_rate as sample_rate
    if (is.null(resample_rate)) {
        resample_rate <- sample_rate
    }
    method <- match.arg(method)

    ## calculate resampling parameters  ===================================
    resample_time <- 1 / resample_rate
    colnames <- names(data)
    time_vec <- data[[time_channel]]
    time_range <- range(time_vec, na.rm = TRUE) * resample_rate
    time_range <- c(floor(time_range[1L]), round(time_range[2L])) / 
        resample_rate
    resampled_times <- seq(time_range[1L], time_range[2L], by = resample_time)
    result <- setNames(data.frame(resampled_times), time_channel)

    ## identify numeric cols; integers and `time_channel` excluded
    numeric_cols <- vapply(data, \(.x) {
        is.numeric(.x) && !is.integer(.x)
    }, logical(1))
    numeric_cols[time_channel] <- FALSE
    non_numeric_cols <- colnames[!numeric_cols & colnames != time_channel]

    ## nearest-match index for "none" method =============================
    if (method == "none") {
        idx <- max.col(
            -abs(outer(resampled_times, time_vec, `-`)),
            ties.method = "first"
        )
        tol <- resample_time * 0.5
        match_ok <- abs(resampled_times - time_vec[idx]) < tol
        idx[!match_ok] <- NA
    }

    ## interpolate numeric columns (locf/linear only) ==================
    if (method != "none" && any(numeric_cols)) {
        result[colnames[numeric_cols]] <- lapply(
            data[numeric_cols], \(.x) {
                replace_missing(
                    .x, time_vec, method = method, xout = resampled_times
                )
            }
        )
    }

    ## index-based fill for non-numeric columns ===========================
    idx_cols <- if (method == "none") {
        colnames[colnames != time_channel]
    } else {
        non_numeric_cols
    }

    if (length(idx_cols) > 0) {
        if (method == "none") {
            ## tolerance-matched index computed above
            result[idx_cols] <- lapply(data[idx_cols], \(.x) .x[idx])
        } else if (length(resampled_times) < length(time_vec)) {
            ## down-sample: first non-NA per bin
            idx <- findInterval(
                time_vec, resampled_times, rightmost.closed = FALSE
            )
            idx[idx == 0] <- 1
            result[idx_cols] <- lapply(data[idx_cols], \(.x) {
                unname(unsplit(
                    lapply(split(.x, idx), \(.v) .v[!is.na(.v)][1L]),
                    unique(idx)
                ))
            })
        } else {
            ## up-sample / regularise: forward fill
            idx <- findInterval(
                resampled_times, time_vec, rightmost.closed = FALSE
            )
            idx[idx == 0] <- 1
            result[idx_cols] <- lapply(data[idx_cols], \(.x) .x[idx])
        }
    }

    ## Metadata =================================
    metadata$time_channel <- time_channel
    metadata$sample_rate <- resample_rate

    if (verbose) {
        cli_inform(c("i" = "Output is resampled at {.val {resample_rate}} Hz."))
    }

    ## column order same as input
    result <- result[colnames]
    return(create_mnirs_data(result, metadata))
}

#' Re-scale data range
#'
#' Expand or reduce the range (min and max values) of data channels to a new
#' amplitude/dynamic range, e.g. re-scale the range of NIRS data to `c(0, 100)`.
#'
#' @usage
#' rescale_mnirs(
#'   data,
#'   nirs_channels = list(),
#'   range,
#'   verbose = TRUE
#' )
#' 
#' @param nirs_channels A `list()` of character vectors indicating grouping
#'   structure of mNIRS channel names to operate on (see *Details*). Must
#'   match column names in `data` exactly. Retrieved from metadata if not
#'   defined explicitly.
#'   \describe{
#'      \item{`list("A", "B", "C")`}{Will operate on each channel independently,
#'      losing the relative scaling between channels.}
#'      \item{`list(c("A", "B", "C"))`}{Will operate on all channels together,
#'      preserving the relative scaling between channels.}
#'      \item{`list(c("A", "B"), c("C", "D"))`}{Will operate on channels `A`
#'      & `B` in one group, and `C` & `D` in another group, preserving
#'      relative scaling within, but not between groups.}
#'   }
#' @param range A numeric vector in the form `c(min, max)`, indicating the
#'   range of output values to which data channels will be re-scaled.
#' @inheritParams validate_mnirs
#'
#' @details
#' `nirs_channels = list()` can be used to group data channels (column names)
#'   to preserve absolute or relative scaling.
#'
#' - Channels grouped together in a vector (e.g. `list(c("A", "B"))`) will be
#'   re-scaled to a common range, and the relative scaling within that group
#'   will be preserved.
#'
#' - Channels in separate list vectors (e.g. `list("A", "B")`) will be
#'   re-scaled independently, and relative scaling between groups will be lost.
#'
#' - A single vector of channel names (e.g. `c("A", "B")`) will group
#'   channels together.
#'
#' - Channels (columns) in `data` not explicitly defined in `nirs_channels`
#'   will be passed through untouched to the output data frame.
#'
#' `nirs_channels` can be retrieved automatically from `data` of class
#'   *"mnirs"* which has been processed with `{mnirs}`, if not defined
#'   explicitly. This will default to returning all `nirs_channels` grouped
#'   together, and should be defined explicitly for other grouping arrangements.
#'
#' @returns
#' A [tibble][tibble::tibble-package] of class *"mnirs"* with metadata
#'   available with `attributes()`.
#'
#' @examples
#' ## read example data
#' data <- read_mnirs(
#'     file_path = example_mnirs("moxy_ramp"),
#'     nirs_channels = c(smo2_left = "SmO2 Live",
#'                       smo2_right = "SmO2 Live(2)"),
#'     time_channel = c(time = "hh:mm:ss"),
#'     verbose = FALSE
#' ) |>
#'     rescale_mnirs(        ## un-grouped nirs channels to rescale separately 
#'         nirs_channels = list(smo2_left, smo2_right), 
#'         range = c(0, 100) ## rescale to a 0-100% functional exercise range
#'     )
#' 
#' data
#'
#' \donttest{
#'     if (requireNamespace("ggplot2", quietly = TRUE)) {
#'         plot(data, time_labels = TRUE) +
#'             ggplot2::geom_hline(yintercept = c(0, 100), linetype = "dotted")
#'     }
#' }
#'
#' @export
rescale_mnirs <- function(
    data,
    nirs_channels = list(),
    range,
    verbose = TRUE
) {
    ## validate =================================
    validate_mnirs_data(data, ncol = 1)
    metadata <- attributes(data)
    if (missing(verbose)) {
        verbose <- getOption("mnirs.verbose", default = TRUE)
    }
    nirs_channels <- validate_nirs_channels(
        enquo(nirs_channels), data, verbose, as_list = TRUE
    )
    validate_numeric(
        range, 2, 
        msg1 = "two-element", 
        msg2 = "between {col_blue('range[1], range[2]]')}."
    )

    if (
        verbose &&
            is.null(attr(data, "nirs_channels")) &&
            !is.list(nirs_channels)
    ) {
        cli_inform(c(
            "!" = "{.fn rescale_mnirs} accepts {.arg nirs_channels} = \\
            {col_blue('list()')} for channel grouping. See `?rescale_mnirs`."
        ))
    }

    ## rescale range ================================
    ## this is actually a fast modify-in-place for loop
    for (group in nirs_channels) {
        group_data <- as.matrix(data[, group, drop = FALSE])
        group_data_range <- range(group_data, na.rm = TRUE)
        range_diff <- diff(group_data_range)

        if (range_diff != 0) {
            # fmt: skip
            data[, group] <- (group_data - group_data_range[1]) /
                range_diff * diff(range) + range[1]
        }
    }

    ## Metadata =================================
    metadata$nirs_channels <- unique(unlist(nirs_channels))

    return(create_mnirs_data(data, metadata))
}

#' Read *{mnirs}* data from file
#'
#' Import time-series data exported from common muscle NIRS (mNIRS) devices and
#' return a tibble of class `"mnirs"` with the selected signal channels and
#' metadata.
#'
#' @param file_path Path of the data file to import. Supported file extensions
#'   are `".xlsx"`, `".xls"`, and `".csv"`.
#'
#' @param nirs_channels A character vector of one or more column names
#'   containing mNIRS signals to import. Names must match the file header
#'   exactly.
#'
#'   - If `NULL` (default), `read_mnirs()` attempts to detect the device from
#'     the file contents and use a known `nirs_channel` name.
#'   - A *named* character vector can be used to rename columns on import, in
#'     the form `c(renamed = "original_name")`.
#'
#' @param time_channel A character string giving the name of the time
#'   (or sample) column to import. The name must match the file header exactly.
#'
#'   - If `NULL` (default), `read_mnirs()` attempts to identify a time-like
#'     column automatically (by known device defaults and/or time-formatted
#'     values).
#'   - A *named* character vector can be used to rename the column on import,
#'     in the form `c(time = "original_name")`.
#'
#' @param event_channel An *optional* character string giving the name of an
#'   event/lap column to import. Names must match the file header exactly.
#'   A named character vector can be used to rename the column on import in
#'   the form `c(event = "original_name")`.
#'
#' @param sample_rate An *optional* numeric sample rate in Hz. If left blank
#'   (`NULL`), the sample rate is estimated from `time_channel` (see *Details*).
#'
#' @param add_timestamp A logical. Default is `FALSE`. If `TRUE` and if the
#'   source data contain an absolute date-time (POSIXct) time value, will add
#'   a `"timestamp"` column in addition to the specified `time_channel` as a
#'   numeric time column.
#'
#' @param zero_time Logical. Default is `FALSE`. If `TRUE`, re-calculates
#'   numeric `time_channel` values to start from zero.
#'
#' @param keep_all Logical. Default is `FALSE`. Will keep only the channels
#'   explicitly specified in `nirs_channels`, `time_channel`, and
#'   `event_channel`. If `TRUE` will keep all columns found in the file
#'   data table.
#'
#'   - If no `nirs_channels` are specified and the file format is recognised,
#'     all columns in the file data table will be returned, as an exploratory
#'     option.
#'
#' @param verbose Logical. Default is `TRUE`. Display or silence (if `FALSE`)
#'   warnings and information messages helpful for troubleshooting. Ad
#'   global default can be set via `options(mnirs.verbose = FALSE)`.
#'
#' @details
#' ## Header detection
#' `read_mnirs()` searches the file for a header row containing the requested
#' channel names. The header row does not need to be the first row in the file.
#'
#' - If duplicate column names exist, columns are matched in the order they
#'   appear and renamed with unique strings.
#' - Columns without a header name in the source file will be renamed to
#'   `col_*`, where `*` is the numeric column number in which they appear in
#'   the file (e.g. `col_6`). This applies to *Artinis Oxysoft* event label
#'   columns, which do not have a column header and must be identified manually.
#'
#' ## Renaming channels
#' A named character vector can be specified to rename `nirs_channels`,
#' `time_channel`, and `event_channel`, in the form
#' `c(renamed = "original_name")`. The `"original_name"` must match the
#' contents of the file data table header row exactly.
#'
#' ## Time parsing
#' `time_channel` will be converted to numeric for analysis.
#'
#' - If `time_channel` is a date-time (POSIXct) format, it will be converted
#'   to numeric and re-based to start from 0, regardless of `zero_time`.
#' - Some devices export a sample index rather than time values. In those
#'   cases, if an export `sample_rate` is detected in the file metadata (e.g.
#'   *Artinis Oxysoft* exports), `read_mnirs()` will create or overwrite a
#'   `"time"` column in seconds derived from the sample index and the detected
#'   `sample_rate`.
#'
#' ## Sample rate
#' If `sample_rate` is not specified, it is estimated from differences in
#' `time_channel`. If `time_channel` is actually a sample index, as described
#' above, this may erroneously be estimated at 1 Hz. `sample_rate` should be
#' specified explicitly in this case.
#'
#' ## Data cleaning
#' Entirely empty rows and columns are removed. Invalid values (e.g.
#' `c(NaN, Inf)`) are standardized to `NA`. A warning will be displayed when
#' irregular sampling is detected (e.g. non-monotonic, repeated, or unequal
#' time values), if `verbose = TRUE`.
#'
#' @returns
#' A [tibble][tibble::tibble-package] of class `"mnirs"`. Metadata are stored
#'   as attributes and can be accessed with `attributes(data)`.
#'
#' @examples
#' read_mnirs(
#'     file_path = example_mnirs("moxy_ramp"), ## call an example data file
#'     nirs_channels = c(
#'         smo2_left = "SmO2 Live",            ## identify and rename channels
#'         smo2_right = "SmO2 Live(2)"
#'     ),
#'     time_channel = c(time = "hh:mm:ss"),    ## date-time format will be converted to numeric
#'     event_channel = NULL,                   ## leave blank if unused
#'     sample_rate = NULL,                     ## if blank, will be estimated from time_channel
#'     add_timestamp = FALSE,                  ## omit a date-time timestamp column
#'     zero_time = TRUE,                       ## recalculate time values from zero
#'     keep_all = FALSE,                       ## return only the specified data channels
#'     verbose = TRUE                          ## show warnings & messages
#' )
#'
#' @export
read_mnirs <- function(
    file_path,
    nirs_channels = NULL,
    time_channel = NULL,
    event_channel = NULL,
    sample_rate = NULL,
    add_timestamp = FALSE,
    zero_time = FALSE,
    keep_all = FALSE,
    verbose = TRUE
) {
    ## global options overrides implicit but not explicit `verbose`
    if (missing(verbose)) {
        verbose <- getOption("mnirs.verbose", default = TRUE)
    }

    ## import data_raw from either excel or csv
    data <- read_file(file_path)

    ## detect mNIRS device from raw data. Returns NULL if not found
    detected_list <- detect_mnirs_device(data)
    nirs_device <- detected_list$nirs_device
    header_row <- detected_list$header_row

    ## resolve channels: use user input if provided, otherwise detect from
    ## known device channel names. Errors if neither available.
    channels <- detect_device_channels(
        data,
        header_row,
        nirs_device,
        nirs_channels,
        time_channel,
        keep_all,
        verbose
    )
    nirs_channels <- channels$nirs_channels
    time_channel <- channels$time_channel
    keep_all <- channels$keep_all ## TRUE when `nirs_channels` unspecified

    ## extract the data_table, and name by header row
    table_list <- read_data_table(data, nirs_channels, header_row)
    data <- table_list$data_table
    file_header <- table_list$file_header

    ## extract start time from file header
    start_timestamp <- extract_start_timestamp(file_header)

    ## attempt to detect `time_channel` automatically
    time_channel <- detect_time_channel(
        data, time_channel, nirs_device, verbose
    )

    ## rename from channel names, make duplicates unique, keep columns
    ## return list(data_renamed, nirs_renamed, time_renamed, event_renamed)
    renamed_list <- select_rename_data(
        data, nirs_channels, time_channel, event_channel, keep_all, verbose
    )
    data <- renamed_list$data
    nirs_renamed <- renamed_list$nirs_channel
    time_renamed <- renamed_list$time_channel
    event_renamed <- renamed_list$event_channel

    ## remove empty (NA) columns and rows
    data <- remove_empty_rows_cols(data)
    ## convert char decimal "," to "." and convert column types
    data <- convert_type(data, time_renamed, event_renamed, verbose)
    ## convert POSIXct to numeric and/or recalc time from zero
    ## return list(data, start_timestamp) — start_timestamp from time_channel POSIXct
    time_list <- parse_time_channel(
        data, time_renamed, start_timestamp, add_timestamp, zero_time
    )
    data <- time_list$data

    ## extract start_timestamp from data if not already found in header
    if (is.null(start_timestamp)) {
        start_timestamp <- time_list$start_timestamp
    }

    ## validate and estimate sample rate
    ## will write new "time" column if Oxysoft export rate detected
    ## return list(data_sampled, time_renamed, sample_rate)
    sample_list <- parse_sample_rate(
        data, file_header, time_renamed, sample_rate, nirs_device, verbose
    )
    data <- sample_list$data
    time_renamed <- sample_list$time_channel
    sample_rate <- sample_list$sample_rate

    ## print warnings for irregular samples
    detect_irregular_samples(data[[time_renamed]], time_renamed, verbose)

    ## assign metadata to attributes(data)
    metadata <- list(
        nirs_device = nirs_device,
        nirs_channels = nirs_renamed,
        time_channel = time_renamed,
        event_channel = event_renamed,
        sample_rate = sample_rate,
        start_timestamp = start_timestamp,
        verbose = verbose
    )

    return(create_mnirs_data(data, metadata))
}

#' Metadata names  of class `"mnirs"`, retrieved with `attr()`
#' @keywords internal
mnirs_metadata <- c(
    "nirs_device",
    "nirs_channels",
    "time_channel",
    "event_channel",
    "sample_rate",
    "start_timestamp",
    "interval_times",
    "interval_span"
)


#' Create an *{mnirs}* data frame with metadata
#'
#' Manually add class `"mnirs"` and metadata to an existing data frame.
#'
#' @param data A data frame with existing metadata (accessed with
#'   `attributes(data)`).
#'
#' @param ... Additional arguments with metadata to add to the data frame.
#'   Can be either seperate named arguments or a list of named values.
#'   - nirs_device
#'   - nirs_channels
#'   - time_channel
#'   - event_channel
#'   - sample_rate
#'   - start_timestamp
#'   - interval_times
#'   - interval_span
#'
#' @details
#' Typically will only be called internally, but can be used to inject
#'   *{mnirs}* metadata into any data frame.
#'
#' @returns
#' A [tibble][tibble::tibble-package] of class `"mnirs"`. Metadata are stored
#'   as attributes and can be accessed with `attributes(data)`.
#'
#' @examples
#' data <- data.frame(
#'     A = 1:3,
#'     B = seq(10, 30, 10),
#'     C = seq(11, 33, 11)
#' )
#'
#' attributes(data)
#'
#' ## inject metadata
#' nirs_data <- create_mnirs_data(
#'     data,
#'     nirs_channels = c("B", "C"),
#'     time_channel = "A",
#'     sample_rate = 1
#' )
#'
#' attributes(nirs_data)
#'
#' @export
create_mnirs_data <- function(data, ...) {
    validate_mnirs_data(data, 1L)

    ## tidy eval ========================================================
    ## capture quosures so bare symbols / tidyselect resolve against `data`
    dots <- rlang::enquos(...)
    args <- Map(\(.q, .nm) {
        if (.nm %in% c("nirs_channels", "time_channel", "event_channel")) {
            parse_channel_name(.q, data)
        } else {
            rlang::eval_tidy(.q)
        }
    }, dots, names(dots) %||% rep("", length(dots)))

    ## overwrite existing attributes and add from incoming metadata
    ## incoming metadata from `...` can be either listed or un-listed
    incoming_metadata <- if (length(args) == 1L && is.list(args[[1L]])) {
        args[[1L]]
    } else {
        args
    }

    metadata <- utils::modifyList(attributes(data), incoming_metadata)

    nirs_data <- tibble::new_tibble(
        data,
        class = "mnirs",
        nirs_device = metadata$nirs_device,
        nirs_channels = metadata$nirs_channels,
        time_channel = metadata$time_channel,
        event_channel = metadata$event_channel,
        sample_rate = metadata$sample_rate,
        start_timestamp = metadata$start_timestamp,
        interval_times = metadata$interval_times,
        interval_span = metadata$interval_span,
    )

    tibble::validate_tibble(nirs_data)

    return(nirs_data)
}


#' Get path to *{mnirs}* example files
#'
#' @param file Name of file as character string. If `NULL`, returns a vector
#' of all available file names.
#'
#' @returns
#' A file path character string for selected example files stored in this
#'   package.
#'
#' @examples
#' ## lists all files
#' example_mnirs()
#'
#' ## partial matching will error if matches multiple
#' try(example_mnirs("moxy"))
#'
#' example_mnirs("moxy_ramp")
#'
#' @export
example_mnirs <- function(file = NULL) {
    dir_files <- list.files(
        system.file("extdata", package = "mnirs"),
        pattern = "^[^~]" ## exclude open files
    )

    if (is.null(file)) {
        return(dir_files)
    }

    matches <- grep(file, dir_files, fixed = TRUE, value = TRUE)
    if (length(matches) > 1L) {
        cli_abort(c(
            "x" = "Multiple files match {.val {file}}:",
            "i" = "Matching files: {.val {matches}}"
        ))
    }

    file <- match.arg(file, choices = dir_files)
    system.file("extdata", file, package = "mnirs", mustWork = TRUE)
}

test_that("rescale_mnirs rescales single channel correctly", {
    data <- tibble(A = c(0, 50, 100), B = c(10, 20, 30))

    result <- rescale_mnirs(
        data,
        nirs_channels = list("A"),
        range = c(0, 1)
    )

    expect_equal(result$A, c(0, 0.5, 1))
    expect_equal(result$B, data$B) # unchanged
})

test_that("rescale_mnirs preserves relative scaling across grouped channels", {
    data <- tibble(A = c(0, 50, 100), B = c(25, 25, 50))

    result <- rescale_mnirs(
        data,
        nirs_channels = list(c("A", "B")),
        range = c(0, 1)
    )

    expect_equal(result$A, c(0, 0.5, 1))
    expect_equal(result$B, c(0.25, 0.25, 0.5))
})

test_that("rescale_mnirs handles multiple separate groups", {
    data <- tibble(A = c(10, 100), B = c(0, 50), C = c(10, 200))

    result <- rescale_mnirs(
        data,
        nirs_channels = list(c("A", "B"), "C"),
        range = c(0, 10)
    )

    expect_equal(result$A, c(1, 10))
    expect_equal(result$B, c(0, 5))
    # C scaled independently
    expect_equal(result$C, c(0, 10))
})

test_that("rescale_mnirs handles negative ranges", {
    data <- tibble(A = c(0, 50, 100))

    result <- rescale_mnirs(
        data,
        nirs_channels = list("A"),
        range = c(-1, 1)
    )

    expect_equal(result$A, c(-1, 0, 1))
})

test_that("rescale_mnirs handles NA values", {
    data <- tibble(A = c(0, NA, 100))

    result <- rescale_mnirs(
        data,
        nirs_channels = list("A"),
        range = c(0, 1)
    )

    expect_equal(result$A, c(0, NA, 1))
})

test_that("rescale_mnirs errors for nirs_channels", {
    data <- tibble(A = c(0, 50, 100))

    expect_error(
        rescale_mnirs(data, nirs_channels = NULL, range = c(0, 1)),
        "nirs_channels.*not detected"
    )
    expect_error(
        rescale_mnirs(data, range = c(0, 1)),
        "nirs_channels.*not detected"
    )
    expect_error(
        rescale_mnirs(data, nirs_channels = "doesn't exist", range = c(0, 1)),
        "nirs_channels.*match exactly"
    )
})

test_that("rescale_mnirs errors with invalid range", {
    data <- tibble(A = c(0, 50, 100))

    expect_error(
        rescale_mnirs(data, nirs_channels = list("A"), range = c(0, 1, 2)),
        "range.*numeric"
    )
    expect_error(
        rescale_mnirs(data, nirs_channels = list("A"), range = "invalid"),
        "range.*numeric"
    )
})

test_that("rescale_mnirs returns unmodified column when values are constant", {
    data <- tibble(A = c(50, 50, 50), B = c(0, 100, 200))

    ## ungrouped should not change
    result <- rescale_mnirs(
        data,
        nirs_channels = list("A", "B"),
        range = c(0, 100)
    )

    expect_equal(result$A, c(50, 50, 50))
    expect_equal(result$B, c(0, 50, 100))

    ## grouped should change
    result <- rescale_mnirs(
        data,
        nirs_channels = list(c("A", "B")),
        range = c(0, 100)
    )

    # When grouped, if one channel is constant and another varies,
    # the constant channel should remain unchanged
    expect_equal(result$A, data$A / 2)
    expect_equal(result$B, data$B / 2)
})

test_that("rescale_mnirs informs when nirs_channels is not a list()", {
    data <- tibble(A = c(0, 50, 100))

    ## fires: verbose=TRUE, no metadata, non-list channels
    expect_message(
        rescale_mnirs(
            data,
            nirs_channels = "A",
            range = c(0, 1),
            verbose = TRUE
        ),
        "list\\(\\).*channel grouping"
    )

    ## silent: verbose=FALSE
    expect_no_message(
        rescale_mnirs(
            data,
            nirs_channels = "A",
            range = c(0, 1),
            verbose = FALSE
        )
    )

    ## silent: nirs_channels already a list()
    expect_no_message(
        rescale_mnirs(
            data,
            nirs_channels = list("A"),
            range = c(0, 1),
            verbose = TRUE
        )
    )

    ## silent: data has nirs_channels metadata
    attr(data, "nirs_channels") <- "A"
    expect_no_message(
        rescale_mnirs(
            data,
            nirs_channels = "A",
            range = c(0, 1),
            verbose = TRUE
        )
    )
})

test_that("rescale_mnirs updates metadata correctly", {
    data <- tibble(A = c(50, 50, 50), B = c(0, 100, 200))

    result <- rescale_mnirs(
        data,
        nirs_channels = list("A", "B"),
        range = c(0, 100),
        verbose = FALSE
    )

    expect_true(all(c("A", "B") %in% attr(result, "nirs_channels")))

    attr(data, "nirs_channels") <- "A"
    result <- rescale_mnirs(
        data,
        nirs_channels = list("A", "B"),
        range = c(0, 100),
        verbose = FALSE
    )

    expect_true(all(c("A", "B") %in% attr(result, "nirs_channels")))
})

test_that("rescale_mnirs works on Moxy", {
    file_path <- example_mnirs("moxy_ramp.xlsx")

    df <- read_mnirs(
        file_path = file_path,
        nirs_channels = c(smo2_left = "SmO2 Live", smo2_right = "SmO2 Live(2)"),
        time_channel = c(time = "hh:mm:ss"),
        verbose = FALSE
    ) |>
        dplyr::mutate(
            dplyr::across(
                dplyr::matches("smo2"),
                \(.x) {
                    replace_invalid(
                        .x,
                        invalid_values = c(0, 100),
                        method = "none"
                    )
                }
            )
        )

    result <- rescale_mnirs(
        df,
        nirs_channels = list(c("smo2_left", "smo2_right")),
        range = c(0, 100)
    )

    # plot(df) + ggplot2::ylim(0, 100) + geom_hline(yintercept = c(0, 100))
    # plot(result) + ggplot2::ylim(0, 100) + geom_hline(yintercept = c(0, 100))

    ## check grouping together: min & max value should come from smo2_right
    expect_false(any(result$smo2_left %in% c(0, 100), na.rm = TRUE))
    expect_true(any(result$smo2_right %in% c(0, 100), na.rm = TRUE))

    ## check grouped apart
    result <- rescale_mnirs(
        df,
        nirs_channels = list("smo2_left", "smo2_right"),
        range = c(0, 100)
    )

    # plot(df) + ggplot2::ylim(0, 100) + geom_hline(yintercept = c(0, 100))
    # plot(result) + ggplot2::ylim(0, 100) + geom_hline(yintercept = c(0, 100))

    ## check grouping together: min value should come from smo2_right
    expect_true(any(result$smo2_left %in% c(0, 100), na.rm = TRUE))
    expect_true(any(result$smo2_right %in% c(0, 100), na.rm = TRUE))
})

test_that("rescale_mnirs works on Train.Red", {
    file_path <- example_mnirs("train.red_intervals.csv")

    df <- read_mnirs(
        file_path = file_path,
        nirs_channels = c(
            smo2_left = "SmO2 unfiltered",
            smo2_right = "SmO2 unfiltered",
            o2hb_left = "O2HB unfiltered",
            o2hb_right = "O2HB unfiltered"
        ),
        time_channel = c(time = "Timestamp (seconds passed)"),
        verbose = FALSE
    )

    result <- rescale_mnirs(
        df,
        nirs_channels = list(
            "smo2_left",
            "smo2_right",
            c("o2hb_left", "o2hb_right")
        ),
        range = c(0, 100)
    )

    # plot(df) + ggplot2::ylim(0, 100) + geom_hline(yintercept = c(0, 100))
    # plot(result) + ggplot2::ylim(0, 100) + geom_hline(yintercept = c(0, 100))

    ## check grouping together: min value should come from each group
    expect_true(any(result$smo2_left %in% c(0, 100), na.rm = TRUE))
    expect_true(any(result$smo2_right %in% c(0, 100), na.rm = TRUE))
    expect_true(any(result$o2hb_left == 0, na.rm = TRUE))
    expect_false(any(result$o2hb_right == 0, na.rm = TRUE))
    expect_false(any(result$o2hb_left == 100, na.rm = TRUE))
    expect_true(any(result$o2hb_right == 100, na.rm = TRUE))
})

test_that("shift_mnirs requires either to or by", {
    data <- tibble(
        time = 1:10,
        ch1 = rnorm(10)
    )

    expect_error(
        shift_mnirs(data, nirs_channels = list("ch1"), time_channel = "time"),
        "to.*by"
    )
})

test_that("shift_mnirs shifts by constant correctly", {
    data <- tibble(
        time = 1:10,
        ch1 = 1:10,
        ch2 = 11:20
    )

    result <- shift_mnirs(
        data,
        nirs_channels = list("ch1", "ch2"),
        time_channel = "time",
        by = 5,
        verbose = FALSE
    )

    expect_equal(result$ch1, data$ch1 + 5)
    expect_equal(result$ch2, data$ch2 + 5)
})

test_that("shift_mnirs preserves relative scaling within groups", {
    data <- tibble(
        time = 1:10,
        ch1 = 1:10,
        ch2 = 11:20
    )

    result <- shift_mnirs(
        data,
        nirs_channels = list(c("ch1", "ch2")),
        time_channel = "time",
        to = 0,
        width = 1,
        position = "min",
        verbose = FALSE
    )

    # ch1 min is 1, so shift both by -1
    expect_equal(result$ch1, 0:9)
    expect_equal(result$ch2, 10:19)
    # Relative difference preserved
    expect_equal(result$ch2[1] - result$ch1[1], 10)
})

test_that("shift_mnirs loses relative scaling across groups", {
    data <- tibble(
        time = 1:10,
        ch1 = 1:10,
        ch2 = 11:20
    )

    result <- shift_mnirs(
        data,
        nirs_channels = list("ch1", "ch2"),
        time_channel = "time",
        to = 0,
        width = 1,
        position = "min",
        verbose = FALSE
    )

    # Each shifted independently to min = 0
    expect_equal(result$ch1, 0:9)
    expect_equal(result$ch2, 0:9)
    # Relative scaling lost
    expect_equal(result$ch2[1] - result$ch1[1], 0)
})

test_that("shift_mnirs handles position = 'max' correctly", {
    data <- tibble(
        time = 1:10,
        ch1 = 1:10
    )

    result <- shift_mnirs(
        data,
        nirs_channels = list("ch1"),
        time_channel = "time",
        to = 100,
        width = 1,
        position = "max",
        verbose = FALSE
    )

    # Max is 10, shift to 100 means add 90
    expect_equal(result$ch1, 91:100)
})

test_that("shift_mnirs handles position = 'first' with span and width", {
    data <- tibble(
        time = seq(0, 9, by = 1),
        ch1 = c(rep(10, 3), 1:7)
    )

    result <- shift_mnirs(
        data,
        nirs_channels = list("ch1"),
        time_channel = "time",
        to = 0,
        position = "first",
        span = 2,
        verbose = FALSE
    )

    # Mean of first 3 values (time 0-2) is 10
    expect_equal(result$ch1, c(rep(0, 3), -9:-3))

    result <- shift_mnirs(
        data,
        nirs_channels = list("ch1"),
        time_channel = "time",
        to = 0,
        position = "first",
        width = 3,
        verbose = FALSE
    )
    expect_equal(result$ch1, c(rep(0, 3), -9:-3))
})

test_that("shift_mnirs preserves unshifted channels", {
    data <- tibble(
        time = 1:10,
        ch1 = 1:10,
        ch2 = 11:20,
        ch3 = 21:30
    )

    result <- shift_mnirs(
        data,
        nirs_channels = list("ch1"),
        time_channel = "time",
        by = 5,
        width = 1,
        verbose = FALSE
    )

    expect_equal(result$ch1, 6:15)
    expect_equal(result$ch2, 11:20)
    expect_equal(result$ch3, 21:30)
})

test_that("shift_mnirs updates metadata correctly", {
    data <- tibble(time = 1:10, ch1 = 1:10, ch2 = 1:10)

    result <- shift_mnirs(
        data,
        nirs_channels = list("ch1"),
        time_channel = "time",
        by = 5,
        width = 1,
        verbose = FALSE
    )

    expect_true("ch1" %in% attr(result, "nirs_channels"))
    expect_equal(attr(result, "time_channel"), "time")

    attr(data, "time_channel") <- NULL

    result <- shift_mnirs(
        data,
        nirs_channels = list("ch2"),
        time_channel = "time",
        by = 5,
        width = 1,
        verbose = FALSE
    )

    expect_true("ch2" %in% attr(result, "nirs_channels"))
    expect_equal(attr(result, "time_channel"), "time")
})

test_that("shift_mnirs handles mixed channel groups", {
    data <- tibble(
        time = 1:10,
        ch1 = 1:10,
        ch2 = 11:20,
        ch3 = 31:40,
        ch4 = 41:50
    )

    result <- shift_mnirs(
        data,
        nirs_channels = list(c("ch1", "ch2"), c("ch3", "ch4")),
        time_channel = "time",
        to = 0,
        width = 1,
        position = "min",
        verbose = FALSE
    )

    # Group 1: min is 1, both shifted by -1
    expect_equal(result$ch1, 0:9)
    expect_equal(result$ch2, 10:19)
    # Group 2: min is 31, both shifted by -31
    expect_equal(result$ch3, 0:9)
    expect_equal(result$ch4, 10:19)
    # Scaling preserved within, not between groups
    expect_equal(result$ch2[1] - result$ch1[1], 10)
    expect_equal(result$ch4[1] - result$ch3[1], 10)
})

test_that("shift_mnirs handles NA values correctly", {
    data <- tibble(
        time = 1:10,
        ch1 = c(NA, 2:10)
    )

    result <- shift_mnirs(
        data,
        nirs_channels = list("ch1"),
        time_channel = "time",
        to = 0,
        width = 1,
        position = "min",
        verbose = FALSE
    )

    # Min of non-NA values is 2
    expect_true(is.na(result$ch1[1]))
    expect_equal(result$ch1[2:10], 0:8)
})

test_that("shift_mnirs informs when nirs_channels is not a list()", {
    data <- tibble(time = 1:5, ch1 = 1:5)

    ## fires: verbose=TRUE, no metadata, non-list channels
    expect_message(
        shift_mnirs(
            data,
            nirs_channels = "ch1",
            time_channel = "time",
            by = 1,
            verbose = TRUE
        ),
        "list\\(\\).*channel grouping"
    )

    ## silent: verbose=FALSE
    expect_no_message(
        shift_mnirs(
            data,
            nirs_channels = "ch1",
            time_channel = "time",
            by = 1,
            verbose = FALSE
        )
    )

    ## silent: nirs_channels already a list()
    expect_no_message(
        shift_mnirs(
            data,
            nirs_channels = list("ch1"),
            time_channel = "time",
            by = 1,
            verbose = TRUE
        )
    )

    ## silent: data has nirs_channels metadata
    attr(data, "nirs_channels") <- "ch1"
    expect_no_message(
        shift_mnirs(
            data,
            nirs_channels = "ch1",
            time_channel = "time",
            by = 1,
            verbose = TRUE
        )
    )
})

test_that("shift_mnirs prioritises to over by", {
    data <- tibble(
        time = 1:10,
        ch1 = 1:10
    )

    expect_message(
        result <- shift_mnirs(
            data,
            nirs_channels = list("ch1"),
            time_channel = "time",
            to = 0,
            by = 100,
            width = 1,
            position = "min",
            verbose = TRUE
        ),
        "to.*overrides.*by"
    )

    # Should use 'to', not 'by'
    expect_equal(result$ch1, 0:9)
})

test_that("shift_mnirs handles unevenly sampled data with span", {
    set.seed(13)
    # 0.5 Hz target (2 second intervals) with ±0.1s error
    base_times <- seq(0, 10, by = 2)
    time_error <- runif(length(base_times), -0.1, 0.1)

    data <- data.frame(
        time = round(base_times + time_error, 2),
        ch1 = c(10, 5, 20, 15, 25, 30)
    )

    # Test with 4-second span (should include ~2 samples per window)
    result <- shift_mnirs(
        data,
        list("ch1"),
        time_channel = "time",
        to = 0,
        span = 4,
        position = "first"
    )

    # First window should average first 3 values within 4s of start
    first_window_idx <- which(data$time <= data$time[1] + 4)
    expect_equal(which(data$time <= 4), first_window_idx)
    expected_mean <- mean(data$ch1[first_window_idx])
    expect_equal(result$ch1[1], data$ch1[1] - expected_mean)

    ## width = 3
    result <- shift_mnirs(
        data,
        list("ch1"),
        time_channel = "time",
        to = 0,
        width = 3,
        position = "first"
    )

    # First window should average first 3 values
    expect_equal(result$ch1[1], data$ch1[1] - mean(data$ch1[1:3]))

    ## unequal sampling and position = "minimum"
    result_min <- shift_mnirs(
        data,
        list("ch1"),
        time_channel = "time",
        to = 0,
        span = 4,
        position = "min"
    )

    x <- data$time
    min_mean <- Inf
    idx <- NA_real_
    for (i in seq_along(x)) {
        window_idx <- which(x >= x[i] & x <= x[i] + 4)
        if (length(window_idx) > 1) {
            # print(window_idx)
            window_mean <- mean(data$ch1[window_idx])
            # print(window_mean)
            min_mean <- min(min_mean, window_mean)
            if (min_mean == window_mean) idx <- window_idx
        }
    }
    expect_equal(idx, first_window_idx)
})

test_that("shift_mnirs handles multiple channel groups", {
    data <- data.frame(
        time = 1:2,
        ch1 = c(10, 20),
        ch2 = c(15, 25),
        ch3 = c(5, 35)
    )
    nirs_channels <- list(c("ch1", "ch2"), "ch3")
    result <- shift_mnirs(data, nirs_channels, time, to = 0, width = 1)

    ## check grouping together: min shuold come from ch1 and ch3
    expect_true(any(result$ch1 == 0, na.rm = TRUE))
    expect_false(any(result$ch2 == 0, na.rm = TRUE))
    expect_true(any(result$ch3 == 0, na.rm = TRUE))
    ## check both shifted together maintaining relative scaling
    expect_equal(result$ch1 - result$ch2, data$ch1 - data$ch2)
    ## check both shifted independently
    expect_false(isTRUE(all.equal(
        result$ch1 - result$ch3,
        data$ch1 - data$ch3
    )))
})

test_that("shift_mnirs preserves non-channel columns", {
    data <- data.frame(
        time = 1:3,
        ch1 = c(10, 20, 30),
        other = c("A", "B", "C")
    )
    result <- shift_mnirs(data, list("ch1"), "time", by = 5)

    expect_equal(result$time, c(1, 2, 3))
    expect_equal(result$other, c("A", "B", "C"))
    expect_equal(result$ch1, c(15, 25, 35))
})

test_that("shift_mnirs validates position argument", {
    data <- data.frame(ch1 = c(10, 20))
    expect_error(
        shift_mnirs(data, list("ch1"), position = "invalid"),
        "to.*by"
    )
})

test_that("shift_mnirs handles empty channel list", {
    data <- data.frame(time = 1:3, value = c(10, 20, 30))
    expect_error(shift_mnirs(data, list(), by = 5), "`nirs_channels` not detected")
})

## shift_mnirs() tidy eval integration ========================================
## helper to create test data with metadata
create_test_data <- function(
    time_max = 10,
    sample_rate = 10,
    add_metadata = TRUE
) {
    time <- seq(0, time_max, by = 1 / sample_rate)
    nrow <- length(time)
    data <- tibble(
        time = time,
        nirs1 = rnorm(nrow, 50, 5),
        nirs2 = rnorm(nrow, 60, 5),
        nirs3 = rnorm(nrow, 80, 5),
        event = c(1, rep(NA, nrow - 2), 2),
    )
    class(data) <- c("mnirs", class(data))

    if (add_metadata) {
        attr(data, "time_channel") <- "time"
        attr(data, "nirs_channels") <- c("nirs1", "nirs2")
        attr(data, "event_channel") <- "event"
        attr(data, "sample_rate") <- sample_rate
    }

    return(data)
}

test_that("shift_mnirs() works with quoted character strings", {
    data <- create_test_data()
    result <- shift_mnirs(
        data,
        nirs_channels = c("nirs1", "nirs2"),
        time_channel = "time",
        to = 0,
        width = 5,
        position = "first",
        verbose = FALSE
    )

    expect_s3_class(result, "mnirs")
    expect_true(all(c("nirs1", "nirs2") %in% names(result)))
    expect_equal(
        result$nirs1[1],
        data$nirs1[1] - mean(c(data$nirs1[1:5], data$nirs2[1:5]))
    )
})

test_that("shift_mnirs() works with bare symbol column names", {
    data <- create_test_data()
    result <- shift_mnirs(
        data,
        nirs_channels = c(nirs1, nirs2),
        time_channel = time,
        to = 0,
        width = 5,
        position = "first",
        verbose = FALSE
    )
    expect_s3_class(result, "mnirs")
    expect_true(all(c("nirs1", "nirs2") %in% names(result)))
    expect_equal(
        result$nirs1[1],
        data$nirs1[1] - mean(c(data$nirs1[1:5], data$nirs2[1:5]))
    )
})

test_that("shift_mnirs() works with external character vector", {
    data <- create_test_data()
    channels <- c("nirs1", "nirs2")
    time_col <- "time"
    result <- shift_mnirs(
        data,
        nirs_channels = channels,
        time_channel = time_col,
        to = 0,
        width = 5,
        position = "first",
        verbose = FALSE
    )
    expect_s3_class(result, "mnirs")
    expect_true(all(c("nirs1", "nirs2") %in% names(result)))
    first_mean <- mean(c(data$nirs1[1:5], data$nirs2[1:5]))
    expect_equal(
        c(result$nirs1[1], result$nirs2[1]),
        c(
            data$nirs1[1] - first_mean,
            data$nirs2[1] - first_mean
        )
    )
})

test_that("shift_mnirs() works with external list object", {
    data <- create_test_data()
    channels <- list(c("nirs1", "nirs2"))
    result <- shift_mnirs(
        data,
        nirs_channels = channels,
        time_channel = "time",
        to = 0,
        width = 5,
        position = "first",
        verbose = FALSE
    )
    expect_s3_class(result, "mnirs")
    first_mean <- mean(c(data$nirs1[1:5], data$nirs2[1:5]))
    expect_equal(
        c(result$nirs1[1], result$nirs2[1]),
        c(
            data$nirs1[1] - first_mean,
            data$nirs2[1] - first_mean
        )
    )
})

test_that("shift_mnirs() works with list() grouping and bare symbols", {
    data <- create_test_data()
    result <- shift_mnirs(
        data,
        nirs_channels = list(c(nirs1, nirs2)),
        time_channel = time,
        to = 0,
        width = 5,
        position = "first",
        verbose = FALSE
    )
    expect_s3_class(result, "mnirs")
    first_mean <- mean(c(data$nirs1[1:5], data$nirs2[1:5]))
    expect_equal(
        c(result$nirs1[1], result$nirs2[1]),
        c(
            data$nirs1[1] - first_mean,
            data$nirs2[1] - first_mean
        )
    )
})

test_that("shift_mnirs() works with list() separate groups", {
    data <- create_test_data()
    result <- shift_mnirs(
        data,
        nirs_channels = list(nirs1, nirs2),
        time_channel = time,
        to = 0,
        width = 5,
        position = "first",
        verbose = FALSE
    )
    expect_s3_class(result, "mnirs")
    expect_equal(
        result$nirs1[1],
        data$nirs1[1] - mean(c(data$nirs1[1:5]))
    )
    expect_equal(
        result$nirs2[1],
        data$nirs2[1] - mean(c(data$nirs2[1:5]))
    )
})

test_that("shift_mnirs() works with tidyselect starts_with()", {
    data <- create_test_data()
    result <- shift_mnirs(
        data,
        nirs_channels = tidyselect::starts_with("nirs"),
        time_channel = time,
        to = 0,
        width = 5,
        position = "first",
        verbose = FALSE
    )
    expect_s3_class(result, "mnirs")
    expect_true(all(c("nirs1", "nirs2") %in% names(result)))
    first_mean <- mean(c(data$nirs1[1:5], data$nirs2[1:5], data$nirs3[1:5]))
    expect_equal(
        c(result$nirs1[1], result$nirs2[1], result$nirs3[1]),
        c(
            data$nirs1[1] - first_mean,
            data$nirs2[1] - first_mean,
            data$nirs3[1] - first_mean
        )
    )
})

test_that("shift_mnirs() works with tidyselect in list()", {
    data <- data.frame(
        time = 1:10,
        smo2_left = runif(10, 50, 70),
        smo2_right = runif(10, 50, 70),
        thb = runif(10, 12, 14)
    )
    data <- create_mnirs_data(
        data,
        list(
            nirs_channels = c("smo2_left", "smo2_right", "thb"),
            time_channel = "time"
        )
    )

    result <- shift_mnirs(
        data,
        nirs_channels = list(tidyselect::starts_with("smo2"), thb),
        time_channel = time,
        to = 0,
        width = 5,
        verbose = FALSE
    )
    expect_s3_class(result, "mnirs")
    expect_true(all(c("smo2_left", "smo2_right", "thb") %in% names(result)))
})

test_that("shift_mnirs() uses metadata when channels NULL", {
    data <- create_test_data()
    result <- shift_mnirs(
        data,
        nirs_channels = NULL,
        time_channel = NULL,
        to = 0,
        width = 5,
        verbose = FALSE
    )
    expect_s3_class(result, "mnirs")
    expect_true(all(c("nirs1", "nirs2") %in% names(result)))
})

test_that("shift_mnirs() errors with non-existent column name", {
    data <- create_test_data()
    expect_error(
        shift_mnirs(
            data,
            nirs_channels = nonexistent,
            time_channel = time,
            to = 0,
            width = 5,
            verbose = FALSE
        ),
        "not detected"
    )
})

test_that("shift_mnirs() mixed quoted and unquoted in list()", {
    data <- create_test_data()
    result <- shift_mnirs(
        data,
        nirs_channels = list("nirs1", nirs2),
        time_channel = "time",
        to = 0,
        width = 5,
        verbose = FALSE
    )
    expect_s3_class(result, "mnirs")
    expect_true(all(c("nirs1", "nirs2") %in% names(result)))
})

test_that("shift_mnirs() preserves grouping with external list", {
    data <- data.frame(
        time = 1:10,
        nirs1 = c(10, 20, 30, 40, 50, 60, 70, 80, 90, 100),
        nirs2 = c(15, 25, 35, 45, 55, 65, 75, 85, 95, 105),
        nirs3 = c(5, 15, 25, 35, 45, 55, 65, 75, 85, 95)
    )
    data <- create_mnirs_data(
        data,
        list(
            nirs_channels = c("nirs1", "nirs2", "nirs3"),
            time_channel = "time"
        )
    )

    channels <- list(c("nirs1", "nirs2"), "nirs3")
    result <- shift_mnirs(
        data,
        channels,
        "time",
        to = 0,
        width = 5,
        position = "first",
        verbose = FALSE
    )

    ## nirs1 and nirs2 grouped:  shifted by same amount
    ## ch3 independent: shifted separately
    expect_equal(
        result$nirs1[1] - result$nirs2[1],
        data$nirs1[1] - data$nirs2[1]
    )
    first_mean <- mean(c(data$nirs1[1:5], data$nirs2[1:5]))
    expect_equal(
        c(result$nirs1[1], result$nirs2[1]),
        c(
            data$nirs1[1] - first_mean,
            data$nirs2[1] - first_mean
        )
    )
    expect_equal(
        result$nirs3[1],
        data$nirs3[1] - mean(c(data$nirs3[1:5]))
    )
})


## integration tests ================================================
test_that("shift_mnirs works on Moxy", {
    data <- read_mnirs(
        file_path = example_mnirs("moxy_ramp.xlsx"),
        nirs_channels = c(smo2_left = "SmO2 Live", smo2_right = "SmO2 Live(2)"),
        time_channel = c(time = "hh:mm:ss"),
        verbose = FALSE
    ) |>
        dplyr::mutate(
            dplyr::across(
                dplyr::matches("smo2"),
                \(.x) {
                    replace_invalid(
                        .x,
                        invalid_values = c(0, 100),
                        method = "none"
                    )
                }
            )
        )

    data_shifted <- shift_mnirs(
        data,
        nirs_channels = c("smo2_left", "smo2_right"),
        time_channel = NULL,
        to = 0,
        by = NULL,
        span = 0,
        position = c("min", "max", "first"),
        verbose = FALSE
    )

    # plot(data) + ggplot2::ylim(0, 100) + geom_hline(yintercept = c(0, 100))
    # plot(data_shifted) + ggplot2::ylim(0, 100) + geom_hline(yintercept = c(0, 100))

    ## check grouping together: min value should come from smo2_right
    expect_false(any(data_shifted$smo2_left == 0, na.rm = TRUE))
    expect_true(any(data_shifted$smo2_right == 0, na.rm = TRUE))
    ## check both shifted together maintaining relative scaling
    expect_equal(
        data_shifted$smo2_left - data_shifted$smo2_right,
        data$smo2_left - data$smo2_right
    )
})

test_that("shift_mnirs(position = 'first') works on Moxy", {
    data <- read_mnirs(
        file_path = example_mnirs("moxy_ramp.xlsx"),
        nirs_channels = c(smo2_left = "SmO2 Live", smo2_right = "SmO2 Live(2)"),
        time_channel = c(time = "hh:mm:ss"),
        zero_time = TRUE,
        verbose = FALSE
    ) |>
        dplyr::mutate(
            dplyr::across(
                dplyr::matches("smo2"),
                \(.x) {
                    replace_invalid(
                        .x,
                        invalid_values = c(0, 100),
                        method = "none"
                    )
                }
            ),
            dplyr::across(
                dplyr::matches("smo2"),
                \(.x) replace_missing(.x, )
            )
        )

    data_shifted <- shift_mnirs(
        data,
        nirs_channels = c("smo2_left", "smo2_right"),
        time_channel = NULL,
        to = 0,
        by = NULL,
        span = 120,
        position = "first",
        verbose = FALSE
    )

    # plot(data) + ggplot2::ylim(0, 100)
    # plot(data_shifted) + ggplot2::geom_ribbon(
    #     data = ~ dplyr::filter(.x, time <= 120),
    #     aes(ymax = 10, ymin = -10), alpha = 0.1, linewidth = NA) +
    #     ggplot2::geom_segment(aes(x = -Inf, xend = 120, y = 0, yend = 0),
    #                           linetype = "dotted", colour = "black")

    first_mean <- data_shifted |>
        dplyr::filter(time <= 120) |>
        dplyr::summarise(
            mean = mean(c(smo2_left, smo2_right), na.rm = TRUE)
        ) |>
        dplyr::pull(mean)

    expect_equal(first_mean, 0)
})

test_that("shift_mnirs works on Train.Red", {
    data <- read_mnirs(
        file_path = example_mnirs("train.red_intervals.csv"),
        nirs_channels = c(
            smo2_left = "SmO2 unfiltered",
            smo2_right = "SmO2 unfiltered",
            o2hb_left = "O2HB unfiltered",
            o2hb_right = "O2HB unfiltered"
        ),
        time_channel = c(time = "Timestamp (seconds passed)"),
        verbose = FALSE
    )

    data_shifted <- shift_mnirs(
        data,
        nirs_channels = list(
            "smo2_left",
            "smo2_right",
            c("o2hb_left", "o2hb_right")
        ),
        time_channel = NULL,
        to = 0,
        by = NULL,
        span = 0, ## should default to 1 sample?
        position = "min",
        verbose = FALSE
    )

    data$time[which(diff(data$time) < 0)]

    # plot(data) + ggplot2::ylim(0, 100)
    # plot(data_shifted) + ggplot2::ylim(0, 100) + geom_hline(yintercept = c(0))

    ## check grouping together: min value should come from each group
    expect_true(any(data_shifted$smo2_left == 0, na.rm = TRUE))
    expect_true(any(data_shifted$smo2_right == 0, na.rm = TRUE))
    expect_true(any(data_shifted$o2hb_left == 0, na.rm = TRUE))
    expect_false(any(data_shifted$o2hb_right == 0, na.rm = TRUE))
    ## check both shifted together maintaining relative scaling
    expect_equal(
        data_shifted$o2hb_left - data_shifted$o2hb_right,
        data$o2hb_left - data$o2hb_right
    )
    ## check both shifted independently
    expect_false(isTRUE(all.equal(
        data_shifted$smo2_left - data_shifted$smo2_right,
        data$smo2_left - data$smo2_right
    )))
})

test_that("resample_mnirs upsamples correctly", {
    data <- data.frame(
        time = c(0, 1, 2),
        value = c(10, 20, 30)
    )

    result <- resample_mnirs(data, "time", 1, 4, verbose = FALSE)

    expect_equal(nrow(result), 9) # 0 to 2 at 0.25 intervals
    expect_equal(result$time, seq(0, 2, by = 0.25))
    expect_equal(result$value[1], 10)
    expect_equal(result$value[5], 20)
    expect_equal(result$value[9], 30)

    result <- resample_mnirs(
        data,
        time_channel = "time",
        sample_rate = 1,
        resample_rate = 4,
        method = "none", ## return only existing samples
        verbose = FALSE
    )
    expect_equal(nrow(result), 9) # 0 to 2 at 0.25 intervals
    expect_equal(result$time, seq(0, 2, by = 0.25))
    expect_equal(result$value[1], 10)
    expect_equal(result$value[5], 20)
    expect_equal(result$value[9], 30)
    expect_true(all(is.na(result$value[c(2:4, 6:8)])))
})

test_that("resample_mnirs downsamples correctly", {
    data <- data.frame(
        time = seq(0, 2, by = 0.1),
        value = seq(10, 30, length.out = 21)
    )

    result <- resample_mnirs(data, "time", 10, 1, verbose = FALSE)

    expect_equal(nrow(result), 3) # 0, 1, 2
    expect_equal(result$time, c(0, 1, 2))
    expect_equal(result$value[1], 10)
    expect_equal(result$value[2], 20)
    expect_equal(result$value[3], 30)

    result <- resample_mnirs(
        data,
        "time",
        10,
        1,
        method = "none",
        verbose = FALSE
    )

    expect_equal(nrow(result), 3) # 0, 1, 2
    expect_equal(result$time, c(0, 1, 2))
    expect_equal(result$value[1], 10)
    expect_equal(result$value[2], 20)
    expect_equal(result$value[3], 30)
})

test_that("resample_mnirs handles resample_rate == sample_rate", {
    data <- data.frame(time = 1:3, value = c(10, 20, 30))
    expect_equal(
        resample_mnirs(data, "time", 1, 1, verbose = FALSE),
        data,
        ignore_attr = TRUE
    )
    expect_equal(
        resample_mnirs(data, "time", 1, verbose = FALSE),
        data,
        ignore_attr = TRUE
    )
    expect_equal(
        resample_mnirs(data, "time", 1, resample_rate = NULL, verbose = FALSE),
        data,
        ignore_attr = TRUE
    )

    ## silent sample_rate and resample_rate
    attr(data, "sample_rate")
    expect_equal(
        resample_mnirs(data, "time", verbose = FALSE),
        data,
        ignore_attr = TRUE
    )

    ## uneven sampling
    set.seed(13)
    data <- data.frame(time = 1:3 + rnorm(3, 0, 0.1), value = c(10, 20, 30))
    result <- resample_mnirs(data, "time", 1, 1, verbose = FALSE)
    expect_true(all.equal(result$value, data$value, tolerance = 1, scale = 1))

    result <- resample_mnirs(data, "time", 1, verbose = FALSE)
    expect_true(all.equal(result$value, data$value, tolerance = 1, scale = 1))

    result <- resample_mnirs(
        data,
        "time",
        1,
        1,
        method = "none",
        verbose = FALSE
    )
    expect_equal(result$value, data$value)

    result <- resample_mnirs(data, "time", 1, method = "none", verbose = FALSE)
    expect_equal(result$value, data$value)
})

test_that("resample_mnirs handles repeated samples", {
    data <- data.frame(
        time = c(1:2, 2, 3:9, 9, 10:17, 17, 18:21) / 10 + 0.1,
        value = seq(10, by = 1, len = 24)
    )

    result <- resample_mnirs(data, "time", 10, 10, "linear", verbose = FALSE)
    expect_equal(range(result$time), floor(range(data$time) * 10) / 10)
    expect_equal(result$value[2], mean(data$value[2:3]))
    expect_equal(result$value[9], mean(data$value[10:11]))
    expect_equal(result$value[17], mean(data$value[19:20]))

    result <- resample_mnirs(
        data, "time", 10, 10, "none", verbose = FALSE
    )
    expect_equal(range(result$time), floor(range(data$time) * 10) / 10)
    expect_equal(result$value[2], data$value[2])
    expect_equal(result$value[9], data$value[10])
    expect_equal(result$value[17], data$value[19])
})

test_that("resample_mnirs handles missing samples", {
    data <- data.frame(
        time = c(1, 3:8, 10:16, 18:21) / 10 + 0.1,
        value = seq(10, by = 1, len = 18)
    )

    result <- resample_mnirs(
        data,
        "time",
        10,
        10,
        method = "linear",
        verbose = FALSE
    )
    expect_equal(range(result$time), floor(range(data$time) * 10) / 10)
    expect_equal(result$value[2], mean(data$value[1:2]))
    expect_equal(result$value[9], mean(data$value[7:8]))
    expect_equal(result$value[17], mean(data$value[14:15]))

    result <- resample_mnirs(
        data,
        time_channel = "time",
        sample_rate = 10,
        method = "none",
        verbose = FALSE
    )
    expect_equal(range(result$time), floor(range(data$time) * 10) / 10)
    expect_true(all(is.na(result$value[c(2, 9, 17)])))
})

test_that("resample_mnirs handles multiple numeric columns", {
    data <- data.frame(
        time = c(0, 1, 2),
        value1 = c(10, 20, 30),
        value2 = c(5, 15, 25)
    )

    result <- resample_mnirs(data, "time", 1, 2, verbose = FALSE)

    expect_equal(ncol(result), 3)
    expect_true(all(c("time", "value1", "value2") %in% names(result)))
    expect_equal(result$value1[1], 10)
    expect_equal(result$value1[3], 20)
    expect_equal(result$value1[5], 30)
    expect_equal(result$value2[1], 5)
    expect_equal(result$value2[3], 15)
    expect_equal(result$value2[5], 25)
})

test_that("resample_mnirs handles categorical columns", {
    data <- data.frame(
        time = c(0, 1, 2),
        value = c(10, 20, 30),
        category = c("A", "B", "C")
    )

    result <- resample_mnirs(data, "time", 1, 4, "linear", verbose = FALSE)
    expect_true("category" %in% names(result))
    expect_equal(result$category[1], "A") # forward fill
    expect_equal(result$category[5], "B") # at t=1
    expect_equal(result$category[9], "C") # at t=2

    result <- resample_mnirs(data, "time", 1, 0.5, "linear", verbose = FALSE)
    expect_true("category" %in% names(result))
    expect_equal(result$category[1], "A") # forward fill
    expect_equal(result$category[2], "C") # at t=2

    result <- resample_mnirs(data, "time", 1, 4, "none", verbose = FALSE)
    expect_true("category" %in% names(result))
    expect_equal(result$category[1], "A") # forward fill
    expect_equal(result$category[5], "B") # at t=1
    expect_equal(result$category[9], "C") # at t=2
    expect_true(all(is.na(result$category[c(2:4, 6:8)])))

    data <- data.frame(
        time = seq(0, 4, 0.5),
        category = NA
    )
    data$category[c(1, 2, 5, 8, 9)] <- c("A", "B", "C", "D", "E")

    result <- resample_mnirs(
        data, "time", 2, 1, "linear", verbose = FALSE
    )
    expect_equal(result$category, c("A", NA, "C", "D", "E"))
})

test_that("non-numeric columns: down-sampling all NA in interval", {
    data <- tibble(
        time = seq(0, 4, by = 0.5),
        value = 1:9,
        category = c("A", NA, NA, NA, "C", NA, NA, NA, "E")
    )

    result <- resample_mnirs(
        data,
        time_channel = "time",
        sample_rate = 2,
        resample_rate = 1,
        method = "linear",
        verbose = FALSE
    )

    expect_equal(
        result$category,
        c("A", NA, "C", NA, "E")
    )
})

test_that("non-numeric columns: method = 'NA' tolerance matching", {
    data <- tibble(
        time = c(0, 0.5, 1.0, 1.5, 2.0),
        value = 1:5,
        category = c("A", "B", "C", "D", "E")
    )

    result <- resample_mnirs(
        data,
        time_channel = "time",
        sample_rate = 2,
        resample_rate = 1,
        method = "none",
        verbose = FALSE
    )

    expect_equal(
        result$category,
        c("A", "C", "E")
    )
})

test_that("non-numeric columns: character and factor types", {
    data <- tibble(
        time = seq(0, 2, by = 0.5),
        value = 1:5,
        char_col = c("A", "B", "C", "D", "E"),
        factor_col = factor(c("low", "high", "low", "high", "low"))
    )

    result <- resample_mnirs(
        data,
        time_channel = "time",
        sample_rate = 2,
        resample_rate = 1,
        method = "linear",
        verbose = FALSE
    )

    expect_equal(result$char_col, c("A", "C", "E"))
    expect_equal(
        as.character(result$factor_col),
        c("low", "low", "low")
    )
    expect_s3_class(result$factor_col, "factor")
})

test_that("resample_mnirs forward-fills when rates exceed actual rate", {
    data <- data.frame(
        time = c(0, 1, 2), ## actual sample_rate == 1
        value = c(10, 20, 30),
        category = c("A", "B", "C")
    )

    result <- resample_mnirs(
        data,
        time_channel = "time",
        sample_rate = 5,
        resample_rate = 2,
        method = "linear",
        verbose = FALSE
    )

    expect_equal(nrow(result), length(seq(0, 2, by = 1 / 2)))
    expect_false(anyNA(result$category))
    expect_equal(result$category, c("A", "A", "B", "B", "C"))
})


test_that("resample_mnirs handles edge cases", {
    ## Intentional that this should return error, even though technically
    ## resampling from one sample to one sample should be ok
    ## It just doesn't make sense for most applications
    # Single row
    data <- data.frame(time = 1, value = 10)
    expect_error(resample_mnirs(data, "time", 1), "valid.*numeric")
})


test_that("resample_mnirs works on Moxy", {
    file_path <- example_mnirs("moxy_ramp.xlsx")

    df <- read_mnirs(
        file_path = file_path,
        nirs_channels = c(smo2 = "SmO2 Live(2)"),
        time_channel = c(time = "hh:mm:ss"),
        verbose = FALSE
    )[1:15, ]

    ## works with metadata
    expect_message(
        result <- resample_mnirs(df, resample_rate = 1),
        "Output is resampled at .*1.*Hz"
    )
    expect_equal(result$time, 0:8)
    expect_s3_class(result, "mnirs")

    ## time-weighted average
    df2 <- df |>
        dplyr::mutate(
            diff = c(diff(time), diff(time)[length(diff(time))]),
            time = round(time * 1) / 1,
        ) |>
        dplyr::summarise(
            .by = time,
            smo2 = stats::weighted.mean(smo2, diff)
        )

    ## expect close enough to time-weighted average
    expect_true(all.equal(
        result, df2, tolerance = 1, scale = 1, check.attributes = FALSE
    ))

    ## should overwrite metadata
    df3 <- resample_mnirs(
        df,
        sample_rate = 2,
        resample_rate = 1.1,
        verbose = FALSE
    )
    expect_equal(attributes(df3)$sample_rate, 1.1)

    ## method = "none"
    result <- resample_mnirs(
        df,
        sample_rate = 2,
        resample_rate = 4,
        method = "none",
        verbose = FALSE
    )

    expect_true(all(result$smo2[!is.na(result$smo2)] == df$smo2))
    ## differences between original and interpolated time samples
    ## should be less than half of the resample_rate = 4 Hz = 0.125 sec
    expect_true(
        all(diff(sort(c(result$time[is.na(result$smo2)], df$time))) > 0.125)
    )
})


test_that("resample_mnirs updates metadata correctly", {
    data <- read_mnirs(
        file_path = example_mnirs("moxy_ramp"),
        nirs_channels = c(smo2_left = "SmO2 Live", smo2_right = "SmO2 Live(2)"),
        time_channel = c(time = "hh:mm:ss"),
        sample_rate = 2,
        verbose = FALSE
    ) |>
        resample_mnirs(
            resample_rate = 1,
            verbose = FALSE
        )
    expect_equal(attr(data, "nirs_channels"), c("smo2_left", "smo2_right"))
    expect_equal(attr(data, "time_channel"), "time")
    expect_equal(attr(data, "sample_rate"), 1)
})


test_that("resample_mnirs works visually on moxy data", {
    skip_if_not_installed("ggplot2")
    skip("visual check of ggplots")
    data <- read_mnirs(
        file_path = example_mnirs("moxy_ramp.xlsx"),
        nirs_channels = c(smo2 = "SmO2 Live"),
        time_channel = c(time = "hh:mm:ss"),
        verbose = FALSE
    )

    (p <- plot(data) +
        ggplot2::scale_colour_manual(
            breaks = c("smo2", "resample"),
            values = palette_mnirs(2)
        ) +
        ggplot2::ylim(0, 100) +
        ggplot2::xlim(0, 100))

    p +
        ggplot2::geom_line(
            data = resample_mnirs(data, method = "none"),
            ggplot2::aes(y = smo2, colour = "resample")
        )

    p +
        ggplot2::geom_line(
            data = resample_mnirs(data, method = "linear"),
            ggplot2::aes(y = smo2, colour = "resample")
        )

    p +
        ggplot2::geom_line(
            data = resample_mnirs(data, method = "locf"),
            ggplot2::aes(y = smo2, colour = "resample")
        )
})

#' Computes rolling local values
#'
#' `compute_local_windows()`: Compute a list of rolling window indices along a
#' time variable `t`.
#'
#' @param idx A numeric vector of indices of `t` at which to calculate local
#'   windows. All indices of `t` by *default*, or can be used to only calculate
#'   for known indices, such as invalid values of `x`.
#' @param width An integer defining the local window in number of samples
#'   around `idx` in which to perform the operation, according to `align`.
#' @param span A numeric value defining the local window time span around `idx`
#'   in which to perform the operation, according to `align`. In units of
#'   `time_channel` or `t`.
#' @param align Window alignment as *"centre"/"center"* (the *default*),
#'   *"left"*, or *"right"*. Where *"left"* is *forward looking*, and *"right"*
#'   is *backward looking* from the current sample.
#' @inheritParams replace_invalid
#'
#' @returns
#' `compute_local_windows()`: A list the same length as `idx` and the same or
#'   shorter length as `t` with numeric vectors of sample indices of length
#'   `width` samples or `span` units of time `t`.
#'
#' @details
#' The local rolling window can be specified by either `width` as the number of
#'   samples, or `span` as the time span in units of `t`. Specifying `width`
#'   is often faster than `span`.
#'
#' `align` defaults to *"centre"* the local window around `idx` between
#'   `[idx - floor((width-1)/2),` `idx + floor(width/2)]` when `width` is
#'   specified. Even `width` values will bias `align` to *"left"*, with the
#'   unequal sample forward of `idx`, effectively returning `NA` at the last
#'   sample index. When `span` is specified, the local window is between
#'   `[t - span/2, t + span/2]`.
#'
#' @rdname compute_helpers
#' @keywords internal
compute_local_windows <- function(
    t,
    idx = seq_along(t),
    width = NULL,
    span = NULL,
    align = c("centre", "left", "right")
) {
    align <- sub("^center$", "centre", align)
    align <- match.arg(align)
    n <- length(t)

    if (!is.null(width)) {
        ## right = backward looking; left = forward looking
        ## centre = left-biased
        offsets <- switch(
            align,
            centre = c(-floor((width - 1L) / 2L), floor(width / 2L)),
            left = c(0L, width - 1L),
            right = c(-(width - 1L), 0L)
        )
        start_idx <- pmax.int(1L, idx + offsets[1L])
        end_idx <- pmin.int(n, idx + offsets[2L])
    } else {
        # fmt: skip
        offsets <- switch(
            align,
            centre = c(-0.5, 0.5),
            left = c(0, 1),
            right = c(-1, 0)
        ) * span
        # fmt: skip
        start_idx <- findInterval(
            t[idx] + offsets[1L], t, left.open = TRUE
        ) + 1L
        end_idx <- findInterval(t[idx] + offsets[2L], t)
    }

    ## inclusive of x[i] for detect outliers
    Map(`:`, start_idx, end_idx)
}


#' @description
#' `compute_local_fun()`: Compute a rolling function along `x` from a list of
#' rolling sample windows.
#'
#' @param window_idx A list the same or shorter length as `x` with numeric
#'   vectors for the sample indices of local rolling windows.
#' @param fn A function to pass through for local rolling calculation.
#' @param ... Additional arguments.
#'
#' @returns
#' `compute_local_fun()`: A numeric vector the same length as `x`.
#'
#' @rdname compute_helpers
#' @keywords internal
compute_local_fun <- function(x, window_idx, fn, ...) {
    vapply(seq_along(window_idx), \(.i) {
        fn(x[window_idx[[.i]]], ...)
    }, numeric(1))
}


#' @description
#' `compute_outliers()`: Computes a vector of local medians and logicals 
#' indicating outliers of `x` within a list of rolling sample windows 
#' `window_idx`.
#'
#' @returns
#' `compute_outliers()`: A `list()` with vectors the same length as `x` for 
#' with numeric local medians and logical identifying where `is_outlier`.
#'
#' @rdname compute_helpers
#' @keywords internal
compute_outliers <- function(
    x,
    window_idx,
    outlier_cutoff
) {
    n <- length(x)
    L <- 1.4826 ## 1 / qnorm(0.75): MAD at the 75% percentile of |Z|
    # MAD = median(|x - median(x)|) within each window
    ## median of absolute local residuals from the local median
    local_stats <- vapply(seq_len(n), \(.i) {
        w <- x[window_idx[[.i]]]
        local_median <- median(w, na.rm = TRUE)
        local_mad <- median(abs(w - local_median), na.rm = TRUE)

        c(local_median, local_mad)
    }, numeric(2))

    local_medians <- local_stats[1L, ]
    local_mad <- local_stats[2L, ]

    ## robust variance threshold based on minimum sample difference
    abs_diffs <- abs(diff(x[!is.na(x)]))
    smallest_var <- suppressWarnings(min(abs_diffs[abs_diffs > 1e-5]))

    ## logical outlier positions
    abs_dev <- abs(x - local_medians)
    is_outlier <- abs_dev > smallest_var &
        abs_dev > (L * outlier_cutoff * local_mad)
    ## NAs from is_outlier check should return FALSE
    is_outlier[is.na(is_outlier)] <- FALSE
    
    ## return list of vectors w/ local logicals and medians
    return(list(
        local_medians = local_medians,
        is_outlier = is_outlier
    ))
}


#' @description
#' `compute_valid_neighbours()`: Compute a list of rolling window indices along
#' `x` to either side of `NA`s.
#'
#' @returns
#' `compute_valid_neighbours()`: A list the same length as the `NA` values in
#'   `x` with numeric vectors of sample indices of length `width` samples or
#'   `span` units of time `t` for valid values neighbouring split to either
#'   side of the invalid `NA`s.
#'
#' @rdname compute_helpers
#' @keywords internal
compute_valid_neighbours <- function(
    x,
    t = seq_along(x),
    width = NULL,
    span = NULL,
    verbose = TRUE
) {
    na_idx <- which(is.na(x))
    valid_idx <- which(!is.na(x))
    n_valid <- length(valid_idx)
    n_na <- length(na_idx)

    if (!is.null(width)) {
        ## find position to the left of each NA in valid_idx sequence
        pos <- findInterval(na_idx, valid_idx)
        half_width <- floor(width / 2L)

        window_idx <- vector("list", n_na)
        for (i in seq_len(n_na)) {
            ## extract width samples before and after
            left <- max(1L, pos[i] - half_width + 1L):pos[i]
            right <- min(n_valid, pos[i] + 1L):min(n_valid, pos[i] + half_width)
            window_idx[[i]] <- valid_idx[sort(unique(c(left, right)))]
        }
        return(window_idx)
    }

    ## pre-compute for span approach
    t_valid <- t[valid_idx]
    t_na <- t[na_idx]
    half_span <- span * 0.5

    ## build per-NA valid neighbours with binary search on sorted `t_valid`
    ## falls back to naerest bracketing pair when no valid samples within `span`
    window_idx <- lapply(seq_len(n_na), \(.i) {
        lo <- findInterval(t_na[.i] - half_span, t_valid, left.open = TRUE) + 1L
        hi <- findInterval(t_na[.i] + half_span, t_valid)
        if (lo > hi) {
            pos <- findInterval(na_idx[.i], valid_idx)
            return(unique(valid_idx[c(pos, min(n_valid, pos + 1L))]))
        }
        valid_idx[lo:hi]
    })

    ## window of valid values exclusive around `x`
    return(window_idx)
}


#' Preserve and restore NA information within a vector
#'
#' `preserve_na()` stores `NA` vector positions and extracts valid non-`NA`
#' values for later restoration with `restore_na()`.
#'
#' @param x A vector containing missing `NA` values.
#'
#' @returns
#' `preserve_na()` returns a list `na_info` with components:
#'   - `na_info$x_valid`: A vector with `NA` values removed.
#'   - `na_info$x_length`: A numeric value of the original input vector length.
#'   - `na_info$na_idx`: A logical vector preserving `NA` positions.
#'
#' `restore_na()` returns a vector `y` the same length as the original
#'   input vector `x` with `NA` values restored to their original positions.
#'
#' @keywords internal
preserve_na <- function(x) {
    na_info <- list(
        x_valid = x[!is.na(x)],
        x_length = length(x),
        na_idx = is.na(x)
    )
    return(na_info)
}


#' Preserve and restore NA information within a vector
#'
#' `restore_na()` restores `NA` values to their original vector positions
#' after processing valid non-`NA` values returned from `preserve_na()`.
#'
#' @param y A vector of valid non-`NA` values returned from `preserve_na()`.
#' @param na_info A list returned from `preserve_na()`.
#'
#' @rdname preserve_na
#' @keywords internal
restore_na <- function(y, na_info) {
    if (all(!na_info$na_idx)) {
        return(y)
    }
    ## fill original length of NAs
    result <- rep(NA, na_info$x_length)
    if (all(na_info$na_idx)) {
        return(result)
    }
    ## replace non-NA with processed output values
    result[!na_info$na_idx] <- y
    return(result)
}

## Test compute_local_windows() =========================================
test_that("compute_local_windows returns correct structure", {
    t <- 1:10
    width <- 3
    result <- compute_local_windows(t, width = width, span = NULL)

    expect_type(result, "list")
    expect_length(result, length(t))
    expect_true(all(vapply(result, is.numeric, logical(1))))

    ## with span
    t <- seq(0, 10, by = 0.5)
    span <- 1
    result <- compute_local_windows(t, width = NULL, span = span)

    expect_type(result, "list")
    expect_length(result, length(t))
})

test_that("compute_local_windows respects width boundaries", {
    t <- 1:10
    width <- 3
    result <- compute_local_windows(t, width = width)

    # First element: can only look forward
    expect_equal(result[[1]], 1:2)
    # Last element: can only look backward
    expect_equal(result[[10]], 9:10)
    # Middle element: symmetrical window
    expect_equal(result[[5]], c(4:6))

    ## even width align centreed
    width <- 2
    result <- compute_local_windows(t, width = width, align = "centre")

    # left-biased forward looking
    expect_equal(result[[1]], 1:2)
    # Last element only at idx
    expect_equal(result[[10]], 10)
    # left-biased forward looking
    expect_equal(result[[5]], c(5:6))

    ## with span
    t <- seq(0, 10, by = 0.5)
    span <- 2
    result <- compute_local_windows(t, span = span)

    # First element: can only look forward
    expect_equal(result[[1]], 1:3)
    expect_equal(t[result[[1]]], t[1:3])
    # Last element: can only look backward
    expect_equal(result[[21]], 19:21)
    expect_equal(t[result[[21]]], t[19:21])
    # Middle element: symmetrical window
    expect_equal(result[[10]], c(8:12))
    expect_equal(t[result[[10]]], t[c(8:12)])
})

test_that("compute_local_windows align = 'center'", {
    t <- 1:10
    width <- 2
    result <- compute_local_windows(t, width = width, align = "center")

    expect_equal(result[[1]], 1:2)
    expect_equal(result[[10]], 10)
    expect_equal(result[[5]], c(5:6))
})

test_that("compute_local_windows width and span >= length(x) works", {
    t <- 1:10
    result <- compute_local_windows(t, width = 20)
    expect_equal(result, rep(list(t), length(t)))

    result <- compute_local_windows(t, span = 20)
    expect_equal(result, rep(list(t), length(t)))
})

test_that("compute_local_windows works with width = 1 and span = 0", {
    t <- 1:10
    result <- compute_local_windows(t, width = 1)
    expect_equal(result, as.list(t))

    result <- compute_local_windows(t, span = 0)
    expect_equal(result, as.list(t))

    compute_local_windows(t, width = 1, span = 0)
})

## test compute_valid_neighbours() ===========================
test_that("compute_valid_neighbours() works with width", {
    x <- c(1, 2, NA, 4, 5, NA, 7)

    # width = 1
    result <- compute_valid_neighbours(x, width = 3)
    expect_length(result, 2) # Two NAs
    expect_equal(result[[1]], c(2, 4))
    expect_equal(result[[2]], c(5, 7))

    # width = 3
    x <- c(1, 2, 3, NA, 5, 6, 7)
    result <- compute_valid_neighbours(x, width = 4)
    expect_equal(result[[1]], c(2, 3, 5, 6))

    # Edge case: NA at start
    x <- c(NA, 20, 30, 40)
    result <- compute_valid_neighbours(x, width = 3)
    expect_equal(result[[1]], 2)

    # Edge case: NA at end
    x <- c(10, 20, 30, NA)
    result <- compute_valid_neighbours(x, width = 3)
    expect_equal(result[[1]], 3)
})

test_that("compute_valid_neighbours() works with span", {
    t <- 0:4
    x <- c(1, 2, NA, 4, 5)

    # span = 1 (includes t[2] = 1 and t[4] = 3)
    result <- compute_valid_neighbours(x, t = t, span = 1)
    expect_length(result, 1)
    expect_equal(result[[1]], c(2, 4))

    # span = 0.5 (no values of t within t ± span should return same as width = 1)
    result <- compute_valid_neighbours(x, t = t, span = 0.5)
    expect_equal(result[[1]], c(2, 4))
    expect_equal(result, compute_valid_neighbours(x, width = 3))

    # Multiple NAs
    x <- c(1, NA, 3, NA, 5)
    result <- compute_valid_neighbours(x, t = t, span = 1)
    expect_length(result, 2)
    expect_equal(result[[1]], c(1, 3))
    expect_equal(result[[2]], c(3, 5))

    # Edge case: NA at start
    x <- c(NA, 2, 3, 4)
    expect_equal(compute_valid_neighbours(x, span = 1)[[1]], 2)
    expect_equal(compute_valid_neighbours(x, span = 0.5)[[1]], 2)

    # Edge case: NA at end
    x <- c(1, 2, 3, NA)
    expect_equal(compute_valid_neighbours(x, span = 1)[[1]], 3)
    expect_equal(compute_valid_neighbours(x, span = 0.5)[[1]], 3)
})

test_that("compute_valid_neighbours width and span >= length(x) works", {
    x <- c(1, 2, NA, 4, 5)
    result <- compute_valid_neighbours(x, width = 8)
    expect_equal(result, list(x[!is.na(x)]))

    result <- compute_valid_neighbours(x, span = 8)
    expect_equal(result, list(x[!is.na(x)]))
})

test_that("compute_valid_neighbours works with width = 1 and span = 0", {
    x <- c(1, 2, NA, 4, 5)
    result <- compute_valid_neighbours(x, width = 1)
    expect_equal(result[[1]], c(2, 4))

    result <- compute_valid_neighbours(x, span = 0)
    expect_equal(result[[1]], c(2, 4))
})

## Test compute_local_fun() =========================================
test_that("compute_local_fun returns correct length", {
    x <- c(1, 2, 3, 4, 5)
    window_idx <- compute_local_windows(x, width = 1)
    result <- compute_local_fun(x, window_idx, fn = median)

    expect_type(result, "double")
    expect_length(result, length(x))
})

test_that("compute_local_fun calculates correct medians", {
    x <- c(10, 20, 30, 40, 50)
    # Manual windows: each looks at neighbours
    window_idx <- compute_local_windows(x, width = 3)
    result <- compute_local_fun(x, window_idx, fn = median)

    expect_equal(result[1], median(x[1:2])) # median of x[2]
    expect_equal(result[2], median(x[1:3])) # median of x[c(1,3)]
    expect_equal(result[5], median(x[4:5])) # median of x[4]
})

test_that("compute_local_fun handles NA values", {
    ## this shouldn't happen with handle_na, but just in case
    x <- c(1, NA, 3, 4, 5)
    window_idx <- compute_local_windows(x, width = 3)
    result <- compute_local_fun(x, window_idx, fn = median, na.rm = TRUE)

    expect_false(all(is.na(result[1:3])))
    expect_equal(result[1], median(x[1]))
    expect_equal(result[2], median(x[c(1, 3)]))
    expect_equal(result[3], median(x[c(3, 4)]))
})

test_that("compute_local_fun() handles empty windows", {
    x <- c(1, 2, 3)
    window_idx <- list(integer(0))
    result <- compute_local_fun(x, window_idx, fn = median, na.rm = TRUE)
    expect_true(is.na(result))
})

test_that("compute_local_fun() handles single values", {
    x <- c(10, 20, 30)
    window_idx <- list(1, 2, 3)
    result <- compute_local_fun(x, window_idx, fn = median)
    expect_equal(result, x)
})


## Test compute_outliers() ==============================================
test_that("compute_outliers returns list of logical and numeric vectors", {
    x <- c(1, 2, 3, 100, 5)
    t <- 1:5
    window_idx <- compute_local_windows(t, width = 3, span = NULL)
    result <- compute_outliers(x, window_idx, outlier_cutoff = 3)

    expect_type(result$local_medians, "double")
    expect_type(result$is_outlier, "logical")
    expect_length(result$local_medians, length(x))
    expect_length(result$is_outlier, length(x))
    expect_true(result$is_outlier[4L]) # 100 should be flagged
    expect_false(any(result$is_outlier[-4L])) # Normal values should not be flagged
})

test_that("compute_outliers threshold sensitivity via outlier_cutoff", {
    x <- c(1, 2, 3, 10, 5)
    t <- 1:5
    window_idx <- compute_local_windows(t, width = 3, span = NULL)

    # Strict threshold
    strict <- compute_outliers(x, window_idx, outlier_cutoff = 1)
    # Lenient threshold
    lenient <- compute_outliers(x, window_idx, outlier_cutoff = 10)

    expect_true(sum(strict$is_outlier) > sum(lenient$is_outlier))
    expect_equal(strict$local_medians, lenient$local_medians)
})

test_that("compute_outliers handles no outliers", {
    x <- 1:5
    t <- 1:5
    window_idx <- compute_local_windows(t, width = 3, span = NULL)
    result <- compute_outliers(x, window_idx, outlier_cutoff = 3)

    expect_true(all(!result$is_outlier))
})

test_that("compute_outliers handles NA", {
    ## this shouldn't happen with handle_na, but just in case
    x <- c(1, 2, NA, 100, 5)
    t <- 1:5
    window_idx <- compute_local_windows(t, width = 3, span = NULL)
    result <- compute_outliers(x, window_idx, outlier_cutoff = 3)

    expect_true(all(!result$is_outlier))
})

library(hexSticker)
library(sysfonts)
library(showtext)

font_families_google()[which(grepl("Merriweather", font_families_google()))]
font_add_google("Merriweather Sans", "merriweather_sans")
font_add_google("Merriweather", "merriweather")

showtext_auto()

hexSticker::sticker(
    ## imported image
    subplot = r"(C:\Users\Jem\Downloads\mnirs-transparent.png)",
    s_x = 1, # centred horizontally
    s_y = 0.95, # slightly below centre to leave room for text
    s_width = 1.0, # fill most of the hex width
    s_height = 0.7, # constrain height proportionally

    ## package name parameters
    package = "mnirs",
    p_x = 1,
    p_y = 1.4,
    p_size = 12,
    p_color = "#004992",
    p_family = "merriweather_sans",
    p_fontface = 2, # "plain", "bold", "italic", "bolditalic"

    ## hex geometry
    h_fill = "#bcced7",    ## light blue-grey background
    # h_color = "#fb653a", ## orange of muscle icon
    # h_color = "#1a284f", ## dark navy
    h_color = "#004992",   ## blue
    h_size = 1.5,
    white_around_sticker = FALSE,

    ## spotlight / white lines echo — leave off unless you want glow
    # spotlight = TRUE,
    # l_x = 1,
    # l_y = 0.35,

    ## ggsave
    ## `*mnirs-hex*.png` will be gitignored
    filename = "man/figures/mnirs-hex.svg",
    dpi = 300
)

#' Extract intervals from *{mnirs}* data
#'
#' Extract intervals from *"mnirs"* time series data, specifying interval
#' start and end boundaries by time value, event label, lap number, or sample 
#' index.
#'
#' @param nirs_channels A character vector or a `list()` of character vectors
#'   of mNIRS channel names to operate on within each interval (see *Details*).
#'   Names must match column names in `data` exactly.
#'   - Must only be specified when `event_groups` contains *"ensemble"*-
#'     averaged intervals. If `event_groups = "distinct"` no channel processing
#'     occurs.
#'   - If `NULL` (default), channels are retrieved from *"mnirs"* metadata.
#'
#' @param event_channel An *optional* character string giving the name of an
#'   event/lap column. The column may contain character event labels or integer 
#'   lap numbers. 
#'   - Required when using [by_label()] or [by_lap()] for `start` or `end`.
#'   - Retrieved from metadata if not defined explicitly.
#'
#' @param sample_rate An *optional* numeric sample rate (Hz) used to bin time
#'   values for ensemble-averaging. If `NULL`, will be estimated from
#'   `time_channel` (see *Details*).
#'
#' @param start Specifies where intervals begin. Either raw values -- numeric
#'   for time values, character for event labels, explicit integer (e.g. `2L`)
#'   for lap numbers -- or created with [by_time()], [by_label()], [by_lap()],
#'   or [by_sample()].
#'
#' @param end Specifies where intervals end. Either raw values -- numeric for
#'   time values, character for event labels, explicit integer (e.g. `2L`)
#'   for lap numbers -- or created with [by_time()], [by_label()], [by_lap()],
#'   or [by_sample()].
#'
#' @param span A one- or two-element numeric vector `c(before, after)` in units 
#'   of `time_channel`, or a `list()` of such vectors. (*default* 
#'   `span = c(-60, 60)`. Applied additively to interval boundaries:
#'   - When both `start` and `end` are specified: `span[1]` shifts start times,
#'     `span[2]` shifts end times.
#'   - When only `start` or only `end` is specified: both `span[1]` and
#'     `span[2]` apply as a window around the event).
#'   - A single *positive* value is recycled to shift the end times (e.g. 
#'     `span = 60` -> `c(0, 60)`).
#'   - A single *negative* value is recycled to shift the start times (e.g. 
#'     `span = -60` -> `c(-60, 0)`).
#'
#' @param event_groups Either a character string or a `list()` of integer
#'   vectors specifying how to group intervals (see *Details*).
#'   \describe{
#'     \item{`"distinct"`}{The default. Extract each interval as an independent
#'     data frame.}
#'     \item{`"ensemble"`}{Ensemble-average each specified `nirs_channel` across
#'     all detected intervals, returning a single data frame.}
#'     \item{`list(c(1, 2), c(3, 4))`}{Ensemble-average each specified
#'     `nirs_channel` within each group and return one data frame per group.}
#'   }
#'
#' @param zero_time Logical. Default is `FALSE`. If `TRUE`, re-calculates
#'   numeric `time_channel` values to start from zero within each interval
#'   data frame.
#'
#' @inheritParams validate_mnirs
#'
#' @details
#' ## Interval specification
#'
#' Interval `start` and `end` boundaries are specified using helper functions, 
#' or by passing raw values directly:
#'
#' \describe{
#'   \item{[by_time()]}{Time values in units of `time_channel`.}
#'   \item{[by_label()]}{Strings to match in `event_channel`.
#'   All matching occurrences are returned.}
#'   \item{[by_lap()]}{Lap numbers to match in `event_channel`. Resolves to the
#'   first sample of each lap for `start`, and the last lap sample for `end`}
#'   \item{[by_sample()]}{Integer sample indices (row numbers).}
#' }
#'
#' Raw values supplied to `start`/`end` are auto-coerced: 
#'   - Numeric -> [by_time()]
#'   - Character -> [by_label()],
#'   - Explicit integer (e.g. `2L`) -> [by_lap()]. 
#'   - Use [by_sample()] explicitly for sample indices.
#'
#' `start` and `end` can use different specification types (e.g., start by
#' label, end by time). When lengths differ, the shorter is recycled.
#'
#' ## Time span window
#'
#' `span` additively expands the time span window around interval boundaries. 
#' 
#' - A two-value vector expands the `start` and `end`, respectively:
#'   `span = c(-60, 60)` expands the `start` earlier by `60`, and the `end` 
#'   later by `60`. For example, 
#'   `start = by_time(30), end = by_time(60), span = c(-5, 10)` returns an
#'   interval of `[25, 70]`.
#' - A single numeric value is recycled according to the sign: `span = -60` 
#'   becomes `c(-60, 0)` to expand the `start` earlier. `span = 60` becomes 
#'   `c(0, 60)` to expand the `end` later.
#' - If only `start` is specified alone, both span values expand the single
#'   boundary window: `start = by_time(30), span = c(-5, 60)` returns 
#'   `[25, 90]`.
#'
#' ## Per-interval nirs_channels for ensemble-averaging
#'
#' When `event_groups = "ensemble"` or a list of numeric grouped intervals,
#' `nirs_channels` can be specified as a list of column names to override
#' ensemble-averaging across interval. For example, to exclude a channel
#' in one interval:
#'
#' ```r
#' nirs_channels = list(
#'   c(A, B, C),
#'   c(A, C) ## channel "B" is excluded
#' )
#' ```
#'
#' If all grouped intervals can include all `nirs_channels`, or if
#' `event_groups = "distinct"`, a single `nirs_channels` character vector can
#' be supplied and recycled to all groups, or left as `NULL` for channels to
#' be taken from *"mnirs"* metadata.
#'
#' ## Grouping intervals
#'
#' `event_groups` controls whether extracted intervals are returned as distinct
#' data frames or ensemble-averaged.
#'
#' \describe{
#'    \item{`"distinct"`}{The default. Extract each interval and return a
#'    list of independent data frames.}
#'    \item{`"ensemble"`}{Ensemble-average each specified `nirs_channel` across
#'    all detected intervals and return a one-item list with a single data
#'    frame.}
#'    \item{`list(c(1, 2), c(3, 4))`}{Ensemble-average each specified
#'    `nirs_channel` within each group and return a list with one data frame
#'    for each group. Any intervals detected but not specified in
#'    `event_groups` are returned as distinct.}
#' }
#'
#' `event_groups` lists can be named (e.g.
#' `list(low = c(1, 2), high = c(3, 4))`) and will pass those names to the
#' returned list of data frames.
#'
#' When `event_groups` is a list of numeric interval numbers, list items in
#' `nirs_channels` and `span` are recycled to the number of groups. If lists
#' are only partially specified, the final item is recycled forward as needed.
#' Extra items are ignored.
#'
#' @returns A named `list()` of [tibbles][tibble::tibble-package] of class
#'   *"mnirs"*, each with metadata available via `attributes()`.
#'
#' @examples
#' ## read example data
#' data <- read_mnirs(
#'     example_mnirs("train.red"),
#'     nirs_channels = c(
#'         smo2_left = "SmO2 unfiltered",
#'         smo2_right = "SmO2 unfiltered"
#'     ),
#'     time_channel = c(time = "Timestamp (seconds passed)"),
#'     zero_time = TRUE,
#'     verbose = FALSE
#' ) |>
#'     ## avoid issues ensemble-averaging irregular samples
#'     resample_mnirs(method = "linear", verbose = FALSE) 
#'
#' ## ensemble-average across multiple intervals
#' interval_list <- extract_intervals(
#'     data,                       ## channels recycled to all intervals by default
#'     nirs_channels = c(smo2_left, smo2_right),
#'     start = by_time(368, 1084), ## manually identified interval start times
#'     span = c(-20, 90),          ## include the last 180-sec of each interval (recycled)
#'     event_groups = "ensemble",  ## ensemble-average across two intervals
#'     zero_time = TRUE            ## re-calculate common time to start from `0`
#' )
#'
#' interval_list[[1L]]
#'
#' \donttest{
#'   if (requireNamespace("ggplot2", quietly = TRUE)) {
#'     plot(interval_list[[1L]], time_labels = TRUE) +
#'       ggplot2::geom_vline(xintercept = 0, linetype = "dotted")
#'   }
#' }
#'
#' @export
extract_intervals <- function(
    data,
    nirs_channels = NULL,
    time_channel = NULL,
    event_channel = NULL,
    sample_rate = NULL,
    start = NULL,
    end = NULL,
    span = list(c(-60, 60)),
    event_groups = c("distinct", "ensemble"),
    zero_time = FALSE,
    verbose = TRUE
) {
    ## validation ==============================================
    validate_mnirs_data(data)
    metadata <- attributes(data)
    if (missing(verbose)) {
        verbose <- getOption("mnirs.verbose", default = TRUE)
    }
    nirs_channels <- validate_nirs_channels(
        enquo(nirs_channels), data, verbose, as_list = TRUE
    )
    time_channel <- validate_time_channel(enquo(time_channel), data)
    ## avoid floating point precision issues downstream with findIntervals()
    time_vec <- round(data[[time_channel]], 6)
    ## estimate sample_rate for appropriate time binning
    sample_rate <- validate_sample_rate(
        data, time_channel, sample_rate, verbose
    )

    ## coerce raw values to mnirs_interval objects
    start <- as_mnirs_interval(start, "start")
    end <- as_mnirs_interval(end, "end")

    ## validate interval specs
    if (is.null(c(start, end))) {
        cli_abort(c(
            "x" = "No interval specification provided.",
            "i" = "Specify {.arg start} and/or {.arg end} using \\
            {.fn by_time}, {.fn by_sample}, {.fn by_label}, or {.fn by_lap}."
        ))
    }

    ## resolve event_channel if by_label or by_lap is used
    uses_event_channel <- (
        inherits(start, "mnirs_interval") && start$type %in% c("label", "lap")
    ) || (
        inherits(end, "mnirs_interval") && end$type %in% c("label", "lap")
    )
    
    if (uses_event_channel) {
        event_channel <- tryCatch(
            validate_event_channel(event_channel, data, required = TRUE),
            error = function(e) {
                cli_abort(c(
                    "x" = "{.arg event_channel} is required when using \\
                    {.fn by_label} or {.fn by_lap}.",
                    "i" = "Specify column name containing {.cls character} \\
                    event labels or {.cls integer} lap numbers."
                ))
            }
        )
        event_vec <- data[[event_channel]]
    } else {
        event_channel <- validate_event_channel(
            event_channel, data, required = FALSE
        )
        event_vec <- NULL
    }

    if (
        verbose && event_groups[1L] != "distinct" &&
            is.null(attr(data, "nirs_channels")) && !is.list(nirs_channels)
    ) {
        cli_inform(c(
            "!" = "{.fn extract_intervals} accepts {.arg nirs_channels} = \\
            {col_blue('list()')} for channel grouping when \\
            ensemble-averaging. See `?extract_intervals`."
        ))
    }

    ## expand parameters ====================================
    event_groups <- make_list(event_groups)
    span <- make_list(span)
    span <- lapply(span, recycle_span)

    ## resolve interval boundary times ==========================
    interval_list <- resolve_interval(start, end, time_vec, event_vec)
    n_events <- length(interval_list$start_time)

    ## recycle params to match number of intervals
    nirs_channels <- recycle_param(
        nirs_channels,
        n_events,
        event_groups,
        verbose
    )
    span <- recycle_param(span, n_events, event_groups, verbose)

    ## apply span and build interval spec ======================
    interval_spec <- apply_span(interval_list, time_vec, span, verbose)

    ## extract interval data ===================================
    df_list <- extract_df_list(data, time_vec, interval_spec, nirs_channels)

    ## apply grouping logic ====================================
    result <- group_intervals(
        df_list,
        nirs_channels,
        metadata,
        event_groups,
        zero_time,
        verbose
    )

    ## add class "mnirs" ========================================
    class(result) <- c("mnirs", class(result))

    return(result)
}

---
title: "Reading and Cleaning Data with {mnirs}"
vignette: >
    %\VignetteIndexEntry{Reading and Cleaning Data with {mnirs}}
    %\VignetteEngine{quarto::html}
    %\VignetteEncoding{UTF-8}
format:
    html:
        embed-resources: true
        toc: true
knitr:
    opts_chunk:
        collapse: true
        comment: '#>'
        dpi: 300
        out.width: "100%"
        fig.align: "centre"
        eval: requireNamespace("ggplot2", quietly = TRUE)
---

# Introduction

Modern wearable muscle near-infrared spectroscopy (mNIRS) devices make it easier than ever to monitor local muscle oxygenation during dynamic activities.

The real challenge comes with deciding how to clean, filter, process, and eventually interpret those data.

The `{mnirs}` package aims to provide standardised, reproducible methods for reading, processing, and analysing mNIRS data. The goal is to help practitioners detect meaningful signal from noise, and improve our confidence in interpreting and applying information to the clients we work with.

In this vignette we will demonstrate how to:

* 📂 Read data files exported from commercial wearable NIRS devices, and import NIRS channels into a standard data frame format with metadata, ready for further processing.

* 📊 Plot and visualise data frames of class *`"mnirs"`*.

* 🔍 Retrieve metadata stored with data frames of class *`"mnirs"`* to avoid repetitively specifying which channels to process.

* ⏱️ Resample data to a higher or lower sample rate, to correct irregular sampling periods, or to match the frequency of other data sources for synchronisation.

* 🧹 Detect and replace local outliers, invalid values, and interpolate across missing data.

* 📈️ Apply digital filtering to optimise signal-to-noise ratio for the responses observed in our data.

* ⚖️ Shift and rescale across multiple NIRS channels, to normalise signal dynamic range while preserving absolute or relative scaling between muscle sites and NIRS channels.

* 🔀 Demonstrate a complete data wrangling process using pipe-friendly functions.

* 🧮 Detect and extract intervals for further analysis.

*`mnirs`* is designed to process NIRS data, but it can be used to read, clean, and process other time series data which require many of the same processing steps. Enjoy!


# 📂 Read data from file

We will read an example data file with two NIRS channels from an incremental ramp cycling assessment recorded with *Moxy* muscle oxygen monitor.

First, install and load the *`mnirs`* package and other required libraries.

*`mnirs`* can be installed with `install.packages("mnirs")`.

```{r}
#| label: setup
#| warning: false

# pak::pak("jemarnold/mnirs") ## install development version
library(ggplot2)   ## for plotting
library(mnirs)
```

The first function called will often be `read_mnirs()`. This is used to read data from *.csv* or *.xls(x)* files exported from common wearable NIRS devices.

Exported data files will often have multiple rows of file header metadata before the data table with NIRS recordings begins. `read_mnirs()` can extract and return this data table along with the file metadata for further processing and analysis.

See `read_mnirs()` for more details.


## `read_mnirs()`

* `file_path`

    Specify the location of the data file, including file extension. e.g. `"./my_file.xlsx"` or `"C:/myfolder/my_file.csv"`.

::: {.callout-note}
### Example data files

A few example data files are included in the *`mnirs`* package. File paths can be accessed with `example_mnirs()`.
:::

* `nirs_channels`
    
    Multiple *NIRS* channels can be specified from the data file as a vector of names. If no `nirs_channels` are specified, `read_mnirs()` will attempt to recognise the NIRS device file format, and return the full data frame with all detected columns. This can be useful for file exploration, to find the target channel names. However, best practice is to specify the desired `nirs_channels` explicitly.

* `time_channel`

    A *time* or *sample* channel name from the data table can be specified. If left blank, the function will attempt to identify the time column automatically, however, best practice is to specify the `time_channel` explicitly.

* `event_channel`

    Optionally, a channel can be specified which indicates character *event* labels or integer *lap* values in the data table.\
\
These channel names are used to detect the data table within the file, and must match exactly with text strings in the file on the same row. We can rename these channels when reading data by specifying a named character vector: 

```{r}
nirs_channels = c(
    renamed1 = "original_name1", 
    renamed2 = "original_name2"
)
```

* `sample_rate`

    The sample rate (in Hz) of the exported data can either be specified explicitly, or it will be estimated from `time_channel`.\
\
Automatic detection usually works well unless there is irregular sampling, or the `time_channel` is a count of samples rather than a time value. For example, *Oxysoft* exports a column of sample numbers rather than time values. For *Oxysoft* files specifically, the function will recognise and read the correct sample rate from the file metadata. However, in most cases `sample_rate` should be defined explicitly if known.

* `add_timestamp`

    `FALSE` by default; if `time_channel` is in date-time (POSIXct) format (e.g., *hh:mm:ss*), by default it will be converted to numeric time in seconds.\
\
If `add_timestamp = TRUE`, the date-time start time value in `time_channel` or in the file metadata will be extracted and a `"timestamp"` column will be added to the returned data frame. This can be useful for synchronising devices based on system time.

* `zero_time`

    `FALSE` by default; if `time_channel` values start at a non-zero value, `zero_time = TRUE` will re-calculate time starting from zero. If `time_channel` was converted from date-time format, it will always be re-calculated from zero regardless of this option.

* `keep_all`

    `FALSE` by default; only the channels explicitly specified will be returned in a data frame. `keep_all = TRUE` will return all columns detected from the data table in the file.\
\
Blank/empty columns will be omitted. Duplicate column names will be repaired by appending a suffix `"_n"`, and empty column names will be renamed as `col_n`; where `n` is equal to the column number in the data file. Renamed columns should be checked to confirm correct naming if duplicates are present.

* `verbose`

    `TRUE` by default; this and most *`mnirs`* functions will return warnings and informational messages which are useful for troubleshooting and data validation. This option can be used to silence those messages. *`mnirs`* messages can be silenced globally for a session by setting `options(mnirs.verbose = FALSE)`.

```{r}
## {mnirs} includes sample files from a few NIRS devices
example_mnirs()

## partial matching will error if matches multiple
try(example_mnirs("moxy"))

data_raw <- read_mnirs(
    file_path = example_mnirs("moxy_ramp"), ## call an example data file
    nirs_channels = c(
        smo2_left = "SmO2 Live",    ## identify and rename channels
        smo2_right = "SmO2 Live(2)"
    ),
    time_channel = c(time = "hh:mm:ss"), ## date-time format will be converted to numeric
    event_channel = NULL,           ## leave blank if unused
    sample_rate = NULL,             ## if blank, will be estimated from time_channel
    add_timestamp = FALSE,          ## omit a date-time timestamp column
    zero_time = TRUE,               ## recalculate time values from zero
    keep_all = FALSE,               ## return only the specified data channels
    verbose = TRUE                  ## show warnings & messages
)

## Note the above info message that sample_rate was estimated correctly at 2 Hz ☝
## ignore the warnings about irregular sampling for now, we will resample later

data_raw
```


# 📊 Plot *`mnirs`* data

*`mnirs`* data can be easily viewed by calling `plot()` (explicitly documented as `plot.mnirs()`). This generic plot function uses *`ggplot2`* and will work on data frames generated or read by *`mnirs`* functions where the metadata contains `class = *"mnirs"*`.

## `plot.mnirs`

* `data`

    This function takes in a data frame of class *`mnirs`* and returns a formatted *`ggplot2`* plot.

* `points`

    `FALSE` by default; a quick way to plot individual points for more precise sample visualisation.

* `time_labels`

    `FALSE` by default; time values on the x-axis will be plotted as numeric by default. `time_labels = TRUE` will instead plot time values as `h:mm:ss` format.

* `na.omit`

    `FALSE` by default; missing data (`NA`s) will be plotted as gaps in the time-series data. If `na.omit = TRUE`, `NA`s will be omitted from the plotted data, effectively plotting across these gaps, making them less visible, but making the general data trend easier to see if there are lots of missing values.

```{r}
## note the `time_labels` plot argument to display time values as `h:mm:ss`
plot(
    data_raw,
    points = FALSE,
    time_labels = TRUE,
    na.omit = FALSE
)
```

*`mnirs`* includes a custom *`ggplot2`* theme and colour palette available with `theme_mnirs()` and `palette_mnirs()`. See those documentation references for more details.


# 🔍 Metadata stored in *`mnirs`* data frames

Data frames generated or read by *`mnirs`* functions will return `class = *"mnirs"*` and contain metadata, which can be retrieved with `attributes()`. 

Instead of re-defining values like our channel names or sample rate, certain *`mnirs`* functions can automatically retrieve them from metadata. They can always be overwritten manually in subsequent functions, or by using a helper function `create_mnirs_data()`.

See `create_mnirs_data()` for more details about metadata.

```{r}
## view metadata
attr(data_raw, "nirs_channels")

attr(data_raw, "time_channel")

attr(data_raw, "sample_rate")
```


# ⏱️ Resample data

mNIRS devices may have sample rates anywhere from 0.5 to 100+ Hz. Higher sample rates gives greater precision. However, issues can arise with irregular sampling where time samples occur at inconsistent frequencies, are repeated, or are lost. Resampling to a consistent sample rate can be used to regularise timestamps to the nearest time value according to the known data sample rate. This can make subsequent processing steps easier, as well as synchronisation with other devices.

Downsampling can be useful to reduce data size and processing requirement, although should not be used primarily as a data filtering/smoothing strategy (see **Digital filtering** section below). Upsampling can be done with interpolation over newly created samples, although caution should always be taken with creating new samples without any additional information. New samples created with resampling will be left blank (`NA`) by default.


## `resample_mnirs()`

* `data`

    This function takes in a data frame, applies processing to all channels specified explicitly or implicitly from *`mnirs`* metadata, then returns the processed data frame. *`mnirs`* metadata will be passed to and from this function.\
\
*`mnirs`* functions are also pipe-friendly for Base R 4.1+ (`|>`) or `{magrittr}` (`%>%`) pipes to chain operations together (see below).

* `time_channel`

    The column name of time series values in `data`. If the data contain *`mnirs`* metadata, this channel will be detected automatically, or it can be specified explicitly.\
\
This function processes all columns in `data`, so `nirs_channels` does not need to be specified.

* `sample_rate`

    The sample rate (in Hz) of `data`. If the data contain *`mnirs`* metadata, this channel will be detected automatically, or it can be specified explicitly.

* `resample_rate`

    Resampling is specified as the desired number of samples per second (Hz). The default `resample_rate` will resample back to the existing `sample_rate` of the data. This can be useful to accommodate irregular sampling with unequal time values. Linear interpolation is used to resample `time_channel` to round values of the `sample_rate`.

* `method`

    Any new samples created by resampling (i.e. samples fill in between gaps, or up-sampled indices) can be filled in or left blank as `NA`. The default `method` is to leave new samples blank, and interpolation must be explicitly specified as either *"locf"* to repeat the last observation carried forward, or *"linear"* interpolation.

```{r}
data_resampled <- resample_mnirs(
    data_raw,            ## blank channels will be retrieved from metadata
    time_channel = time, ## channels can be left blank or specified explicitly
    sample_rate = NULL,  ## blank by default will be retrieved from metadata
    resample_rate = 2,   ## blank by default will resample to `sample_rate`
    method = "linear"    ## linear interpolation across resampled indices
)

## note the altered "time" values from the original data frame 👇
data_resampled
```


::: {.callout-note}
## Use `resample_mnirs()` to smooth over irregular or skipped samples

If we see a warning from `read_mnirs()` about duplicated or irregular samples like we saw above, we can use `resample_mnirs()` to restore `time_channel` to a regular sample rate and interpolate across skipped samples.

Note: if we perform both of these steps in a piped function call, we will still see the warning appear about irregular sampling at the end of the pipe. This warning is returned by `read_mnirs()`. But viewing the data frame can confirm that `resample_mnirs()` has resolved the issue.
:::


# 🧹 Replace local outliers, invalid values, and missing values

We can see some data issues in the plot above, so let's clean those with a single wrangling function `replace_mnirs()`, to prepare our data for digital filtering and smoothing.

*`mnirs`* tries to include basic functions which work on vector data, and convenience wrappers which combine functionality and can be used on multiple channels in a data frame at once.

Let's explain the functionality of this data-wide function, and for more details about the vector-specific functions see `replace_mnirs()`.


## `replace_mnirs()`

* `data`

    This function takes in a data frame, resamples all data frame columns to the new sample rate, then returns the processed data frame. *`mnirs`* metadata will be passed to and from this function.

* `nirs_channels`

    Specify which column names in `data` to process, i.e. the response variables. If the data contain *`mnirs`* metadata, these channels will be detected automatically. Channels not explicitly specified will be passed through unprocessed to the returned data frame.

* `time_channel` & `sample_rate`

    If the data contain *`mnirs`* metadata, these will be detected automatically, or they can be specified explicitly.

* `invalid_values`, `invalid_above`, or `invalid_below`

    Specific invalid values can be replaced, e.g. if a NIRS device exports `0`, `100`, or some other fixed value when signal recording is in error. If spikes or drops are present in the data, these can be replaced by specifying values above or below which to consider invalid. If left as `NULL`, no values will be replaced.

* `outlier_cutoff`

    Local outliers can be detected using a cutoff calculated from the local median value. A default value of `3` is recommended and corresponds to Pearson's rule (i.e., ± 3 SD about the local median). If left as `NULL`, no outliers will be replaced.

* `width` or `span`

    Local outlier detection and median interpolation use a rolling local window specified by one of either `width` or `span`. `width` defines a number of samples centred around the local index being evaluated (`idx`), whereas `span` defines a range of time in units of `time_channel`.

* `method`

    Missing data (`NA`), invalid values, and local outliers specified above can be replaced via interpolation or fill methods; either `"linear"` interpolation (the default), fill with local `"median"`, or `"locf"` (*"last observation carried forward"*).\
\
`NA`s can be passed through to the returned data frame with `method = "none"`. However, subsequent processing & analysis steps may return errors when `NA`s are present. Therefore, it is good practice to deal with missing data early during data processing.

* `verbose`

    As above, an option to toggle warnings and info messages.

```{r}
data_cleaned <- replace_mnirs(
    data_resampled,     ## blank channels will be retrieved from metadata
    invalid_values = 0, ## known invalid values in the data
    invalid_above = 90, ## remove data spikes above 90
    outlier_cutoff = 3, ## recommended default value
    width = 7,          ## window to detect and replace outliers/missing values
    method = "linear"   ## linear interpolation over `NA`s
)

plot(data_cleaned, time_labels = TRUE)
```

That cleaned up all the obvious data issues.


# 📈 Digital filtering

To improve the signal-to-noise ratio in our dataset without losing information, we should apply digital filtering to smooth the data.



## Choosing a digital filter

There are a few digital filtering methods available in *`mnirs`*. Which option is best for *you* will depend in large part on the sample rate of your data and the frequency of the response signal or phenomenon you are interested in observing.

Choosing filter parameters is an important processing step to improve signal-to-noise ratio and enhance our subsequent interpretations. Over-filtering the data can introduce data artefacts which can negatively influence signal analysis and interpretations, just as much as trying to analyse overly-noisy raw data.

It is perfectly valid to choose a digital filter by iteratively testing filter parameters until the signal or response of interest appears to be visually optimised with minimal data artefacts, to your satisfaction.

We will discuss the process of choosing a digital filter in more depth in another article coming soon.


## `filter_mnirs()`

* `data`

    This function takes in a data frame, applies processing to all channels specified, then returns the processed data frame. *`mnirs`* metadata will be passed to and from this function.

* `nirs_channels`, `time_channel`, & `sample_rate`

    If the data contain *`mnirs`* metadata, these will be detected automatically, or they can be specified explicitly.

* `na.rm`

    `FALSE` by default; `filter_mnirs()` will return an error if any missing data (`NA`) are detected in the response variables (`nirs_channels`). Setting `na.rm = TRUE` will ignore these `NA`s and pass them through to the returned data frame, but this must be opted into explicitly in this function and elsewhere.


### Smoothing-spline

* `method = "smooth_spline"`

    The default non-parametric cubic smoothing spline is often a good first filtering option when exploring the data, and works well over longer time spans. For more rapid and repeated responses, a smoothing-spline may not work as well.

* `spar`

    The smoothing parameter of the cubic spline will be determined automatically by default, or can be specified explicitly. See `stats::smooth.spline()`.


### Butterworth digital filter

* `method = "butterworth"`

    A Butterworth low-pass digital filter (specified by `type = "low"`) is probably the most common method used in mNIRS research (whether appropriately or not). For certain applications such as identifying a signal with a known frequency (e.g. cycling/running cadence or heart rate), a pass-band or a different filter type may be better suited.

* `order`

    The filter order number, specifying the number of passes the filter performs over the data. The default `order = 2` will often noticeably improve the filter over a single pass, however, higher orders above ~`4` can begin to introduce artefacts, particularly at sharp transition points.

* `W` or `fc`

    The cutoff frequency can be specified either as `W`; a fraction (between `[0, 1]`) of the [*Nyquist frequency*](https://www.youtube.com/watch?v=IZJQXlbm2dU), which is equal to half of the `sample_rate` of the data. Or as `fc`; the cutoff frequency in Hz, where this absolute frequency should still be between 0 Hz and the Nyquist frequency.
    
* `type`

    The filter type is specified as either `c("low", "high", "stop", "pass")`.


For filtering vector data and more details about Butterworth filter parameters, see `filter_butter()`.


### Moving average

* `method = "moving_average"`

    The simplest smoothing method is a moving average filter applied over a specified number of samples or time span. Commonly, this might be a 5- or 15-second centred moving average.

* `width` or `span`

    Moving average filtering occurs within a rolling local window specified by one of either `width` or `span`. `width` defines a number of samples centred around the local index being evaluated (`idx`), whereas `span` defines a range of time in units of `time_channel`.

For filtering vector data and more details about the moving average filter, see `filter_ma()`.


## Apply the filter

Let's try a *Butterworth low-pass* filter, and we'll specify some empirically chosen filter parameters for these data. See `filter_mnirs()` for further details on each of these filtering methods and their respective parameters.


```{r}
data_filtered <- filter_mnirs(
    data_cleaned,           ## blank channels will be retrieved from metadata
    method = "butterworth", ## Butterworth digital filter is a common choice
    order = 2,              ## filter order number
    W = 0.02,               ## filter fractional critical frequency `[0, 1]`
    type = "low",           ## specify a "low-pass" filter
    na.rm = TRUE            ## explicitly ignore NAs
)

## we will add the non-filtered data back to the plot to compare
plot(data_filtered, time_labels = TRUE) +
    geom_line(
        data = data_cleaned, 
        aes(y = smo2_left, colour = "smo2_left"), alpha = 0.4
    ) +
    geom_line(
        data = data_cleaned, 
        aes(y = smo2_right, colour = "smo2_right"), alpha = 0.4
    )
```


That did a nice job reducing the high-frequency noise in our data, while preserving the sharp edges, such as the rapid reoxygenation after maximal exercise.

# ⚖️ Shift and rescale data

NIRS values are not measured on an absolute scale (arguably not even percent (%) saturation/SmO<sub>2</sub>). Therefore, we may need to adjust or calibrate our data to normalise NIRS signal values between muscle sites, individuals, trials, etc. depending on our intended comparison.

For example, we may want to set our mean baseline value to zero for all NIRS signals at the start of a recording. Or we may want to compare signal kinetics (the rate of change or time course of a response) after rescaling signal amplitudes to a common dynamic range.

These functions allow us to either shift NIRS values up or down while preserving the dynamic range (the absolute amplitude from minimum to maximum values) of our NIRS channels, or rescale the data to a new dynamic range with larger or smaller amplitude.

We can group NIRS channels together to preserve the absolute and relative scaling among certain channels, and modify that scaling between other groups of channels.



## `shift_mnirs()`

* `data`

    This function takes in a data frame, applies processing to all channels specified, then returns the processed data frame. *`mnirs`* metadata will be passed to and from this function.

* `nirs_channels`

    Unlike previous functions, `nirs_channels` should provided as a list (e.g. `list(c(A, B), c(C))`). Each list item represents a group to be shifted together to a common scale. Separate list items will be shifted to separate scales. The relative scaling between channels will be preserved within each group, but lost between groups.\
\
`nirs_channels` should be specified explicitly to ensure the intended grouping structure is returned. The default *`mnirs`* metadata will group all NIRS channels together.

* `time_channel`

    If the data contain *`mnirs`* metadata, this channel will be detected automatically, or it can be specified explicitly.

* `to` or `by`

    The shift amplitude can be specified by either shifting signals `to` a new value, or shifting signals `by` a fixed amplitude, given in units of the NIRS signals.

* `width` or `span`

    Shifting can be performed on the mean value within a window specified by one of either `width` or `span`. `width` defines a number of samples centred around the local index being evaluated (`idx`), whereas `span` defines a range of time in units of `time_channel`.\
\
e.g. `width = 1` will shift from the single minimum sample, whereas `span = 1` would shift from the minimum mean value of all samples within one second.

* `position`

    Specifies the reference point used to shift the data; either `"min"`, `"max"`, or `"first"` sample(s).


For this data set, we want to shift each NIRS channel so that the mean of the 2-minute baseline is equal to zero, which would then give us a change in deoxygenation from baseline during the incremental cycling protocol.

::: {.callout-note}
## `{tidyverse}`-style channel name specification

This is a good time to note that here and in most *`mnirs`* functions, data channels can be specified using `{tidyverse}`-style naming. Data frame column names can be specified either with quotes as a character string (`"smo2"`), or as a direct symbol (`smo2`).

`{tidyselect}` support functions such as `starts_with()`, `matches()` can also be used.
:::

```{r}
data_shifted <- shift_mnirs(
    data_filtered,     ## un-grouped nirs channels to shift separately 
    nirs_channels = list(smo2_left, smo2_right), ## 👈 channels grouped separately
    to = 0,            ## NIRS values will be shifted to zero
    span = 120,        ## shift the *first* 120 sec of data to zero
    position = "first"
)

plot(data_shifted, time_labels = TRUE) +
    geom_hline(yintercept = 0, linetype = "dotted")
```

Before shifting, the minimum (end of exercise) values for *smo2_left* and *smo2_right* were similar, but the starting baseline values were different. 

When we shift both baseline values to zero, however, we can see that the *smo2_left* signal has a smaller deoxygenation amplitude compared to the *smo2_right* signal, and a (slightly) greater hyperaemic reoxygenation peak.

We have to consider how our assumptions and processing decisions will influence our interpretations; by shifting both starting values, we are normalising for, and implicitly assuming that the baseline represents the same starting condition for the tissues in both legs.

This may or may not be an appropriate assumption for your research question; for example, this may be appropriate when we are more interested in the relative change (delta) in each leg during an intervention or exposure (often referred to as `"∇SmO2"`), but not if we were interested in asymmetries that could influence SmO<sub>2</sub> at rest.



## `rescale_mnirs()`

We may also want to rescale our data to a new dynamic range, changing the signal amplitude.

* `data`

    This function takes in a data frame, applies processing to all channels specified, then returns the processed data frame. *`mnirs`* metadata will be passed to and from this function.

* `nirs_channels`

    Unlike previous functions, `nirs_channels` should provided as a list (e.g. `list(c(A, B), c(C))`). Each list item represents a group to be rescale together to a common range. Separate list items will be rescaled to separate ranges. The relative scaling between channels will be preserved within each group, but lost between groups.\
\
`nirs_channels` should be specified explicitly to ensure the intended grouping structure is returned. The default *`mnirs`* metadata will group all NIRS channels together.

* `range`

    Specifies the new dynamic range in the form `c(min, max)`. For example, if we want to calibrate each NIRS signal to its own observed 'functional range' during exercise, we could rescale them to 0-100%. 

```{r}
data_rescaled <- rescale_mnirs(
    data_filtered,    ## un-grouped nirs channels to rescale separately 
    nirs_channels = list(smo2_left, smo2_right), ## 👈 channels grouped separately
    range = c(0, 100) ## rescale to a 0-100% functional exercise range
)

plot(data_rescaled, time_labels = TRUE) +
    geom_hline(yintercept = c(0, 100), linetype = "dotted")
```

Here, our assumption is that during a maximal exercise task, the minimum and maximum values represent the functional capacity of each tissue volume being observed. So we rescale the functional dynamic range in each leg.

By normalising this way, we might lose meaningful differences captured by the different amplitudes between *smo2_left* and *smo2_right*, but we might be more interested in the trend or time course of each response.

Our interpretation may be that *smo2_right* appears to start at a slightly higher percent of its functional range, deoxygenates faster toward a minimum, and reaches a quasi-plateau near maximal exercise. While *smo2_left* deoxygenates slightly slower and continues to deoxygenate until maximal task tolerance. 

Additionally, the left leg reoxygenates slightly faster than the right during recovery. This might, for example, indicate exercise capacity differences between the limbs (although these differences are marginal and only discussed as representative for influence on interpretations).



# 🔀 Pipe-friendly functions

Most *`mnirs`* functions can be piped together using Base R 4.1+ (`|>`) or `{magrittr}` (`%>%`). The entire pre-processing stage can easily be performed in a sequential pipe.

To demonstrate this, we'll read a different example file recorded with *Train.Red FYER* muscle oxygen sensor and pipe it through each processing stage straight to a plot.

This is also a good time to demonstrate how to use the global `mnirs.verbose` option to silence all warning & information messages, for example when dealing with a familiar dataset. We recommend leaving `verbose = TRUE` by default whenever reading and exploring a new file.

```{r}
options(mnirs.verbose = FALSE)

nirs_data <- read_mnirs(
    example_mnirs("train.red"),
    nirs_channels = c(
        smo2_left = "SmO2 unfiltered",
        smo2_right = "SmO2 unfiltered"
    ),
    time_channel = c(time = "Timestamp (seconds passed)"),
    zero_time = TRUE
) |>
    resample_mnirs(method = "linear") |> ## default settings will resample to the same `sample_rate`
    replace_mnirs(
        invalid_above = 73,
        outlier_cutoff = 3,
        span = 7
    ) |>
    filter_mnirs(
        method = "butterworth",
        order = 2,
        W = 0.02,
        na.rm = TRUE
    ) |>
    shift_mnirs(
        nirs_channels = list(smo2_left, smo2_right), ## 👈 channels grouped separately
        to = 0,
        span = 60,
        position = "first"
    ) |>
    rescale_mnirs(
        nirs_channels = list(c(smo2_left, smo2_right)), ## 👈 channels grouped together
        range = c(0, 100)
    )

plot(nirs_data, time_labels = TRUE)
```

We have two exercise intervals in this data set. Let's demonstrate some of the common analysis methods currently available with *`mnirs`*.

# 🧮 Interval extraction

After the NIRS signal has been cleaned and filtered, it should be ready for further processing and analysis. 

*`mnirs`* is under development to include functionality for processing discrete intervals and events, e.g. reoxygenation kinetics, slope calculations for post-occlusion microvascular responsiveness, and critical oxygenation breakpoints.



## `extract_intervals()`

Often we will need to locate and extract smaller intervals from a NIRS recording for further processing and analysis. For example, if we want to extract the last three minutes from repeated exercise trials.

We may have marked event labels or have incremental lap numbers in the specified `event_channel`, or we may simply have to manually specify the event markers by `time_channel` value or row number in the data frame. See `extract_intervals()` for more details.

* `data`

    This function takes in a data frame, applies processing to all channels specified, then returns a list of processed data frames. *`mnirs`* metadata will be passed to and from this function.

* `nirs_channels`

    If returning a list of *"distinct"* intervals (see `event_groups` below), `nirs_channels` does not have to be specified, as no channels are processed.\
\
Only when *"ensemble"*-averaging, `nirs_channels` should be specified by providing a list of column names (e.g. `list(c(A, B), c(A))`), where each list item specifies the channels to be ensemble-averaged within the respective group (ensemble-groups are specified by `event_groups` below), in the order in which they are returned. The default *`mnirs`* metadata will ensemble-average all `nirs_channels`.

* `time_channel`, `event_channel`, & `sample_rate`

    If the data contain *`mnirs`* metadata, these will be detected automatically, or they can be specified explicitly.

* `start` & `end`

    Interval boundaries are specified using helper functions: `by_time()` for `time_channel` values; `by_sample()` for sample indices (row numbers); `by_label()` for `event_channel` labels; or `by_lap()` for `event_channel` lap integers. Provide both `start` and `end` to define precise intervals, or provide `start` alone with `span` to extract windows around events. Multiple values can be passed to extract multiple intervals.

* `event_groups`

    Events can be extracted and returned as a list of `"distinct"` intervals, or `"ensemble"`-averaged into a single data frame. Custom grouping structure for ensemble-averaging can be specified by event number, in order of appearance within the original data.\
\
e.g. `event_groups = list(c(1, 2), c(3, 4))` would return a list of two intervals, each ensemble-averaged from the respective events, in sequential order from the original data.

* `span`

    When only `start` (or `end`) is provided, `span` specifies a time window in units of `time_channel` as `c(before, after)`, where positive values indicate time after the event and negative values indicate time before. When both `start` and `end` are provided, `span` shifts boundaries additively: `span[1]` adjusts starts, `span[2]` adjusts ends.\
\
A list of unique `span` vectors can be specified for each interval, otherwise a single `span` vector will be recycled to all intervals.

* `zero_time`

    `FALSE` by default; because `time_channel` values within each returned interval likely start at a non-zero value, `zero_time = TRUE` will re-calculate time starting from zero at the specified event, so the start time (or however an event is specified) will become `0`.\
\
When ensemble-averaging across intervals, time will always be re-calculated from `0`, since the time values have lost their meaning.

```{r}
#| fig.width: 8
#| fig.asp: 0.55

## return each interval independently with `event_groups = "distinct"`
distinct <- extract_intervals(
    nirs_data,                  ## channels blank for "distinct" grouping
    start = by_time(348, 1064), ## manually identified interval start times
    end = by_time(458, 1174),   ## interval end time (start + 150 sec)
    event_groups = "distinct",  ## return a list of data frames for each (2) event
    span = c(0, 0),             ## no boundary modification
    zero_time = FALSE           ## return original time values
)

plot(distinct, time_labels = TRUE)
```

```{r}
## ensemble average both intervals with `event_groups = "ensemble"`
ensemble <- extract_intervals(
    nirs_data,                  ## channels recycled to all intervals by default
    nirs_channels = c(smo2_left, smo2_right),
    start = by_time(368, 1084), ## alternatively specify start times + span
    event_groups = "ensemble",  ## ensemble-average across two intervals
    span = c(-20, 90),          ## span recycled to all intervals by default
    zero_time = TRUE            ## re-calculate common time to start from `0`
)

plot(ensemble, time_labels = TRUE) + 
    geom_vline(xintercept = 0, linetype = "dotted")
```

The ensemble-averaged kinetics capture a more representative 'average' response during the two interval deoxygenation events.

Ensemble-averaging can help mitigate data dropouts and other quality issues from measurement error or biological variability. For example, ensemble-averaging is common for evaluating systemic VO<sub>2</sub> kinetics, where responses can be highly variable trial to trial.


# Conclusion

This vignette walks through the core functionality of *`mnirs`* to read, clean, and pre-process data in preparation for analysis.

Future development and articles will cover standardised analysis methods, including deoxygenation & reoxygenation kinetics and critical oxygenation breakpoint analysis.

---
title: "Reading and Cleaning Data with mnirs"
vignette: >
    %\VignetteIndexEntry{Reading and Cleaning Data with mnirs}
    %\VignetteEngine{quarto::html}
    %\VignetteEncoding{UTF-8}
format:
    html:
        embed-resources: true
        toc: true
knitr:
    opts_chunk:
        collapse: true
        comment: '#>'
        dpi: 300
        out.width: "100%"
        fig.align: "centre"
        eval: requireNamespace("ggplot2", quietly = TRUE)
---

# Introduction

Modern wearable muscle near-infrared spectroscopy (mNIRS) devices make it easier than ever to monitor local muscle oxygenation during dynamic activities.

The real challenge comes with deciding how to clean, filter, process, and eventually interpret those data.

The `{mnirs}` package aims to provide standardised, reproducible methods for reading, processing, and analysing mNIRS data. The goal is to help practitioners detect meaningful signal from noise, and improve our confidence in interpreting and applying information to the clients we work with.

In this vignette we will demonstrate how to:

* 📂 Read files exported from wearable NIRS devices, and import NIRS channels into a standard data frame format with metadata, ready for further processing.

* 📊 Plot and visualise data frames of class *`"mnirs"`*.

* 🔍 Retrieve metadata stored with data frames of class *`"mnirs"`* to avoid repetitively specifying which channels to process.

* ⏱️ Resample data to a higher or lower sample rate, to correct irregular sampling periods, or to match the frequency of other data sources for synchronisation.

* 🧹 Detect and replace local outliers, invalid values, and interpolate across missing data.

* 📈️ Apply digital filtering to optimise signal-to-noise ratio for the responses observed in our data.

* ⚖️ Shift and rescale across multiple NIRS channels, to transform signal dynamic range while preserving absolute or relative scaling between muscle sites and NIRS channels.

* 🔀 Demonstrate a complete data wrangling process using pipe-friendly functions.

* 🧮 Detect and extract intervals for further analysis.

*`mnirs`* is designed to process NIRS data, but it can be used to read, clean, and process other time series data which require many of the same processing steps. Enjoy!


# 📂 Read data from file

We will read an example data file with two NIRS channels from an incremental ramp cycling assessment recorded with *Moxy* muscle oxygen monitor.

First, install and load the *`mnirs`* package and other required libraries.

*`mnirs`* can be installed with `install.packages("mnirs")`.

```{r}
#| label: setup
#| warning: false

# pak::pak("jemarnold/mnirs") ## install development version
library(ggplot2)   ## for plotting
library(mnirs)
```

The first function called will often be `read_mnirs()`. This is used to read data files exported from mNIRS devices in *.xls(x)*, *.csv*, or *.txt* formats.

Exported data files will often have multiple rows of file header metadata before the data table with NIRS recordings begins. `read_mnirs()` can extract and return this data table along with the file metadata for further processing and analysis.

See `read_mnirs()` for more details.


## `read_mnirs()`

* `file_path`

    Specify the location of the data file, including file extension. e.g. `"my_file.xlsx"`, `"C:/myfolder/my_file.csv"`, or `here::here("data", "my_file.csv")`.

::: {.callout-note}
### Example data files

A few example data files are included in the *`mnirs`* package. File paths can be accessed with `example_mnirs()`.
:::

* `nirs_channels`
    
    Multiple *NIRS* channels can be specified from the data file as a vector of names. If no `nirs_channels` are specified, `read_mnirs()` will attempt to recognise the NIRS device file format, and return the full data frame with all detected columns. This can be useful for file exploration, to find the target channel names. However, best practice is to specify the desired `nirs_channels` explicitly.

* `time_channel`

    A *time* or *sample* channel name from the data table can be specified. If left blank, the function will attempt to identify the time column automatically, however, best practice is to specify the `time_channel` explicitly.

* `event_channel`

    Optionally, a channel can be specified which indicates character *event* labels or integer *lap* values in the data table.\
\
These channel names are used to detect the data table within the file, and must match exactly with text strings in the file on the same row. We can rename these channels when reading data by specifying a named character vector: 

```{r}
nirs_channels = c(
    renamed1 = "original_name1", 
    renamed2 = "original_name2"
)
```

* `sample_rate`

    The sample rate (in Hz) of the exported data can either be specified explicitly, or it will be estimated from `time_channel`.\
\
Automatic detection usually works well unless there is irregular sampling, or the `time_channel` is a count of samples rather than a time value. For example, *Oxysoft* exports a column of sample numbers rather than time values. For *Oxysoft* files specifically, the function will recognise and read the correct sample rate from the file metadata. However, in most cases `sample_rate` should be defined explicitly if known.

* `add_timestamp`

    `FALSE` by default; if `time_channel` is in date-time (POSIXct) format (e.g., *hh:mm:ss*), by default it will be converted to numeric time in seconds.\
\
If `add_timestamp = TRUE`, the date-time start time value in `time_channel` or in the file metadata will be extracted and a `"timestamp"` column will be added to the returned data frame. This can be useful for synchronising devices based on system time.

* `zero_time`

    `FALSE` by default; if `time_channel` values start at a non-zero value, `zero_time = TRUE` will re-calculate time starting from zero. If `time_channel` was converted from date-time format, it will always be re-calculated from zero regardless of this option.

* `keep_all`

    `FALSE` by default; only the channels explicitly specified will be returned in a data frame. `keep_all = TRUE` will return all columns detected from the data table in the file.\
\
Blank/empty columns will be omitted. Duplicate column names will be repaired by appending a suffix `"_n"`, and empty column names will be renamed as `col_n`; where `n` is equal to the column number in the data file. Renamed columns should be checked to confirm correct naming if duplicates are present.

* `verbose`

    `TRUE` by default; this and most *`mnirs`* functions will return warnings and informational messages which are useful for troubleshooting and data validation. This option can be used to silence those messages. *`mnirs`* messages can be silenced globally for a session by setting `options(mnirs.verbose = FALSE)`.

```{r}
## {mnirs} includes sample files from a few NIRS devices
example_mnirs()

## partial matching will error if matches multiple
try(example_mnirs("moxy"))

data_raw <- read_mnirs(
    file_path = example_mnirs("moxy_ramp"), ## call an example data file
    nirs_channels = c(
        smo2_left = "SmO2 Live",    ## identify and rename channels
        smo2_right = "SmO2 Live(2)"
    ),
    time_channel = c(time = "hh:mm:ss"), ## date-time format will be converted to numeric
    event_channel = NULL,           ## leave blank if unused
    sample_rate = NULL,             ## if blank, will be estimated from time_channel
    add_timestamp = FALSE,          ## omit a date-time timestamp column
    zero_time = TRUE,               ## recalculate time values from zero
    keep_all = FALSE,               ## return only the specified data channels
    verbose = TRUE                  ## show warnings & messages
)

## Note the above info message that sample_rate was estimated correctly at 2 Hz ☝
## ignore the warnings about irregular sampling for now, we will resample later

data_raw
```


# 📊 Plot *`mnirs`* data

*`mnirs`* data can be easily viewed by calling `plot()` (explicitly documented as `plot.mnirs()`). This generic plot function uses *`ggplot2`* and will work on data frames (or lists of data frames) processed by *`mnirs`* functions, where the metadata contains `class = *"mnirs"*`.

:::{.callout-note}
### Processing lists of *`mnirs`* data frames

As of `mnirs v0.7.0`, most processing functions will accept either a single data frame, a list of data frames, or a grouped data frame (requiring `{dplyr}`), and returns a single or list of data frames, respectively. This allows separate processing of NIRS interval data, e.g. filtering and shifting NIRS channels within discrete intervals, rather than on the global data frame.
:::

## `plot.mnirs`

* `data`

    This function takes in a data frame or list of data frames of class *`mnirs`* and returns a formatted *`ggplot2`* plot. Lists of *`mnirs`* data frames are returned as plot facets.

* `points`

    `FALSE` by default; a quick way to plot individual points for more precise sample visualisation.

* `time_labels`

    `FALSE` by default; time values on the x-axis will be plotted as numeric by default. `time_labels = TRUE` will instead plot time values as `h:mm:ss` format.

* `na.omit`

    `FALSE` by default; missing data (`NA`s) will be plotted as gaps in the time-series data. If `na.omit = TRUE`, `NA`s will be omitted from the plotted data, effectively plotting across these gaps, making them less visible, but making the general data trend easier to see if there are lots of missing values.

```{r}
## note the `time_labels` plot argument to display time values as `h:mm:ss`
plot(
    data_raw,
    points = FALSE,
    time_labels = TRUE,
    na.omit = FALSE
)
```

*`mnirs`* includes a custom *`ggplot2`* theme and colour palette available with `theme_mnirs()` and `palette_mnirs()`. See those documentation references for more details.


# 🔍 Metadata stored in *`mnirs`* data frames

Data frames generated or read by *`mnirs`* functions will return `class = *"mnirs"*` and contain metadata, which can be retrieved with `attributes()`. 

Instead of re-defining values like our channel names or sample rate, certain *`mnirs`* functions can automatically retrieve them from metadata. They can always be overwritten manually in subsequent functions, or by using a helper function `create_mnirs_data()`.

See `create_mnirs_data()` for more details about metadata.

```{r}
## view metadata
attr(data_raw, "nirs_channels")

attr(data_raw, "time_channel")

attr(data_raw, "sample_rate")
```


# ⏱️ Resample data

mNIRS devices may have sample rates anywhere from 0.5 to 100+ Hz. Higher sample rates gives greater precision. However, issues can arise with irregular sampling where time samples occur at inconsistent frequencies, are repeated, or are lost. Resampling to a consistent sample rate can be used to regularise timestamps to the nearest time value according to the known data sample rate. This can make subsequent processing steps easier, as well as synchronisation with other devices.

Downsampling can be useful to reduce data size and processing requirement, although should not be used primarily as a data filtering/smoothing strategy (see **Digital filtering** section below). Upsampling can be done with interpolation over newly created samples, although caution should always be taken with creating new samples without any additional information. New samples created with resampling will be left blank (`NA`) by default.


## `resample_mnirs()`

* `data`

    This function takes in a data frame or list of data frames, applies processing to all channels specified explicitly or implicitly from *`mnirs`* metadata, then returns the processed data frame (or list). *`mnirs`* metadata will be passed to and from this function.\
\
*`mnirs`* functions are also pipe-friendly for Base R 4.1+ (`|>`) or `{magrittr}` (`%>%`) pipes to chain operations together (see below).

* `time_channel`

    The column name of time series values in `data`. If the data contain *`mnirs`* metadata, this channel will be detected automatically, or it can be specified explicitly.\
\
This function processes all columns in `data`, so `nirs_channels` does not need to be specified.

* `sample_rate`

    The sample rate (in Hz) of `data`. If the data contain *`mnirs`* metadata, this channel will be detected automatically, or it can be specified explicitly.

* `resample_rate`

    Resampling is specified as the desired number of samples per second (Hz). The default `resample_rate` will resample back to the existing `sample_rate` of the data. This can be useful to accommodate irregular sampling with unequal time values. Linear interpolation is used to resample `time_channel` to round values of the `sample_rate`.

* `method`

    Any new samples created by resampling (i.e. samples fill in between gaps, or up-sampled indices) can be filled in or left blank as `NA`. The default `method` is to leave new samples blank, and interpolation must be explicitly specified as either *"locf"* to repeat the last observation carried forward, or *"linear"* interpolation.

```{r}
data_resampled <- resample_mnirs(
    data_raw,            ## blank channels will be retrieved from metadata
    time_channel = time, ## channels can be left blank or specified explicitly
    sample_rate = NULL,  ## blank by default will be retrieved from metadata
    resample_rate = 2,   ## blank by default will resample to `sample_rate`
    method = "linear"    ## linear interpolation across resampled indices
)

## note the altered "time" values from the original data frame 👇
data_resampled
```


::: {.callout-note}
## Use `resample_mnirs()` to smooth over irregular or skipped samples

If we see a warning from `read_mnirs()` about duplicated or irregular samples like we saw above, we can use `resample_mnirs()` to restore `time_channel` to a regular sample rate and interpolate across skipped samples.

Note: if we perform both of these steps in a piped function call, we will still see the warning appear about irregular sampling at the end of the pipe. This warning is returned by `read_mnirs()`. But viewing the data frame can confirm that `resample_mnirs()` has resolved the issue.
:::


# 🧹 Replace local outliers, invalid values, and missing values

We can see some data issues in the plot above, so let's clean those with a single wrangling function `replace_mnirs()`, to prepare our data for digital filtering and smoothing.

*`mnirs`* tries to include basic functions which work on vector data, and convenience wrappers which combine functionality and can be used on multiple channels in a data frame at once.

Let's explain the functionality of this data-wide function, and for more details about the vector-specific functions see `replace_mnirs()`.


## `replace_mnirs()`

* `nirs_channels`

    Specify which column names in `data` to process, i.e. the response variables. If the data contain *`mnirs`* metadata, these channels will be detected automatically. Channels not explicitly specified will be passed through unprocessed to the returned data frame.

* `time_channel` & `sample_rate`

    If the data contain *`mnirs`* metadata, these will be detected automatically, or they can be specified explicitly.

* `invalid_values`, `invalid_above`, or `invalid_below`

    Specific invalid values can be replaced, e.g. if a NIRS device exports `0`, `100`, or some other fixed value when signal recording is in error. If spikes or drops are present in the data, these can be replaced by specifying values above or below which to consider invalid. If left as `NULL`, no values will be replaced.

* `outlier_cutoff`

    Local outliers can be detected using a cutoff calculated from the local median value. A default value of `3` is recommended and corresponds to Pearson's rule (i.e., ± 3 SD about the local median). If left as `NULL`, no outliers will be replaced.

* `width` or `span`

    Local outlier detection and median interpolation use a rolling local window specified by one of either `width` or `span`. `width` defines a number of samples centred around the local index being evaluated (`idx`), whereas `span` defines a range of time in units of `time_channel`.

* `method`

    Missing data (`NA`), invalid values, and local outliers specified above can be replaced via interpolation or fill methods; either `"linear"` interpolation (the default), fill with local `"median"`, or `"locf"` (*"last observation carried forward"*).\
\
`NA`s can be passed through to the returned data frame with `method = "none"`. However, subsequent processing & analysis steps may return errors when `NA`s are present. Therefore, it is good practice to deal with missing data early during data processing.

* `verbose`

    As above, an option to toggle warnings and info messages.

```{r}
data_cleaned <- replace_mnirs(
    data_resampled,     ## blank channels will be retrieved from metadata
    invalid_values = 0, ## known invalid values in the data
    invalid_above = 90, ## remove data spikes above 90
    outlier_cutoff = 3, ## recommended default value
    width = 7,          ## window to detect and replace outliers/missing values
    method = "linear"   ## linear interpolation over `NA`s
)

plot(data_cleaned, time_labels = TRUE)
```

That cleaned up all the obvious data issues.


# 📈 Digital filtering

To improve the signal-to-noise ratio in our dataset without losing information, we should apply digital filtering to smooth the data.



## Choosing a digital filter

There are a few digital filtering methods available in *`mnirs`*. Which option is best for *you* will depend in large part on the sample rate of your data and the frequency of the response signal or phenomenon you are interested in observing.

Choosing filter parameters is an important processing step to improve signal-to-noise ratio and enhance our subsequent interpretations. Over-filtering the data can introduce data artefacts which can negatively influence signal analysis and interpretations, just as much as trying to analyse overly-noisy raw data.

It is perfectly valid to choose a digital filter by iteratively testing filter parameters until the signal or response of interest appears to be visually optimised with minimal data artefacts, to your satisfaction.

We will discuss the process of choosing a digital filter in more depth in another article coming soon.


## `filter_mnirs()`

* `na.rm`

    `FALSE` by default; `filter_mnirs()` will return an error if any missing data (`NA`) are detected in the response variables (`nirs_channels`). Setting `na.rm = TRUE` will ignore these `NA`s and pass them through to the returned data frame, but this must be opted into explicitly in this function and elsewhere.


### Smoothing-spline

* `method = "smooth_spline"`

    The default non-parametric cubic smoothing spline is often a good first filtering option when exploring the data, and works well over longer time spans. For more rapid and repeated responses, a smoothing-spline may not work as well.

* `spar`

    The smoothing parameter of the cubic spline will be determined automatically by default, or can be specified explicitly. See `stats::smooth.spline()`.


### Butterworth digital filter

* `method = "butterworth"`

    A Butterworth low-pass digital filter (specified by `type = "low"`) is probably the most common method used in mNIRS research (whether appropriately or not). For certain applications such as identifying a signal with a known frequency (e.g. cycling/running cadence or heart rate), a pass-band or a different filter type may be better suited.

* `order`

    The filter order number, specifying the number of passes the filter performs over the data. The default `order = 2` will often noticeably improve the filter over a single pass, however, higher orders above ~`4` can begin to introduce artefacts, particularly at sharp transition points.

* `W` or `fc`

    The cutoff frequency can be specified either as `W`; a fraction (between `[0, 1]`) of the [*Nyquist frequency*](https://www.youtube.com/watch?v=IZJQXlbm2dU), which is equal to half of the `sample_rate` of the data. Or as `fc`; the cutoff frequency in Hz, where this absolute frequency should still be between 0 Hz and the Nyquist frequency.
    
* `type`

    The filter type is specified as either `c("low", "high", "stop", "pass")`.


For filtering vector data and more details about Butterworth filter parameters, see `filter_butterworth()`.


### Moving average

* `method = "moving_average"`

    The simplest smoothing method is a moving average filter applied over a specified number of samples or time span. Commonly, this might be a 5- or 15-second centred moving average.

* `width` or `span`

    Moving average filtering occurs within a rolling local window specified by one of either `width` or `span`. `width` defines a number of samples centred around the local index being evaluated (`idx`), whereas `span` defines a range of time in units of `time_channel`.

For filtering vector data and more details about the moving average filter, see `filter_moving_average()`.


## Apply the filter

Let's try a *Butterworth low-pass* filter, and we'll specify some empirically chosen filter parameters for these data. See `filter_mnirs()` for further details on each of these filtering methods and their respective parameters.


```{r}
data_filtered <- filter_mnirs(
    data_cleaned,           ## blank channels will be retrieved from metadata
    method = "butterworth", ## Butterworth digital filter is a common choice
    order = 2,              ## filter order number
    W = 0.02,               ## filter fractional critical frequency `[0, 1]`
    type = "low",           ## specify a "low-pass" filter
    na.rm = TRUE            ## explicitly ignore NAs
)

## we will add the non-filtered data back to the plot to compare
plot(data_filtered, time_labels = TRUE) +
    geom_line(
        data = data_cleaned, 
        aes(y = smo2_left, colour = "smo2_left"), alpha = 0.4
    ) +
    geom_line(
        data = data_cleaned, 
        aes(y = smo2_right, colour = "smo2_right"), alpha = 0.4
    )
```


That did a nice job reducing the high-frequency noise in our data, while preserving the sharp edges, such as the rapid reoxygenation after maximal exercise.

# ⚖️ Shift and rescale data

NIRS values are not measured on an absolute scale (arguably not even percent (%) saturation/SmO<sub>2</sub>). Therefore, we may need to adjust or calibrate our data to normalise NIRS signal values between muscle sites, individuals, trials, etc. depending on our intended comparison.

For example, we may want to set our mean baseline value to zero for all NIRS signals at the start of a recording. Or we may want to compare signal kinetics (the rate of change or time course of a response) after rescaling signal amplitudes to a common dynamic range.

These functions allow us to either shift NIRS values up or down while preserving the dynamic range (the absolute amplitude from minimum to maximum values) of our NIRS channels, or rescale the data to a new dynamic range with larger or smaller amplitude.

We can group NIRS channels together to preserve the absolute and relative scaling among certain channels, and modify that scaling between other groups of channels.



## `shift_mnirs()`

* `group_channels`

    Controls how `nirs_channels` are grouped, to preserve absolute or relative scaling. Either a character keyword or a `list()` of channel-name vectors:\
\
`"ensemble"` (the default) shifts all channels together to a common value, preserving the relative scaling between channels.\
\
`"distinct"` shifts each channel independently, losing the relative scaling between channels.\
\
A `list()` of channel-name vectors (e.g. `list(c(A, B), c(C))`) shifts channels `A` & `B` together and `C` separately, preserving relative scaling within, but not between groups.

* `to` or `by`

    The shift amplitude can be specified by either shifting signals `to` a new value, or shifting signals `by` a fixed amplitude, given in units of the NIRS signals.

* `width` or `span`

    Shifting can be performed on the mean value within a window specified by one of either `width` or `span`. `width` defines a number of samples centred around the local index being evaluated (`idx`), whereas `span` defines a range of time in units of `time_channel`.\
\
e.g. `width = 1` will shift from the single minimum sample, whereas `span = 1` would shift from the minimum mean value of all samples within one second.

* `position`

    Specifies the reference point used to shift the data; either `"min"`, `"max"`, or `"first"` sample(s).


For this data set, we want to shift each NIRS channel so that the mean of the 2-minute baseline is equal to zero, which would then give us a change in deoxygenation from baseline during the incremental cycling protocol.

::: {.callout-note}
## `{tidyverse}`-style channel name specification

This is a good time to note that here and in most *`mnirs`* functions, data channels can be specified using `{tidyverse}`-style naming. Data frame column names can be specified either with quotes as a character string (`"smo2"`), or as a direct symbol (`smo2`).

`{tidyselect}` support functions such as `starts_with()`, `matches()` can also be used.
:::

```{r}
data_shifted <- shift_mnirs(
    data_filtered,
    group_channels = list(smo2_left, smo2_right), ## 👈 channels shifted separately
    to = 0,            ## NIRS values will be shifted to zero
    span = 120,        ## shift the *first* 120 sec of data to zero
    position = "first"
)

plot(data_shifted, time_labels = TRUE) +
    geom_hline(yintercept = 0, linetype = "dotted")
```

Before shifting, the minimum (end of exercise) values for *smo2_left* and *smo2_right* were similar, but the starting baseline values were different. 

When we shift both baseline values to zero, however, we can see that the *smo2_left* signal has a smaller deoxygenation amplitude compared to the *smo2_right* signal, and a (slightly) greater hyperaemic reoxygenation peak.

We have to consider how our assumptions and processing decisions will influence our interpretations; by shifting both starting values, we are normalising for, and implicitly assuming that the baseline represents the same starting condition for the tissues in both legs.

This may or may not be an appropriate assumption for your research question; for example, this may be appropriate when we are more interested in the relative change (delta) in each leg during an intervention or exposure (often referred to as `"∇SmO2"`), but not if we were interested in asymmetries that could influence SmO<sub>2</sub> at rest.



## `rescale_mnirs()`

We may also want to rescale our data to a new dynamic range, changing the signal amplitude.

* `group_channels`

    Controls how `nirs_channels` are grouped, to preserve absolute or relative scaling. Either a character keyword or a `list()` of channel-name vectors:\
\
`"ensemble"` (the default) rescales all channels together to a common range, preserving the relative scaling between channels.\
\
`"distinct"` rescales each channel independently, losing the relative scaling between channels.\
\
A `list()` of channel-name vectors (e.g. `list(c(A, B), c(C))`) rescales channels `A` & `B` together and `C` separately, preserving relative scaling within, but not between groups.

* `range`

    Specifies the new dynamic range in the form `c(min, max)`. For example, if we want to calibrate each NIRS signal to its own observed 'functional range' during exercise, we could rescale them to 0-100%. 

```{r}
data_rescaled <- rescale_mnirs(
    data_filtered,
    group_channels = list(smo2_left, smo2_right), ## 👈 channels rescaled separately
    range = c(0, 100) ## rescale to a 0-100% functional exercise range
)

plot(data_rescaled, time_labels = TRUE) +
    geom_hline(yintercept = c(0, 100), linetype = "dotted")
```

Here, our assumption is that during a maximal exercise task, the minimum and maximum values represent the functional capacity of each tissue volume being observed. So we rescale the functional dynamic range in each leg.

By normalising this way, we might lose meaningful differences captured by the different amplitudes between *smo2_left* and *smo2_right*, but we might be more interested in the trend or time course of each response.

Our interpretation may be that *smo2_right* appears to start at a slightly higher percent of its functional range, deoxygenates faster toward a minimum, and reaches a quasi-plateau near maximal exercise. While *smo2_left* deoxygenates slightly slower and continues to deoxygenate until maximal task tolerance. 

Additionally, the left leg reoxygenates slightly faster than the right during recovery. This might, for example, indicate exercise capacity differences between the limbs (although these differences are marginal and only discussed as representative for influence on interpretations).



# 🔀 Pipe-friendly functions

Most *`mnirs`* functions can be piped together using Base R 4.1+ (`|>`) or `{magrittr}` (`%>%`). The entire processing workflow can easily be performed in a sequential pipe.

To demonstrate this, we'll read a different example file recorded with *Train.Red FYER* muscle oxygen sensor and pipe it through each processing stage straight to a plot.

This is also a good time to demonstrate how to use the global `mnirs.verbose` option to silence all warning & information messages, for example when dealing with a familiar dataset. We recommend leaving `verbose = TRUE` by default whenever reading and exploring a new file.

```{r}
options(mnirs.verbose = FALSE)

nirs_data <- read_mnirs(
    example_mnirs("train.red"),
    nirs_channels = c(
        smo2_left = "SmO2 unfiltered",
        smo2_right = "SmO2 unfiltered"
    ),
    time_channel = c(time = "Timestamp (seconds passed)"),
    zero_time = TRUE
) |>
    resample_mnirs(method = "linear") |> ## default settings will resample to the same `sample_rate`
    replace_mnirs(
        invalid_above = 73,
        outlier_cutoff = 3,
        span = 7
    ) |>
    filter_mnirs(
        method = "butterworth",
        order = 2,
        W = 0.02,
        na.rm = TRUE
    ) |>
    shift_mnirs(
        group_channels = list(smo2_left, smo2_right), ## 👈 channels grouped separately
        to = 0,
        span = 60,
        position = "first"
    ) |>
    rescale_mnirs(
        group_channels = list(c(smo2_left, smo2_right)), ## 👈 channels grouped together
        range = c(0, 100)
    )

plot(nirs_data, time_labels = TRUE)
```

We have two exercise intervals in this data set. Let's demonstrate some of the common analysis methods currently available with *`mnirs`*.

# 🧮 Interval extraction

After the NIRS signal has been cleaned and filtered, it should be ready for further processing and analysis. 

*`mnirs`* is under development to include functionality for processing discrete intervals and events, e.g. reoxygenation kinetics, slope calculations for post-occlusion microvascular responsiveness, and critical oxygenation breakpoints.



## `extract_intervals()`

Often we will need to locate and extract smaller intervals from a NIRS recording for further processing and analysis. For example, if we want to extract the last three minutes from repeated exercise trials.

We may have marked event labels or have incremental lap numbers in the specified `event_channel`, or we may simply have to manually specify the event markers by `time_channel` value or row number in the data frame. See `extract_intervals()` for more details.

* `data`

    This function takes in a single data frame, detects and extracts specified intervals, and returns a list of processed data frames. *`mnirs`* metadata will be passed to and from this function.

* `nirs_channels`

    If returning a list of *"distinct"* intervals (see `group_intervals` below), `nirs_channels` does not have to be specified, as no channels are processed.\
\
Only when *"ensemble"*-averaging, `nirs_channels` should be specified by providing a list of column names (e.g. `list(c(A, B), c(A))`), where each list item specifies the channels to be ensemble-averaged within the respective group (ensemble-groups are specified by `group_intervals` below), in the order in which they are returned. The default *`mnirs`* metadata will ensemble-average all `nirs_channels`.

* `start` & `end`

    Interval boundaries are specified using helper functions: `by_time()` for `time_channel` values; `by_sample()` for sample indices (row numbers); `by_label()` for `event_channel` labels; or `by_lap()` for `event_channel` lap integers. Provide both `start` and `end` to define precise intervals, or provide `start` alone with `span` to extract windows around events. Multiple values can be passed to extract multiple intervals.

* `group_intervals`

    Events can be extracted and returned as a list of `"distinct"` intervals, or `"ensemble"`-averaged into a single data frame. Custom grouping structure for ensemble-averaging can be specified by event number, in order of appearance within the original data.\
\
e.g. `group_intervals = list(c(1, 2), c(3, 4))` would return a list of two intervals, each ensemble-averaged from the respective events, in sequential order from the original data.

* `span`

    When only `start` (or `end`) is provided, `span` specifies a time window in units of `time_channel` as `c(before, after)`, where positive values indicate time after the event and negative values indicate time before. When both `start` and `end` are provided, `span` shifts boundaries additively: `span[1]` adjusts starts, `span[2]` adjusts ends.\
\
A list of unique `span` vectors can be specified for each interval, otherwise a single `span` vector will be recycled to all intervals.

* `zero_time`

    `FALSE` by default; because `time_channel` values within each returned interval likely start at a non-zero value, `zero_time = TRUE` will re-calculate time starting from zero at the specified event, so the start time (or however an event is specified) will become `0`.\
\
When ensemble-averaging across intervals, time will always be re-calculated from `0`, since the time values have lost their meaning.

```{r}
#| fig.width: 8
#| fig.asp: 0.55

## return each interval independently with `group_intervals = "distinct"`
distinct <- extract_intervals(
    nirs_data,                    ## channels blank for "distinct" grouping
    group_intervals = "distinct", ## return a list of data frames for each (2) event
    start = by_time(348, 1064),   ## manually identified interval start times
    end = by_time(458, 1174),     ## interval end time (start + 150 sec)
    span = c(0, 0),               ## no boundary modification
    zero_time = FALSE             ## return original time values
)

plot(distinct, time_labels = TRUE)
```

```{r}
## ensemble average both intervals with `group_intervals = "ensemble"`
ensemble <- extract_intervals(
    nirs_data,                    ## channels recycled to all intervals by default
    group_intervals = "ensemble", ## ensemble-average across two intervals
    start = by_time(368, 1084),   ## alternatively specify start times + span
    span = c(-20, 90),            ## span recycled to all intervals by default
    zero_time = TRUE              ## re-calculate common time to start from `0`
)

plot(ensemble, time_labels = TRUE) + 
    geom_vline(xintercept = 0, linetype = "dotted")
```

The ensemble-averaged kinetics capture a more representative 'average' response during the two interval deoxygenation events.

Ensemble-averaging can help mitigate data dropouts and other quality issues from measurement error or biological variability. For example, ensemble-averaging is common for evaluating systemic VO<sub>2</sub> kinetics, where responses can be highly variable trial to trial.


# Conclusion

This vignette walks through the core functionality of *`mnirs`* to read, clean, and pre-process data in preparation for analysis.

Future development and articles will cover standardised analysis methods, including deoxygenation & reoxygenation kinetics and critical oxygenation breakpoint analysis.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract_interval_helpers.R
\name{group_intervals}
\alias{group_intervals}
\title{Apply grouping to intervals}
\usage{
group_intervals(
  df_list,
  nirs_channels,
  metadata,
  event_groups,
  zero_time = FALSE,
  verbose = TRUE
)
}
\description{
Apply grouping to intervals
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{moxy_intervals.csv}
\alias{moxy_intervals.csv}
\title{0.5 Hz Moxy onboard export}
\format{
.csv file with seven columns and 936 rows:
\describe{
\item{mm-dd}{Recording date (dd-MMM format).}
\item{hh:mm:ss}{Recording time of day (hh:mm:ss format).}
\item{SmO2 Live}{Muscle oxygen saturation, raw signal (\%).}
\item{SmO2 Averaged}{Muscle oxygen saturation, rolling average (\%).}
\item{THb}{Total haemoglobin (arbitrary units).}
\item{Lap}{Lap marker (integer). Not typically in use.}
\item{Session Ct}{Session count of recordings.}
}

Channel mapping for \code{\link[=read_mnirs]{read_mnirs()}}:
\itemize{
\item \code{nirs_channels = c("SmO2 Live", "SmO2 Averaged", "THb")}
\item \code{time_channel = c("hh:mm:ss")}
\item \code{interval_times = list(start = c(124, 486, 848, 1210), end = c(364, 726, 1088, 1450))}
}
}
\source{
Moxy Monitor (Fortiori Design LLC), exported via Moxy Portal App.
(https://www.moxymonitor.com/)
}
\description{
Exported from Moxy onboard recording at 0.5 Hz
no smoothing. Containing four 4-minute cycling work intervals,
placed on the vastus lateralis muscle site.
}
\examples{
example_mnirs("moxy_intervals")

}
\seealso{
\code{\link[=read_mnirs]{read_mnirs()}}, \code{\link[=example_mnirs]{example_mnirs()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validate_mnirs.R
\name{validate_mnirs}
\alias{validate_mnirs}
\alias{validate_numeric}
\alias{validate_mnirs_data}
\alias{validate_nirs_channels}
\alias{validate_time_channel}
\alias{validate_event_channel}
\alias{estimate_sample_rate}
\alias{validate_sample_rate}
\alias{validate_width_span}
\alias{validate_x_t}
\title{Validate \code{{mnirs}} parameters}
\usage{
validate_numeric(
  x,
  elements = Inf,
  range = NULL,
  inclusive = c("left", "right"),
  integer = FALSE,
  allow_na = FALSE,
  msg1 = "",
  msg2 = ""
)

validate_mnirs_data(data, ncol = 2L)

validate_nirs_channels(
  nirs_channels,
  data,
  verbose = FALSE,
  as_list = FALSE,
  env = rlang::caller_env()
)

validate_time_channel(time_channel, data, env = rlang::caller_env())

validate_event_channel(
  event_channel,
  data,
  required = TRUE,
  env = rlang::caller_env()
)

estimate_sample_rate(x)

validate_sample_rate(data, time_channel, sample_rate, verbose = TRUE)

validate_width_span(width = NULL, span = NULL, verbose = TRUE, msg = "")

validate_x_t(x, t, allow_na = FALSE)
}
\arguments{
\item{x}{A numeric vector.}

\item{elements}{An integer. Default is \code{Inf}. The number of numeric elements
expected in \code{x}.}

\item{range}{A two-element numeric vector giving the valid range for \code{x}.}

\item{inclusive}{A character vector specifying which boundaries of \code{range}
are included. Any of \code{"left"}, \code{"right"} (default is both). Use \code{FALSE} to
exclude both endpoints.}

\item{integer}{Logical. Default is \code{FALSE}. If \code{TRUE}, validate \code{x} as
integer-like values using \code{\link[rlang:is_integerish]{rlang::is_integerish()}}. Otherwise tested as a
numeric value.}

\item{allow_na}{Logical. Default is \code{FALSE}. If \code{TRUE}, allows pass through
of \code{NA} to the returned numeric/integer vector.}

\item{msg1, msg2}{A character string appended to the \code{\link[cli:cli_abort]{cli::cli_abort()}}
message when numeric validation fails.}

\item{data}{A data frame of class \emph{"mnirs"} containing time series data and
metadata.}

\item{nirs_channels}{A character vector giving the names of mNIRS columns to
operate on. Must match column names in \code{data} exactly.
\itemize{
\item If \code{NULL} (default), the \code{nirs_channels} metadata attribute of \code{data} is
used.
}}

\item{verbose}{Logical. Default is \code{TRUE}. Display or silence (if \code{FALSE})
warnings and information messages helpful for troubleshooting. Ad
global default can be set via \code{options(mnirs.verbose = FALSE)}.}

\item{as_list}{Logical. Default is \code{FALSE}. If \code{nirs_channels} is specified
as a list, it will be coerced to a flat character vector and an
information message is displayed (when \code{verbose = TRUE}). If \code{TRUE},
\code{nirs_channels} is returned as-is, i.e. as a list for callers which
require it.}

\item{time_channel}{A character string naming the time or sample column.
Must match a column name in \code{data} exactly.
\itemize{
\item If \code{NULL} (default), the \code{time_channel} metadata attribute of \code{data} is
used.
}}

\item{event_channel}{A character string naming the event/lap column. Must
match a column name in \code{data} exactly.
\itemize{
\item If \code{NULL} (default), the \code{event_channel} metadata attribute of \code{data} is
used.
}}

\item{required}{Logical. Default is \code{TRUE}. \code{event_channel} must be
present or detected in metadata. If \code{FALSE}, \code{event_channel} may be \code{NULL}.}

\item{sample_rate}{A numeric sample rate in Hz.
\itemize{
\item If \code{NULL} (default), the \code{sample_rate} metadata attribute of \code{data} will
be used if detected, or the sample rate will be estimated from
\code{time_channel}.
}}
}
\value{
Returns the validated object (e.g. a resolved \code{time_channel}
string), or invisibly returns \code{NULL} for successful validations. On
failure, an error is thrown via \code{\link[cli:cli_abort]{cli::cli_abort()}}.
}
\description{
Resolve and validate \emph{{mnirs}} metadata and perform basic data quality
checks.
}
\details{
\code{validate_mnirs()} is an internal documentation topic for a set of
validators used throughout the package. These validators:
\itemize{
\item Prefer explicit user-supplied arguments.
\item Fall back to \emph{"mnirs"} metadata attributes when available.
\item Fail fast with informative \code{\link[cli:cli_abort]{cli::cli_abort()}} messages when values are
missing or invalid.
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.mnirs.R
\name{breaks_timespan}
\alias{breaks_timespan}
\title{Breaks for time span data}
\usage{
breaks_timespan(unit = c("secs", "mins", "hours", "days", "weeks"), n = 5)
}
\arguments{
\item{unit}{The time unit used to interpret numeric data input (\emph{defaults}
to \emph{"secs"}).}

\item{n}{Desired number of breaks. You may get slightly more or fewer breaks
than requested.}
}
\value{
Returns a function for generating breaks.
}
\description{
Pretty time span breaks for plotting in units of 5, 15, 30, 60 sec, etc.
Modified from \code{\link[scales:breaks_timespan]{scales::breaks_timespan()}}.
}
\examples{
\dontshow{if (rlang::is_installed("ggplot2")) withAutoprint(\{ # examplesIf}
x <- 0:120
y <- sin(2 * pi * x / 15) + rnorm(length(x), 0, 0.2)

ggplot2::ggplot(data.frame(x, y), ggplot2::aes(x, y)) +
    theme_mnirs() +
    ggplot2::scale_x_continuous(breaks = breaks_timespan()) +
    ggplot2::geom_line()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_mnirs_helpers.R
\name{select_rename_data}
\alias{select_rename_data}
\title{Select data table columns and rename from channels, handling duplicates}
\usage{
select_rename_data(
  data,
  nirs_channels,
  time_channel,
  event_channel = NULL,
  keep_all = FALSE,
  verbose = TRUE
)
}
\description{
Select data table columns and rename from channels, handling duplicates
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_mnirs_helpers.R
\name{detect_irregular_samples}
\alias{detect_irregular_samples}
\title{Report warnings for unbalanced time_channel samples}
\usage{
detect_irregular_samples(x, time_channel, verbose = TRUE)
}
\description{
Report warnings for unbalanced time_channel samples
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract_interval_helpers.R
\name{recycle_span}
\alias{recycle_span}
\title{recycle a single-element span to c(before, after)
positive -> c(0, x), negative -> c(x, 0)}
\usage{
recycle_span(span)
}
\description{
recycle a single-element span to c(before, after)
positive -> c(0, x), negative -> c(x, 0)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract_interval_helpers.R
\name{zero_offset_data}
\alias{zero_offset_data}
\title{Recalculate time_channel values with zero offset at event time (t0)}
\usage{
zero_offset_data(data, time_channel, t0)
}
\description{
Recalculate time_channel values with zero offset at event time (t0)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract_interval_helpers.R
\name{extract_df_list}
\alias{extract_df_list}
\title{Extract interval data by time range}
\usage{
extract_df_list(data, time_vec, interval_spec, nirs_channels)
}
\description{
Extract interval data by time range
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mnirs_methods.R
\name{print.mnirs}
\alias{print.mnirs}
\title{Methods for mnirs objects}
\usage{
\method{print}{mnirs}(x, ...)
}
\arguments{
\item{x}{Object of class \code{"mnirs"}.}

\item{...}{Additional arguments passed to \code{\link[=print]{print()}}.}
}
\value{
\item{\code{print}}{Returns \code{x} without class attributes.}
}
\description{
Generic methods for objects of class \code{"mnirs"}.
}
\examples{
x <- read_mnirs(
    example_mnirs("train.red"),
    nirs_channels = c(smo2 = "SmO2"),
    time_channel = c(time = "Timestamp (seconds passed)"),
    verbose = FALSE
) |>
    resample_mnirs(method = "linear", verbose = FALSE) |>
    extract_intervals(
        start = by_time(2452, 3168),
        span = c(-60, 120),
        verbose = FALSE
    )

print(x)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{portamon-oxcap.xlsx}
\alias{portamon-oxcap.xlsx}
\title{10 Hz Artinis Oxysoft export recorded with Portamon}
\format{
.xlsx file with header metadata and six columns and 7943 rows:
\describe{
\item{Column 1}{Sample index (divide by sample rate for seconds).}
\item{Column 2}{tHb: total haemoglobin concentration change
(\eqn{\mu}M).}
\item{Column 3}{HHb: deoxyhaemoglobin concentration change (\eqn{\mu}M).}
\item{Column 4}{O2Hb: oxyhaemoglobin concentration change (\eqn{\mu}M).}
\item{Column 5}{Event marker (character).}
\item{Column 6}{Unmarked event label (character).}
}

Channel mapping for \code{\link[=read_mnirs]{read_mnirs()}}:
\itemize{
\item \code{nirs_channels = c(THb = 2, HHb = 3, O2Hb = 4)}
\item \code{time_channel = c(sample = 1)}
\item \code{event_channel = c(event = 5, label = "col_6")}
}
}
\source{
Artinis Medical Systems. Portamon, exported via Oxysoft desktop
software (https://artinis.com/)
}
\description{
Exported from Artinis Oxysoft, recorded on
Portamon at 10 Hz on the vastus lateralis muscle. Containing
two trials of repeated occlusion oxidative capacity testing,
each with 17 occlusions.
}
\examples{
example_mnirs("portamon")

}
\seealso{
\code{\link[=read_mnirs]{read_mnirs()}}, \code{\link[=example_mnirs]{example_mnirs()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.mnirs.R
\name{scale_colour_mnirs}
\alias{scale_colour_mnirs}
\alias{scale_color_mnirs}
\alias{scale_fill_mnirs}
\title{Scales for custom \emph{{mnirs}} palette}
\usage{
scale_colour_mnirs(..., aesthetics = "colour")

scale_color_mnirs(..., aesthetics = "colour")

scale_fill_mnirs(..., aesthetics = "fill")
}
\arguments{
\item{...}{Arguments passed to \code{ggplot2::discrete_scale()}.}

\item{aesthetics}{A character vector with aesthetic(s) passed to
\code{ggplot2::discrete_scale()}. \emph{Default} is \code{"colour"}.}
}
\value{
A \link[ggplot2:ggplot]{ggplot2} scale object.
}
\description{
Scales for custom \emph{{mnirs}} palette
}
\examples{
\dontshow{if (rlang::is_installed(c("ggplot2", "scales"))) withAutoprint(\{ # examplesIf}
## plot example data
data <- read_mnirs(
    file_path = example_mnirs("moxy_ramp"),
    nirs_channels = c(smo2_left = "SmO2 Live",
                      smo2_right = "SmO2 Live(2)"),
    time_channel = c(time = "hh:mm:ss"),
    verbose = FALSE
)

ggplot2::ggplot(data, ggplot2::aes(x = time)) +
    theme_mnirs() +
    scale_colour_mnirs(name = NULL) +
    ggplot2::geom_line(ggplot2::aes(y = smo2_left, colour = "smo2_left")) +
    ggplot2::geom_line(ggplot2::aes(y = smo2_right, colour = "smo2_right"))
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=theme_mnirs]{theme_mnirs()}}, \code{\link[=palette_mnirs]{palette_mnirs()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract_interval_helpers.R
\name{find_interval_time}
\alias{find_interval_time}
\title{resolve a single mnirs_interval object to time values}
\usage{
find_interval_time(
  interval,
  time_vec,
  event_vec = NULL,
  position = c("first", "last")
)
}
\description{
resolve a single mnirs_interval object to time values
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_mnirs_helpers.R
\name{is_empty}
\alias{is_empty}
\title{Detect empty or NA strings}
\usage{
is_empty(x)
}
\description{
Detect empty or NA strings
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract_interval_helpers.R
\name{apply_span}
\alias{apply_span}
\title{apply span to resolved times and build interval_spec data frame}
\usage{
apply_span(interval_list, time_vec, span, verbose = TRUE)
}
\description{
apply span to resolved times and build interval_spec data frame
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.mnirs.R
\name{theme_mnirs}
\alias{theme_mnirs}
\title{Custom \emph{{mnirs}} ggplot2 theme}
\usage{
theme_mnirs(
  base_size = 14,
  base_family = "sans",
  border = c("partial", "full"),
  ink = "black",
  paper = "white",
  accent = "#0080ff",
  ...
)
}
\arguments{
\item{base_size}{Base font size, given in pts.}

\item{base_family}{Base font family.}

\item{border}{Define either a \emph{partial} or \emph{full} border around plots.}

\item{ink}{Colour for text and lines. \emph{Default} is \emph{"black"}.}

\item{paper}{Background colour. \emph{Default} is \emph{"white"}.}

\item{accent}{Accent colour for highlights. \emph{Default} is \emph{"#0080ff"}.}

\item{...}{Additional arguments to add to \verb{[ggplot2::theme()]}.}
}
\value{
A \link[ggplot2:ggplot]{ggplot2} theme object.
}
\description{
A \verb{[ggplot2][ggplot2::ggplot2-package]} theme for display.
}
\details{
\itemize{
\item \code{axis.title = element_text(face = "bold")} by \emph{default} Modify to \emph{"plain"}.
\item \code{panel.grid.major} & \code{panel.grid.major} set to blank. Modify to
\verb{= element_line()} for visible grid lines.
\item \code{legend.position = "top"} by \emph{default} Modify \code{"none"} to remove legend
entirely.
\item \code{border = "partial"} uses \code{panel.border = element_blank()} and
\code{axis.line = element_line()}.
\item \code{border = "full"} uses \verb{panel.border = element_rect(colour = "black",}
\verb{linewidth = 1)} and \code{axis.line = element_line()}.
\item \code{base_family = "sans"} by \emph{default}.
}
}
\examples{
\dontshow{if (rlang::is_installed(c("ggplot2", "scales"))) withAutoprint(\{ # examplesIf}
## plot example data
read_mnirs(
    file_path = example_mnirs("moxy_ramp"),
    nirs_channels = c(smo2_left = "SmO2 Live",
                      smo2_right = "SmO2 Live(2)"),
    time_channel = c(time = "hh:mm:ss"),
    verbose = FALSE
) |>
    plot(time_labels = TRUE)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=palette_mnirs]{palette_mnirs()}}, \code{\link[=scale_colour_mnirs]{scale_colour_mnirs()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_mnirs.R
\name{create_mnirs_data}
\alias{create_mnirs_data}
\title{Create an \emph{{mnirs}} data frame with metadata}
\usage{
create_mnirs_data(data, ...)
}
\arguments{
\item{data}{A data frame with existing metadata (accessed with
\code{attributes(data)}).}

\item{...}{Additional arguments with metadata to add to the data frame.
Can be either seperate named arguments or a list of named values.
\itemize{
\item nirs_device
\item nirs_channels
\item time_channel
\item event_channel
\item sample_rate
\item start_timestamp
\item interval_times
\item interval_span
}}
}
\value{
A \link[tibble:tibble-package]{tibble} of class \code{"mnirs"}. Metadata are stored
as attributes and can be accessed with \code{attributes(data)}.
}
\description{
Manually add class \code{"mnirs"} and metadata to an existing data frame.
}
\details{
Typically will only be called internally, but can be used to inject
\emph{{mnirs}} metadata into any data frame.
}
\examples{
data <- data.frame(
    A = 1:3,
    B = seq(10, 30, 10),
    C = seq(11, 33, 11)
)

attributes(data)

## inject metadata
nirs_data <- create_mnirs_data(
    data,
    nirs_channels = c("B", "C"),
    time_channel = "A",
    sample_rate = 1
)

attributes(nirs_data)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter_mnirs.R
\name{filter_mnirs}
\alias{filter_mnirs}
\alias{filter_mnirs.smooth_spline}
\alias{filter_mnirs.butterworth}
\alias{filter_mnirs.moving_average}
\title{Filter a data frame}
\usage{
filter_mnirs(
  data,
  nirs_channels = NULL,
  time_channel = NULL,
  method = c("smooth_spline", "butterworth", "moving_average"),
  na.rm = FALSE,
  verbose = TRUE,
  ...
)
}
\arguments{
\item{data}{A data frame of class \emph{"mnirs"} containing time series data and
metadata.}

\item{nirs_channels}{A character vector giving the names of mNIRS columns to
operate on. Must match column names in \code{data} exactly.
\itemize{
\item If \code{NULL} (default), the \code{nirs_channels} metadata attribute of \code{data} is
used.
}}

\item{time_channel}{A character string naming the time or sample column.
Must match a column name in \code{data} exactly.
\itemize{
\item If \code{NULL} (default), the \code{time_channel} metadata attribute of \code{data} is
used.
}}

\item{method}{A character string indicating how to filter the data (see
\emph{Details}).
\describe{
\item{\code{"smooth_spline"}}{Fits a cubic smoothing spline.}
\item{\code{"butterworth"}}{Uses a centred Butterworth digital filter.}
\item{\code{"moving_average"}}{Uses a centred moving average filter.}
}}

\item{na.rm}{Logical; default is \code{FALSE}, propagates any \code{NA}s to the
returned vector. If \code{TRUE}, ignores \code{NA}s and processes available valid
samples within the local window. May return errors or warnings. (see
\emph{Details}).}

\item{verbose}{Logical. Default is \code{TRUE}. Display or silence (if \code{FALSE})
warnings and information messages helpful for troubleshooting. Ad
global default can be set via \code{options(mnirs.verbose = FALSE)}.}

\item{...}{Additional method-specific arguments must be specified
(see \emph{Details}).}
}
\value{
A \link[tibble:tibble-package]{tibble} of class \emph{"mnirs"} with metadata
available with \code{attributes()}.
}
\description{
Apply digital filtering/smoothing to numeric vector data within a data frame
using either:
\enumerate{
\item A cubic smoothing spline.
\item A Butterworth digital filter.
\item A simple moving average.
}
}
\details{
\subsection{method = "smooth_spline"}{

Aliases: \code{method = c("smooth spline", "spline")}

Applies a non-parametric cubic smoothing spline from
\code{\link[stats:smooth.spline]{stats::smooth.spline()}}. Smoothing is defined by the parameter \code{spar},
which can be left as \code{NULL} and automatically determined via penalised
log likelihood. This usually works well for responses occurring on the
order of minutes or longer. \code{spar} can be specified typically, but not
necessarily, in the range \verb{spar = [0, 1]}.

Additional arguments (\code{...}) accepted when \code{method = "smooth_spline"}:

\describe{
\item{\code{spar}}{A numeric smoothing parameter passed to
\code{\link[stats:smooth.spline]{stats::smooth.spline()}}. If \code{NULL} (\emph{default}), automatically
determined via penalised log likelihood.}
}
}

\subsection{method = "butterworth"}{

Aliases: \code{method = c("butter")}

Applies a centred (two-pass symmetrical) Butterworth digital filter
from \code{\link[signal:butter]{signal::butter()}} and \code{\link[signal:filtfilt]{signal::filtfilt()}}.

Filter \code{type} defines how the desired signal frequencies are either
passed or rejected from the output signal. \emph{Low-pass} and \emph{high-pass}
filters allow only frequencies \emph{lower} or \emph{higher} than the cutoff
frequency, respectively to be passed through to the output signal.
\emph{Stop-band} defines a critical range of frequencies which are rejected
from the output signal. \emph{Pass-band} defines a critical range of
frequencies which are passed through as the output signal.

The filter order (number of passes) is defined by \code{order}, typically
in the range \verb{order = [1, 10]}. Higher filter order tends to capture
more rapid changes in amplitude, but also causes more distortion
around those change points in the signal. General advice is to use
the lowest filter order which sufficiently captures the desired rapid
responses in the data.

The critical (cutoff) frequency can be defined by \code{W}, a numeric value
for \emph{low-pass} and \emph{high-pass} filters, or a two-element vector
\code{c(low, high)} defining the lower and upper bands for \emph{stop-band}
and \emph{pass-band} filters. \code{W} represents the desired fractional cutoff
frequency in the range \verb{W = [0, 1]}, where \code{1} is the Nyquist
frequency, i.e., half the \code{sample_rate} of the data in Hz.

Alternatively, the cutoff frequency can be defined by \code{fc} and
\code{sample_rate} together. \code{fc} represents the desired cutoff frequency
directly in Hz, and \code{sample_rate} is the sample rate of the recorded data
in Hz. Where \code{W = fc / (sample_rate / 2)}.

Only one of either \code{W} or \code{fc} should be defined. If both are
defined, \code{W} will be preferred over \code{fc}.

Additional arguments (\code{...}) accepted when \code{method = "butterworth"}:

\describe{
\item{\code{order}}{An integer for the filter order (\emph{default} \code{2}).}
\item{\code{W}}{A numeric fractional cutoff frequency within \verb{[0, 1]}. One
of either \code{W} or \code{fc} must be specified.}
\item{\code{fc}}{A numeric absolute cutoff frequency in Hz. Used with
\code{sample_rate} to compute \code{W}.}
\item{\code{sample_rate}}{A numeric sample rate in Hz. Will be taken from
metadata or estimated from \code{time_channel} if not defined.}
\item{\code{type}}{A character string specifying filter type, one of:
\code{c("low", "high", "stop", "pass")} (\code{"low"} is the default).}
\item{\code{edges}}{A character string specifying the edge padding, one of:
\code{c("rev", "rep1", "none")} (\code{"rev"} is the default).
See \code{\link[=filter_butter]{filter_butter()}}.}
}
}

\subsection{method = "moving_average"}{

Aliases: \code{method = c("moving average", "ma")}

Applies a centred (symmetrical) moving average filter in a local
window, defined by either \code{width} as the number of samples around
\code{idx} between \verb{[idx - floor(width/2), idx + floor(width/2)]}. Or by
\code{span} as the timespan in units of \code{time_channel} between
\verb{[t - span/2, t + span/2]}.

Additional arguments (\code{...}) accepted when \code{method = "moving_average"}:

\describe{
\item{\code{width} or \code{span}}{Either an integer number of samples, or a
numeric time duration in units of \code{time_channel} within the local
window. One of either \code{width} or \code{span} must be specified.}
\item{\code{partial}}{Logical; \code{FALSE} by default, only returns values
where a full window of valid (non-\code{NA}) samples are available.
If \code{TRUE}, ignores \code{NA} and allows calculation over partial windows
at the edges of the data.}
}
}

\subsection{Missing values}{

Missing values (\code{NA}) in \code{nirs_channels} will cause an error for
\code{method = "smooth_spline"} or \code{"butterworth"}, unless \code{na.rm = TRUE}.
Then \code{NA}s will be ignored and passed through to the returned data.

For \code{method = "moving_average"}, \code{na.rm} controls whether \code{NA}s within
each local window are either propagated to the returned vector when
\code{na.rm = FALSE} (the default), or ignored before processing if
\code{na.rm = TRUE}.
}
}
\examples{
## read example data and clean for outliers
data <- read_mnirs(
    file_path = example_mnirs("moxy_ramp"),
    nirs_channels = c(smo2 = "SmO2 Live"),
    time_channel = c(time = "hh:mm:ss"),
    verbose = FALSE
) |>
    replace_mnirs(
        invalid_values = c(0, 100),
        outlier_cutoff = 3,
        width = 7,
        verbose = FALSE
    )

data

data_filtered <- filter_mnirs(
    data,                   ## blank channels will be retrieved from metadata
    method = "butterworth", ## Butterworth digital filter is a common choice
    order = 2,              ## filter order number
    W = 0.02,               ## filter fractional critical frequency `[0, 1]`
    type = "low",           ## specify a "low-pass" filter
    na.rm = TRUE            ## explicitly ignore NAs
)

## note the smoothed `smo2` values
data_filtered

\donttest{
    if (requireNamespace("ggplot2", quietly = TRUE)) {
        ## plot filtered data and add the raw data back to the plot to compare
        plot(data_filtered, time_labels = TRUE) +
            ggplot2::geom_line(
                data = data,
                ggplot2::aes(y = smo2, colour = "smo2"), alpha = 0.4
            )
    }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{moxy_ramp.xlsx}
\alias{moxy_ramp.xlsx}
\title{2 Hz PerfPro export of Moxy data}
\format{
.xlsx file with five columns and 2202 rows:
\describe{
\item{mm-dd}{Recording date (dd-MMM format).}
\item{hh:mm:ss}{Time of day (hh:mm:ss format).}
\item{SmO2 Live}{Muscle oxygen saturation, left leg (\%). Contains
simulated erroneous and missing samples.}
\item{SmO2 Live(2)}{Muscle oxygen saturation, right leg (\%).}
\item{Lap}{Lap marker (integer).}
}

Channel mapping for \code{\link[=read_mnirs]{read_mnirs()}}:
\itemize{
\item \code{nirs_channels = c("SmO2 Live", "SmO2 Live(2)")}
\item \code{time_channel = c("hh:mm:ss")}
\item \code{event_channel = c("Lap")}
\item \code{interval_times = list(start = c(204, 878))} (start and end of exercise)
}
}
\source{
Moxy Monitor (Fortiori Design LLC), exported via PerfPro Studio
desktop software (https://perfprostudio.com/).
}
\description{
Exported from PerfPro Studio software, recorded
at 0.5 Hz no smoothing and exported at 2 Hz. Containing a
ramp incremental cycling protocol, placed on bilateral
vastus lateralis muscle sites. Intentional data errors
(outliers, invalid values, and missing \code{NA} values) have
been introduced to demonstrate \emph{{mnirs}} cleaning
functions.
}
\examples{
example_mnirs("moxy_ramp")

}
\seealso{
\code{\link[=read_mnirs]{read_mnirs()}}, \code{\link[=example_mnirs]{example_mnirs()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/replace_helpers.R
\name{compute_local_windows}
\alias{compute_local_windows}
\alias{compute_local_fun}
\alias{compute_outliers}
\alias{compute_valid_neighbours}
\title{Computes rolling local values}
\usage{
compute_local_windows(
  t,
  idx = seq_along(t),
  width = NULL,
  span = NULL,
  align = c("centre", "left", "right")
)

compute_local_fun(x, window_idx, fn, ...)

compute_outliers(x, window_idx, outlier_cutoff)

compute_valid_neighbours(
  x,
  t = seq_along(x),
  width = NULL,
  span = NULL,
  verbose = TRUE
)
}
\arguments{
\item{t}{An \emph{optional} numeric vector of the predictor variable (e.g. time).
Default is \code{seq_along(x)}.}

\item{idx}{A numeric vector of indices of \code{t} at which to calculate local
windows. All indices of \code{t} by \emph{default}, or can be used to only calculate
for known indices, such as invalid values of \code{x}.}

\item{width}{An integer defining the local window in number of samples
around \code{idx} in which to perform the operation, according to \code{align}.}

\item{span}{A numeric value defining the local window time span around \code{idx}
in which to perform the operation, according to \code{align}. In units of
\code{time_channel} or \code{t}.}

\item{align}{Window alignment as \emph{"centre"/"center"} (the \emph{default}),
\emph{"left"}, or \emph{"right"}. Where \emph{"left"} is \emph{forward looking}, and \emph{"right"}
is \emph{backward looking} from the current sample.}

\item{x}{A numeric vector of the response variable.}

\item{window_idx}{A list the same or shorter length as \code{x} with numeric
vectors for the sample indices of local rolling windows.}

\item{fn}{A function to pass through for local rolling calculation.}

\item{...}{Additional arguments.}

\item{outlier_cutoff}{A numeric value for the local outlier threshold, as
the number of standard deviations from the local median.
\itemize{
\item Default \code{NULL} will not replace outliers.
\item Lower values are more sensitive and flag more outliers; higher values
are more conservative.
\item \code{outlier_cutoff = 3} Pearson's 3 sigma edit rule.
\code{outlier_cutoff = 2} approximates a Tukey-style 1.5*IQR rule.
\code{outlier_cutoff = 0} Tukey's median filter.
}}

\item{verbose}{Logical. Default is \code{TRUE}. Display or silence (if \code{FALSE})
warnings and information messages helpful for troubleshooting. Ad
global default can be set via \code{options(mnirs.verbose = FALSE)}.}
}
\value{
\code{compute_local_windows()}: A list the same length as \code{idx} and the same or
shorter length as \code{t} with numeric vectors of sample indices of length
\code{width} samples or \code{span} units of time \code{t}.

\code{compute_local_fun()}: A numeric vector the same length as \code{x}.

\code{compute_outliers()}: A \code{list()} with vectors the same length as \code{x} for
with numeric local medians and logical identifying where \code{is_outlier}.

\code{compute_valid_neighbours()}: A list the same length as the \code{NA} values in
\code{x} with numeric vectors of sample indices of length \code{width} samples or
\code{span} units of time \code{t} for valid values neighbouring split to either
side of the invalid \code{NA}s.
}
\description{
\code{compute_local_windows()}: Compute a list of rolling window indices along a
time variable \code{t}.

\code{compute_local_fun()}: Compute a rolling function along \code{x} from a list of
rolling sample windows.

\code{compute_outliers()}: Computes a vector of local medians and logicals
indicating outliers of \code{x} within a list of rolling sample windows
\code{window_idx}.

\code{compute_valid_neighbours()}: Compute a list of rolling window indices along
\code{x} to either side of \code{NA}s.
}
\details{
The local rolling window can be specified by either \code{width} as the number of
samples, or \code{span} as the time span in units of \code{t}. Specifying \code{width}
is often faster than \code{span}.

\code{align} defaults to \emph{"centre"} the local window around \code{idx} between
\verb{[idx - floor((width-1)/2),} \verb{idx + floor(width/2)]} when \code{width} is
specified. Even \code{width} values will bias \code{align} to \emph{"left"}, with the
unequal sample forward of \code{idx}, effectively returning \code{NA} at the last
sample index. When \code{span} is specified, the local window is between
\verb{[t - span/2, t + span/2]}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter_mnirs.R
\name{filter_ma}
\alias{filter_ma}
\alias{filter_moving_average}
\title{Apply a moving average filter}
\usage{
filter_ma(
  x,
  t = seq_along(x),
  width = NULL,
  span = NULL,
  partial = FALSE,
  na.rm = FALSE,
  verbose = TRUE,
  ...
)

filter_moving_average(
  x,
  t = seq_along(x),
  width = NULL,
  span = NULL,
  partial = FALSE,
  na.rm = FALSE,
  verbose = TRUE,
  ...
)
}
\arguments{
\item{x}{A numeric vector of the response variable.}

\item{t}{An \emph{optional} numeric vector of the predictor variable (e.g. time).
Default is \code{seq_along(x)}.}

\item{width}{An integer defining the local window in number of samples
centred on \code{idx}, between
\verb{[idx - floor(width/2), idx + floor(width/2)]}.}

\item{span}{A numeric value defining the local window time span around
\code{idx} in units of \code{time_channel} or \code{t}, between
\verb{[t - span/2, t + span/2]}.}

\item{partial}{Logical; default is \code{FALSE}, requires local windows to have
complete number of samples specified by \code{width} or \code{span}. If \code{TRUE},
processes available samples within the local window. See \emph{Details}.}

\item{na.rm}{Logical; default is \code{FALSE}, propagates any \code{NA}s to the
returned vector. If \code{TRUE}, ignores \code{NA}s and processes available valid
samples within the local window. May return errors or warnings. (see
\emph{Details}).}

\item{verbose}{Logical. Default is \code{TRUE}. Display or silence (if \code{FALSE})
warnings and information messages helpful for troubleshooting. Ad
global default can be set via \code{options(mnirs.verbose = FALSE)}.}

\item{...}{Additional arguments.}
}
\value{
A numeric vector the same length as \code{x}.
}
\description{
Apply a simple moving average smoothing filter to vector data.
\code{filter_moving_average()} is an alias of \code{filter_ma()}.
}
\details{
\subsection{Rolling window}{

Applies a centred (symmetrical) moving average filter in a local
window, defined by either \code{width} as the number of samples around
\code{idx} between \verb{[idx - floor(width/2), idx + floor(width/2)]}. Or by
\code{span} as the timespan in units of \code{time_channel} between
\verb{[t - span/2, t + span/2]}.
}

\subsection{Partial windows}{

The default \code{partial = FALSE} requires a complete number of samples
specified by \code{width} or \code{span} (estimated from the sample rate of \code{t} when
\code{span} is used). \code{NA} is returned if fewer samples are present in the
local window.

Setting \code{partial = TRUE} allows computation with only a single valid sample,
such as at edge conditions. But these values will be more sensitive to
noise and should be used with caution.
}

\subsection{Missing values}{

\code{na.rm} controls whether missing values (\code{NA}s) within each local window are
either propagated to the returned vector when \code{na.rm = FALSE} (the default),
or ignored before processing if \code{na.rm = TRUE}.
}
}
\examples{
x <- c(1, 3, 2, 5, 4, 6, 5, 7)
t <- c(0, 1, 2, 4, 5, 6, 7, 10)  ## irregular time with gaps

## width: centred window of 3 samples
filter_ma(x, width = 3)

## partial = TRUE fills edge values with a narrower window
filter_ma(x, width = 3, partial = TRUE)

## span: centred window of 2 time-units (accounts for irregular sampling)
filter_ma(x, t, span = 2)

## na.rm = FALSE (default): any NA in the window propagates to the result
x_na <- c(1, NA, 3, 4, 5, NA, 7, 8)
filter_ma(x_na, width = 3, na.rm = FALSE)

## na.rm = TRUE: skip NAs and return the local mean of local valid values
filter_ma(x_na, width = 3, partial = TRUE, na.rm = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract_interval_helpers.R
\name{resolve_interval}
\alias{resolve_interval}
\title{resolve start/end into time value vectors (no span applied)}
\usage{
resolve_interval(start, end, time_vec, event_vec = NULL)
}
\description{
resolve start/end into time value vectors (no span applied)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/replace_mnirs.R
\name{replace_mnirs}
\alias{replace_mnirs}
\alias{replace_invalid}
\alias{replace_outliers}
\alias{replace_missing}
\title{Replace outliers, invalid, and missing values in \emph{{mnirs}} data}
\usage{
replace_mnirs(
  data,
  nirs_channels = NULL,
  time_channel = NULL,
  invalid_values = NULL,
  invalid_above = NULL,
  invalid_below = NULL,
  outlier_cutoff = NULL,
  width = NULL,
  span = NULL,
  method = c("linear", "median", "locf", "none"),
  verbose = TRUE
)

replace_invalid(
  x,
  t = seq_along(x),
  invalid_values = NULL,
  invalid_above = NULL,
  invalid_below = NULL,
  width = NULL,
  span = NULL,
  method = c("median", "none"),
  verbose = TRUE,
  ...
)

replace_outliers(
  x,
  t = seq_along(x),
  outlier_cutoff = 3,
  width = NULL,
  span = NULL,
  method = c("median", "none"),
  verbose = TRUE,
  ...
)

replace_missing(
  x,
  t = seq_along(x),
  width = NULL,
  span = NULL,
  method = c("linear", "median", "locf"),
  verbose = TRUE,
  ...
)
}
\arguments{
\item{data}{A data frame of class \emph{"mnirs"} containing time series data and
metadata.}

\item{nirs_channels}{A character vector giving the names of mNIRS columns to
operate on. Must match column names in \code{data} exactly.
\itemize{
\item If \code{NULL} (default), the \code{nirs_channels} metadata attribute of \code{data} is
used.
}}

\item{time_channel}{A character string naming the time or sample column.
Must match a column name in \code{data} exactly.
\itemize{
\item If \code{NULL} (default), the \code{time_channel} metadata attribute of \code{data} is
used.
}}

\item{invalid_values}{A numeric vector of invalid values to be replaced,
e.g. \code{invalid_values = c(0, 100, 102.3)}. Default \code{NULL} will not replace
invalid values.}

\item{invalid_above, invalid_below}{Numeric values each specifying cutoff
values, above or below which (respectively) will be replaced, \emph{inclusive}
of the specified cutoff values.}

\item{outlier_cutoff}{A numeric value for the local outlier threshold, as
the number of standard deviations from the local median.
\itemize{
\item Default \code{NULL} will not replace outliers.
\item Lower values are more sensitive and flag more outliers; higher values
are more conservative.
\item \code{outlier_cutoff = 3} Pearson's 3 sigma edit rule.
\code{outlier_cutoff = 2} approximates a Tukey-style 1.5*IQR rule.
\code{outlier_cutoff = 0} Tukey's median filter.
}}

\item{width}{An integer defining the local window in number of samples
centred on \code{idx}, between
\verb{[idx - floor(width/2), idx + floor(width/2)]}.}

\item{span}{A numeric value defining the local window time span around
\code{idx} in units of \code{time_channel} or \code{t}, between
\verb{[t - span/2, t + span/2]}.}

\item{method}{A character string indicating how to handle \code{NA}
replacement (see \emph{Details}):
\describe{
\item{\code{"linear"}}{Replaces \code{NA}s via linear interpolation (the
\emph{default}) using \code{\link[stats:approxfun]{stats::approx()}}.}
\item{\code{"median"}}{Replaces \code{NA}s with the local median of valid
values within a centred window defined by \code{width} or \code{span}.}
\item{\code{"locf"}}{\emph{"Last observation carried forward"}. Replaces
\code{NA}s with the most recent valid value to the left for trailing
\code{NA}s or to the right for leading \code{NA}s, using
\code{\link[stats:approxfun]{stats::approx()}}.}
\item{\code{"none"}}{Returns \code{NA}s without replacement.}
}}

\item{verbose}{Logical. Default is \code{TRUE}. Display or silence (if \code{FALSE})
warnings and information messages helpful for troubleshooting. Ad
global default can be set via \code{options(mnirs.verbose = FALSE)}.}

\item{x}{A numeric vector of the response variable.}

\item{t}{An \emph{optional} numeric vector of the predictor variable (e.g. time).
Default is \code{seq_along(x)}.}

\item{...}{Additional arguments.}
}
\value{
\code{replace_mnirs()} return a \link[tibble:tibble-package]{tibble} of
class \code{"mnirs"} with metadata available via \code{attributes()}.

\code{replace_invalid()} returns a numeric vector the same length as
\code{x} with invalid values replaced.

\code{replace_outliers()} returns a numeric vector the same length as
\code{x} with outliers replaced.

\code{replace_missing()} returns a numeric vector the same length as
\code{x} with missing values replaced.
}
\description{
Detect and replace local outliers, specified invalid values, and missing
\code{NA} values across \code{nirs_channels} within an \emph{"mnirs"} data frame.
\code{replace_mnirs()} operates on a data frame, extending the vectorised
functions:.

\code{replace_invalid()} detects specified invalid values or range cutoffs in a
numeric vector and replace them with the local median value or \code{NA}.

\code{replace_outliers()} detects local outliers in a numeric vector using a Hampel filter and replaces with the local median value or \code{NA}.

\code{replace_missing()} detects missing (\code{NA}) values in a numeric vector and
replaces via interpolation.
}
\details{
\subsection{Automatic channel detection}{

\code{nirs_channels} and \code{time_channel} are retrieved automatically from
\emph{"mnirs"} metadata if not specified explicitly. Columns in \code{data} not
listed in \code{nirs_channels} are passed through unprocessed.
}

\subsection{The rolling window}{

\code{replace_outliers()} and \code{replace_missing()} (when \code{method = "median"})
operate over a local rolling window for outlier detection and median
interpolation. The window is specified by either \code{width} as the number
of samples, or \code{span} as the time span in units of \code{time_channel}.
A partial window is calculated at the edges of the data.
}

\subsection{Replace invalid values with with replace_invalid()}{

Specific \code{invalid_values} can be replaced, such as \code{c(0, 100, 102.3)}.
Data ranges can be replaced with cutoff values specified by \code{invalid_above}
and \code{invalid_below}, where any values higher or lower than the specified
cutoff values (respectively) will be replaced, \emph{inclusive} of the cutoff
values themselves.
}

\subsection{Outlier detection with replace_outliers()}{

Rolling local medians are computed across \code{x} within a window defined
by \code{width} (number of samples) or \code{span} (time span in units of \code{t}).

Outliers are detected with robust median absolute deviation (MAD),
adapted from \code{pracma::hampel()}. Deviations equal to or less than the
smallest absolute time series difference in \code{x} are excluded, to avoid
flagging negligible differences where local data have minimal or zero
variation.
}

\subsection{Replacement behaviour}{

Values of \code{x} outside the local bounds defined by \code{outlier_cutoff} are
identified as outliers and either replaced with the local median
(\code{method = "median"}, the \emph{default}) or set to \code{NA} (\code{method = "none"}).

Existing \code{NA} values in \code{x} are \emph{not} replaced. They are passed
through to the returned vector. See \code{\link[=replace_missing]{replace_missing()}}.
}

\subsection{Choosing outlier_cutoff}{

\code{outlier_cutoff} is the number of (MAD-normalised) standard deviations
from the local median. Higher values are more conservative; lower
values flag more outliers.
\itemize{
\item \code{outlier_cutoff = 3} -- Pearson's 3 sigma edit rule (default).
\item \code{outlier_cutoff = 2} -- approximately Tukey-style 1.5*IQR rule.
\item \code{outlier_cutoff = 0} -- Tukey's median filter (every point
replaced by local median).
}
}

\subsection{Interpolation with replace_missing()}{

\code{method = "linear"} and \code{method = "locf"} use \code{\link[stats:approxfun]{stats::approx()}} with
\code{rule = 2}, so leading \code{NA}s are filled by \emph{"nocb"}
(\emph{"next observation carried backward"}) and trailing \code{NA}s by \emph{"locf"}.

\code{method = "median"} calculates the local median of valid (non-\code{NA})
values to either side of \code{NA}s, within a window defined by \code{width}
(number of samples) or \code{span} (time span in units of \code{t}). Sequential
\code{NA}s are all replaced by the same median value.
}

\subsection{Edge behaviour for method = "median"}{

If there are no valid values within \code{span} to one side of the \code{NA},
the median of the other side is used (i.e. for leading and trailing
\code{NA}s). If there are no valid values within either side, the first
valid sample on either side is used (equivalent to
\code{replace_missing(x, width = 1)}).
}
}
\examples{
## vectorised operations
x <- c(1, 999, 3, 4, 999, 6)
replace_invalid(x, invalid_values = 999, width = 3, method = "median")

(x_na <- replace_outliers(x, outlier_cutoff = 3, width = 3, method = "none"))

replace_missing(x_na, method = "linear")

## read example data
data <- read_mnirs(
    file_path = example_mnirs("moxy_ramp"),
    nirs_channels = c(smo2 = "SmO2 Live"),
    time_channel = c(time = "hh:mm:ss"),
    verbose = FALSE
)

## clean data
data_clean <- replace_mnirs(
    data,                  ## channels retrieved from metadata
    invalid_values = 0,    ## known invalid values in the data
    invalid_above = 90,    ## remove data spikes above 90
    outlier_cutoff = 3,    ## Pearson's 3 sigma edit rule
    width = 7,            ## window for outlier detection and interpolation
    method = "linear"      ## linear interpolation over NAs
)

\donttest{
  if (requireNamespace("ggplot2", quietly = TRUE)) {
    ## plot original and show where values have been replaced
    ## ignore warning about replacing the existing colour scale
    plot(data, time_labels = TRUE) +
      ggplot2::scale_colour_manual(
        name = NULL,
        breaks = c("smo2", "replaced"),
        values = palette_mnirs(2)
      ) +
      ggplot2::geom_point(
        data = data[data_clean$smo2 != data$smo2, ],
        ggplot2::aes(y = smo2, colour = "replaced"),
        na.rm = TRUE
      ) +
      ggplot2::geom_line(
        data = {
          data_clean[!is.na(data$smo2), "smo2"] <- NA
          data_clean
        },
        ggplot2::aes(y = smo2, colour = "replaced"),
        linewidth = 1, na.rm = TRUE
      )
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shift_mnirs.R
\name{shift_mnirs}
\alias{shift_mnirs}
\title{Shift data range}
\usage{
shift_mnirs(
  data,
  nirs_channels = list(),
  time_channel = NULL,
  to = NULL,
  by = NULL,
  width = NULL,
  span = NULL,
  position = c("min", "max", "first"),
  verbose = TRUE
)
}
\arguments{
\item{data}{A data frame of class \emph{"mnirs"} containing time series data and
metadata.}

\item{nirs_channels}{A character vector giving the names of mNIRS columns to
operate on. Must match column names in \code{data} exactly.
\itemize{
\item If \code{NULL} (default), the \code{nirs_channels} metadata attribute of \code{data} is
used.
}}

\item{time_channel}{A character string naming the time or sample column.
Must match a column name in \code{data} exactly.
\itemize{
\item If \code{NULL} (default), the \code{time_channel} metadata attribute of \code{data} is
used.
}}

\item{to}{A numeric value in units of \code{nirs_channels} to which the data
channels will be shifted, e.g. shift the minimum value to zero.}

\item{by}{A numeric value in units of \code{nirs_channels} by which the data
channels will be shifted, e.g. shift all values up by 10 units.}

\item{width}{An integer defining the local window in number of samples
centred on \code{idx}, between
\verb{[idx - floor(width/2), idx + floor(width/2)]}.}

\item{span}{A numeric value defining the local window time span around
\code{idx} in units of \code{time_channel} or \code{t}, between
\verb{[t - span/2, t + span/2]}.}

\item{position}{Indicates where the reference values will be shifted from.
\describe{
\item{\code{"min"}}{(The \emph{default}) will shift the minimum value(s) \code{to}
or \code{by} the specified value.}
\item{\code{"max"}}{Will shift the maximum value(s) \code{to} or \code{by} the
specified values.}
\item{\code{"first"}}{Will shift first value(s) \code{to} or \code{by} the specified
values.}
}}

\item{verbose}{Logical. Default is \code{TRUE}. Display or silence (if \code{FALSE})
warnings and information messages helpful for troubleshooting. Ad
global default can be set via \code{options(mnirs.verbose = FALSE)}.}
}
\value{
A \link[tibble:tibble-package]{tibble} of class \emph{"mnirs"} with metadata
available with \code{attributes()}.
}
\description{
Move the range of data channels in a data frame up or down, while preserving
the absolute amplitude/dynamic range of each channel, and the relative
scaling across channels. e.g. shift the minimum data value to zero for all
positive values, or shift the mean of the first time span in a recording
to zero.
}
\details{
\code{nirs_channels = list()} can be used to group data channels (column names)
to preserve absolute or relative scaling.
\itemize{
\item Channels grouped together in a vector (e.g. \code{list(c("A", "B"))}) will be
shifted to a common value, and the relative scaling within that group
will be preserved.
\item Channels in separate list vectors (e.g. \code{list("A", "B")}) will be
shifted independently, and relative scaling between groups will be lost.
\item A single vector of channel names (e.g. \code{c("A", "B")}) will group
channels together.
\item Channels (columns) in \code{data} not explicitly defined in \code{nirs_channels}
will be passed through untouched to the output data frame.
}

\code{nirs_channels} and \code{time_channel} can be retrieved automatically from
\code{data} of class \emph{"mnirs"} which has been processed with \code{{mnirs}},
if not defined explicitly. This will default to returning all
\code{nirs_channels} grouped together, and should be defined explicitly
for other grouping arrangements.

Only one of either \code{to} or \code{by} and one of either \code{width} or \code{span} should
be defined. If both of either pairing are defined, \code{to} will be preferred
over \code{by}, and \code{width} will be preferred over \code{span}.
}
\examples{
## read example data
data <- read_mnirs(
    file_path = example_mnirs("moxy_ramp"),
    nirs_channels = c(smo2_left = "SmO2 Live",
                      smo2_right = "SmO2 Live(2)"),
    time_channel = c(time = "hh:mm:ss"),
    verbose = FALSE
) |>
    shift_mnirs(        ## un-grouped nirs channels to shift separately 
        nirs_channels = list(smo2_left, smo2_right), 
        to = 0,         ## NIRS values will be shifted to zero
        span = 120,     ## shift the *first* 120 sec of data to zero
        position = "first"
    )

data

\donttest{
    if (requireNamespace("ggplot2", quietly = TRUE)) {
        plot(data, time_labels = TRUE) +
            ggplot2::geom_hline(yintercept = 0, linetype = "dotted")
    }
}  

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{artinis_intervals.xlsx}
\alias{artinis_intervals.xlsx}
\title{10 Hz Artinis Oxysoft export recorded with Oxymon MKIII}
\format{
.xlsx file with header metadata and five columns and 20919 rows:
\describe{
\item{Column 1}{Sample index (divide by sample rate for seconds).}
\item{Column 2}{O2Hb: oxyhaemoglobin concentration change (\eqn{\mu}M).}
\item{Column 3}{HHb: deoxyhaemoglobin concentration change (\eqn{\mu}M).}
\item{Column 4}{Event marker (character).}
\item{Column 5}{Unmarked event label (character).}
}

Channel mapping for \code{\link[=read_mnirs]{read_mnirs()}}:
\itemize{
\item \code{nirs_channels = c(O2Hb = 2, HHb = 3)}
\item \code{time_channel = c(sample = 1)}
\item \code{event_channel = c(event = 4, label = "col_5")}
\item \verb{interval_times = list(start = c(158, 999, 1750) end = c(493, 1333, 1961))} two intervals and post-exercise occlusion
}
}
\source{
Artinis Medical Systems. Oxymon MKIII, exported via Oxysoft desktop
software (https://artinis.com/)
}
\description{
Exported from Artinis Oxysoft, recorded on Oxymon MKIII at
50 Hz and exported at 10 Hz. Containing two 5-minute cycling work
intervals and an ischaemic occlusion, placed on the vastus lateralis
muscle site.
}
\examples{
example_mnirs("artinis_intervals")

}
\seealso{
\code{\link[=read_mnirs]{read_mnirs()}}, \code{\link[=example_mnirs]{example_mnirs()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract_interval_helpers.R
\name{as_mnirs_interval}
\alias{as_mnirs_interval}
\title{coerce raw values to mnirs_interval objects}
\usage{
as_mnirs_interval(x, arg = "start")
}
\arguments{
\item{x}{A raw value or mnirs_interval object.}

\item{arg}{Name of the argument for error messages.}
}
\description{
coerce raw values to mnirs_interval objects
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validate_mnirs.R
\name{abort_validation}
\alias{abort_validation}
\title{validate_numeric abort message construction}
\usage{
abort_validation(name, integer = FALSE, msg1 = "", msg2 = "")
}
\description{
validate_numeric abort message construction
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_mnirs.R
\name{example_mnirs}
\alias{example_mnirs}
\title{Get path to \emph{{mnirs}} example files}
\usage{
example_mnirs(file = NULL)
}
\arguments{
\item{file}{Name of file as character string. If \code{NULL}, returns a vector
of all available file names.}
}
\value{
A file path character string for selected example files stored in this
package.
}
\description{
Get path to \emph{{mnirs}} example files
}
\examples{
## lists all files
example_mnirs()

## partial matching will error if matches multiple
try(example_mnirs("moxy"))

example_mnirs("moxy_ramp")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_mnirs_helpers.R
\name{rename_duplicates}
\alias{rename_duplicates}
\title{Rename duplicate strings in a vector with \code{make.unique()}}
\usage{
rename_duplicates(x)
}
\description{
Rename duplicate strings in a vector with \code{make.unique()}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mnirs-package.R
\docType{package}
\name{mnirs-package}
\alias{mnirs}
\alias{mnirs-package}
\title{mnirs: Muscle Near-Infrared Spectroscopy Processing and Analysis}
\description{
Read, process, and analyse data from muscle near-infrared spectroscopy (mNIRS) devices. Import raw data from .csv or .xls(x) files and return time-series data and metadata. Includes standardised methods for cleaning, filtering, and pre-processing mNIRS data for subsequent analysis. Also includes a custom plot theme and colour palette. Intended for mNIRS researchers and practitioners in exercise physiology, sports science, and clinical rehabilitation with minimal coding experience required.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://jemarnold.github.io/mnirs/}
  \item \url{https://github.com/jemarnold/mnirs}
  \item Report bugs at \url{https://github.com/jemarnold/mnirs/issues}
}

}
\author{
\strong{Maintainer}: Jem Arnold \email{jem.arnold@gmail.com} (\href{https://orcid.org/0000-0003-3908-9447}{ORCID}) [copyright holder]

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract_interval_helpers.R
\name{preserve_metadata}
\alias{preserve_metadata}
\title{Zero-offset time values and add metadata}
\usage{
preserve_metadata(data, metadata, zero_time = FALSE)
}
\description{
Zero-offset time values and add metadata
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_mnirs_helpers.R
\docType{data}
\name{datetime_formats}
\alias{datetime_formats}
\title{Datetime format strings for POSIXct parsing}
\format{
An object of class \code{character} of length 7.
}
\usage{
datetime_formats
}
\description{
Datetime format strings for POSIXct parsing
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_mnirs_helpers.R
\name{parse_time_channel}
\alias{parse_time_channel}
\title{Parse time_channel character or dttm to numeric}
\usage{
parse_time_channel(
  data,
  time_channel,
  start_timestamp = NULL,
  add_timestamp = FALSE,
  zero_time = FALSE
)
}
\description{
Parse time_channel character or dttm to numeric
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.mnirs.R
\name{as_plot_data}
\alias{as_plot_data}
\title{Validate and bind a list of mnirs data frames for plotting}
\usage{
as_plot_data(x)
}
\description{
Validate and bind a list of mnirs data frames for plotting
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_mnirs_helpers.R
\name{parse_sample_rate}
\alias{parse_sample_rate}
\title{Validate and Estimate Sample Rate}
\usage{
parse_sample_rate(
  data,
  file_header,
  time_channel,
  sample_rate = NULL,
  nirs_device = NULL,
  verbose = TRUE
)
}
\description{
Validate and Estimate Sample Rate
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract_interval_helpers.R
\name{recycle_param}
\alias{recycle_param}
\title{Recycle parameter to match number of events}
\usage{
recycle_param(param, n_events, event_groups, verbose = TRUE)
}
\description{
Recycle an argument vector to a list or repeat the last list item to match
the number of events.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validate_mnirs.R
\name{within}
\alias{within}
\title{Detect if numeric values fall within range of a vector}
\usage{
within(x, vec, inclusive = c("left", "right"))
}
\arguments{
\item{x}{A numeric vector.}

\item{vec}{A numeric vector from which \code{left} and \code{right} boundary values
for \code{x} will be taken.}

\item{inclusive}{A character vector to specify which of \code{left} and/or
\code{right} boundary values should be included in the range, or both (the
default), or excluded if \code{FALSE}.}
}
\value{
A logical vector the same length as \code{x}.
}
\description{
Vectorised check for \code{x \%in\% vec}, inclusive or exclusive of left and right
boundary values, specified independently.
}
\details{
\code{inclusive = FALSE} can be used to test for positive non-zero values:
\code{within(x, c(0, Inf), inclusive = FALSE)}.
}
\seealso{
\code{\link[dplyr:between]{dplyr::between()}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_mnirs_helpers.R
\name{name_channels}
\alias{name_channels}
\title{Force names on character strings}
\usage{
name_channels(x)
}
\description{
Force names on character strings
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_mnirs_helpers.R
\name{read_data_table}
\alias{read_data_table}
\title{Read data table from raw data}
\usage{
read_data_table(data, nirs_channels = NULL, header_row = 1L)
}
\description{
Read data table from raw data
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract_interval_helpers.R
\name{recycle_to_length}
\alias{recycle_to_length}
\title{Recycle parameter list to target length}
\usage{
recycle_to_length(param, n, name = c("event", "group"), verbose = TRUE)
}
\description{
Recycle parameter list to target length
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_mnirs_helpers.R
\name{convert_type}
\alias{convert_type}
\title{Standardise comma decimals to periods in character columns}
\usage{
convert_type(data, time_channel, event_channel = NULL, verbose = TRUE)
}
\description{
Standardise comma decimals to periods in character columns
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validate_mnirs.R
\name{make_list}
\alias{make_list}
\title{Validate if an item is a list}
\usage{
make_list(x)
}
\description{
Validate if an item is a list
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_mnirs.R
\name{read_mnirs}
\alias{read_mnirs}
\title{Read \emph{{mnirs}} data from file}
\usage{
read_mnirs(
  file_path,
  nirs_channels = NULL,
  time_channel = NULL,
  event_channel = NULL,
  sample_rate = NULL,
  add_timestamp = FALSE,
  zero_time = FALSE,
  keep_all = FALSE,
  verbose = TRUE
)
}
\arguments{
\item{file_path}{Path of the data file to import. Supported file extensions
are \code{".xlsx"}, \code{".xls"}, and \code{".csv"}.}

\item{nirs_channels}{A character vector of one or more column names
containing mNIRS signals to import. Names must match the file header
exactly.
\itemize{
\item If \code{NULL} (default), \code{read_mnirs()} attempts to detect the device from
the file contents and use a known \code{nirs_channel} name.
\item A \emph{named} character vector can be used to rename columns on import, in
the form \code{c(renamed = "original_name")}.
}}

\item{time_channel}{A character string giving the name of the time
(or sample) column to import. The name must match the file header exactly.
\itemize{
\item If \code{NULL} (default), \code{read_mnirs()} attempts to identify a time-like
column automatically (by known device defaults and/or time-formatted
values).
\item A \emph{named} character vector can be used to rename the column on import,
in the form \code{c(time = "original_name")}.
}}

\item{event_channel}{An \emph{optional} character string giving the name of an
event/lap column to import. Names must match the file header exactly.
A named character vector can be used to rename the column on import in
the form \code{c(event = "original_name")}.}

\item{sample_rate}{An \emph{optional} numeric sample rate in Hz. If left blank
(\code{NULL}), the sample rate is estimated from \code{time_channel} (see \emph{Details}).}

\item{add_timestamp}{A logical. Default is \code{FALSE}. If \code{TRUE} and if the
source data contain an absolute date-time (POSIXct) time value, will add
a \code{"timestamp"} column in addition to the specified \code{time_channel} as a
numeric time column.}

\item{zero_time}{Logical. Default is \code{FALSE}. If \code{TRUE}, re-calculates
numeric \code{time_channel} values to start from zero.}

\item{keep_all}{Logical. Default is \code{FALSE}. Will keep only the channels
explicitly specified in \code{nirs_channels}, \code{time_channel}, and
\code{event_channel}. If \code{TRUE} will keep all columns found in the file
data table.
\itemize{
\item If no \code{nirs_channels} are specified and the file format is recognised,
all columns in the file data table will be returned, as an exploratory
option.
}}

\item{verbose}{Logical. Default is \code{TRUE}. Display or silence (if \code{FALSE})
warnings and information messages helpful for troubleshooting. Ad
global default can be set via \code{options(mnirs.verbose = FALSE)}.}
}
\value{
A \link[tibble:tibble-package]{tibble} of class \code{"mnirs"}. Metadata are stored
as attributes and can be accessed with \code{attributes(data)}.
}
\description{
Import time-series data exported from common muscle NIRS (mNIRS) devices and
return a tibble of class \code{"mnirs"} with the selected signal channels and
metadata.
}
\details{
\subsection{Header detection}{

\code{read_mnirs()} searches the file for a header row containing the requested
channel names. The header row does not need to be the first row in the file.
\itemize{
\item If duplicate column names exist, columns are matched in the order they
appear and renamed with unique strings.
\item Columns without a header name in the source file will be renamed to
\verb{col_*}, where \code{*} is the numeric column number in which they appear in
the file (e.g. \code{col_6}). This applies to \emph{Artinis Oxysoft} event label
columns, which do not have a column header and must be identified manually.
}
}

\subsection{Renaming channels}{

A named character vector can be specified to rename \code{nirs_channels},
\code{time_channel}, and \code{event_channel}, in the form
\code{c(renamed = "original_name")}. The \code{"original_name"} must match the
contents of the file data table header row exactly.
}

\subsection{Time parsing}{

\code{time_channel} will be converted to numeric for analysis.
\itemize{
\item If \code{time_channel} is a date-time (POSIXct) format, it will be converted
to numeric and re-based to start from 0, regardless of \code{zero_time}.
\item Some devices export a sample index rather than time values. In those
cases, if an export \code{sample_rate} is detected in the file metadata (e.g.
\emph{Artinis Oxysoft} exports), \code{read_mnirs()} will create or overwrite a
\code{"time"} column in seconds derived from the sample index and the detected
\code{sample_rate}.
}
}

\subsection{Sample rate}{

If \code{sample_rate} is not specified, it is estimated from differences in
\code{time_channel}. If \code{time_channel} is actually a sample index, as described
above, this may erroneously be estimated at 1 Hz. \code{sample_rate} should be
specified explicitly in this case.
}

\subsection{Data cleaning}{

Entirely empty rows and columns are removed. Invalid values (e.g.
\code{c(NaN, Inf)}) are standardized to \code{NA}. A warning will be displayed when
irregular sampling is detected (e.g. non-monotonic, repeated, or unequal
time values), if \code{verbose = TRUE}.
}
}
\examples{
read_mnirs(
    file_path = example_mnirs("moxy_ramp"), ## call an example data file
    nirs_channels = c(
        smo2_left = "SmO2 Live",            ## identify and rename channels
        smo2_right = "SmO2 Live(2)"
    ),
    time_channel = c(time = "hh:mm:ss"),    ## date-time format will be converted to numeric
    event_channel = NULL,                   ## leave blank if unused
    sample_rate = NULL,                     ## if blank, will be estimated from time_channel
    add_timestamp = FALSE,                  ## omit a date-time timestamp column
    zero_time = TRUE,                       ## recalculate time values from zero
    keep_all = FALSE,                       ## return only the specified data channels
    verbose = TRUE                          ## show warnings & messages
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validate_mnirs.R
\name{parse_channel_name}
\alias{parse_channel_name}
\title{Parse channel expressions for NSE}
\usage{
parse_channel_name(channel, data, env = rlang::caller_env())
}
\arguments{
\item{channel}{A quosure from \code{rlang::enquo()}.}

\item{data}{A data frame for tidyselect context.}

\item{env}{Environment for symbol evaluation.}
}
\value{
A character vector, list of character vectors, or \code{NULL}.
}
\description{
Converts quosures to character vectors, handling bare symbols, character
strings, lists, and tidyselect expressions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.mnirs.R
\name{palette_mnirs}
\alias{palette_mnirs}
\title{Custom \emph{{mnirs}} colour palette}
\usage{
palette_mnirs(...)
}
\arguments{
\item{...}{Either a single numeric specifying the number of colours to
return, or character strings specifying colour names. If empty, all
colours are returned.}
}
\value{
Named (when selecting by name) or unnamed character vector of
hex colours.
}
\description{
Custom \emph{{mnirs}} colour palette
}
\examples{
\dontshow{if (rlang::is_installed("scales")) withAutoprint(\{ # examplesIf}
scales::show_col(palette_mnirs())
scales::show_col(palette_mnirs(2))
scales::show_col(palette_mnirs("red", "orange"))
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=theme_mnirs]{theme_mnirs()}}, \code{\link[=scale_colour_mnirs]{scale_colour_mnirs()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_mnirs.R
\docType{data}
\name{mnirs_metadata}
\alias{mnirs_metadata}
\title{Metadata names  of class \code{"mnirs"}, retrieved with \code{attr()}}
\format{
An object of class \code{character} of length 8.
}
\usage{
mnirs_metadata
}
\description{
Metadata names  of class \code{"mnirs"}, retrieved with \code{attr()}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract_interval_helpers.R
\name{by_time}
\alias{by_time}
\alias{by_label}
\alias{by_lap}
\alias{by_sample}
\title{Specify interval boundaries by time, label, lap, or sample}
\usage{
by_time(...)

by_label(...)

by_lap(...)

by_sample(...)
}
\arguments{
\item{...}{Specify start or end boundaries.
\describe{
\item{\code{by_time(...)}}{Numeric time values in units of \code{time_channel}.}
\item{\code{by_label(...)}}{Character strings to match in \code{event_channel}.
All matching occurrences are returned.}
\item{\code{by_lap(...)}}{Integer lap numbers to match in \code{event_channel}.
For \code{start}, resolves to the first sample of each lap. For \code{end},
resolves to the last sample.}
\item{\code{by_sample(...)}}{Integer sample indices (row numbers).}
}}
}
\value{
An object of class \code{"mnirs_interval"} for use with the \code{start}
and \code{end} arguments of \code{\link[=extract_intervals]{extract_intervals()}}.
}
\description{
Helper functions to define interval start or end boundaries for
\code{\link[=extract_intervals]{extract_intervals()}}.
}
\details{
These helpers can be used explicitly for arguments \code{start}/\code{end}, or raw
values can be passed directly:
\itemize{
\item Numeric -> \code{\link[=by_time]{by_time()}}
\item Character -> \code{\link[=by_label]{by_label()}},
\item Explicit integer (e.g. \code{2L}) -> \code{\link[=by_lap]{by_lap()}}.
\item Use \code{\link[=by_sample]{by_sample()}} explicitly for sample indices.
}
}
\examples{
## read example data
data <- read_mnirs(
    example_mnirs("train.red"),
    nirs_channels = c(
        smo2_left = "SmO2 unfiltered",
        smo2_right = "SmO2 unfiltered"
    ),
    time_channel = c(time = "Timestamp (seconds passed)"),
    event_channel = c(lap = "Lap/Event"),
    zero_time = TRUE,
    verbose = FALSE
)

## start and end by time
extract_intervals(data, start = by_time(66), end = by_time(357))

## start by lap
extract_intervals(data, start = by_lap(2, 4), span = 0)

## introduce event_channel with "start" string
data$event <- NA_character_
data$event[1000] <- "start"
data <- create_mnirs_data(data, event_channel = "event")

## start by label, end by time
extract_intervals(data, start = by_label("start"), end = by_time(1500))

## multiple intervals by sample index
extract_intervals(data, start = by_sample(1000, 1500), end = by_sample(2000, 2600))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{train.red_intervals.csv}
\alias{train.red_intervals.csv}
\title{10 Hz Train.Red App export}
\format{
.csv file with header metadata and 10 columns and 11995 rows:
\describe{
\item{Timestamp (seconds passed)}{Elapsed time (s).}
\item{Lap/Event}{Lap number (numeric).}
\item{SmO2}{Muscle oxygen saturation, filtered (\%). Two channels have
duplicated names. If both are called, the second will be renamed to
\code{SmO2_1}.}
\item{SmO2 unfiltered}{Muscle oxygen saturation, raw signal (\%). Two
channels have duplicated names. If both are called, the second will
be renamed to \verb{SmO2 unfiltered_1}.}
\item{O2HB unfiltered}{Oxyhaemoglobin concentration, raw signal
(arbitrary units). Two channels have duplicated names. If both are
called, the second will be renamed to \verb{O2HB unfiltered_1}.}
\item{HHB unfiltered}{Deoxyhaemoglobin concentration, raw signal
(arbitrary units). Two channels have duplicated names. If both are
called, the second will be renamed to \verb{HHb unfiltered_1}.}
}

Channel mapping for \code{\link[=read_mnirs]{read_mnirs()}}:
\itemize{
\item \code{nirs_channels = c("SmO2", "SmO2 unfiltered", "O2HB unfiltered", "HHb unfiltered")}
\item \code{time_channel = c("Timestamp (seconds passed)")}
\item \code{event_channel = c("Lap/Event")}
\item \code{interval_times = list(start = c(2150.09, 2872.28), end = c(2452.26, 3167.98))}
\item \code{interval_times = list(start = c(65.94, 788.13), end = c(368.11, 1083.83))} from zero_time
}
}
\source{
Train.Red (Train.Red B.V.), exported via Train.Red app
(https://train.red/)
}
\description{
Exported from Train.Red app, recorded at 10 Hz.
Containing two 5-minute cycling work intervals, placed on
bilateral vastus lateralis muscle sites. Some data channels
have been omitted to reduce file size.
}
\examples{
example_mnirs("train.red")

}
\seealso{
\code{\link[=read_mnirs]{read_mnirs()}}, \code{\link[=example_mnirs]{example_mnirs()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_mnirs_helpers.R
\name{detect_device_channels}
\alias{detect_device_channels}
\title{Detect known channels for a device}
\usage{
detect_device_channels(
  data,
  header_row = 1L,
  nirs_device = NULL,
  nirs_channels = NULL,
  time_channel = NULL,
  keep_all = FALSE,
  verbose = TRUE
)
}
\description{
Detect known channels for a device
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rescale_mnirs.R
\name{rescale_mnirs}
\alias{rescale_mnirs}
\title{Re-scale data range}
\usage{
rescale_mnirs(
  data,
  nirs_channels = list(),
  range,
  verbose = TRUE
)
}
\arguments{
\item{data}{A data frame of class \emph{"mnirs"} containing time series data and
metadata.}

\item{nirs_channels}{A \code{list()} of character vectors indicating grouping
structure of mNIRS channel names to operate on (see \emph{Details}). Must
match column names in \code{data} exactly. Retrieved from metadata if not
defined explicitly.
\describe{
\item{\code{list("A", "B", "C")}}{Will operate on each channel independently,
losing the relative scaling between channels.}
\item{\code{list(c("A", "B", "C"))}}{Will operate on all channels together,
preserving the relative scaling between channels.}
\item{\code{list(c("A", "B"), c("C", "D"))}}{Will operate on channels \code{A}
& \code{B} in one group, and \code{C} & \code{D} in another group, preserving
relative scaling within, but not between groups.}
}}

\item{range}{A numeric vector in the form \code{c(min, max)}, indicating the
range of output values to which data channels will be re-scaled.}

\item{verbose}{Logical. Default is \code{TRUE}. Display or silence (if \code{FALSE})
warnings and information messages helpful for troubleshooting. Ad
global default can be set via \code{options(mnirs.verbose = FALSE)}.}
}
\value{
A \link[tibble:tibble-package]{tibble} of class \emph{"mnirs"} with metadata
available with \code{attributes()}.
}
\description{
Expand or reduce the range (min and max values) of data channels to a new
amplitude/dynamic range, e.g. re-scale the range of NIRS data to \code{c(0, 100)}.
}
\details{
\code{nirs_channels = list()} can be used to group data channels (column names)
to preserve absolute or relative scaling.
\itemize{
\item Channels grouped together in a vector (e.g. \code{list(c("A", "B"))}) will be
re-scaled to a common range, and the relative scaling within that group
will be preserved.
\item Channels in separate list vectors (e.g. \code{list("A", "B")}) will be
re-scaled independently, and relative scaling between groups will be lost.
\item A single vector of channel names (e.g. \code{c("A", "B")}) will group
channels together.
\item Channels (columns) in \code{data} not explicitly defined in \code{nirs_channels}
will be passed through untouched to the output data frame.
}

\code{nirs_channels} can be retrieved automatically from \code{data} of class
\emph{"mnirs"} which has been processed with \code{{mnirs}}, if not defined
explicitly. This will default to returning all \code{nirs_channels} grouped
together, and should be defined explicitly for other grouping arrangements.
}
\examples{
## read example data
data <- read_mnirs(
    file_path = example_mnirs("moxy_ramp"),
    nirs_channels = c(smo2_left = "SmO2 Live",
                      smo2_right = "SmO2 Live(2)"),
    time_channel = c(time = "hh:mm:ss"),
    verbose = FALSE
) |>
    rescale_mnirs(        ## un-grouped nirs channels to rescale separately 
        nirs_channels = list(smo2_left, smo2_right), 
        range = c(0, 100) ## rescale to a 0-100\% functional exercise range
    )

data

\donttest{
    if (requireNamespace("ggplot2", quietly = TRUE)) {
        plot(data, time_labels = TRUE) +
            ggplot2::geom_hline(yintercept = c(0, 100), linetype = "dotted")
    }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_mnirs_helpers.R
\name{read_file}
\alias{read_file}
\title{Read raw data frame from file path}
\usage{
read_file(file_path)
}
\description{
Read raw data frame from file path
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.mnirs.R
\name{plot.mnirs}
\alias{plot.mnirs}
\title{Plot \emph{{mnirs}} objects}
\usage{
\method{plot}{mnirs}(
  x,
  points = FALSE,
  time_labels = FALSE,
  n.breaks = 5,
  na.omit = FALSE,
  ...
)
}
\arguments{
\item{x}{Data frame or list of data frames of class \emph{"mnirs"} (e.g. from
\code{\link[=extract_intervals]{extract_intervals()}}). List input produces a faceted plot with one
panel per element.}

\item{points}{Logical. Default is \code{FALSE}. If \code{TRUE} displays
\code{ggplot2::geom_points()}. Otherwise displays \code{ggplot2::geom_lines()}.}

\item{time_labels}{Logical. Default is \code{FALSE}. If \code{TRUE} displays x-axis
time values formatted as \emph{"hh:mm:ss"} using \code{\link[=format_hmmss]{format_hmmss()}}. Otherwise,
x-axis values are displayed as numeric.}

\item{n.breaks}{A numeric value specifying the number of breaks in both
x- and y-axes. Default is \code{5}.}

\item{na.omit}{Logical. Default is \code{FALSE}. If \code{TRUE} omits missing (\code{NA})
and non-finite \code{c(Inf, -Inf, NaN)} from display.}

\item{...}{Additional arguments.}
}
\value{
A \link[ggplot2:ggplot]{ggplot2} object.
}
\description{
Create a base plot for data frames or lists of data frames with class
\emph{"mnirs"}.
}
\details{
When \code{x} is a named list of \emph{"mnirs"} data frames, elements are bound into a
single data frame and displayed as faceted panels via
\code{\link[ggplot2:facet_wrap]{ggplot2::facet_wrap()}}.

Arguments in \code{...} are currently passed to \code{\link[ggplot2:facet_wrap]{ggplot2::facet_wrap()}}
formals, such as \code{nrow}, \code{ncol}, and \code{scales} for more precise control.
}
\examples{
\dontshow{if (rlang::is_installed(c("ggplot2", "scales"))) withAutoprint(\{ # examplesIf}
data <- read_mnirs(
    example_mnirs("train.red"),
    nirs_channels = c(smo2 = "SmO2"),
    time_channel = c(time = "Timestamp (seconds passed)"),
    verbose = FALSE
)

## plot time labels as "hh:mm:ss"
plot(data, time_labels = TRUE)

data_list <- extract_intervals(
    data,
    start = by_time(2452, 3168),
    span = c(-60, 120),
    verbose = FALSE
)

## plot a list of mnirs data frames as faceted panels
plot(data_list, time_labels = TRUE)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_mnirs_helpers.R
\docType{data}
\name{device_patterns}
\alias{device_patterns}
\title{Known channel names and detection patterns for supported mNIRS devices}
\format{
An object of class \code{list} of length 5.
}
\usage{
device_patterns
}
\description{
Known channel names and detection patterns for supported mNIRS devices
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.mnirs.R
\name{format_hmmss}
\alias{format_hmmss}
\title{Format time span data as h:mm:ss}
\usage{
format_hmmss(x)
}
\arguments{
\item{x}{A numeric vector.}
}
\value{
A character vector the same length as \code{x}.
}
\description{
Convert numeric time span data to \code{h:mm:ss} format for pretty plotting.
Inspired by \code{\link[ggplot2:scale_date]{ggplot2::scale_x_time()}}.
}
\details{
If all values are less than 3600 (1 hour), then format is returned as
\code{mm:ss}. If any value is greater than 3600, format is returned as
\code{h:mm:ss} with leading zeroes.
}
\examples{
\dontshow{if (rlang::is_installed("ggplot2")) withAutoprint(\{ # examplesIf}
x <- 0:120
y <- sin(2 * pi * x / 15) + rnorm(length(x), 0, 0.2)

ggplot2::ggplot(data.frame(x, y), ggplot2::aes(x, y)) +
    theme_mnirs() +
    ggplot2::scale_x_continuous(
        breaks = breaks_timespan(),
        labels = format_hmmss
    ) +
    ggplot2::geom_line()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_mnirs_helpers.R
\name{detect_mnirs_device}
\alias{detect_mnirs_device}
\title{Detect mnirs device from file metadata}
\usage{
detect_mnirs_device(data)
}
\description{
Detect mnirs device from file metadata
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_mnirs_helpers.R
\name{detect_time_channel}
\alias{detect_time_channel}
\title{Detect time_channel from header row}
\usage{
detect_time_channel(
  data,
  time_channel = NULL,
  nirs_device = NULL,
  verbose = TRUE
)
}
\description{
Detect time_channel from header row
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_mnirs_helpers.R
\name{remove_empty_rows_cols}
\alias{remove_empty_rows_cols}
\title{Remove Empty Rows and Columns}
\usage{
remove_empty_rows_cols(data)
}
\description{
Remove Empty Rows and Columns
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract_interval_helpers.R
\name{ensemble_intervals}
\alias{ensemble_intervals}
\title{Ensemble average multiple intervals}
\usage{
ensemble_intervals(df_list, nirs_channels, metadata, verbose = TRUE)
}
\description{
Ensemble average multiple intervals
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/replace_helpers.R
\name{preserve_na}
\alias{preserve_na}
\alias{restore_na}
\title{Preserve and restore NA information within a vector}
\usage{
preserve_na(x)

restore_na(y, na_info)
}
\arguments{
\item{x}{A vector containing missing \code{NA} values.}

\item{y}{A vector of valid non-\code{NA} values returned from \code{preserve_na()}.}

\item{na_info}{A list returned from \code{preserve_na()}.}
}
\value{
\code{preserve_na()} returns a list \code{na_info} with components:
\itemize{
\item \code{na_info$x_valid}: A vector with \code{NA} values removed.
\item \code{na_info$x_length}: A numeric value of the original input vector length.
\item \code{na_info$na_idx}: A logical vector preserving \code{NA} positions.
}

\code{restore_na()} returns a vector \code{y} the same length as the original
input vector \code{x} with \code{NA} values restored to their original positions.
}
\description{
\code{preserve_na()} stores \code{NA} vector positions and extracts valid non-\code{NA}
values for later restoration with \code{restore_na()}.

\code{restore_na()} restores \code{NA} values to their original vector positions
after processing valid non-\code{NA} values returned from \code{preserve_na()}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_mnirs_helpers.R
\name{extract_start_timestamp}
\alias{extract_start_timestamp}
\title{Extract earliest POSIXct value from file header metadata}
\usage{
extract_start_timestamp(file_header)
}
\description{
Extract earliest POSIXct value from file header metadata
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter_mnirs.R
\name{filter_butter}
\alias{filter_butter}
\title{Apply a Butterworth digital filter}
\usage{
filter_butter(
  x,
  order = 2L,
  W,
  type = c("low", "high", "stop", "pass"),
  edges = c("rev", "rep1", "none"),
  na.rm = FALSE,
  ...
)
}
\arguments{
\item{x}{A numeric vector.}

\item{order}{An integer defining the filter order (\emph{default} \code{order = 2}).}

\item{W}{A one- or two-element numeric vector defining the filter cutoff
frequency(ies) as a fraction of the Nyquist frequency (see \emph{Details}).}

\item{type}{A character string indicating the digital filter type (see
\emph{Details}).
\describe{
\item{\code{"low"}}{For a \emph{low-pass} filter (the \emph{default}).}
\item{\code{"high"}}{For a \emph{high-pass} filter.}
\item{\code{"stop"}}{For a \emph{stop-band} (band-reject) filter.}
\item{\code{"pass"}}{For a \emph{pass-band} filter.}
}}

\item{edges}{A character string indicating edge detection padding for \code{x}.
\describe{
\item{\code{"rev"}}{Will pad \code{x} with the preceding 5\% data in reverse
sequence (\emph{the default}).}
\item{\code{"rep1"}}{Will pad \code{x} by repeating the last preceding value.}
\item{\code{"none"}}{Will return the unpadded \code{\link[signal:filtfilt]{signal::filtfilt()}} output.}
}}

\item{na.rm}{Logical; default is \code{FALSE}, propagates any \code{NA}s to the
returned vector. If \code{TRUE}, ignores \code{NA}s and processes available valid
samples within the local window. May return errors or warnings. (see
\emph{Details}).}

\item{...}{Additional method-specific arguments must be specified
(see \emph{Details}).}
}
\value{
A numeric vector the same length as \code{x}.
}
\description{
Apply a Butterworth digital filter to vector data with \code{\link[signal:butter]{signal::butter()}}
and \code{\link[signal:filtfilt]{signal::filtfilt()}} which handles 'edges' better at the start and end
of the data.
}
\details{
Applies a centred (two-pass symmetrical) Butterworth digital filter from
\code{\link[signal:butter]{signal::butter()}} and \code{\link[signal:filtfilt]{signal::filtfilt()}}.

Filter \code{type} defines how the desired signal frequencies are either
passed or rejected from the output signal. \emph{Low-pass} and \emph{high-pass}
filters allow only frequencies \emph{lower} or \emph{higher} than the cutoff
frequency \code{W} to be passed through as the output signal, respectively.
\emph{Stop-band} defines a critical range of frequencies which are rejected
from the output signal. \emph{Pass-band} defines a critical range of
frequencies which are passed through as the output signal.

The filter order (number of passes) is defined by \code{order}, typically in
the range \verb{order = [1, 10]}. Higher filter order tends to capture more
rapid changes in amplitude, but also causes more distortion around
those change points in the signal. General advice is to use the
lowest filter order which sufficiently captures the desired rapid
responses in the data.

The critical (cutoff) frequency is defined by \code{W}, a numeric value for
\emph{low-pass} and \emph{high-pass} filters, or a two-element vector
\code{c(low, high)} defining the lower and upper bands for \emph{stop-band} and
\emph{pass-band} filters. \code{W} represents the desired fractional cutoff
frequency in the range \verb{W = [0, 1]}, where \code{1} is the Nyquist
frequency, i.e., half the sample rate of the data in Hz.

Missing values (\code{NA}) in \code{x} will cause an error unless \code{na.rm = TRUE}.
Then \code{NA}s will be ignored and passed through to the returned vector.
}
\examples{
set.seed(13)
sin <- sin(2 * pi * 1:150 / 50) * 20 + 40
noise <- rnorm(150, mean = 0, sd = 6)
noisy_sin <- sin + noise
without_edge_detection <- filter_butter(
    x = noisy_sin,
    order = 2,
    W = 0.1,
    edges = "none"
)
with_edge_detection <- filter_butter(
    x = noisy_sin,
    order = 2,
    W = 0.1,
    edges = "rep1"
)

\dontshow{if (rlang::is_installed("ggplot2")) withAutoprint(\{ # examplesIf}
ggplot2::ggplot(data.frame(), ggplot2::aes(x = seq_along(noise))) +
    theme_mnirs() +
    scale_colour_mnirs(name = NULL) +
    ggplot2::geom_line(ggplot2::aes(y = noisy_sin)) +
    ggplot2::geom_line(
        ggplot2::aes(y = without_edge_detection, colour = "without_edge_detection")
    ) +
    ggplot2::geom_line(
        ggplot2::aes(y = with_edge_detection, colour = "with_edge_detection")
    )
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[signal:filtfilt]{signal::filtfilt()}}, \code{\link[signal:butter]{signal::butter()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/resample_mnirs.R
\name{resample_mnirs}
\alias{resample_mnirs}
\title{Re-sample an \emph{{mnirs}} data frame}
\usage{
resample_mnirs(
  data,
  time_channel = NULL,
  sample_rate = NULL,
  resample_rate = sample_rate,
  method = c("none", "linear", "locf"),
  verbose = TRUE
)
}
\arguments{
\item{data}{A data frame of class \emph{"mnirs"} containing time series data and
metadata.}

\item{time_channel}{A character string naming the time or sample column.
Must match a column name in \code{data} exactly.
\itemize{
\item If \code{NULL} (default), the \code{time_channel} metadata attribute of \code{data} is
used.
}}

\item{sample_rate}{A numeric sample rate in Hz.
\itemize{
\item If \code{NULL} (default), the \code{sample_rate} metadata attribute of \code{data} will
be used if detected, or the sample rate will be estimated from
\code{time_channel}.
}}

\item{resample_rate}{An \emph{optional} sample rate (Hz) for the output data
frame. If \code{NULL} (\emph{default}) resamples to the existing \code{sample_rate},
which regularises any irregular samples without changing the rate.}

\item{method}{A character string specifying how new samples are filled.
Default is \emph{"none"}. Filling must be opted into explicitly (see \emph{Details}):
\describe{
\item{\code{"none"}}{Matches each new sample to the nearest original
\code{time_channel} value without any interpolation, to within tolerance of
half a sample-interval. New samples are returned as \code{NA}.}
\item{\code{"locf"}}{(\emph{"Last observation carried forward"}). Fills new and
missing samples with the most recent valid non-\code{NA} value to the left,
or the nearest valid value to the right for leading \code{NA}s. Safe for
numeric, integer, and character columns.}
\item{\code{"linear"}}{Fills new and missing samples via linear interpolation
using \code{\link[stats:approxfun]{stats::approx()}}. Suitable for numeric columns only; non-numeric
columns will fall back to \code{"locf"} behaviour.}
}}

\item{verbose}{Logical. Default is \code{TRUE}. Display or silence (if \code{FALSE})
warnings and information messages helpful for troubleshooting. Ad
global default can be set via \code{options(mnirs.verbose = FALSE)}.}
}
\value{
A \link[tibble:tibble-package]{tibble} of class \code{"mnirs"}. Metadata are
stored as attributes and can be accessed with \code{attributes(data)}.
}
\description{
Up- or down-sample an \emph{"mnirs"} data frame to a new sample rate, filling
new samples via nearest-neighbour matching or interpolation.
}
\details{
This function uses \code{\link[=replace_missing]{replace_missing()}} (based on \code{\link[stats:approxfun]{stats::approx()}}) to
interpolate across new samples in the resampled data range.
\subsection{Sample rate and time channel}{

\code{time_channel} and \code{sample_rate} are retrieved automatically from \code{data}
of class \emph{"mnirs"}, if not defined explicitly.

Otherwise, \code{sample_rate} will be estimated from the values in \code{time_channel}.
However, this may return unexpected values, and it is safer to define
\code{sample_rate} explicitly or retrieve it from \emph{"mnirs"} metadata.
}

\subsection{Default behaviour}{

When \code{resample_rate} is omitted, the output has the same \code{sample_rate} as
the input but with a regular, evenly-spaced \code{time_channel}. This is useful
for regularising data that contains missing or repeated samples without
changing the nominal rate.
}

\subsection{Column handling}{

Numeric columns are interpolated according to \code{method} (see
\code{?replace_missing}). Non-numeric columns (e.g. character event labels,
integer lap numbers) are always filled by last-observation-carried-forward,
regardless of \code{method}:
\itemize{
\item For \code{method = "none"}, existing rows are matched to the nearest original
values of \code{time_channel} without interpolation or filling, meaning newly
created samples and any \code{NA}s in the original data are returned as \code{NA}.
\item When down-sampling, numeric columns use time-weighted averaging.
Non-numeric columns use the first valid value in each output bin.
}
}
}
\examples{
## read example data
data <- read_mnirs(
    file_path = example_mnirs("moxy_ramp"),
    nirs_channels = c(smo2 = "SmO2 Live"),
    time_channel = c(time = "hh:mm:ss"),
    verbose = TRUE
)

## note warning about irregular sampling
data

data_resampled <- resample_mnirs(
    data,               ## blank channels will be retrieved from metadata
    resample_rate = 2,  ## blank by default will resample to `sample_rate`
    method = "linear",  ## linear interpolation across resampled indices
    verbose = TRUE      
)

## note the altered `time` values resolving the above warning
data_resampled

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract_intervals.R
\name{extract_intervals}
\alias{extract_intervals}
\title{Extract intervals from \emph{{mnirs}} data}
\usage{
extract_intervals(
  data,
  nirs_channels = NULL,
  time_channel = NULL,
  event_channel = NULL,
  sample_rate = NULL,
  start = NULL,
  end = NULL,
  span = list(c(-60, 60)),
  event_groups = c("distinct", "ensemble"),
  zero_time = FALSE,
  verbose = TRUE
)
}
\arguments{
\item{data}{A data frame of class \emph{"mnirs"} containing time series data and
metadata.}

\item{nirs_channels}{A character vector or a \code{list()} of character vectors
of mNIRS channel names to operate on within each interval (see \emph{Details}).
Names must match column names in \code{data} exactly.
\itemize{
\item Must only be specified when \code{event_groups} contains \emph{"ensemble"}-
averaged intervals. If \code{event_groups = "distinct"} no channel processing
occurs.
\item If \code{NULL} (default), channels are retrieved from \emph{"mnirs"} metadata.
}}

\item{time_channel}{A character string naming the time or sample column.
Must match a column name in \code{data} exactly.
\itemize{
\item If \code{NULL} (default), the \code{time_channel} metadata attribute of \code{data} is
used.
}}

\item{event_channel}{An \emph{optional} character string giving the name of an
event/lap column. The column may contain character event labels or integer
lap numbers.
\itemize{
\item Required when using \code{\link[=by_label]{by_label()}} or \code{\link[=by_lap]{by_lap()}} for \code{start} or \code{end}.
\item Retrieved from metadata if not defined explicitly.
}}

\item{sample_rate}{An \emph{optional} numeric sample rate (Hz) used to bin time
values for ensemble-averaging. If \code{NULL}, will be estimated from
\code{time_channel} (see \emph{Details}).}

\item{start}{Specifies where intervals begin. Either raw values -- numeric
for time values, character for event labels, explicit integer (e.g. \code{2L})
for lap numbers -- or created with \code{\link[=by_time]{by_time()}}, \code{\link[=by_label]{by_label()}}, \code{\link[=by_lap]{by_lap()}},
or \code{\link[=by_sample]{by_sample()}}.}

\item{end}{Specifies where intervals end. Either raw values -- numeric for
time values, character for event labels, explicit integer (e.g. \code{2L})
for lap numbers -- or created with \code{\link[=by_time]{by_time()}}, \code{\link[=by_label]{by_label()}}, \code{\link[=by_lap]{by_lap()}},
or \code{\link[=by_sample]{by_sample()}}.}

\item{span}{A one- or two-element numeric vector \code{c(before, after)} in units
of \code{time_channel}, or a \code{list()} of such vectors. (\emph{default}
\code{span = c(-60, 60)}. Applied additively to interval boundaries:
\itemize{
\item When both \code{start} and \code{end} are specified: \code{span[1]} shifts start times,
\code{span[2]} shifts end times.
\item When only \code{start} or only \code{end} is specified: both \code{span[1]} and
\code{span[2]} apply as a window around the event).
\item A single \emph{positive} value is recycled to shift the end times (e.g.
\code{span = 60} -> \code{c(0, 60)}).
\item A single \emph{negative} value is recycled to shift the start times (e.g.
\code{span = -60} -> \code{c(-60, 0)}).
}}

\item{event_groups}{Either a character string or a \code{list()} of integer
vectors specifying how to group intervals (see \emph{Details}).
\describe{
\item{\code{"distinct"}}{The default. Extract each interval as an independent
data frame.}
\item{\code{"ensemble"}}{Ensemble-average each specified \code{nirs_channel} across
all detected intervals, returning a single data frame.}
\item{\code{list(c(1, 2), c(3, 4))}}{Ensemble-average each specified
\code{nirs_channel} within each group and return one data frame per group.}
}}

\item{zero_time}{Logical. Default is \code{FALSE}. If \code{TRUE}, re-calculates
numeric \code{time_channel} values to start from zero within each interval
data frame.}

\item{verbose}{Logical. Default is \code{TRUE}. Display or silence (if \code{FALSE})
warnings and information messages helpful for troubleshooting. Ad
global default can be set via \code{options(mnirs.verbose = FALSE)}.}
}
\value{
A named \code{list()} of \link[tibble:tibble-package]{tibbles} of class
\emph{"mnirs"}, each with metadata available via \code{attributes()}.
}
\description{
Extract intervals from \emph{"mnirs"} time series data, specifying interval
start and end boundaries by time value, event label, lap number, or sample
index.
}
\details{
\subsection{Interval specification}{

Interval \code{start} and \code{end} boundaries are specified using helper functions,
or by passing raw values directly:

\describe{
\item{\code{\link[=by_time]{by_time()}}}{Time values in units of \code{time_channel}.}
\item{\code{\link[=by_label]{by_label()}}}{Strings to match in \code{event_channel}.
All matching occurrences are returned.}
\item{\code{\link[=by_lap]{by_lap()}}}{Lap numbers to match in \code{event_channel}. Resolves to the
first sample of each lap for \code{start}, and the last lap sample for \code{end}}
\item{\code{\link[=by_sample]{by_sample()}}}{Integer sample indices (row numbers).}
}

Raw values supplied to \code{start}/\code{end} are auto-coerced:
\itemize{
\item Numeric -> \code{\link[=by_time]{by_time()}}
\item Character -> \code{\link[=by_label]{by_label()}},
\item Explicit integer (e.g. \code{2L}) -> \code{\link[=by_lap]{by_lap()}}.
\item Use \code{\link[=by_sample]{by_sample()}} explicitly for sample indices.
}

\code{start} and \code{end} can use different specification types (e.g., start by
label, end by time). When lengths differ, the shorter is recycled.
}

\subsection{Time span window}{

\code{span} additively expands the time span window around interval boundaries.
\itemize{
\item A two-value vector expands the \code{start} and \code{end}, respectively:
\code{span = c(-60, 60)} expands the \code{start} earlier by \code{60}, and the \code{end}
later by \code{60}. For example,
\verb{start = by_time(30), end = by_time(60), span = c(-5, 10)} returns an
interval of \verb{[25, 70]}.
\item A single numeric value is recycled according to the sign: \code{span = -60}
becomes \code{c(-60, 0)} to expand the \code{start} earlier. \code{span = 60} becomes
\code{c(0, 60)} to expand the \code{end} later.
\item If only \code{start} is specified alone, both span values expand the single
boundary window: \verb{start = by_time(30), span = c(-5, 60)} returns
\verb{[25, 90]}.
}
}

\subsection{Per-interval nirs_channels for ensemble-averaging}{

When \code{event_groups = "ensemble"} or a list of numeric grouped intervals,
\code{nirs_channels} can be specified as a list of column names to override
ensemble-averaging across interval. For example, to exclude a channel
in one interval:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{nirs_channels = list(
  c(A, B, C),
  c(A, C) ## channel "B" is excluded
)
}\if{html}{\out{</div>}}

If all grouped intervals can include all \code{nirs_channels}, or if
\code{event_groups = "distinct"}, a single \code{nirs_channels} character vector can
be supplied and recycled to all groups, or left as \code{NULL} for channels to
be taken from \emph{"mnirs"} metadata.
}

\subsection{Grouping intervals}{

\code{event_groups} controls whether extracted intervals are returned as distinct
data frames or ensemble-averaged.

\describe{
\item{\code{"distinct"}}{The default. Extract each interval and return a
list of independent data frames.}
\item{\code{"ensemble"}}{Ensemble-average each specified \code{nirs_channel} across
all detected intervals and return a one-item list with a single data
frame.}
\item{\code{list(c(1, 2), c(3, 4))}}{Ensemble-average each specified
\code{nirs_channel} within each group and return a list with one data frame
for each group. Any intervals detected but not specified in
\code{event_groups} are returned as distinct.}
}

\code{event_groups} lists can be named (e.g.
\code{list(low = c(1, 2), high = c(3, 4))}) and will pass those names to the
returned list of data frames.

When \code{event_groups} is a list of numeric interval numbers, list items in
\code{nirs_channels} and \code{span} are recycled to the number of groups. If lists
are only partially specified, the final item is recycled forward as needed.
Extra items are ignored.
}
}
\examples{
## read example data
data <- read_mnirs(
    example_mnirs("train.red"),
    nirs_channels = c(
        smo2_left = "SmO2 unfiltered",
        smo2_right = "SmO2 unfiltered"
    ),
    time_channel = c(time = "Timestamp (seconds passed)"),
    zero_time = TRUE,
    verbose = FALSE
) |>
    ## avoid issues ensemble-averaging irregular samples
    resample_mnirs(method = "linear", verbose = FALSE) 

## ensemble-average across multiple intervals
interval_list <- extract_intervals(
    data,                       ## channels recycled to all intervals by default
    nirs_channels = c(smo2_left, smo2_right),
    start = by_time(368, 1084), ## manually identified interval start times
    span = c(-20, 90),          ## include the last 180-sec of each interval (recycled)
    event_groups = "ensemble",  ## ensemble-average across two intervals
    zero_time = TRUE            ## re-calculate common time to start from `0`
)

interval_list[[1L]]

\donttest{
  if (requireNamespace("ggplot2", quietly = TRUE)) {
    plot(interval_list[[1L]], time_labels = TRUE) +
      ggplot2::geom_vline(xintercept = 0, linetype = "dotted")
  }
}

}
